/*
 *  PROGRAM: XdmfGen V0.9
 *   
 *  PURPOSE: generate Xdfm "light data" to describe "heavy"
 *           HDF5 data as written by Cronos (mandatory for
 *           visualization using e.g. Paraview).
 *
 *  REQUIRES: text file FN_XBODY
 *
 *  WRITTEN: (first version) Feb 2011 by Jens Kleimann <jk@tp4.rub.de>
 *           (novel version) Jul 2011 (output to single *.xmf)
 */

#define VERSION  "XdmfGen V0.9"
#define USAGE    "XdmfGen <poub> <pname> <itime_1> [<itime_2> [...] ]"
//#define FN_XBODY "xdmf_body.partxmf"

#include <constants.H>
#include <iostream>
#include <vector>
#include <cstdlib>
#include "H5Cpp.h"
#include <sys/stat.h>
#include <fstream>
using namespace std;
#ifndef H5_NO_NAMESPACE
using namespace H5;
#endif

bool file_missing (string filename) {
	FILE *file;
	if ((file = fopen( filename.c_str(), "r"))) {
		fclose (file);
		return false;
	} else {
		cerr << "Error, file \"" + filename + "\" not found!" << endl;
		return true;
	}
}

int listpos (const string & str, const vector<string> & strvect) {
	// --> return position of <str>'s first occurrence in <strvect>
	for (int pos = 0; pos < int(strvect.size()); ++pos)
		if  (strvect[pos] == str)
			return pos;
	return -1;  // report item_not_found
}

void rem_item (const string & str, vector<string> & strvect) {
	// --> remove <str> from <strvect>
	int pos = listpos (str, strvect);
	strvect.erase (strvect.begin() + pos);
}

void paste_line (FILE * xmf, string line) {
	fprintf (xmf, (line + "\n").c_str() );
}

void paste_block (FILE * xmf, string tag,
                  string in1 = "", string in2 = "") {
	// --> copy tagged <text> from stencil file
	//     pattern: [tag]:<text>
	//              012345678
	ifstream xbody ( FN_XBODY );
	bool tag_found = false;
	if (xbody.is_open()) {
		while (! (xbody.eof()) ) {
			string line;
			getline (xbody, line);
			if (line.compare (0, 6, "["+ tag +"]:" ) == 0) {
				tag_found = true;
				size_t pos1 = line.find ("%1%");
				if (pos1 != string::npos)
					line = line.substr(0,pos1) + in1 + line.substr(pos1+3);
				size_t pos2 = line.find ("%2%");
				if (pos2 != string::npos)
					line = line.substr(0,pos2) + in2 + line.substr(pos2+3);
				fprintf (xmf, (line.substr(6)+"\n").c_str() );
			}
		}
		if (tag_found == false)
			cerr << "Warning, tag '" << tag << "' not found." << endl;
		xbody.close();
	} else {
		cerr << "Error: xdmf stencil file '"
		     << FN_XBODY << "' not found." << endl;
		exit(1);
	}
}

int main (int argc, char *argv[]) {
	cout << " ** " << VERSION << " ** " << endl;
	const int Nf = argc-3;   // total number of frames
	if (Nf < 1) {
		cerr << "Usage: " << USAGE << endl;
		exit(1);
	}
	const string poub  (argv[1]);
	const string pname (argv[2]);
	const string gridfile = pname + "_grid/" + pname + "_grid.h5";
	const string xdmffile = pname            +         ".xmf"    ;
	const string x_dirfile = poub + "/" + xdmffile;
	const string g_dirfile = poub + "/" + gridfile;
	int idummy;
	
	// --> check for grid file & get name of coordinate system
	if (file_missing (g_dirfile)) {
		cout << "press <Return> key to proceed...";
		cin >> idummy;
	}
	
	int iiret;   // all-purpose dummy integer return value
	char *attr_value;        // buffer for names of fields
	hid_t id_file, id_group, id_dset, id_space, id_attr;
	herr_t status;
	
	id_file  = H5Fopen  (g_dirfile.c_str(), H5P_DEFAULT, H5F_ACC_RDONLY);
	id_group = H5Gopen2 (id_file, "/Data" , H5P_DEFAULT);
	id_attr  = H5Aopen_name (id_group, "Geom_Type");
	status   = H5Aread  (id_attr, H5Aget_type(id_attr), &attr_value);
	status   = H5Aclose (id_attr);	
	status   = H5Gclose (id_group); 
	status   = H5Fclose (id_file);
	string geom_name = string(attr_value);
	
	// (Car)tesian label subscripts (may) get replaced by specific ones
	const string sub_Car[] = {  "x",   "y",   "z"};
	const string sub_Cyl[] = {  "r", "phi",   "z"};
	const string sub_Sph[] = {"rad", "tet", "phi"};
	const string sub_gen[] = {  "1",   "2",   "3"};
	vector<string> subs;
	if      (geom_name == "Cartesian"  ) subs.assign (sub_Car, sub_Car+3);
	else if (geom_name == "Cylindrical") subs.assign (sub_Cyl, sub_Cyl+3);
	else if (geom_name == "Spherical"  ) subs.assign (sub_Sph, sub_Sph+3);
	else                                 subs.assign (sub_gen, sub_gen+3);
	
	string * itime    = new string [Nf];  // itime strings (from argv[])
	string * time     = new string [Nf];  // simulation timestamps
	string * hdf5file = new string [Nf];  // HDF5 file names 
	int    * rim      = new int    [Nf];  // number of ghost cells
	
	hsize_t fulldimsZYX[3];  // (full) array dimensions
	vector<string> name_S;   // names of (S)calar (om) fields
	vector<string> name_V;   // names of (V)ector fields
	vector<string> name_f;   // names of fields (read-in, both S and V)
	int Nfgrid[3] = {-1};    // grid extension (w/ ghosts)
	bool same_grid = true;
	bool same_vars = true;
	bool same_rim  = true;
	
	for (int it = 0; it < Nf; it++) {  // loop over HDF files
		itime   [it] = argv[3+it];
		hdf5file[it] = (pname + "_float/"   +
		                pname + "_flt_step" + itime[it] + ".h5");
		const string h_dirfile = poub + "/" + hdf5file[it];
		
		cout << "Opening datafile \"" << h_dirfile << "\"... ";
		if (file_missing (h_dirfile)) exit(1);
		else  cout << "OK." << endl;
		
		id_file  = H5Fopen (h_dirfile.c_str(), H5P_DEFAULT, H5F_ACC_RDONLY);
		id_group = H5Gopen2 (id_file, "/Data", H5P_DEFAULT);
		
		// --> get { rim, time } from datafile
		double r_time;
		id_attr = H5Aopen_name (id_group, "rim");
		status  = H5Aread      (id_attr, H5Aget_type(id_attr), &(rim[it]));
		status  = H5Aclose     (id_attr);
		id_attr = H5Aopen_name (id_group, "time");
		status  = H5Aread      (id_attr, H5Aget_type(id_attr), &r_time);
		status  = H5Aclose     (id_attr);
		
		char time_cbuf [20]; // buffer for timestamp conversion (dbl->string)
		idummy = sprintf (time_cbuf, "%f", r_time);
		time[it] = string(time_cbuf);
		
		// --> generate list of all om fields in /Data/
		//     (both scalars and vector comp.s)
		const int n_attr = H5Aget_num_attrs (id_group);
		const size_t maxlen = 20;
		char attr_name[maxlen];  // for name of current attribute
		int ient = 0;            // index for Name_om<XX> entries
		name_f.clear();
		for (int iattr = 0; iattr < n_attr; ++iattr) {
			id_attr = H5Aopen_idx (id_group, iattr);
			iiret = int( H5Aget_name (id_attr, maxlen, attr_name) );
			const string astr = string (attr_name) + "      ";
			if (astr.compare (0,7,string("Name_om")) == 0) {
				hid_t id_memtype = H5Aget_type (id_attr);
				status = H5Aread (id_attr, id_memtype, &attr_value);
				name_f.push_back (string(attr_value)); // add, treat as scalar
				cout << " field #" << ient << " : " << name_f[ient] << endl;
				++ient;
			}
			status = H5Aclose (id_attr);
		}
		
		id_dset  = H5Dopen1 (id_file, ("/Data/" + name_f[0]).c_str());
		id_space = H5Dget_space (id_dset);
		iiret    = H5Sget_simple_extent_dims (id_space, fulldimsZYX, NULL);
		status   = H5Dclose (id_dset);
		status   = H5Fclose (id_file);
		
		int Nfc[3];   // current (full) grid size
		for (short comp = 0; comp < 3; comp++) {
			Nfc[comp] = int(fulldimsZYX[comp]);
			if (it == 0) { Nfgrid[comp] = Nfc[comp]; }
			else same_grid = (Nfc[comp] == Nfgrid[comp]) and same_grid;
		}
		if (it == 0)  { name_S = name_f; }
		else same_vars = (name_S == name_f) and same_vars;
		
		cout << "rim = " << rim[it] << ", time = " << time[it] << endl;
		printf (" fulldimsXYZ = [ %d %d %d ]\n", Nfc[2], Nfc[1], Nfc[0]);
		
		if (it == 0)   // --> sort out vector components into seperate list
			for (int ient = 0; ient < int(name_S.size()); ++ient) {
				string cu_name = name_S[ient];  // current name
				string last_2c = cu_name.substr (cu_name.size()-2, 2);
				if (last_2c == "_x") {     // => look for (y,z) components
					string vecname = cu_name.substr (0, cu_name.size()-2);
					if (listpos (vecname+"_y", name_S) > -1 &&  // all comp.s
					    listpos (vecname+"_z", name_S) > -1)   //  present?
						name_V.push_back (vecname);   // => add to vector list
				}
			}
	}
	cout << "Reading done." << endl;
	
	// --> check consistency
	string res[2];
	res[true] = "--OK--"; res[false] = "-FAIL-";
	for (int it = 1; it < Nf; it++)
			same_rim = same_rim and (rim[it] == rim[0]);
	cout << "[consistency] grid size: " << res[same_grid] << endl;
	cout << "               boundary: " << res[same_rim ] << endl;
	cout << "              variables: " << res[same_vars] << endl;	
	
	int n_vec = int(name_V.size());
	if (n_vec == 0) { cout << "No vectors found." << endl; }
	else            { cout << "Set of vectors found: { ";
		for (int iv = 0; iv < n_vec; ++iv) 
			cout << name_V[iv] << " ";
		cout << "}" << endl;
	}
	
	// --> remove vector components from scalar list
	for (int iv = 0; iv < n_vec; ++iv)
		for (int isub = 0; isub < 3; ++isub)
			rem_item (name_V[iv] + "_" + sub_Car[isub], name_S);
	
	// --> staggered B needs at least one ghost layer!
	const bool mag = (listpos ("B", name_V) > 0);
	if (mag and rim == 0) {
		cout << "Warning: datafile lacks boundary,"
		     << " B field access disabled." << endl;
		rem_item ("B", name_V);
		--n_vec;
	}
	
	string ts("        <!-- ====== time slice separator ====== -->\n        ");
	FILE *xmf = 0;   // --> create and open Xdmf output file
	xmf = fopen (x_dirfile.c_str(), "w");
	
	paste_block (xmf, "beg");  // write Xdmf head(er) part
	paste_line (xmf, "<!ENTITY grid_file   \""+gridfile+"\">");
	for (int it = 0; it < Nf; it++)
		paste_line (xmf, "<!ENTITY data_file" + itime[it] + " \""
		            + hdf5file[it] + "\">");
	fprintf (xmf, "<!ENTITY fullDimsZYX \"%d %d %d\">\n",
	         Nfgrid[0], Nfgrid[1], Nfgrid[2]);
	fprintf (xmf, "<!ENTITY cellDimsZYX \"%d %d %d\">\n",
	         Nfgrid[0]-2*rim[0], Nfgrid[1]-2*rim[0], Nfgrid[2]-2*rim[0]);
	fprintf (xmf, "<!ENTITY nodeDimsZYX \"%d %d %d\">\n",
	         Nfgrid[0]-2*rim[0]+1, Nfgrid[1]-2*rim[0]+1, Nfgrid[2]-2*rim[0]+1);
	fprintf (xmf, "<!ENTITY ghost_cells \"%d\">\n", rim[0]  );
	fprintf (xmf, "<!ENTITY ghostminus1 \"%d\">\n", rim[0]-1);
	paste_line (xmf, "<!ENTITY coordName_1 \"" + subs[0] + "\">");
	paste_line (xmf, "<!ENTITY coordName_2 \"" + subs[1] + "\">");
	paste_line (xmf, "<!ENTITY coordName_3 \"" + subs[2] + "\">");
	paste_line (xmf, "]>");
	
	paste_block (xmf, "dTA");  // def. of transfo matrix + auxilliary terms 
	
	// --> write definition for vector components
	for (int it = 0; it < Nf; it++) {
		string sep = "{define vector components, (" + itime[it] + ")=itime}";
		paste_line (xmf, "  <!-- BEGIN " + sep + " -->\n  ");
		for (int iv = 0; iv < n_vec; ++iv)
			for (int isub = 0; isub < 3; ++isub) {
				string label, compname = name_V[iv] + "_" + sub_Car[isub];
				if (name_V[iv] == "B") label = "dB" + sub_Car[isub]; // B vector
				else                   label = "gVe";          // generic vector
				paste_block (xmf, label, compname, itime[it]);
			}
		paste_line (xmf, ("  <!-- END " + sep + " -->\n  "));
	}
	paste_block (xmf, "do1"); // start domain declaration, begin Cgrid
	
	// --> C grid: print vector attributes, looping over time
	for (int it = 0; it < Nf; it++) {
		if (it > 0)  { paste_line (xmf, ts); }
		paste_block (xmf, "Cg1", time[it]);
		for (int iv = 0; iv < n_vec; ++iv)
			paste_block (xmf, "AtV", name_V[iv], itime[it]);
		paste_block (xmf, "Atu"); // attributes for unit vectors
	}
	paste_block (xmf, "do2");    // end Cgrid, begin Ngrid
	
	// --> N grid: print scalar attributes, looping over time
	for (int it = 0; it < Nf; it++) {
		if (it > 0)  { paste_line (xmf, ts); }
		paste_block (xmf, "Ng1", time[it]);
		for (int is = 0; is < int(name_S.size()); ++is) {
			paste_block (xmf, "AtS", name_S[is], itime[it]);
		}
		paste_line (xmf, "        </Grid>\n        ");
	}
	paste_block (xmf, "end");   // closing tags
	fclose (xmf);
	cout << "Light data written to " << x_dirfile << endl;
	
	delete[] itime;
	delete[] time;
	delete[] hdf5file;
	delete[] rim;
	return 0;
}
