#include "constants.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)
#include "Absorption.H"

#define USE_CROSS_SECTION_INTERPOLATION

using namespace ABSORPTION;

REAL crossSectionFunctional(
		REAL pairBeta
)
{
	assert(pairBeta > 0);
	assert(pairBeta < 1);

	REAL beta2 = sqr(pairBeta);
	return pi * (1 - beta2) * (
			(3. - sqr(beta2)) * log((1. + pairBeta) / (1. - pairBeta))
			- 2 * pairBeta * (2. - beta2)
	);
}

CrossSectionInterpolation::CrossSectionInterpolation(int nPoints):
	nPoints(nPoints)
{
	dPairBeta = 1./(nPoints - 1);

	// + 1 to get rid of the if branching in the interpolation (for the last point)
	values = new REAL[nPoints + 1];
	values[nPoints] = 0.;

	// Set first and last value manually to avoid numerical problems
	values[0] = 0.;
	values[nPoints - 1] = 0.;

	for (int idx=1; idx < nPoints - 1; ++idx) {
		REAL pairBeta = dPairBeta * idx;
		values[idx] = crossSectionFunctional(pairBeta);
	}
}

CrossSectionInterpolation::~CrossSectionInterpolation()
{
	delete [] values;
}

REAL CrossSectionInterpolation::operator()(REAL pairBeta) const
{
	assert(pairBeta > 0);
	assert(pairBeta < 1);

	REAL ratio = pairBeta / dPairBeta;
	int idx = int(ratio);
	REAL lambda = ratio - idx;

	return values[idx] * (1. - lambda) + lambda * values[idx + 1];
}

#ifdef USE_CROSS_SECTION_INTERPOLATION
const CrossSectionInterpolation crossSectionInterpolation(1000);

REAL ABSORPTION::get_crossSectionFunctional(
		REAL pairBeta
)
{
	return crossSectionInterpolation(pairBeta);
}
#else
REAL ABSORPTION::get_crossSectionFunctional(
		REAL pairBeta
)
{
	return crossSectionFunctional(pairBeta);
}
#endif



Absorption::Absorption(
		std::shared_ptr<RadiationField> _seedPhotonField
	)
{
	seedPhotonField = _seedPhotonField;
	iAbsorption = NULL;

//	if (!seedPhotonField->has_sphericalSymmetry()) {
//		throw CException(" No Absorption implementation available for radiation fields without spherical symmetry. ", __PRETTY_FUNCTION__);
//	}

	if(auto iRF_casted = std::dynamic_pointer_cast<RadiationFieldFactorized>(seedPhotonField))
	{
		ccout << "Using the Absorption implementation for factorized radiation fields " << endl;

		set_iAbsorption(new AbsorptionFactorizedRadiationField(iRF_casted));
	} else {
		throw CException(" No Absorption implementation available for this radiation field. ", __PRETTY_FUNCTION__);
	}

}

Absorption::~Absorption()
{
	delete iAbsorption;
}

void Absorption::set_iAbsorption(
		IAbsorption * _iAbsorption
)
{
	delete iAbsorption;
	iAbsorption = _iAbsorption;
}

//void Absorption::compute_constants(
//		Data & gdata,
//		int ix, int iy, int iz
//	)
//{
//	assert(iAbsorption != NULL);
//	iAbsorption->compute_constants(gdata, ix, iy, iz);
//}

std::string Absorption::get_name(
) const
{
	return "Absorption";
}

REAL Absorption::compute_opticalDepth(
		Data & gdata,
		int ix, int iy, int iz,
		const REAL gammaPhotonEnergy,
		const REAL gammaPhotonDirection[3]
)
{
	assert(iAbsorption != NULL);
	assert(ready);
//	cout << normFactor<< endl;
	const double opticalDepth = iAbsorption->compute_opticalDepth(gdata, ix, iy, iz, gammaPhotonEnergy, gammaPhotonDirection);
	assert(!std::isnan(opticalDepth));
	return opticalDepth * normFactor;
}


REAL Absorption::compute(
		Data & gdata,
		int ix, int iy, int iz,
		REAL photonSolidFlux,
		const REAL gammaPhotonEnergy,
		const REAL seedPhotonDirection[3]
)
{
	const double opticalDepth = compute_opticalDepth(gdata, ix, iy, iz, gammaPhotonEnergy, seedPhotonDirection);
	const double attenuation = exp(-opticalDepth);
	if (!std::isfinite(attenuation)) {
		cout << "Nonfinite attenuation " << attenuation << " " << opticalDepth << endl;
	}
	return attenuation * photonSolidFlux;
}

void Absorption::set_normalisation(
		normalisation & normHD,
		normalisation & normParticle,
		normalisation & normEmission
)
{
	Quantity crossSectionUnit = 0.5 * sqr(CRONOS_CONSTANTS::ClassicalElectronRadius);
	normFactor = (crossSectionUnit
				* normParticle.normQuantity(normParticle.SPEC_NUM_DENS)
				* normParticle.normQuantity(normParticle.ENERGY)
				* normHD.normQuantity(normHD.LEN)
			).normalizeTo(UnitQuantity());

	if(NeedsElectronRestEnerergy * needer = dynamic_cast<NeedsElectronRestEnerergy *>(iAbsorption)) {
		Quantity electronRestEnergy = CRONOS_CONSTANTS::ElectronMass * sqr(CRONOS_CONSTANTS::SpeedOfLight);
		REAL electronRestEnergy_num = normParticle.get_num(electronRestEnergy);
		needer->set_electronRestEnergy(electronRestEnergy_num);
	}

	ready = true;
}

void NeedsElectronRestEnerergy::set_electronRestEnergy(
		REAL electronRestEnergy
)
{
	this->electronRestEnergy = electronRestEnergy;
	electronRestEnergy_sqr = sqr(electronRestEnergy);
	electronRestEnergy_sqr_inv = 1./electronRestEnergy_sqr;
}


// -----------------------------------------------------------------------------------------------------------------
// IMPLEMENTAION


AbsorptionFactorizedRadiationField::AbsorptionFactorizedRadiationField(
	 std::shared_ptr<RadiationFieldFactorized> _radiationField
)
{
		radiationField		= _radiationField;
		iRffNumberDensity	= _radiationField->get_iRadiationFieldFactor_NumberDensity();
		iRffEnergy 			= _radiationField->get_iRadiationFieldFactor_Energy();
		iRffDirection		= _radiationField->get_iRadiationFieldFactor_Direction();

		iRffMonochromatic 	= std::dynamic_pointer_cast<IRadiationFieldFactor_Monochromatic>(iRffEnergy);
		iRffBeamed 			= std::dynamic_pointer_cast<IRadiationFieldFactor_Beamed>(iRffDirection);
		iRffSphericalCoords = std::dynamic_pointer_cast<IRFFD_SphericalCoordinates>(iRffDirection);
		sphericalSource 	= std::dynamic_pointer_cast<SphericalSource>(radiationField);

	// Line of Sight Integration
	if (sphericalSource) {
		ccout << "  - Using psi line of sight" << endl;
		lineOfSightIntegral =
				[this](Data &gdata, const REAL pos[3],
					   const REAL gammaPhotonEnergy, const REAL gammaPhotonDirection[3]) -> REAL
				{
					return lineOfSightIntegral_Psi(gdata, pos, gammaPhotonEnergy, gammaPhotonDirection);
				};
	} else {
		ccout << "  - Using standard line of sight" << endl;
		lineOfSightIntegral =
				[this](Data &gdata, const REAL pos[3],
					   const REAL gammaPhotonEnergy, const REAL gammaPhotonDirection[3]) -> REAL
				{
					return lineOfSightIntegral_Length(gdata, pos, gammaPhotonEnergy, gammaPhotonDirection);
				};
	}

	// Solid Angle Integration
	if (iRffBeamed) {
		ccout << "  - Using delta angular integration" << endl;
		solidAngleIntegral_direction =
			[this](Data &gdata, const REAL pos[3],
				const REAL gammaPhotonEnergy, const REAL gammaPhotonDirection[3]) -> REAL
				{
					return solidAngleIntegral_delta(gdata, pos, gammaPhotonEnergy, gammaPhotonDirection);
				};
	} else
	if (
			// One should imply the other in future
			iRffSphericalCoords && sphericalSource
		) {
		ccout << "  - Using angular integration for spherical sources" << endl;
		solidAngleIntegral_sphericalSource =
			[=, this](Data &gdata, const REAL pos[3],
				const REAL gammaPhotonEnergy, const REAL psi) -> REAL
				{
					return solidAngleIntegral_spherical(gdata, pos, gammaPhotonEnergy, psi);
				};
	} else {
		throw CException("Absorption implementation nor available for this case");
	}

	// Energy Integration
	if (iRffMonochromatic) {
		ccout << "  - Using delta energy integration" << endl;
		energyIntegral =
			[this](Data &gdata, const REAL pos[3],
				const REAL gammaPhotonEnergy, const REAL muGammaGamma) -> REAL
				{
					return energyIntegral_delta(gdata, pos, gammaPhotonEnergy, muGammaGamma);
				};
	} else {
		ccout << "  - Using full energy integration" << endl;
		energyIntegral =
			[this](Data &gdata, const REAL pos[3],
				const REAL gammaPhotonEnergy, const REAL muGammaGamma) -> REAL
				{
					return energyIntegral_full(gdata, pos, gammaPhotonEnergy, muGammaGamma);
				};
	}

	gaussLegendreTableEnergy = gsl_integration_glfixed_table_alloc(8);
	gaussLegendreTableAzimuth = gsl_integration_glfixed_table_alloc(7);
	gaussLegendreTablePolar = gsl_integration_glfixed_table_alloc(7);
	gaussLegendreTablePsi = gsl_integration_glfixed_table_alloc(21);

	qagLimitLineOfSightLength = 100;
	integrationWorkspaceLineOfSightLength = gsl_integration_workspace_alloc(qagLimitLineOfSightLength);
}

AbsorptionFactorizedRadiationField::~AbsorptionFactorizedRadiationField()
{
	gsl_integration_glfixed_table_free(gaussLegendreTableEnergy);
	gsl_integration_glfixed_table_free(gaussLegendreTableAzimuth);
	gsl_integration_glfixed_table_free(gaussLegendreTablePolar);
	gsl_integration_glfixed_table_free(gaussLegendreTablePsi);

	gsl_integration_workspace_free(integrationWorkspaceLineOfSightLength);
}



class UndefinedRadiationFieldException: public runtime_error {
public:
	UndefinedRadiationFieldException(): runtime_error("Undefined Radiation Field") {}
};

class RaySourceCollision: public runtime_error {
public:
	RaySourceCollision(): runtime_error("Ray hits source") {}
};

REAL AbsorptionFactorizedRadiationField::compute_opticalDepth(
		Data & gdata,
		int ix, int iy, int iz,
		const REAL gammaPhotonEnergy,
		const REAL gammaPhotonDirection[3]
)
{
	REAL pos0[3] = {gdata.getCen_x(ix), gdata.getCen_y(iy), gdata.getCen_z(iz)};
	REAL opticalDepth;

	try {
		opticalDepth = lineOfSightIntegral(gdata, pos0, gammaPhotonEnergy, gammaPhotonDirection);
	} catch (RaySourceCollision & e) {
		opticalDepth = 1e30;
	} catch (UndefinedRadiationFieldException & e) {
		opticalDepth = 1e30;
	// } catch (UndefinedInsideSourceException & e) {
	// 	opticalDepth = 1e30;
	}

	return opticalDepth;
}


REAL AbsorptionFactorizedRadiationField::lineOfSightIntegral_Length(
		Data & gdata,
		const REAL pos0[3],
		const REAL gammaPhotonEnergy,
		const REAL gammaPhotonDirection[3]
		)
{
//	assert(solidAngleIntegral_direction);
	assert(iRffNumberDensity);

	// Here we have the problem that gsl_function.function needs a function pointer 'double (*)(double, void *)',
	// therefore just casting a (non-static) member function is not working, due to references to other class memebers.
	// This can be avoided by passing the actual function as an argument to a function evaluator.

	std::function<double(double)> integrand;

	// Check if a spherical integrator is available
	if (solidAngleIntegral_sphericalSource) {
		REAL d0 = 0;
		REAL d0cosPsi0 = 0;
		const vector3D & center = sphericalSource->get_center();

		for (int iDir = 0; iDir < 3; ++iDir) {
			REAL d0Comp = center[iDir] - pos0[iDir];
			d0 += sqr(d0Comp);
			d0cosPsi0 = d0Comp * gammaPhotonDirection[iDir];
		}

		d0 = sqrt(d0);
		REAL cosPsi0 = d0cosPsi0 / d0;
		REAL sinPsi0 = sqrt(1. - sqr(cosPsi0));

		// Check if ray is approaching source
		if( cosPsi0 > 0. )
		// Check if the ray hits the source
		if( d0 * sinPsi0 < sphericalSource->get_radius()) {
			throw RaySourceCollision();
		}

		integrand = [&](double length) -> double {
			REAL pos[3];
			for (int iDir = 0; iDir < 3; ++iDir) {
				pos[iDir] = pos0[iDir] + length * gammaPhotonDirection[iDir];
			}

			if (!radiationField->isDefined(gdata, pos)) {
				throw UndefinedRadiationFieldException();
			}

			REAL psi = atan(d0 * sinPsi0 / (d0 * cosPsi0 - length));
			if (length >= d0 * cosPsi0 ) {
				psi += pi;
			}
			return solidAngleIntegral_spherical(gdata, pos, gammaPhotonEnergy, psi)
				   * iRffNumberDensity->numberDensity(gdata, pos);
		};
	}
	else
	if (solidAngleIntegral_direction) {
		integrand = [&](double length) -> double {
			REAL pos[3];
			for (int iDir = 0; iDir < 3; ++iDir) {
				pos[iDir] = pos0[iDir] + length * gammaPhotonDirection[iDir];
			}

			// e.g. the ray hits the star
			if (! radiationField->isDefined(gdata, pos)){
				throw UndefinedRadiationFieldException();
			}

			return solidAngleIntegral_direction(gdata, pos, gammaPhotonEnergy, gammaPhotonDirection)
				   * iRffNumberDensity->numberDensity(gdata, pos);
		};
	} else{
		throw CException("No suitable direction integrand available ", __PRETTY_FUNCTION__);
	}


	auto evaluator = [](double x, void * arg)-> double { // note thunk is captureless
		return (*static_cast<decltype(integrand)*>(arg))(x);
	};

	static gsl_function gsl_lineOfSightIntegrand;
	gsl_lineOfSightIntegrand.function = evaluator;
	gsl_lineOfSightIntegrand.params = &integrand;

	//	return gsl_integration_glfixed(&gsl_lineOfSightIntegrand, psi0, pi, gaussLegendreTable5);

	gsl_set_error_handler_off();


	double abs_error = 1e2;
	double rel_error = 1e-2;
	double result;
	double error;
	int status;

	try {
		status = gsl_integration_qagiu(&gsl_lineOfSightIntegrand, 0, abs_error, rel_error, qagLimitLineOfSightLength,
									   integrationWorkspaceLineOfSightLength, &result, &error);
	} catch (UndefinedRadiationFieldException e) { //e.g. the ray hits the star
		return 1e30;
	}


	if (status == GSL_SUCCESS) {
		return result;
	} else
	if (status == GSL_EDIVERGE) {
		return 1e30;
	} else
	if (status == GSL_EMAXITER) {
		cerr << "  " << __FUNCTION__
			 << " GSL integration: "
			 << string(gsl_strerror(status)) << " "
			 << "est. err = " << error
			 << endl;
		return result;
	} else
	if (status == GSL_EROUND) { // Has happened for big distances to the source
		cerr << "  " << __FUNCTION__
			 << " GSL integration: "
			 << string(gsl_strerror(status)) << " "
			 << "est. res = " << result << " "
			 << "est. err = " << error << " "
			 << endl;
		return result;
	} else {
		throw CException(" GSL integration error: " + string(gsl_strerror(status)), __PRETTY_FUNCTION__);
	}
}



REAL AbsorptionFactorizedRadiationField::lineOfSightIntegral_Psi(
		Data & gdata,
		const REAL pos0[3],
		const REAL gammaPhotonEnergy,
		const REAL gammaPhotonDirection[3]
)
{
	assert(sphericalSource);

	REAL d0 = 0;
	REAL d0cosPsi0 = 0;
	const vector3D & center = sphericalSource->get_center();

	for (int iDir = 0; iDir < 3; ++iDir) {
		REAL d0Comp = center[iDir] - pos0[iDir];
		d0 += sqr(d0Comp);
		d0cosPsi0 = d0Comp * gammaPhotonDirection[iDir];
	}

	d0 = sqrt(d0);
	REAL cosPsi0 = d0cosPsi0 / d0;

	REAL sinPsi0 = sqrt(1. - sqr(cosPsi0));
	REAL psi0 = acos(cosPsi0);


	// Check if ray is approaching source
	if( cosPsi0 > 0. )
	// Check if the ray hits the source
	if( d0 * sinPsi0 < sphericalSource->get_radius()) {
		throw RaySourceCollision();
	}

	// For radial rays this scheme does not work
	// we switch therefor to the standard scheme
	if (cosPsi0 == -1.) {
//		return lineOfSightIntegral_Length(gdata, pos0, gammaPhotonEnergy, gammaPhotonDirection);
		return lineOfSightIntegral_Length_Radial(gdata, pos0, gammaPhotonEnergy, gammaPhotonDirection); // optimized
	}

	std::function<double(double)> integrand;

	// Check if a spherical integrator is available
	if (solidAngleIntegral_sphericalSource) {
		integrand = [&](double psi) -> double {
			REAL cscPsi = 1. / sin(psi);
			REAL length = d0 * sin(psi - psi0) * cscPsi;
			REAL dLen_dPsi = d0 * sqr(cscPsi) * sinPsi0;

			REAL pos[3];
			for (int iDir = 0; iDir < 3; ++iDir) {
				pos[iDir] = pos0[iDir] + length * gammaPhotonDirection[iDir];
			}

			if (!radiationField->isDefined(gdata, pos)) {
				throw UndefinedRadiationFieldException();
			}

			const double solidAngleIntegral = solidAngleIntegral_spherical(gdata, pos, gammaPhotonEnergy, psi);
			assert(!std::isnan(solidAngleIntegral));
			return solidAngleIntegral
				   * iRffNumberDensity->numberDensity(gdata, pos) * dLen_dPsi;
		};
	}
	else
	if (solidAngleIntegral_direction) {
		integrand = [&](double psi) -> double {
			REAL cscPsi = 1. / sin(psi);
			REAL length = d0 * sin(psi - psi0) * cscPsi;
			REAL dLen_dPsi = d0 * sqr(cscPsi) * sinPsi0;

			REAL pos[3];
			for (int iDir = 0; iDir < 3; ++iDir) {
				pos[iDir] = pos0[iDir] + length * gammaPhotonDirection[iDir];
			}

			if (!radiationField->isDefined(gdata, pos)) {
				throw UndefinedRadiationFieldException();
			}

			const double solidAngleIntegral = solidAngleIntegral_direction(gdata, pos, gammaPhotonEnergy, gammaPhotonDirection);
			assert(!std::isnan(solidAngleIntegral));
			return solidAngleIntegral
				   * iRffNumberDensity->numberDensity(gdata, pos) * dLen_dPsi;
		};
	} else{
		throw CException("No suitable direction integrand available ", __PRETTY_FUNCTION__);
	}

	auto evaluator = [](double x, void * arg)-> double {
		return (*static_cast<decltype(integrand)*>(arg))(x);
	};

	static gsl_function gsl_lineOfSightIntegrand;
	gsl_lineOfSightIntegrand.function = evaluator;
	gsl_lineOfSightIntegrand.params = &integrand;

	const double result = gsl_integration_glfixed(&gsl_lineOfSightIntegrand, psi0, pi, gaussLegendreTablePsi);

	assert(!std::isnan(result));
	return result;
}

REAL AbsorptionFactorizedRadiationField::lineOfSightIntegral_Length_Radial(
		Data & gdata,
		const REAL pos0[3],
		const REAL gammaPhotonEnergy,
		const REAL gammaPhotonDirection[3]
)
{
	const REAL psi=pi;
	std::function<double(double)> integrand;

	// Check if a spherical integrator is available
	if (solidAngleIntegral_sphericalSource) {
		integrand = [&](double length) -> double {
			REAL pos[3];
			for (int iDir = 0; iDir < 3; ++iDir) {
				pos[iDir] = pos0[iDir] + length * gammaPhotonDirection[iDir];
			}

			if (!radiationField->isDefined(gdata, pos)) {
				throw UndefinedRadiationFieldException();
			}

			return solidAngleIntegral_spherical(gdata, pos, gammaPhotonEnergy, psi)
				   * iRffNumberDensity->numberDensity(gdata, pos);
		};
	}
	else
	if (solidAngleIntegral_direction) {
		integrand = [&](double length) -> double {
			REAL pos[3];
			for (int iDir = 0; iDir < 3; ++iDir) {
				pos[iDir] = pos0[iDir] + length * gammaPhotonDirection[iDir];
			}

			if (!radiationField->isDefined(gdata, pos)) {
				throw UndefinedRadiationFieldException();
			}

			return solidAngleIntegral_direction(gdata, pos, gammaPhotonEnergy, gammaPhotonDirection)
				   * iRffNumberDensity->numberDensity(gdata, pos);
		};
	} else{
		throw CException("No suitable direction integrand available ", __PRETTY_FUNCTION__);
	}


	auto evaluator = [](double x, void * arg)-> double { // note thunk is captureless
		return (*static_cast<decltype(integrand)*>(arg))(x);
	};

	static gsl_function gsl_lineOfSightIntegrand;
	gsl_lineOfSightIntegrand.function = evaluator;
	gsl_lineOfSightIntegrand.params = &integrand;

	//	return gsl_integration_glfixed(&gsl_lineOfSightIntegrand, psi0, pi, gaussLegendreTable5);

	gsl_set_error_handler_off();


	double abs_error = 1e2;
	double rel_error = 1e-2;
	double result;
	double error;
	int status;

	status = gsl_integration_qagiu(&gsl_lineOfSightIntegrand, 0, abs_error, rel_error, qagLimitLineOfSightLength,
									   integrationWorkspaceLineOfSightLength, &result, &error);


	if (status == GSL_SUCCESS) {
		assert(!std::isnan(result));
		return result;
	} else
	if (status == GSL_EDIVERGE) {
		return 1e30;
	} else
	if (status == GSL_EMAXITER) {
		cerr << "  " << __FUNCTION__
			 << " GSL integration: "
			 << string(gsl_strerror(status)) << " "
			 << "est. err = " << error
			 << endl;
		assert(!std::isnan(result));
		return result;
	} else
	if (status == GSL_EROUND) { // Has happened for big distances to the source
		cerr << "  " << __FUNCTION__
			 << " GSL integration: "
			 << string(gsl_strerror(status)) << " "
			 << "est. res = " << result << " "
			 << "est. err = " << error << " "
			 << endl;
		assert(!std::isnan(result));
		return result;
	} else {
		throw CException(" GSL integration error: " + string(gsl_strerror(status)), __PRETTY_FUNCTION__);
	}
}


REAL AbsorptionFactorizedRadiationField::solidAngleIntegral_delta(
		Data & gdata,
		const REAL pos[3],
		const REAL gammaPhotonEnergy,
		const REAL gammaPhotonDirection[3]
		)
{
	assert(iRffBeamed != NULL);

	REAL seedPhotonDirection [3];
	iRffBeamed->get_beamDirection(gdata, pos, seedPhotonDirection);

	REAL muGammaGamma = 0;
	for (int iDir = 0; iDir < 3; ++iDir) {
		muGammaGamma += seedPhotonDirection[iDir] * gammaPhotonDirection[iDir];
	}

	if (muGammaGamma == 1) {
		return 0.;
	} else {
		return energyIntegral(gdata, pos, gammaPhotonEnergy, muGammaGamma);
	}
}



//REAL AbsorptionFactorizedRadiationField::solidAngleIntegral_sphericalCoords(
//		Data & gdata,
//		const REAL pos[3],
//		const REAL gammaPhotonEnergy,
//		const REAL gammaPhotonDirection[3]
//)
//{
//	assert(iRffSphericalCoords != NULL);
//	REAL muSeed;
//
//	// polar angle
//
//	auto integrandPolar = [&](double phiSeed) -> double {
//		REAL muGammaGamma = 0;
//		REAL photonDirection[3];
//		iRffSphericalCoords->get_dir(gdata, pos, muSeed, phiSeed, photonDirection);
//		for (int iDir = 0; iDir < 3; ++iDir) {
//			muGammaGamma += photonDirection[iDir] * gammaPhotonDirection[iDir];
//		}
//
//		if (1. - muGammaGamma < 1e-15) {
//			return 0.;
//		} else {
//			return iRffSphericalCoords->fDirection(gdata, pos, muSeed, phiSeed)
//				   * energyIntegral(gdata, pos, gammaPhotonEnergy, muGammaGamma);
//		}
//	};
//
//	auto evaluatorPolar=[](double x, void * arg)-> double {
//		return (*static_cast<decltype(integrandPolar)*>(arg))(x);
//	};
//
//	static gsl_function gsl_integrand_polar;
//	gsl_integrand_polar.function = evaluatorPolar;
//	gsl_integrand_polar.params = &integrandPolar;
//
//	REAL phiMin = iRffSphericalCoords->get_phiMin(gdata, pos);
//	REAL phiMax = iRffSphericalCoords->get_phiMax(gdata, pos);
//
//
//	// azimuthal angle
//
//	auto integrandAzimuth = [&](double _muSeed) -> double {
//		muSeed = _muSeed;
//		return gsl_integration_glfixed(&gsl_integrand_polar, phiMin, phiMax, gaussLegendreTablePolar);
//	};
//
//	auto evaluatorAzimuth=[](double x, void * arg)-> double {
//		return (*static_cast<decltype(integrandAzimuth)*>(arg))(x);
//	};
//
//	static gsl_function gsl_integrand_azimuth;
//	gsl_integrand_azimuth.function = evaluatorAzimuth;
//	gsl_integrand_azimuth.params = &integrandAzimuth;
//
//	REAL muSeedMin = iRffSphericalCoords->get_muMin(gdata, pos);
//	REAL muSeedMax = iRffSphericalCoords->get_muMax(gdata, pos);
//
//	// The source is far enough away and can be effectively treated as pointlike
//	// This is set to be equal by the radiation field implementation
//	if (muSeedMin == muSeedMax) {
//		REAL muGammaGamma = 0;
//		REAL photonDirection[3];
//		iRffSphericalCoords->get_dir(gdata, pos, muSeedMin, 0., photonDirection);
//		for (int iDir = 0; iDir < 3; ++iDir) {
//			muGammaGamma += photonDirection[iDir] * gammaPhotonDirection[iDir];
//		}
//
//		return 2 * pi * energyIntegral(gdata, pos, gammaPhotonEnergy, muGammaGamma);
//	}
//
//	return gsl_integration_glfixed(&gsl_integrand_azimuth, muSeedMin, muSeedMax, gaussLegendreTableAzimuth);
//}

REAL AbsorptionFactorizedRadiationField::solidAngleIntegral_spherical(
		Data & gdata,
		const REAL pos[3],
		REAL gammaPhotonEnergy,
		REAL psi
) {
	assert(sphericalSource);
	assert(iRffSphericalCoords);

	REAL muSeed;

	auto integrandPolar = [&](double phiSeed) -> double {
		REAL sinTheta = sqrt(1. - sqr(muSeed));
		REAL muGammaGamma = -cos(psi) * muSeed - sin(psi) * cos(phiSeed) * sinTheta;
		return iRffSphericalCoords->fDirection(gdata, pos, muSeed, phiSeed)
			   * energyIntegral(gdata, pos, gammaPhotonEnergy, muGammaGamma);
	};

	auto evaluatorPolar = [](double x, void *arg) -> double {
		return (*static_cast<decltype(integrandPolar) *>(arg))(x);
	};

	static gsl_function gsl_integrand_polar;
	gsl_integrand_polar.function = evaluatorPolar;
	gsl_integrand_polar.params = &integrandPolar;

	REAL phiMin = iRffSphericalCoords->get_phiMin(gdata, pos);
	REAL phiMax = iRffSphericalCoords->get_phiMax(gdata, pos);

	auto integrandAzimuth = [&](double _muSeed) -> double {
		muSeed = _muSeed;
		return gsl_integration_glfixed(&gsl_integrand_polar, phiMin, phiMax, gaussLegendreTablePolar);
	};

	auto evaluatorAzimuth = [](double x, void *arg) -> double {
		return (*static_cast<decltype(integrandAzimuth) *>(arg))(x);
	};

	static gsl_function gsl_integrand_azimuth;
	gsl_integrand_azimuth.function = evaluatorAzimuth;
	gsl_integrand_azimuth.params = &integrandAzimuth;

	REAL muSeedMin = iRffSphericalCoords->get_muMin(gdata, pos);
	REAL muSeedMax = iRffSphericalCoords->get_muMax(gdata, pos);

	// The source is far enough away and can be effectively treated as pointlike
	// This is set to be equal by the radiation field implementation
	if (muSeedMin == muSeedMax) {
		REAL muGammaGamma = -cos(psi);
		return energyIntegral(gdata, pos, gammaPhotonEnergy, muGammaGamma);
	}

	const double result = gsl_integration_glfixed(&gsl_integrand_azimuth, muSeedMin, muSeedMax, gaussLegendreTableAzimuth);

	assert(!std::isnan(result));
	return result;
}


REAL AbsorptionFactorizedRadiationField::energyIntegral_delta(
		Data &gdata,
		const REAL pos[3],
		const REAL gammaPhotonEnergy,
		const REAL muGammaGamma
)
{
	REAL seedPhotonEnergy = iRffMonochromatic->get_photonEnergy(gdata, pos);
	REAL sInvariant = gammaPhotonEnergy * seedPhotonEnergy * 0.5 * (1. - muGammaGamma) * electronRestEnergy_sqr_inv;

	if (sInvariant > 1.) {
		REAL beta = sqrt(1. - 1./sInvariant);
		return (1. - muGammaGamma) * get_crossSectionFunctional(beta);
	} else {
		return 0;
	}
}

REAL AbsorptionFactorizedRadiationField::energyIntegral_full(
		Data &gdata,
		const REAL pos[3],
		const REAL gammaPhotonEnergy,
		const REAL muGammaGamma
)
{
	if (muGammaGamma == 1.) {
		return 0.;
	}

	REAL zMu = 0.5 * (1. - muGammaGamma);

	auto integrand = [&](double beta) -> double {
		assert(beta > 0);
		assert(beta < 1);
		REAL sInvariant = 1. / (1. - sqr(beta));
		REAL seedPhotonEnergy =  electronRestEnergy_sqr * sInvariant  / (gammaPhotonEnergy * zMu);
//		REAL dEps_dBeta = 2. * sInvariant * beta * seedPhotonEnergy;
		REAL dEps_dBeta_Times_2zMu =
				4. * sInvariant * beta * electronRestEnergy_sqr * sInvariant  / gammaPhotonEnergy;
		return get_crossSectionFunctional(beta)
			* iRffEnergy->fEnergy(gdata, pos, seedPhotonEnergy)
			* dEps_dBeta_Times_2zMu;
	};

	auto evaluator=[](double x, void * arg)-> double {
		return (*static_cast<decltype(integrand)*>(arg))(x);
	};

	static gsl_function gsl_integrand;
	gsl_integrand.function = evaluator;
	gsl_integrand.params = &integrand;

	const double result = gsl_integration_glfixed(&gsl_integrand, 0., 1., gaussLegendreTableEnergy);

	assert(!std::isnan(result));
	return result;
}

#endif