#ifndef CRONOS_ABSORPTION_H
#define CRONOS_ABSORPTION_H

#include "data.H"
#include "RadiationField.H"
#include "EmissionPostProcess.H"
#include "physical_constants.H"
#include "gsl/gsl_integration.h"
#include "gsl/gsl_errno.h"
#include <functional>

namespace ABSORPTION {
	/**
	 * @brief Absorption Interface Class
	 */
	class IAbsorption {
	public:
		IAbsorption() {};

		virtual ~IAbsorption() {};

		virtual REAL compute_opticalDepth(
				Data &gdata,
				int ix, int iy, int iz,
				const REAL gammaPhotonEnergy,
				const REAL gammaPhotonDirection[3]
		) = 0;
	};
}

/**
 * @brief Absorption Aggregation/Adapter Class
 * @details Delegates the respective method calls to their implementation.
 */
class Absorption: public EmissionPostProcess, public ABSORPTION::IAbsorption {
public:
	Absorption(std::shared_ptr<RadiationField> seedPhotonField);

	~Absorption();

	void set_iAbsorption(
			ABSORPTION::IAbsorption * iAbsorption
	);

//	void compute_constants(
//			Data & gdata,
//			int ix, int iy, int iz
//	) final;

	REAL compute_opticalDepth(
			Data & gdata,
			int ix, int iy, int iz,
			const REAL gammaPhotonEnergy,
			const REAL gammaPhotonDirection[3]
	) final;

	REAL compute(
			Data & gdata,
			int ix, int iy, int iz,
			REAL photonSolidFlux,
			const REAL gammaPhotonEnergy,
			const REAL gammaPhotonDirection[3]
	) final;

	std::string get_name(
	) const final;

	void set_normalisation(
			normalisation & normHD,
			normalisation & normParticle,
			normalisation & normEmission
	) final;


private:
	/** To check if normalisation was called */
	bool ready = false;
	REAL normFactor;
	
	std::shared_ptr<RadiationField> seedPhotonField;
	ABSORPTION::IAbsorption * iAbsorption;
	normalisation * normHD;
	normalisation * normParticle;
};



namespace ABSORPTION {
	/**
	 * @brief implements the gamma-gamma cross section
	 * @details a flag in the implementation can be used to switch to an interpolation
	 * @copydetails from Gould (1967) - 10.1103/PhysRev.155.1404
	 * @return gamma-gamma cross section divided by r0^2/2
	 */
	REAL get_crossSectionFunctional(
			REAL pairBeta
	);

	class CrossSectionInterpolation {
	public:
		CrossSectionInterpolation(int nPoints);
		~CrossSectionInterpolation();
		REAL operator()(REAL pairBeta) const;
	private:
		const int nPoints;
		REAL * values;
		REAL dPairBeta;
	};

	/** Baseclass to express the need for the Electron-Restenergy */
	class NeedsElectronRestEnerergy {
	public:
		/**
		 * @brief Sets the electron rest energy in the respective units
		 * @details Is determined by Absorption::set_normalisation
		 */
		void set_electronRestEnergy(
				REAL electronRestEnergy
		);
	protected:
		REAL electronRestEnergy, electronRestEnergy_sqr, electronRestEnergy_sqr_inv;
	};



	/**
	 * @brief Absorption implementation for factorized radiation fields
	 */
	class AbsorptionFactorizedRadiationField : public IAbsorption, public NeedsElectronRestEnerergy {
	public:
		AbsorptionFactorizedRadiationField(
				std::shared_ptr<RadiationFieldFactorized> radiationFieldFactorized
		);

		~AbsorptionFactorizedRadiationField();

//	void compute_constants(
//			Data & gdata,
//			int ix, int iy, int iz
//	) final;

		REAL compute_opticalDepth(
				Data &gdata,
				int ix, int iy, int iz,
				const REAL gammaPhotonEnergy,
				const REAL gammaPhotonDirection[3]
		) final;

	private:
		gsl_integration_glfixed_table *gaussLegendreTableEnergy;
		gsl_integration_glfixed_table *gaussLegendreTableAzimuth;
		gsl_integration_glfixed_table *gaussLegendreTablePolar;
		gsl_integration_glfixed_table *gaussLegendreTablePsi;

		size_t qagLimitLineOfSightLength;
		gsl_integration_workspace *integrationWorkspaceLineOfSightLength;

		std::shared_ptr<RadiationFieldFactorized> radiationField;
		std::shared_ptr<IRadiationFieldFactor_NumberDensity> iRffNumberDensity;
		std::shared_ptr<IRadiationFieldFactor_Energy> iRffEnergy;
		std::shared_ptr<IRadiationFieldFactor_Direction> iRffDirection;
		std::shared_ptr<IRadiationFieldFactor_Monochromatic> iRffMonochromatic;
		std::shared_ptr<IRadiationFieldFactor_Beamed> iRffBeamed;
		std::shared_ptr<IRFFD_SphericalCoordinates> iRffSphericalCoords;
		std::shared_ptr<SphericalSource> sphericalSource;



		std::function<REAL(Data&, const REAL *, const REAL, const REAL *)> lineOfSightIntegral;

		/**
		 * @details This integrator assumes an arbitrary radiation field (no sph simplifications)
		 * */
		REAL lineOfSightIntegral_Length(
				Data &gdata,
				const REAL pos[3],
				const REAL gammaPhotonEnergy,
				const REAL gammaPhotonDirection[3]
		);

		/**
		* @details This integrator assumes as spherical source
		*/
		REAL lineOfSightIntegral_Psi(
				Data &gdata,
				const REAL pos[3],
				const REAL gammaPhotonEnergy,
				const REAL gammaPhotonDirection[3]
		);

		/** @details LoS Integral in length coordinates assuming
		 *  Psi = pi (radially outgoing gamma ray)
		 *  */
		REAL lineOfSightIntegral_Length_Radial(
				Data &gdata,
				const REAL pos[3],
				const REAL gammaPhotonEnergy,
				const REAL gammaPhotonDirection[3]
		);

		
		// SolidAngle Integration

		std::function<REAL(Data&, const REAL *, const REAL, const REAL *)> solidAngleIntegral_direction;

//		REAL solidAngleIntegral_sphericalCoords(
//				Data &gdata,
//				const REAL pos[3],
//				const REAL gammaPhotonEnergy,
//				const REAL gammaPhotonDirection[3]
//		);

		REAL solidAngleIntegral_delta(
				Data &gdata,
				const REAL pos[3],
				const REAL gammaPhotonEnergy,
				const REAL gammaPhotonDirection[3]
		);

		std::function<REAL(Data&, const REAL *, const REAL, const REAL)> solidAngleIntegral_sphericalSource;

		REAL solidAngleIntegral_spherical(
				Data &gdata,
				const REAL pos[3],
				const REAL gammaPhotonEnergy,
				const REAL psi
		);


		// Energy integration

		std::function<REAL(Data&, const REAL *, const REAL, const REAL)> energyIntegral;

		REAL energyIntegral_delta(
				Data &gdata,
				const REAL pos[3],
				const REAL gammaPhotonEnergy,
				const REAL muGammaGamma
		);

		REAL energyIntegral_full(
				Data &gdata,
				const REAL pos[3],
				const REAL gammaPhotonEnergy,
				const REAL muGammaGamma
		);


	};


}

#endif
