#include "gridgen.H"
#include "DummyHyperbolicSolver.H"
#include <iomanip>

DummyHyperbolicSolver::DummyHyperbolicSolver(Data &gdata, ProblemType &Problem)
{

#if (FLUID_TYPE == CRONOS_MULTIFLUID)
	with_mag = gdata.fluids->with_magField();
	q_rho = gdata.fluids->get_q_rho(0);
	q_sx = gdata.fluids->get_q_sx(0);
	q_sy = gdata.fluids->get_q_sy(0);
	q_sz = gdata.fluids->get_q_sz(0);
	q_Bx = gdata.fluids->get_q_Bx();
	q_By = gdata.fluids->get_q_By();
	q_Bz = gdata.fluids->get_q_Bz();
	q_Eges = gdata.fluids->get_q_Eges(0);
	q_Eadd = gdata.fluids->get_q_Eadd(0);
#else
	with_mag = (FLUID_TYPE == CRONOS_MHD);
	q_rho = gdata.fluid.get_q_rho();
	q_sx = gdata.fluid.get_q_sx();
	q_sy = gdata.fluid.get_q_sy();
	q_sz = gdata.fluid.get_q_sz();
	q_Bx = gdata.fluid.get_q_Bx();
	q_By = gdata.fluid.get_q_By();
	q_Bz = gdata.fluid.get_q_Bz();
	q_Eges = gdata.fluid.get_q_Eges();
	q_Eadd = gdata.fluid.get_q_Eadd();
#endif


#if (FLUID_TYPE == CRONOS_MULTIFLUID)
	numFluids = gdata.fluids->get_numFluids();
	n_om = gdata.fluids->get_N_OM();
	n_omInt = gdata.fluids->get_N_OMINT();
	n_Omega = gdata.fluids->get_N_OMEGA();
	n_omUser = gdata.fluids->get_N_OM_USER();
	n_omIntUser = gdata.fluids->get_N_OMINT_USER();
	n_OmegaUser = gdata.fluids->get_N_OMEGA_USER();
	n_omIntAll = gdata.fluids->get_N_OMINT_ALL();
#else
	n_om = N_OM;
	n_omInt = N_OMINT;
	n_Omega = N_OMEGA;
#if(OMS_USER == TRUE)
	n_omUser = N_OM_USER;
	n_omIntUser = N_OMINT_USER;
	n_OmegaUser = N_OMEGA_USER;
#else
	n_omUser = 0;
	n_omIntUser = 0;
	n_OmegaUser = 0;
#endif
	n_omIntAll = N_OMINT_ALL;
#endif

	init_constants(gdata);

	IntegrateA = true;
	bcVecPotResized = false;

	// Get starting time
	gettimeofday(&tstep, 0);

}


DummyHyperbolicSolver::~DummyHyperbolicSolver() {
}


void DummyHyperbolicSolver::init_constants(Data &gdata)
{
	/*
	  Here all (!) relevant quantities have to be defined.
	*/

	yama = value((char*)"Adiabatic_exponent");

	c2_iso = sqr(value((char*)"Isothermal_Soundspeed")); 
	rho0 = value((char*)"Initial_density");


	denomPres = pow(rho0,1-yama);

	thermal = static_cast<int>(value((char*)"thermal"));

	// Values for phystest:
	rhomin = value((char*)"rhomin");
	Ethmin = value((char*)"pmin");
	vmax = value((char*)"vmax");

}

void DummyHyperbolicSolver::addConstants_toH5(Data &gdata, Hdf5Stream &h5out) {
	//! Add constants to each hdf5 output file

	// Write adiabatic exponent
	h5out.AddGlobalAttr("HyperbolicSolver_Gamma", yama);

	// Write isothermal soundspeed
	h5out.AddGlobalAttr("HyperbolicSolver_Isothermal_Soundspeed", sqrt(c2_iso));

	// Write isothermal soundspeed
	h5out.AddGlobalAttr("HyperbolicSolver_SQR_Soundspeed", c2_iso);

	// Write initial density
	h5out.AddGlobalAttr("HyperbolicSolver_Initial_density", rho0);

	// Write allowed extreme values:
	h5out.AddGlobalAttr("HyperbolicSolver_rhomin",rhomin);
	h5out.AddGlobalAttr("HyperbolicSolver_pmin",Ethmin);
	h5out.AddGlobalAttr("HyperbolicSolver_vmax",vmax);

	// Write chosen Riemann solver:
	// string RiemannSolver = svalue((char*)"RiemannSolver");
	// h5out.AddGlobalAttr("HyperbolicSolver_RiemannSolver",RiemannSolver);

	// Indicate if vector potential is integrated:
	h5out.AddGlobalAttr("HyperbolicSolver_IntegrateA", IntegrateA);
	
}

void DummyHyperbolicSolver::getConstants_fromH5(Data &gdata, Hdf5iStream &h5in) {
	//! Get any necessary constant from hdf5 input file:
	
	// So far we only use this to make sure the correct treatment of
	// the magnetic field is used - can be extended in the future.
	int use_vecpot; // Hack to avoid having to write additional bool
					// input-routine.

	// Need to check if attribute is available:
	if(h5in.doesAttrExist("HyperbolicSolver_IntegrateA")) {
		h5in.ReadGlobalAttr("HyperbolicSolver_IntegrateA", use_vecpot);
		is_OldRestart = false;
	} else {
		is_OldRestart = true;
	}
	IntegrateA = use_vecpot;

}


void DummyHyperbolicSolver::init(Data &gdata, gridFunc &gfunc,
                            ProblemType &Problem)
{

	if(gdata.time < 0.1*gdata.dt) {

		// initial conditions
     
		for (int l = 0; l < n_Omega; l++) {
			gdata.om[l].clear();
		}

		if(gdata.rank==0){
			cout << "calling init (type=" << Problem.get_Type() << ")..." << endl;
			cout << Problem.get_Name() << endl;
			cout << gdata.mx[0] << " " << gdata.mx[1] << " " << gdata.mx[2] << endl;
		}
    
		Problem.init_fields(gdata);

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)

#if (FLUID_TYPE == CRONOS_MULTIFLUID)
		for(int iFluid=0; iFluid<numFluids; ++iFluid) {
		// Obtain indices:
		q_Eadd = gdata.fluids->get_q_Eadd(iFluid);
		q_rho  = gdata.fluids->get_q_rho(iFluid);
#endif

		gdata.om[q_Eadd] = gdata.om[q_Eges];
#if(AUX_ENERGY == ENTROPY)
		for (int k = -B; k <= gdata.mx[2]+B; ++k) {
			for (int j = -B; j <= gdata.mx[1]+B; ++j) {
				for (int i = -B; i <= gdata.mx[0]+B; ++i) {
#if(FLUID_TYPE == CRONOS_RHD)
					gdata.om[q_Eadd](i,j,k) *= (Problem.gamma - 1.) * pow(gdata.om[q_rho](i,j,k), -Problem.gamma);
#else
					gdata.om[q_Eadd](i,j,k) *= pow(gdata.om[q_rho](i,j,k), 1.-Problem.gamma);
#endif
				}
			}
		}
#endif

#if (FLUID_TYPE == CRONOS_MULTIFLUID)
		} // Close loop from above
#endif

#endif


		if(gdata.rank==0) {
			cout << "all ICs set." << endl;
		}

	}

	// if Vector-Potential is given -> Transform to magnetic field
	if (FLUID_TYPE == CRONOS_MHD || with_mag) {
		if(gdata.om[q_Bx].getName() == "A_x" ||
				gdata.om[q_By].getName() == "A_y" ||
				gdata.om[q_Bz].getName() == "A_z") {
			TransPot2Mag(gdata, gfunc, Problem);
			IntegrateA = true;
			if(gdata.rank == 0) {
				cout << " Integrating the vector potential " << endl;
			}
		} else {
			IntegrateA = false;
			if(gdata.rank == 0) {
				cout << " Integrating the magnetic induction " << endl;
			}
		}
	}


	init_general(gdata, gfunc, Problem);
}


void DummyHyperbolicSolver::init_general(Data &gdata, gridFunc &gfunc,
                                    ProblemType &Problem) {

	// General initialisation procedure
	// This routine is executed for normal start AND for a restart
	
#if(FLUID_TYPE==CRONOS_MULTIFLUID)
	int i_magFluid = gdata.fluids->get_i_magFluid();
#else
	int i_magFluid = 0;
#endif

	// Introducing unique ids to all fields om, om_user - can be used
	// for identification (like, e.g., in bcs)
	gdata.set_fieldIds();

	// Check if using thermal energy or temperature as base variable:
#if (ENERGETICS == FULL)
	if(gdata.om[q_Eges].getName() == "Etherm") {
		thermal = true;
	} else {
		thermal = false;
	}
#endif

	// gfunc.boundary(gdata, Problem); 

#if(FLUID_TYPE==CRONOS_MULTIFLUID)
	// Loop over all fluids:
	for(int iFluid=0; iFluid<gdata.fluids->get_numFluids(); ++iFluid) {
		CronosFluid fluid = gdata.fluids->fluids[iFluid];
		q_rho = fluid.get_q_rho_global();
		q_sx = fluid.get_q_sx_global();
		q_sy = fluid.get_q_sy_global();
		q_sz = fluid.get_q_sz_global();
		q_Eges = fluid.get_q_Eges_global();
		q_Eadd= fluid.get_q_Eadd_global();
//		Trafo->reset_Indices(fluid);

	}

	// Check wheter we have a magnetised fluid
	bool with_MagField = gdata.fluids->with_magField();
	if(with_MagField) {
		compute_divB(gdata, gfunc, Problem);
	}
#else

#if(FLUID_TYPE == CRONOS_MHD)
	compute_divB(gdata, gfunc, Problem);
#endif

#endif

#if (USE_COROTATION == CRONOS_ON)
	// Transform back to co-rotating frame
	Trafo->TransInertToCorot(gdata, gfunc, Problem);
#endif

	// Prepare fields to hold changes for magnetic field
#if (FLUID_TYPE == CRONOS_MHD)
	if(IntegrateA) {
		gdata.nom[q_Bx].resize(Index::set(0,0,0),
				Index::set(gdata.mx[0],gdata.mx[1]+1,gdata.mx[2]+1));
		gdata.nom[q_By].resize(Index::set(0,0,0),
				Index::set(gdata.mx[0]+1,gdata.mx[1],gdata.mx[2]+1));
		gdata.nom[q_Bz].resize(Index::set(0,0,0),
				Index::set(gdata.mx[0]+1,gdata.mx[1]+1,gdata.mx[2]));
	} else {
		gdata.nom[q_Bx].resize(Index::set(-1,0,0),
				Index::set(gdata.mx[0],gdata.mx[1],gdata.mx[2]));
		gdata.nom[q_By].resize(Index::set(0,-1,0),
				Index::set(gdata.mx[0],gdata.mx[1],gdata.mx[2]));
		gdata.nom[q_Bz].resize(Index::set(0,0,-1),
				Index::set(gdata.mx[0],gdata.mx[1],gdata.mx[2]));
	}
#endif
	//
	//	Computation of additional aux fields for ICs
	//
	for (int k = -B; k <= gdata.mx[2]+B; ++k) {
		for (int j = -B; j <= gdata.mx[1]+B; ++j) {
			for (int i = -B; i <= gdata.mx[0]+B; ++i) {
				#if (PARTICLE_TRANSPORT == CRONOS_ON)
					gdata.lorentz(i,j,k) = sqrt(1.
							+ sqr(gdata.om[q_sx](i,j,k)) + sqr(gdata.om[q_sy](i,j,k)) + sqr(gdata.om[q_sz](i,j,k)));
				#endif
			}
		}
	}


	//
	//	Particle Initialisation
	//
#if (PARTICLE_TRANSPORT == CRONOS_ON)
	ccout << "---------- Particle Transport ------------------------" << endl;
	Problem.init_particleTransport(gdata);
	Problem.particleTransport.finish_setup(gdata);
	ccout << "------------------------------------------------------" << endl;
#endif

}


void DummyHyperbolicSolver::restart(Data &gdata, gridFunc &gfunc,
                               ProblemType &Problem) {
	// If a vector potential is supplied in the restart process -
	// compute the magnetic induction
//#if (FLUID_TYPE == CRONOS_MHD)
	if(with_mag) {
		// Only needed for old restart files -- otherwise the option should
		// be fixed already.

		if(is_OldRestart) {
			if(N_SUBS > 2 && gdata.om[n_omInt+N_ADD].getName() == "A_x" &&
					gdata.om[n_omInt+N_ADD+1].getName() == "A_y" &&
					gdata.om[n_omInt+N_ADD+2].getName() == "A_z") {
				IntegrateA = true;
				compute_B(gdata, gfunc, Problem);
				if(gdata.rank == 0) {
					cout << " Integrating the vector potential " << endl;
				}
			} else {
				if(gdata.rank == 0) {
					cout << " Integrating the magnetic induction " << endl;
				}
			}
		} else {
			if(IntegrateA) {
				if(gdata.rank == 0) {
					cout << " Integrating the vector potential " << endl;
				}
				// Better recompute magnetic field - to avoid problems in case of grid resize
				compute_B(gdata, gfunc, Problem);
			} else {
				if(gdata.rank == 0) {
					cout << " Integrating the magnetic induction " << endl;
				}
			}
		}

		if(BOUT_DBL < 1) {
			gen_MagBoundValues(gdata, gfunc, Problem);
		}
	}
	init_general(gdata, gfunc, Problem);

}




double DummyHyperbolicSolver::power(double base, int expo)
{
	double ret = 1.;
	for (int mul = 1; mul <= abs(expo); mul++)
		ret *= base;
	return (expo > 0) ? ret : 1./ret;
}




void DummyHyperbolicSolver::CheckNan(NumMatrix<REAL,3> &field, 
                                int q, int rim, int pos,
                                string fieldname)
{
	string message = fieldname + " is NAN ";

	int lo[3], up[3];
	lo[0] = field.getLow(0);
	lo[1] = field.getLow(1);
	lo[2] = field.getLow(2);

	up[0] = field.getHigh(0);
	up[1] = field.getHigh(1);
	up[2] = field.getHigh(2);

	for (int k = lo[2]; k <= up[2]; k++){
		for (int j = lo[1]; j <= up[1]; j++){
			for (int i = lo[0]; i <= up[0]; i++){
				if(std::isnan(field(i,j,k))) {
					throw CException(message,pos,q,i,j,k);
				}
			}
		}
	}
}


void DummyHyperbolicSolver::compute_B(Data &gdata, gridFunc &gfunc,
								 ProblemType &Problem) {

	if(N_SUBS < 3) {
		cerr << " Need extra array for vector potential " << endl;
		exit(-27);
	}

#if(FLUID_TYPE == CRONOS_MULTIFLUID)
	int q_Ax = n_omIntAll+N_ADD;
	int q_Ay = n_omIntAll+N_ADD+1;
	int q_Az = n_omIntAll+N_ADD+2;

//	Pot & Ax = gdata.om[n_omIntAll+N_ADD];
//	Pot & Ay = gdata.om[n_omIntAll+N_ADD+1];
//	Pot & Az = gdata.om[n_omIntAll+N_ADD+2];
#else
	int q_Ax = n_omInt+N_ADD;
	int q_Ay = n_omInt+N_ADD+1;
	int q_Az = n_omInt+N_ADD+2;

//	Pot & Ax = gdata.om[n_omInt+N_ADD];
//	Pot & Ay = gdata.om[n_omInt+N_ADD+1];
//	Pot & Az = gdata.om[n_omInt+N_ADD+2];
#endif
	Pot & Ax = gdata.om[q_Ax];
	Pot & Ay = gdata.om[q_Ay];
	Pot & Az = gdata.om[q_Az];

#if (VEC_POT_BCS == TRUE)
	// If vector-potential boundaries are needed, we need to
	// (slightly) extend the corresponding fields:
	if(!bcVecPotResized) {

// 		Pot ABack;
// 		ABack.resize(gdata.mx);

// 		ABack = Ax;
// 		Ax.resize(Index::set(-B,-B,-B),
// 		          Index::set(gdata.mx[0]+B,gdata.mx[1]+B+1,gdata.mx[2]+B+1));

// 		for (int k = -B; k <= gdata.mx[2]+B; ++k){
// 			for (int j = -B; j <= gdata.mx[1]+B; ++j){
// 				for (int i = -B; i <= gdata.mx[0]+B; ++i){

// 					ABack(i,j,k) = Ax(i,j,k);

// 				}
// 			}
// 		}

		Ax.resize_ll(Index::set(-B,-B,-B),
		             Index::set(gdata.mx[0]+B,gdata.mx[1]+B+1,gdata.mx[2]+B+1));
		Ay.resize_ll(Index::set(-B,-B,-B),
		             Index::set(gdata.mx[0]+B+1,gdata.mx[1]+B,gdata.mx[2]+B+1));
		Az.resize_ll(Index::set(-B,-B,-B),
		             Index::set(gdata.mx[0]+B+1,gdata.mx[1]+B+1,gdata.mx[2]+B));
		bcVecPotResized = true;
	}

	// Need to compute boundary-conditions for vector-potential first:

#endif

#ifdef parallel
	gfunc.BoundaryHandler.boundary(gdata, Problem, &gdata.om[q_Ax], 3);
#endif


	Pot & Bx = gdata.om[q_Bx];
	Pot & By = gdata.om[q_By];
	Pot & Bz = gdata.om[q_Bz];


#if (VEC_POT_BCS == TRUE)
	int imin[3] = {-B,-B,-B};
	int imax[3] = {gdata.mx[0]+B, gdata.mx[1]+B, gdata.mx[2]+B};
#else
	int imin[3] = {-1, 0, 0};
	int imax[3] = {gdata.mx[0], gdata.mx[1], gdata.mx[2]};
#endif

#if (NON_LINEAR_GRID == CRONOS_OFF)
	REAL idx[3]= {(1./gdata.dx[0]), (1./gdata.dx[1]), (1./gdata.dx[2])};
#endif

	for (int k = imin[2]; k <= imax[2]; ++k){
		for (int j = imin[1]; j <= imax[1]; ++j){
			for (int i = imin[0]; i <= imax[0]; ++i){

#if (NON_LINEAR_GRID == CRONOS_ON)
				REAL idx[3] = {(gdata.getCen_idx(0,i)), (gdata.getCen_idx(1,j)),
				               (gdata.getCen_idx(2,k))};
#endif

#if (GEOM > 1)

				REAL f_geom = 1./(gdata.h1(i,j,k,1,0,0)*gdata.h2(i,j,k,1,0,0));
				Bx(i,j,k) = (+(gdata.h2(i,j,k,1, 1,0)*Az(i+1,j+1,k  ) -
				               gdata.h2(i,j,k,1,-1,0)*Az(i+1,j  ,k  ))*idx[1]
				             -(gdata.h1(i,j,k,1,0, 1)*Ay(i+1,j  ,k+1) -
				               gdata.h1(i,j,k,1,0,-1)*Ay(i+1,j  ,k  ))*idx[2])*f_geom;

#else

				Bx(i,j,k) = (+(Az(i+1,j+1,k  ) - Az(i+1,j  ,k  ))*idx[1]
							 -(Ay(i+1,j  ,k+1) - Ay(i+1,j  ,k  ))*idx[2]);

#endif

			}
		}
	}


#if (VEC_POT_BCS == FALSE)
	imin[0] = 0;
	imin[1] = -1;
#endif

	for (int k = imin[2]; k <= imax[2]; ++k){
//		cout << " Machnetisch " << k << endl;
		for (int j = imin[1]; j <= imax[1]; ++j){
			for (int i = imin[0]; i <= imax[0]; ++i){

#if (NON_LINEAR_GRID == CRONOS_ON)
				REAL idx[3] = {(gdata.getCen_idx(0,i)), (gdata.getCen_idx(1,j)),
				               (gdata.getCen_idx(2,k))};
#endif

#if (GEOM > 1)

				REAL f_geom = 1./(gdata.h0(i,j,k,0,1,0)*gdata.h2(i,j,k,0,1,0));
				By(i,j,k) = (+(gdata.h0(i,j,k,0,1, 1)*Ax(i  ,j+1,k+1) -
				               gdata.h0(i,j,k,0,1,-1)*Ax(i  ,j+1,k  ))*idx[2]
				             -(gdata.h2(i,j,k, 1,1,0)*Az(i+1,j+1,k  ) -
				               gdata.h2(i,j,k,-1,1,0)*Az(i  ,j+1,k  ))*idx[0])*f_geom;

#else

				By(i,j,k) = (+(Ax(i  ,j+1,k+1) - Ax(i  ,j+1,k  ))*idx[2]
							 -(Az(i+1,j+1,k  ) - Az(i  ,j+1,k  ))*idx[0]);
//				 if(i==400 && j==0 && k==0) {
//				   cout << " Making Bees " << " ";
//				   cout << Ax(i,j+1,k) << " " << Ax(i,j+1,k+1) << " ";
//				   cout << Az(i+1,j+1,k) << " " << Az(i,j+1,k) << " ";
//				   cout << endl;
//				 }


#endif
			}
		}
	}


#if (VEC_POT_BCS == FALSE)
	imin[1] = 0;
	imin[2] = -1;
#endif

	for (int k = imin[2]; k <= imax[2]; ++k){
		for (int j = imin[1]; j <= imax[1]; ++j){
			for (int i = imin[0]; i <= imax[0]; ++i){

#if (NON_LINEAR_GRID == CRONOS_ON)
				REAL idx[3] = {(gdata.getCen_idx(0,i)), (gdata.getCen_idx(1,j)),
				               (gdata.getCen_idx(2,k))};
#endif

#if (GEOM > 1)

				REAL f_geom = 1./(gdata.h0(i,j,k,0,0,1)*gdata.h1(i,j,k,0,0,1));
				Bz(i,j,k) = (+(gdata.h1(i,j,k, 1,0,1)*Ay(i+1,j  ,k+1) -
				               gdata.h1(i,j,k,-1,0,1)*Ay(i  ,j  ,k+1))*idx[0]
				             -(gdata.h0(i,j,k,0, 1,1)*Ax(i  ,j+1,k+1) -
				               gdata.h0(i,j,k,0,-1,1)*Ax(i  ,j  ,k+1))*idx[1])*f_geom;
				// REAL f_geom = 1./(gdata.h0(i,j,k+0.5)*gdata.h1(i,j,k+0.5));
				// Bz(i,j,k) = (+(gdata.h1(i+0.5,j,k+0.5)*Ay(i+1,j  ,k+1) -
				//                gdata.h1(i-0.5,j,k+0.5)*Ay(i  ,j  ,k+1))*gdata.idx[0]
				//              -(gdata.h0(i,j+0.5,k+0.5)*Ax(i  ,j+1,k+1) -
				//                gdata.h0(i,j-0.5,k+0.5)*Ax(i  ,j  ,k+1))*gdata.idx[1])*f_geom;

#else

				Bz(i,j,k) = (+(Ay(i+1,j  ,k+1) - Ay(i  ,j  ,k+1))*idx[0]
							 -(Ax(i  ,j+1,k+1) - Ax(i  ,j  ,k+1))*idx[1]);
//									if(i==0 && j>390 && j<410 && k==0) {
//										cout << " Mag " << gdata.om[q_Bz](i,j,k) << " " << q_Bz << endl;
//									}

#endif

			}
		}
	}

//	cout << " Bef BC values at (-1,0,0) ";
//	cout << gdata.om[4](-1,0,0) << " ";
//	cout << gdata.om[5](-1,0,0) << " ";
//	cout << gdata.om[6](-1,0,0) << " ";
//	cout << endl;
//	cout << " By at upper border " << " ";
//	cout << gdata.om[5](gdata.mx[0]-1,0,0) << " ";
//	cout << gdata.om[5](gdata.mx[0],0,0) << " ";
//	cout << gdata.mx[0] << " ";
//	cout << endl;
//	cout << " A at upper border " << " ";
//	cout << Ax(gdata.mx[0],1,1) << " ";
//	cout << Ax(gdata.mx[0],1,0) << " ";
//	cout << Az(gdata.mx[0]+1,1,0) << " ";
//	cout << Az(gdata.mx[0],1,0) << " ";
//	cout << Az(gdata.mx[0]-1,1,0) << " ";
//	cout << endl;


	// gfunc.BoundaryHandler.boundary(gdata, Problem, &gdata.om[q_Bx], 3);

//	cout << " Aft BC values at (-1,0,0) ";
//	cout << gdata.om[4](-1,0,0) << " ";
//	cout << gdata.om[5](-1,0,0) << " ";
//	cout << gdata.om[6](-1,0,0) << " ";
//	cout << endl;


}





void DummyHyperbolicSolver::TransPot2Mag(Data &gdata, gridFunc &gfunc,
									ProblemType &Problem) {

	if(gdata.om[q_Bx].getName() == "B_x" ||
	   gdata.om[q_By].getName() == "B_y" ||
	   gdata.om[q_Bz].getName() == "B_z") {
		cerr << " Magnetic field already set " << endl;
		exit(2);
	}

	if(N_SUBS < 3) {
		cerr << " Need extra array for vector potential " << endl;
		exit(-27);
	}

#if(FLUID_TYPE == CRONOS_MULTIFLUID)
	gdata.om[n_omIntAll+N_ADD  ] = gdata.om[q_Bx];
	gdata.om[n_omIntAll+N_ADD+1] = gdata.om[q_By];
	gdata.om[n_omIntAll+N_ADD+2] = gdata.om[q_Bz];
#else
	gdata.om[n_omInt+N_ADD  ] = gdata.om[q_Bx];
	gdata.om[n_omInt+N_ADD+1] = gdata.om[q_By];
	gdata.om[n_omInt+N_ADD+2] = gdata.om[q_Bz];
#endif

	compute_B(gdata, gfunc, Problem);

	gdata.om[q_Bx].rename("B_x");
	gdata.om[q_By].rename("B_y");
	gdata.om[q_Bz].rename("B_z");


}
