#include <algorithm>
#include "EmissionDirection.H"

DirectionList::DirectionList() {}

void DirectionList::push_back(const Direction& _dir) {
  vector3D dir(_dir);
  auto norm = _dir.norm();
  if (norm == 1. || norm == 0.) {
    return directions.push_back(std::move(dir));
  } else {
    dir /= norm;
    return directions.push_back(std::move(dir));
  }
}

NumMatrix<double, 2> DirectionList::convert() const {
  NumMatrix<double, 2> clone(Index::set(0, 0), Index::set(directions.size() - 1, 2));
  clone.clear();
  int n = 0;
  for (const auto& dir : directions) {
    clone(n, 0) = dir[0];
    clone(n, 1) = dir[1];
    clone(n, 2) = dir[2];
    ++n;
  }
  return clone;
}

// Disable parallel write for now
#define HDF_PARALLEL_IO_LOCAL CRONOS_OFF

void EmissionDirections::_writeToH5(Hdf5Stream& h5out, hid_t group) const {
#if !defined(parallel) || (HDF_PARALLEL_IO_LOCAL == CRONOS_OFF)
  int rank = 0;
#ifdef parallel
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#endif

  if (rank == 0) {
    h5out.AddGlobalAttr("nDirections", directions.size(), group);
    int mx[2] = {(int)size(), 3};
    h5out.WriteNDArray("directions", directions.data()->data(), mx, 2, group);

    h5out.AddGlobalAttr("type", type(), group);
  }

#else
#error "NOT IMPLEMENTED"
#endif
}

SphericalDirections::SphericalDirections(const Direction& _axis, const NumArray<double>& _theta,
                                         const NumArray<double>& _phi)
    : axis(_axis.normalised()), theta(_theta), phi(_phi) {
  directions.reserve(theta.getLength() * phi.getLength());

  const Direction ex = {1, 0, 0};

  if (axis.inner(Direction{0, 0, 1}) == 1.) {
    // if the axis is along the z coordinate we will use the std spherical coordinates
    // (i.e. phi starts from e_x)
    e1 = ex;
  } else {
    // otherwise use the rotation vector as origin for phi
    // (i.e. the vector around which one needs to rotate e_z to end up with axis)
    e1 = -1 * axis.cross(vector3D{0, 0, 1}).normalised();
  }
  e2 = axis.cross(e1);

  for (const auto& _theta : theta) {
    for (const auto& _phi : phi) {
      const Direction dir = computeDirection(axis, e1, e2, _theta, _phi);
      push_back(dir);
    }
  }
}

Direction SphericalDirections::computeDirection(const Direction& axis, const Direction& e1,
                                                const Direction& e2, const double& theta,
                                                const double& phi) const {
  return cos(theta) * axis + sin(theta) * (cos(phi) * e1 + sin(phi) * e2);
}

std::string SphericalDirections::type() const { return "spherical"; }

void SphericalDirections::writeToH5(Hdf5Stream& h5out, hid_t group) const {
  _writeToH5(h5out, group);

  // Additional info

  h5out.AddGlobalAttr("nTheta", theta.getLength(), group);
  h5out.WriteNumArray("theta", theta, group);
  h5out.AddGlobalAttr("nPhi", phi.getLength(), group);
  h5out.WriteNumArray("phi", phi, group);

  std::vector<std::array<double, 2> > thetaphi;
  thetaphi.reserve(theta.getLength() * phi.getLength());
  for (const auto& _theta : theta) {
    for (const auto& _phi : phi) {
      thetaphi.push_back({_theta, _phi});
    }
  }
  int mx2[2] = {(int)size(), 2};
  h5out.WriteNDArray("dir2thetaphi", thetaphi.data()->data(), mx2, 2, group);
  h5out.AddAttributeToArray(
      "dir2thetaphi", "comment",
      std::string("Conversion from iDirection (row) to theta and phi (first and second column)"),
      group);

  // Write reversed order due to omit fortran sorting of NumMatrices
  int dim[2] = {phi.getLength(), theta.getLength()};
  NumMatrix<int, 2> thetaphi_index(dim);

  for (int iTheta = 0; iTheta < theta.getLength(); ++iTheta) {
    for (int iPhi = 0; iPhi < phi.getLength(); ++iPhi) {
      thetaphi_index(iPhi, iTheta) = iPhi + iTheta * phi.getLength();
    }
  }

  h5out.WriteNDMatrix("thetaphi2dir", thetaphi_index, group);
  h5out.AddAttributeToArray(
      "thetaphi2dir", "comment",
      std::string("Conversion from iTheta (row) and iPhi (column) to iDirection"), group);
}

RotatingOrbitObserver::RotatingOrbitObserver(const Direction& axis, const NumArray<double>& theta,
                                             const NumArray<double>& phiOffset,
                                             const double infcTrueAnomaly,
                                             const double orbitalPhase, const bool clockwiseOrbit)
    : SphericalDirections(axis, theta,
                          phiOffset + computePhiObs(infcTrueAnomaly, orbitalPhase, clockwiseOrbit))

{}

double RotatingOrbitObserver::computePhiObs(const double infcTrueAnomaly,
                                                   const double orbitalPhase,
                                                   const bool clockwiseOrbit) {
  // In units of pi
  double observerAzimuth = infcTrueAnomaly;
  
  if (clockwiseOrbit) {
    observerAzimuth += 2 * orbitalPhase;
  } else {
    observerAzimuth -= 2 * orbitalPhase;
  }

  // Since the azimuth axis is rotated wrt the periastron axis
  observerAzimuth += 0.5;

  return observerAzimuth * M_PI;
}