#ifndef EMISSION_DIRECTION_H
#define EMISSION_DIRECTION_H

#include <memory>
#include "Hdf5File_cbase.H"
#include "marray.H"
#include "matrix.H"

using Direction = vector3D;

/**
 * @brief List container for directions
 *
 * Implements basic access and iterators
 */
class DirectionList {
 protected:
  std::vector<Direction> directions;
  using iterator = decltype(std::declval<const decltype(directions)>().begin());

 public:
  DirectionList();

  template <class... Args>
  inline Direction& operator[](Args&&... idx) {
    return directions.operator[](std::forward<Args>(idx)...);
  }

  template <class... Args>
  inline const Direction& operator[](Args&&... idx) const {
    return directions.operator[](std::forward<Args>(idx)...);
  }

  void push_back(const Direction&);
  size_t size() const { return directions.size(); }

  iterator begin() const { return directions.begin(); }
  iterator end() const { return directions.end(); }

  NumMatrix<double, 2> convert() const;
};

/**
 * @brief Direction container in which the emission will be computed
 */
class EmissionDirections : public DirectionList {
 public:
  /**
   * @brief Writes the directions to a given hdf5 group
   * Writes also additional data for the postprocessing of the respective subclass
   */
  virtual void writeToH5(Hdf5Stream& h5out, hid_t group) const = 0;

  /**
   * @brief Returns the generator type
   * This is stored as hdf5 attribute for postprocessing identification
   */
  virtual std::string type() const = 0;

 protected:
  /**
   * @brief Basic hdf5 writing strategy for a list of directions
   * Intended for subclass usage
   */
  void _writeToH5(Hdf5Stream& h5out, hid_t group) const;
};

/**
 * @brief Specialization that constructs the emission directions for a grid in spherical coordinates
 */
class SphericalDirections : public EmissionDirections {
  const Direction axis;
  Direction e1, e2;
  const NumArray<double> theta, phi;

  Direction computeDirection(const Direction& axis, const Direction& e1, const Direction& e2,
                             const double& theta, const double& phi) const;

 public:
  /**
   * @brief Construct a new Spherical Directions object
   *
   * @param axis direction that defines the theta angle
   * @param theta array of azimuthal angles
   * @param phi array of polar angles
   */
  SphericalDirections(const Direction& axis, const NumArray<double>& theta,
                      const NumArray<double>& phi);

  virtual void writeToH5(Hdf5Stream&, hid_t group) const override;
  virtual std::string type() const override;
};

/**
 * @brief Specialization that constructs the spherical emission directions centered around the
 * observer location
 */
class RotatingOrbitObserver : public SphericalDirections {
 public:
  /**
   * @brief Standard Constructor
   *
   * @param axis direction that defines the theta angle
   * @param theta array of azimuthal angles
   * @param phiOffset array of polar angle offsets around the observer position
   * @param infcTrueAnomaly the true anomaly at which the observer is located
   * @param orbitalPhase current orbital phase
   * @param clockwiseOrbit flag that sets the orbit to be clockwise
   */
  RotatingOrbitObserver(const Direction& axis, const NumArray<double>& theta,
                        const NumArray<double>& phiOffset, const double infcTrueAnomaly,
                        const double orbitalPhase, const bool clockwiseOrbit = false);

 private:
  static double computePhiObs(const double infcTrueAnomaly, const double orbitalPhase,
                              const bool clockwiseOrbit = false);
};

#endif