#include "EmissionHandler.H"

#include <sys/stat.h>
#include <utime.h>
#include <cassert>
#include <cmath>

#include "CException.H"
#include "CronosTimer.H"
#include "Progress.H"
#include "benchmark.H"

using namespace std;
using namespace CRONOS_CONSTANTS;

void EmissionNormalisation::setup() {
  set_definition(LEN, Meter);
  set_definition(TIME, Second);
  set_definition(ENERGY, ElectronVolt);
  set_definition(TEMP);
  set_definition(MAG_IND, Tesla);
  set_definition(NUM_DENS);

  finish_setup();
}

EmissionHandler::EmissionHandler(ParameterFileReader &_parameters) : parameters(_parameters) {
  normEmission.setup();

  nEmissions = parameters.value("nEmissions");
  solvers.reserve(nEmissions);
}

void EmissionHandler::work(Data &gdata, ProblemType &problem,
                           ParticleTransport &particleTransport) {
  setupPostprocesses(particleTransport);

  ccout << " Computing auxiliary fields" << endl;
  preprocess(gdata, particleTransport);
  computeAuxFields(gdata, particleTransport);

  for (size_t iEmission = 0; iEmission < nEmissions; ++iEmission) {
    ccout << "======================================================" << endl;

    EmissionSolver solver(parameters, gdata, problem, particleTransport, normEmission, iEmission);
    solver.work(gdata, particleTransport);
    solvers.push_back(std::move(solver));
  }

  ccout << " all done!" << endl;
  ccout << "======================================================" << endl;

  ccout << " Waiting for other ranks to finish " << endl;
  for (auto &solver : solvers) {
    solver.mergeData(gdata);
  }
}

void EmissionHandler::setupPostprocesses(ParticleTransport &particleTransport) {
  auto &normHD = particleTransport.get_normHD();
  auto &normParticle = particleTransport.get_normParticle();
  auto &emissionPostProcesses = particleTransport.get_emissionPostProcesses();
  for (auto postProc : emissionPostProcesses) {
    postProc->set_normalisation(normHD, normParticle, normEmission);
  }
}

void EmissionHandler::computeAuxFields(Data &gdata, ParticleTransport &particleTransport) const {
  // const int q_rho = gdata.fluid.get_q_rho();
  const int q_sx = gdata.fluid.get_q_sx();
  const int q_sy = gdata.fluid.get_q_sy();
  const int q_sz = gdata.fluid.get_q_sz();
  // const int q_Bx = gdata.fluid.get_q_Bx();
  // const int q_By = gdata.fluid.get_q_By();
  // const int q_Bz = gdata.fluid.get_q_Bz();
  // const int q_Eges = gdata.fluid.get_q_Eges();
  // const int q_Eadd = gdata.fluid.get_q_Eadd();

  for (int iz = 0; iz <= gdata.mx[2]; ++iz) {
    for (int iy = 0; iy <= gdata.mx[1]; ++iy) {
      for (int ix = 0; ix <= gdata.mx[0]; ++ix) {
        gdata.lorentz(ix, iy, iz) =
            sqrt(1. + sqr(gdata.om[q_sx](ix, iy, iz)) + sqr(gdata.om[q_sy](ix, iy, iz)) +
                 sqr(gdata.om[q_sz](ix, iy, iz)));

      }  // ix
    }    // iy
  }      // iz
}

void EmissionHandler::preprocess(Data &gdata, ParticleTransport &particleTransport) const {
  // const int q_rho = gdata.fluid.get_q_rho();
  const int q_sx = gdata.fluid.get_q_sx();
  const int q_sy = gdata.fluid.get_q_sy();
  const int q_sz = gdata.fluid.get_q_sz();
  // const int q_Bx = gdata.fluid.get_q_Bx();
  // const int q_By = gdata.fluid.get_q_By();
  // const int q_Bz = gdata.fluid.get_q_Bz();
  // const int q_Eges = gdata.fluid.get_q_Eges();
  // const int q_Eadd = gdata.fluid.get_q_Eadd();

  if (value_exists("doppler_off")) {
    for (int iz = 0; iz <= gdata.mx[2]; ++iz) {
      for (int iy = 0; iy <= gdata.mx[1]; ++iy) {
        for (int ix = 0; ix <= gdata.mx[0]; ++ix) {
          gdata.om[q_sx](ix, iy, iz) = 0;
          gdata.om[q_sy](ix, iy, iz) = 0;
          gdata.om[q_sz](ix, iy, iz) = 0;
        }  // ix
      }    // iy
    }      // iz
  }

  // // Throw out some particles for testing
  // if (parameters.value_exists("sky" + ToString(iSky, 2) + "_gamma_min")) {
  //   ccout << "WARNING: We are throwing out particles!! Make sure you know what you are doing!"
  //         << endl;
  //   REAL gamma_min = parameters.value("sky" + ToString(iSky, 2) + "_gamma_min");
  //   bool gamma_min_renorm =
  //       parameters.value_exists("sky" + ToString(iSky, 2) + "_gamma_min_renorm");
  //   for (int iz = 0; iz <= gdata.mx[2]; ++iz) {
  //     for (int iy = 0; iy <= gdata.mx[1]; ++iy) {
  //       for (int ix = 0; ix <= gdata.mx[0]; ++ix) {
  //         REAL densOld = particle->energyDensity(ix, iy, iz);
  //         if (densOld == 0) continue;

  //         for (int ip = particle->getLo(); ip < particle->getHi(); ++ip) {
  //           if (particle->get_momentumCenter(ip) < gamma_min) {
  //             particle(ip, ix, iy, iz) = 0;
  //           }
  //         }

  //         if (gamma_min_renorm) {
  //           REAL densNew = particle->energyDensity(ix, iy, iz);
  //           for (int ip = particle->getLo(); ip < particle->getHi(); ++ip) {
  //             particle(ip, ix, iy, iz) *= densOld / densNew;
  //           }
  //         }
  //       }
  //     }
  //   }
  // }

  // Flooring negative Particles
  for (auto particle_ptr : particleTransport.get_particles()) {
    auto &particle = *particle_ptr;

    int negCounter = 0;
    for (int ip = particle.getLo(); ip <= particle.getHi(); ++ip) {
      for (int iz = 0; iz <= gdata.mx[2]; ++iz) {
        for (int iy = 0; iy <= gdata.mx[1]; ++iy) {
          for (int ix = 0; ix <= gdata.mx[0]; ++ix) {
            if (particle(ip, ix, iy, iz) < 0) {
              particle(ip, ix, iy, iz) = 0;
              ++negCounter;
            }
          }
        }
      }
    }

#ifdef parallel
    gdata.MpiSum(negCounter);
#endif
    ccout << " Flooring " << negCounter << " negative particle cells ["
          << particle.get_particleName() << "]" << endl;
  }
}

// Disable parallel write for now
#define HDF_PARALLEL_IO_LOCAL CRONOS_OFF

void EmissionHandler::dataout(const Data &gdata) const {
  CronosTimer hdfTimer;
  hdfTimer.start();

  // Get Filename
  string dirname, filename;
  dirname = string(getenv("poub")) + "/" + string(getenv("pname")) + "_emission";

  mkdir(dirname.c_str(), 511);
  utime(dirname.c_str(), NULL);

  filename = dirname;
  filename += "/";
  filename += string(getenv("pname")) + "_emission_step" + ToString(gdata.tstep) + ".h5";

  ccout << " Writing data to " << filename << endl;

  int numout = 0;

  // Todo before a data output can be performed in the MPI case we have to
  // 		- sum over all ranks -> done (more or less)
  // 		- define regions for the parallel write

#if !defined(parallel) || (HDF_PARALLEL_IO_LOCAL == CRONOS_OFF)

  if (gdata.rank == 0) {
    Hdf5Stream h5out(filename, numout, gdata.rank);

#if (FLUID_TYPE == CRONOS_HYDRO)
    h5out.AddGlobalAttr("fluid_type", "hydro");
#elif (FLUID_TYPE == CRONOS_MHD)
    h5out.AddGlobalAttr("fluid_type", "mhd");
#elif (FLUID_TYPE == CRONOS_RHD)
    h5out.AddGlobalAttr("fluid_type", "rhd");
#else
    h5out.AddGlobalAttr("fluid_type", "N/A");
#endif

    h5out.AddGlobalAttr("geometry", GEOM);
    h5out.AddGlobalAttr("timestep", gdata.tstep);
    h5out.AddGlobalAttr("time", gdata.time);

    // In the parallel case only rank 0 should write

    double dummy[3];
    hid_t info = h5out.AddGroup("info");
    h5out.AddGlobalAttr("nEmissions", nEmissions, info);

    for (const auto &solver : solvers) {
      ccout << "  Writing " << solver.getEmissionName() << endl;

      hid_t emissionGroup = h5out.AddGroup(solver.getEmissionName());

      h5out.AddGlobalAttr("particle_type", solver.getParticle().get_particleType(), emissionGroup);
      h5out.AddGlobalAttr("particle_name", solver.getParticle().get_particleName(), emissionGroup);

      h5out.AddGlobalAttr("nEnergies", solver.getEnergies().getLength(), emissionGroup);
      h5out.WriteNumArray("energies", solver.getEnergies(), emissionGroup);

      hid_t directionGroup = h5out.AddGroup(solver.getEmissionName() + "/directions");
      solver.getDirections().writeToH5(h5out, directionGroup);
      h5out.CloseGroup(directionGroup);

      h5out.AddGlobalAttr("nChannels", solver.getNChannels(), emissionGroup);
      int iChannel = 0;
      for (const auto &channel : solver.getChannelNames()) {
        h5out.AddGlobalAttr("channel" + ToString(iChannel, 2), channel, emissionGroup);
        ++iChannel;
      }

      int iReducer = 0;
      for (const auto &reducer : solver.getEmissivityReducers()) {
        const string reducerName = reducer->type();
        hid_t reductionGroup =
            h5out.AddGroup(solver.getEmissionName() + "/reduction" + ToString(iReducer, 2));
        reducer->writeToH5(h5out, reductionGroup);
        h5out.CloseGroup(reductionGroup);
        ++iReducer;
      }
      h5out.AddGlobalAttr("nReductions", iReducer, emissionGroup);

      h5out.CloseGroup(emissionGroup);
    }

    h5out.CloseGroup(info);
  }
#else
#error "NOT IMPLEMENTED"
#endif

  hdfTimer.stop();
  ccout << "  Writing done after: " << hdfTimer.time() << " s" << endl;
}