#include "EmissionSolver.H"

#include <sys/stat.h>
#include <utime.h>
#include <cassert>
#include <cmath>
#include <limits>

#include "CException.H"
#include "CronosTimer.H"
#include "Progress.H"
#include "benchmark.H"
#include "range.H"

// #define DLEVEL 0 // Write nothing
#define DLEVEL 1  // Write Direction
// #define DLEVEL 2 // Write Domain Coordinate
// #define DLEVEL 3 // Write Energy

// EmissionSolver::EmissionSolver(std::vector<RadiativeLossProcess*>&,
// std::vector<EmissionPostProcess*>&, const EmissionDirections&,
//     const ComputationFlags& computationFlags)
//     : parameters(_parameters), normEmission(_normEmission) {
//   emissionName = "emission" + ToString(emissionID, 2);
//   const auto& normHD = particleTransport.get_normHD();
//   const auto& normParticle = particleTransport.get_normParticle();

//   // particle
//   const int particleID = parameters.value(emissionName + ".particleID");
//   particle = &particleTransport.get_particle(particleID);
//   set_particle(particleTransport, *particle);

//   nRadiativeLossProcesses = particleTransport.get_radiativeLossProcesses().size();
//   nEmissionPostProcesses = particleTransport.get_emissionPostProcesses().size();
//   nChannels = nRadiativeLossProcesses * (nEmissionPostProcesses + 1);

//   // Prepare energies
//   energies = NumArray<double>::set_log(parameters.value(emissionName + ".energies.b"),
//                                        parameters.value(emissionName + ".energies.e"),
//                                        parameters.value(emissionName + ".energies.n"));

//   // Prepare directions
//   const string dirGeneratorKey = emissionName + ".directions";
//   const string dirGeneratorName = svalue(dirGeneratorKey);
//   if (dirGeneratorName == "spherical") {
//     const Direction axis = {parameters.value(dirGeneratorKey + ".axis.x"),
//                             parameters.value(dirGeneratorKey + ".axis.y"),
//                             parameters.value(dirGeneratorKey + ".axis.z")};
//     const NumArray<double> theta =
//         NumArray<double>::set_linear(parameters.value(dirGeneratorKey + ".theta.b") * M_PI,
//                                      parameters.value(dirGeneratorKey + ".theta.e") * M_PI,
//                                      parameters.value(dirGeneratorKey + ".theta.n"));

//     const NumArray<double> phi =
//         NumArray<double>::set_linear(parameters.value(dirGeneratorKey + ".phi.b") * M_PI,
//                                      parameters.value(dirGeneratorKey + ".phi.e") * M_PI,
//                                      parameters.value(dirGeneratorKey + ".phi.n"));

//     directions_ptr = std::unique_ptr<EmissionDirections>(new SphericalDirections(axis, theta,
//     phi));
//   } else {
//     throw CException("No such direction generator: " + dirGeneratorName);
//   }

//   // Prepare filter
//   computationFlags = get_computationFlags(gdata, normHD, emissionName);

//   // Prepare channel names
//   channelNames.reserve(nChannels);
//   for (const auto &radProc : particleTransport.get_radiativeLossProcesses()) {
//     channelNames.push_back(radProc->get_type());
//   }

//   for (const auto &postProc : particleTransport.get_emissionPostProcesses())
//     for (const auto &radProc : particleTransport.get_radiativeLossProcesses()) {
//       channelNames.push_back(radProc->get_type() + "_" + postProc->get_name());
//     }

//   // Prepare reduction
//   const size_t nReductions = (size_t)value(emissionName + ".reductions.n");
//   emissivityReducers.reserve(nReductions);

//   for (const auto &iReducer : iter::range(0, nReductions)) {
//     const string reducerKey = emissionName + ".reduction" + ToString(iReducer, 2);
//     const string reducerName = svalue(reducerKey);

//     if (reducerName == "spectrum") {
//       emissivityReducers.emplace_back(new SpectrumReduction(gdata, energies.getLength(),
//                                                             directions_ptr->size(), channelNames,
//                                                             normHD, normParticle, normEmission));
//     } else {
//       throw CException("No such emissivity reducer");
//     }
//   }
// }

EmissionSolver::EmissionSolver(ParameterFileReader &_parameters, Data &gdata, ProblemType &problem,
                               ParticleTransport &particleTransport,
                               const EmissionNormalisation &_normEmission, const int emissionID)
    : parameters(_parameters), normEmission(_normEmission) {
  emissionName = "emission" + ToString(emissionID, 2);
  const auto &normHD = particleTransport.get_normHD();
  const auto &normParticle = particleTransport.get_normParticle();

  // particle
  const int particleID = parameters.value(emissionName + ".particleID");
  particle = &particleTransport.get_particle(particleID);
  set_particle(gdata, particleTransport, *particle);

  nRadiativeLossProcesses = particleTransport.get_radiativeLossProcesses().size();
  nEmissionPostProcesses = particleTransport.get_emissionPostProcesses().size();
  nChannels = nRadiativeLossProcesses * (nEmissionPostProcesses + 1);

  // Prepare energies
  energies = NumArray<double>::set_log(parameters.value(emissionName + ".energies.b"),
                                       parameters.value(emissionName + ".energies.e"),
                                       parameters.value(emissionName + ".energies.n"));

  // Prepare directions
  const string dirGeneratorKey = emissionName + ".directions";
  const string dirGeneratorName = svalue(dirGeneratorKey);
  if (dirGeneratorName == "spherical") {
    const Direction axis = {parameters.value(dirGeneratorKey + ".axis.x"),
                            parameters.value(dirGeneratorKey + ".axis.y"),
                            parameters.value(dirGeneratorKey + ".axis.z")};
    const NumArray<double> theta =
        NumArray<double>::set_linear(parameters.value(dirGeneratorKey + ".theta.b") * M_PI,
                                     parameters.value(dirGeneratorKey + ".theta.e") * M_PI,
                                     parameters.value(dirGeneratorKey + ".theta.n"));

    const NumArray<double> phi =
        NumArray<double>::set_linear(parameters.value(dirGeneratorKey + ".phi.b") * M_PI,
                                     parameters.value(dirGeneratorKey + ".phi.e") * M_PI,
                                     parameters.value(dirGeneratorKey + ".phi.n"));

    directions_ptr = std::unique_ptr<EmissionDirections>(new SphericalDirections(axis, theta, phi));
  } else if (dirGeneratorName == "rotatingObserver") {
    const Direction axis = {parameters.value(dirGeneratorKey + ".axis.x"),
                            parameters.value(dirGeneratorKey + ".axis.y"),
                            parameters.value(dirGeneratorKey + ".axis.z")};
    const NumArray<double> theta =
        NumArray<double>::set_linear(parameters.value(dirGeneratorKey + ".theta.b") * M_PI,
                                     parameters.value(dirGeneratorKey + ".theta.e") * M_PI,
                                     parameters.value(dirGeneratorKey + ".theta.n"));

    const NumArray<double> phi =
        NumArray<double>::set_linear(parameters.value(dirGeneratorKey + ".phi.b") * M_PI,
                                     parameters.value(dirGeneratorKey + ".phi.e") * M_PI,
                                     parameters.value(dirGeneratorKey + ".phi.n"));

    const double infcTrueAnomaly = parameters.value(dirGeneratorKey + ".infcTrueAnomaly");

    // This should be refactored when the Problem class has an orbit object!
    const double orbitalPhase = problem.getOrbitalPhase();
    ccout << "rotatingObserver: Orbital Phase = " << orbitalPhase << endl;
    ccout << "rotatingObserver: Phis = " << phi << endl;
    const double clockwiseOrbit = false;

    directions_ptr = std::unique_ptr<EmissionDirections>(
        new RotatingOrbitObserver(axis, theta, phi, infcTrueAnomaly, orbitalPhase, clockwiseOrbit));
  } else {
    throw CException("No such direction generator: " + dirGeneratorName);
  }

  // Prepare filter
  computationFlags = get_computationFlags(gdata, particleTransport, emissionName);

  // Prepare channel names
  channelNames.reserve(nChannels);
  for (const auto &radProc : particleTransport.get_radiativeLossProcesses()) {
    channelNames.push_back(radProc->get_type());
  }

  for (const auto &postProc : particleTransport.get_emissionPostProcesses())
    for (const auto &radProc : particleTransport.get_radiativeLossProcesses()) {
      channelNames.push_back(radProc->get_type() + "_" + postProc->get_name());
    }

  // Prepare reduction
  const size_t nReductions = (size_t)value(emissionName + ".reductions.n");
  emissivityReducers.reserve(nReductions);

  for (const auto &iReducer : iter::range(0, nReductions)) {
    const string reducerKey = emissionName + ".reduction" + ToString(iReducer, 2);
    const string reducerName = svalue(reducerKey);

    if (reducerName == "spectrum") {
      emissivityReducers.emplace_back(new SpectrumReduction(gdata, energies.getLength(),
                                                            directions_ptr->size(), channelNames,
                                                            normHD, normParticle, normEmission));
    } else if (reducerName == "projection") {
      Direction up = {0, 0, 0};
      const string upKey = reducerKey + ".up";
      if (value_exists(upKey + ".x")) {
        up[0] = value(upKey + ".x");
      }
      if (value_exists(upKey + ".y")) {
        up[1] = value(upKey + ".y");
      }
      if (value_exists(upKey + ".z")) {
        up[2] = value(upKey + ".z");
      }

      if (up == Direction{0, 0, 0}) {
        up = Direction{0, 0, 1};
      }

      ccout << "  " << reducerKey << " up: " << up << std::endl;

      double dx = std::numeric_limits<double>::max();
      for (int iDir = 0; iDir < 3; ++iDir) {
        dx = std::min(dx, gdata.get_dx(iDir, 0, 0));
      }

      const vector3D upperCorner = {gdata.global_xe[0], gdata.global_xe[1], gdata.global_xe[2]};
      const vector3D lowerCorner = {gdata.global_xb[0], gdata.global_xb[1], gdata.global_xb[2]};
      const vector3D diff = upperCorner - lowerCorner;
      double length = diff.norm();
      const vector3D offset = 0.5 * (upperCorner + lowerCorner);

      LinearGrid x1Grid(-0.5 * length, 0.5 * length, ceil(length / dx), 0);
      LinearGrid x2Grid(-0.5 * length, 0.5 * length, ceil(length / dx), 0);

      emissivityReducers.emplace_back(
          new ProjectionReduction(gdata, x1Grid, x2Grid, energies.getLength(), *directions_ptr,
                                  channelNames, up, offset, normHD, normParticle, normEmission));
    } else {
      throw CException("No such emissivity reducer");
    }
  }
}

NumMatrix<bool, 3> EmissionSolver::get_computationFlags(Data &gdata,
                                                        ParticleTransport &particleTransport,
                                                        string emissionName) const {
  const auto &normHD = particleTransport.get_normHD();

  // Compute average density
  REAL particleDensitySum = 0;
  int cellCount = 0;
  for (int iz = 0; iz <= gdata.mx[2]; ++iz) {
    for (int iy = 0; iy <= gdata.mx[1]; ++iy) {
      for (int ix = 0; ix <= gdata.mx[0]; ++ix) {
        const REAL localDensity = particle->numberDensity(ix, iy, iz);
        if (localDensity > 0) {
          particleDensitySum += localDensity;
          ++cellCount;
        }
      }
    }
  }
  const int cellCountLocal = cellCount;

#ifdef parallel
  gdata.MpiSum(particleDensitySum);
  gdata.MpiSum(cellCount);
#endif
  const REAL meanParticleDensity = particleDensitySum / cellCount;

  if (cellCount == 0) {
    ccout << " Particle field " << particle->particleID << " contains no particles ["
          << particle->get_particleName() << "]" << endl;
  } else if (meanParticleDensity < 0) {
    throw CException("Negative Particle density");
  } else {
    ccout << " Particle field " << particle->particleID << " has " << cellCount
          << " cells containing particles [" << particle->get_particleName() << "]" << endl;
  }

  auto computationFlags = get_filterFlags(gdata, normHD, emissionName);

  for (int iz = 0; iz <= gdata.mx[2]; ++iz) {
    for (int iy = 0; iy <= gdata.mx[1]; ++iy) {
      for (int ix = 0; ix <= gdata.mx[0]; ++ix) {
        computationFlags(ix, iy, iz) &=
            (particle->numberDensity(ix, iy, iz) > 1e-15 * meanParticleDensity);
      }
    }
  }

  return computationFlags;
}

NumMatrix<bool, 3> EmissionSolver::get_filterFlags(const Data &gdata, const normalisation &normHD,
                                                   string emissionMapName) const {
  NumMatrix<bool, 3> compFlag(Index::set(0, 0, 0),
                              Index::set(gdata.mx[0], gdata.mx[1], gdata.mx[2]));
  compFlag.clear();

  int counter = 0;

  if (value_exists(emissionMapName + ".filter")) {
    const std::string filterType = svalue(emissionMapName + ".filter");

    if (filterType == "spherical") {
      ccout << "  Using filter spherical for " << emissionMapName << endl;
      const Quantity filterRadius =
          value(emissionMapName + ".filter.radius") * normHD.normQuantity(normHD.LEN);
      const Quantity filterCenterX =
          value(emissionMapName + ".filter.centerX") * normHD.normQuantity(normHD.LEN);
      const Quantity filterCenterY =
          value(emissionMapName + ".filter.centerY") * normHD.normQuantity(normHD.LEN);
      const Quantity filterCenterZ =
          value(emissionMapName + ".filter.centerZ") * normHD.normQuantity(normHD.LEN);

      const REAL filterRadius_num = normHD.get_num(filterRadius);
      const REAL filterCenterX_num = normHD.get_num(filterCenterX);
      const REAL filterCenterY_num = normHD.get_num(filterCenterY);
      const REAL filterCenterZ_num = normHD.get_num(filterCenterZ);

      for (int iz = 0; iz <= gdata.mx[2]; ++iz) {
        for (int iy = 0; iy <= gdata.mx[1]; ++iy) {
          for (int ix = 0; ix <= gdata.mx[0]; ++ix) {
            const REAL distance = sqrt(sqr(gdata.getCen_x(ix) - filterCenterX_num) +
                                       sqr(gdata.getCen_y(iy) - filterCenterY_num) +
                                       sqr(gdata.getCen_z(iz) - filterCenterZ_num));
            if (distance < filterRadius_num) {
              compFlag(ix, iy, iz) = true;
              ++counter;
            } else {
              compFlag(ix, iy, iz) = false;
            }
          }
        }
      }

    } else if (filterType == "box") {
      ccout << "  Using filter box for " << emissionMapName << endl;
      const bool useLower[3] = {value_exists(emissionMapName + ".filter.lowerX"),
                                value_exists(emissionMapName + ".filter.lowerY"),
                                value_exists(emissionMapName + ".filter.lowerZ")};
      const bool useUpper[3] = {value_exists(emissionMapName + ".filter.upperX"),
                                value_exists(emissionMapName + ".filter.upperY"),
                                value_exists(emissionMapName + ".filter.upperZ")};

      const REAL lower[3] = {
          normHD.get_num((useLower[0] ? value(emissionMapName + ".filter.lowerX") : 0) *
                         normHD.normQuantity(normHD.LEN)),
          normHD.get_num((useLower[1] ? value(emissionMapName + ".filter.lowerY") : 0) *
                         normHD.normQuantity(normHD.LEN)),
          normHD.get_num((useLower[2] ? value(emissionMapName + ".filter.lowerZ") : 0) *
                         normHD.normQuantity(normHD.LEN))};
      const REAL upper[3] = {
          normHD.get_num((useUpper[0] ? value(emissionMapName + ".filter.upperX") : 0) *
                         normHD.normQuantity(normHD.LEN)),
          normHD.get_num((useUpper[1] ? value(emissionMapName + ".filter.upperY") : 0) *
                         normHD.normQuantity(normHD.LEN)),
          normHD.get_num((useUpper[2] ? value(emissionMapName + ".filter.upperZ") : 0) *
                         normHD.normQuantity(normHD.LEN))};

      int idx[3];
      for (idx[2] = 0; idx[2] <= gdata.mx[2]; ++idx[2]) {
        for (idx[1] = 0; idx[1] <= gdata.mx[1]; ++idx[1]) {
          for (idx[0] = 0; idx[0] <= gdata.mx[0]; ++idx[0]) {
            bool compFlagLoc = true;

            for (int dir = 0; dir < 3; ++dir) {
              if (useLower[dir]) {
                compFlagLoc &= (lower[dir] <= gdata.getCen(dir, idx[dir]));
              }

              if (useUpper[dir]) {
                compFlagLoc &= (upper[dir] > gdata.getCen(dir, idx[dir]));
              }
            }

            compFlag[idx] = compFlagLoc;

            if (compFlagLoc) {
              ++counter;
            }
          }
        }
      }

    } else if (filterType == "plane") {
      ccout << "  Using filter plane for " << emissionMapName << endl;
      const REAL planeOrigin[3] = {value(emissionMapName + ".filter.planeX"),
                                   value(emissionMapName + ".filter.planeY"),
                                   value(emissionMapName + ".filter.planeZ")};
      const int side = value(emissionMapName + ".filter.side");

      int idx[3];
      for (idx[2] = 0; idx[2] <= gdata.mx[2]; ++idx[2]) {
        for (idx[1] = 0; idx[1] <= gdata.mx[1]; ++idx[1]) {
          for (idx[0] = 0; idx[0] <= gdata.mx[0]; ++idx[0]) {
            REAL discriminator = 0;
            for (int dir = 0; dir < 3; ++dir) {
              discriminator +=
                  side * (gdata.getCen(dir, idx[dir]) - planeOrigin[dir]) * planeOrigin[dir];
            }

            const bool compFlagLoc = discriminator > 0;
            compFlag[idx] = compFlagLoc;
            if (compFlagLoc) {
              ++counter;
            }
          }
        }
      }

    } else {
      throw invalid_argument("No such filter type " + filterType);
    }

#ifdef parallel
    gdata.MpiSum(counter);
#endif
    ccout << "  Filtering cells: " << counter << endl;

  } else {
    for (int iz = 0; iz <= gdata.mx[2]; ++iz) {
      for (int iy = 0; iy <= gdata.mx[1]; ++iy) {
        for (int ix = 0; ix <= gdata.mx[0]; ++ix) {
          compFlag(ix, iy, iz) = true;
        }
      }
    }
  }
  return compFlag;
}

void EmissionSolver::work(Data &gdata, ParticleTransport &particleTransport) {
  benchmark(emissionName,
            [&]() {
              do_work(gdata, particleTransport);
              return 1;
            },
            true, false);
}


void EmissionSolver::do_work(Data &gdata, ParticleTransport &particleTransport) {
  CronosTimer emissivityTimer;
  std::vector<CronosTimer> reductionTimer(emissivityReducers.size());

  // Emission Integral
  hsize_t iDirection = 0;
  for (const auto &direction : *directions_ptr) {
    // #if DLEVEL >= 1
    ccout << "  Direction: " << direction << endl;
    // #endif
    set_direction(gdata, particleTransport, direction);

    emissivityTimer.start();
    computeEmissivities(gdata, particleTransport, iDirection, direction);
    emissivityTimer.stop();

    ++iDirection;
  }  // direction

  // Statistics
  emissivityTime = emissivityTimer.time();
}


void EmissionSolver::mergeData(Data &gdata) {
  for (auto &reducer : emissivityReducers) {
    reducer->mpiMerge();
  }
  double mean, max;

  mean = max = emissivityTime;
#ifdef parallel
  gdata.MpiAvergage(mean);
  gdata.MpiMax(max);
#endif
  ccout << "   Emissivity mean=" << mean << " max=" << max << endl;
}

// void EmissionSolver::computeReductions(Data &gdata, ParticleTransport &particleTransport) {
//   // Work estimation
//   // int totalSteps = radiativeLossProcesses.size() * phis.getLength() * thetas.getLength() *
//   //                  energies.getLength() * cellCountLocal;
//   // StepProgress progress(totalSteps);
//   // IntervalProgressPrinter stepPrinter;
//   // TimeIntervalProgressPrinter timePrinter;

//   CronosTimer emissivityTimer;
//   std::vector<CronosTimer> reductionTimer(emissivityReducers.size());

//   EmissivityContainer emissivityContainer(gdata, energies, nChannels);

//   // Emission Integral
//   hsize_t iDirection = 0;
//   for (const auto &direction : *directions_ptr) {
//     // #if DLEVEL >= 1
//     ccout << "  Direction: " << direction << endl;
//     // #endif
//     set_direction(gdata, particleTransport, direction);

//     emissivityTimer.start();
//     computeEmissivities(gdata, particleTransport, direction, emissivityContainer);
//     emissivityTimer.stop();

//     auto timer = reductionTimer.begin();
//     for (auto &reducer : emissivityReducers) {
//       (*timer).start();
//       reducer->reduce(iDirection, emissivityContainer);
//       (*timer).stop();
//       ++timer;
//     }
//     ++iDirection;
//   }  // direction

//   // Statistics
//   emissivityTime = emissivityTimer.time();
//   for (const auto &timer : reductionTimer) {
//     reductionTime.push_back(timer.time());
//   }
// }

void EmissionSolver::computeEmissivities(
    Data &gdata, ParticleTransport &particleTransport, const size_t &iDirection,
    const Direction &direction) {
  auto &radiativeLossProcesses = particleTransport.get_radiativeLossProcesses();
  auto &emissionPostProcesses = particleTransport.get_emissionPostProcesses();

  for (int iz = 0; iz <= gdata.mx[2]; ++iz) {
    for (int iy = 0; iy <= gdata.mx[1]; ++iy) {
      for (int ix = 0; ix <= gdata.mx[0]; ++ix) {
#if DLEVEL >= 2
        ccout << "Domain Coordinate: " << ix << " " << iy << " " << iz << endl;
#endif

        set_position(gdata, particleTransport, ix, iy, iz);

        if (!computationFlags(ix, iy, iz)) {
          continue;
        }

        for (int iEnergy = 0; iEnergy < energies.getLength(); ++iEnergy) {
#if DLEVEL >= 3
          ccout << "Energy: " << iEnergy << endl;
#endif
          const auto &energy = energies[iEnergy];
          set_energy(gdata, particleTransport, energy);

          size_t iRad = 0;
          for (auto &radProc : radiativeLossProcesses) {
            set_radiativeLossProcess(gdata, particleTransport, radProc);
            auto &emissionField = radProc->get_emissionField();

            const auto photonEmissivity =
                emissionField.get_photonEmissivity_LAB(gdata, ix, iy, iz, energy, direction.data());
            applyReductions(iDirection, iChannel(iRad), iEnergy, ix, iy, iz, photonEmissivity);

            size_t iPost = 0;
            for (auto &postProc : emissionPostProcesses) {
              const auto photonEmissivity_post =
                  postProc->compute(gdata, ix, iy, iz, photonEmissivity, energy, direction.data());
              applyReductions(iDirection, iChannel(iRad, iPost), iEnergy, ix, iy, iz,
                              photonEmissivity_post);

              ++iPost;
            }

            // ++progress;
            // stepPrinter.print(ccout, progress, 0.1);
            // timePrinter.print(ccout, progress, 60);

            ++iRad;
          }  // iRad

        }  // iEnergy

      }  // ix
    }    // iy
  }      // iz
}

void EmissionSolver::applyReductions(const size_t &iDirection, const size_t &iChannel, const size_t &iEnergy,
                     const size_t &ix, const size_t &iy, const size_t &iz,
                     const double &emissivity) {
  for (auto &reducer : emissivityReducers) {
    reducer->reduce(iDirection, iChannel, iEnergy, ix, iy, iz, emissivity);
  }
}

void EmissionSolver::set_particle(Data &gdata, ParticleTransport &particleTransport,
                                  Particle &particle) const {
  auto &radiativeLossProcesses = particleTransport.get_radiativeLossProcesses();
  for (auto &radProc : radiativeLossProcesses) {
    radProc->set_particle(particle);
  }
}

void EmissionSolver::set_position(Data &gdata, ParticleTransport &particleTransport, int ix, int iy,
                                  int iz) const {
  auto &radiativeLossProcesses = particleTransport.get_radiativeLossProcesses();
  for (auto &radProc : radiativeLossProcesses) {
    radProc->compute_constants(gdata, ix, iy, iz);
  }

  auto &emissionPostProcesses = particleTransport.get_emissionPostProcesses();
  for (auto &postProc : emissionPostProcesses) {
    postProc->set_position(gdata, ix, iy, iz);
  }
}

void EmissionSolver::set_direction(Data &gdata, ParticleTransport &particleTransport,
                                   const Direction &observerDirection) const {
  auto &emissionPostProcesses = particleTransport.get_emissionPostProcesses();
  for (auto &postProc : emissionPostProcesses) {
    postProc->set_direction(gdata, observerDirection.data());
  }
}

void EmissionSolver::set_energy(Data &gdata, ParticleTransport &particleTransport,
                                REAL photonEnergy) const {
  auto &emissionPostProcesses = particleTransport.get_emissionPostProcesses();
  for (auto &postProc : emissionPostProcesses) {
    postProc->set_energy(gdata, photonEnergy);
  }
}

void EmissionSolver::set_radiativeLossProcess(Data &gdata, ParticleTransport &particleTransport,
                                              RadiativeLossProcess *radiativeLossProcess) const {
  auto &emissionPostProcesses = particleTransport.get_emissionPostProcesses();
  for (auto &postProc : emissionPostProcesses) {
    postProc->set_radiativeLossProcess(gdata, radiativeLossProcess);
  }
}

int EmissionSolver::iChannel(int iRadProc) const { return iRadProc; }

int EmissionSolver::iChannel(int iRadProc, int iPostProc) const {
  return nRadiativeLossProcesses * (iPostProc + 1) + iRadProc;
}
