#ifndef CRONOS_EMISSION_SOLVER_H
#define CRONOS_EMISSION_SOLVER_H

#ifdef parallel
#include "mpi.h"
#endif

#include <vector>

#include "EmissionDirection.H"
#include "EmissivityContainer.H"
#include "ParticleTransport.H"
#include "data.H"
#include "problem.H"
#include "normalisation.H"

class EmissionHandler;

class EmissionNormalisation : public normalisation {
 public:
  virtual void setup();
};

class EmissionSolver {
 public:
  using ComputationFlags = NumMatrix<bool, 3>;
  using reducer_ptr_t = std::unique_ptr<EmissivityReduction>;

  EmissionSolver(ParameterFileReader& parameters, Data& gdata, ProblemType&,
                 ParticleTransport& particleTransport, const EmissionNormalisation& normEmission,
                 const int emissionID);

  void work(Data& gdata, ParticleTransport&);

  void mergeData(Data& gdata);

	const NumArray<double>& getEnergies() const {
		return energies;
	}

	const EmissionDirections& getDirections() const {
		return *directions_ptr;
	}

	size_t getNChannels() const {
		return nChannels;
	}

	const std::vector<reducer_ptr_t>& getEmissivityReducers() const {
		return emissivityReducers;
	}
	const std::vector<string>& getChannelNames() const {
		return channelNames;
	}

  const Particle& getParticle() const {
    return *particle;
  }

  std::string getEmissionName() const {
    return emissionName;
  }

 private:
  // void computeReductions(Data&, ParticleTransport&);
  void do_work(Data& gdata, ParticleTransport&);
  void computeEmissivities(Data&, ParticleTransport&, const size_t& iDirection, const Direction&);

  void applyReductions(const size_t& iDirection, const size_t& iChannel, const size_t& iEnergy,
                      const size_t& ix, const size_t& iy, const size_t& iz,
                      const double& emissivity);

  NumMatrix<bool, 3> get_computationFlags(Data&, ParticleTransport&,
                                          const string emissionName) const;

  /**
   * @brief Checks for which cells the emission has to be computed
   */
  NumMatrix<bool, 3> get_filterFlags(const Data& gdata, const normalisation& normHD,
                                     string emissionName) const;

  /**
   * @brief Passes parameters through to Processes
   */
  void set_particle(Data&, ParticleTransport&, Particle&) const;

  void set_position(Data&, ParticleTransport&, const int ix, const int iy, const int iz) const;

  void set_direction(Data&, ParticleTransport&, const Direction&) const;

  void set_energy(Data&, ParticleTransport&, REAL photonEnergy) const;

  void set_radiativeLossProcess(Data&, ParticleTransport&, RadiativeLossProcess*) const;

  int iChannel(int iRadProc) const;
  int iChannel(int iRadProc, int iPostProc) const;

  ParameterFileReader& parameters;

  int nRadiativeLossProcesses;
  int nEmissionPostProcesses;

  const EmissionNormalisation& normEmission;

  string emissionName;
  NumArray<double> energies;
  std::unique_ptr<EmissionDirections> directions_ptr;
  NumMatrix<bool, 3> computationFlags;
  size_t nChannels;

  std::vector<reducer_ptr_t> emissivityReducers;

  std::vector<string> channelNames;
  Particle* particle;

  // Statistics
  double emissivityTime;
};

#endif
