#include "CronosTimer.H"
#include "EmissivityContainer.H"

// EmissivityContainer::EmissivityContainer(const Grid& grid, const NumArray<double>&
// emissionEnergies,
//                                          const size_t nChannels) {
//   resize(Index::set(0, 0, 0, 0, 0), Index::set(nChannels - 1, emissionEnergies.getLength() - 1,
//                                                grid.mx[0], grid.mx[1], grid.mx[2]));
//   clear();
// }

template <class T, int N>
void EmissivityReduction::_mpiMerge(DataT<T, N>& dataList) {
#ifdef parallel
  CronosTimer mergeTimer;
  mergeTimer.start();

  ccout << " Merging data of all ranks " << endl;
  for (auto& data : dataList) {
    MPI_Allreduce(MPI_IN_PLACE, data.data(), data.getSize(), MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  }
  mergeTimer.stop();
  ccout << "   mpi merge done after: " << mergeTimer.time() << " s" << endl;

#endif
}

// Disable parallel write for now
#define HDF_PARALLEL_IO_LOCAL CRONOS_OFF

template <class T, int N>
void EmissivityReduction::_writeToH5(Hdf5Stream& h5out, hid_t group,
                                     const DataT<T, N>& data) const {
#if !defined(parallel) || (HDF_PARALLEL_IO_LOCAL == CRONOS_OFF)
  int rank = 0;
#ifdef parallel
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#endif

  if (rank == 0) {
    for (const auto& channelData : data) {
      h5out.WriteNDMatrix(channelData.getName(), channelData, group);
    }
    h5out.AddGlobalAttr("type", type(), group);
  }

#else
#error "NOT IMPLEMENTED"
#endif
}

SpectrumReduction::SpectrumReduction(const Grid& _grid, const size_t _nEnergies,
                                     const size_t _nDirections,
                                     const std::vector<string>& channelNames,
                                     const normalisation& normHD, const normalisation& normParticle,
                                     const normalisation& normEmission)
    : grid(_grid),
      nEnergies(_nEnergies),
      nDirections(_nDirections),
      nChannels(channelNames.size()) {
  data.resize(nChannels);
  for (auto& channelData : data) {
    channelData.resize(Index::set(0, 0), Index::set(nEnergies - 1, nDirections - 1));
    channelData.clear();
  }

  size_t iChannel = 0;
  for (auto& channelData : data) {
    channelData.rename(channelNames[iChannel]);
    ++iChannel;
  }

  FluxRenormFac = normEmission.renormalise(normEmission.DIFF_NUM_EMISS, normParticle, 1.);
  Volume2Density = cube(normHD.normQuantity(normHD.LEN))
                       .normalizeTo(1. / normEmission.normQuantity(normEmission.NUM_DENS));
}

// void SpectrumReduction::reduce(const size_t& iDirection, const EmissivityContainer& emissivity) {
//   for (int iz = 0; iz <= grid.mx[2]; ++iz) {
//     for (int iy = 0; iy <= grid.mx[1]; ++iy) {
//       for (int ix = 0; ix <= grid.mx[0]; ++ix) {
//         for (int iChannel = 0; iChannel < nChannels; ++iChannel) {
//           for (int iEnergy = 0; iEnergy < nEnergies; ++iEnergy) {
//             data[iChannel](iEnergy, iDirection) += emissivity(iChannel, iEnergy, ix, iy, iz) *
//                                                    grid.get_CellVolume(ix, iy, iz) *
//                                                    Volume2Density * FluxRenormFac;
//           }
//         }
//       }
//     }
//   }
// }

std::string SpectrumReduction::type() const { return "spectrum"; }

void SpectrumReduction::mpiMerge() { _mpiMerge(data); }

void SpectrumReduction::writeToH5(Hdf5Stream& h5, hid_t group) const {
  _writeToH5(h5, group, data);
}

ProjectionReduction::ProjectionReduction(
    const Grid& _grid, const Grid1D& _x1Grid, const Grid1D& _x2Grid, const size_t _nEnergies,
    const DirectionList& _directions, const std::vector<string>& channelNames, const Direction& _up,
    const vector3D& _offset, const normalisation& normHD, const normalisation& normParticle,
    const normalisation& normEmission)
    : grid(_grid),
      x1Grid(_x1Grid),
      x2Grid(_x2Grid),
      up(_up),
      offset(_offset),
      nEnergies(_nEnergies),
      directions(_directions),
      nChannels(channelNames.size()) {
  data.resize(nChannels);
  for (auto& channelData : data) {
    channelData.resize(Index::set(0, 0, 0, 0), Index::set(x1Grid.get_mx(), x2Grid.get_mx(),
                                                          nEnergies - 1, directions.size() - 1));
    channelData.clear();
  }

  size_t iChannel = 0;
  for (auto& channelData : data) {
    channelData.rename(channelNames[iChannel]);
    ++iChannel;
  }

  FluxRenormFac = normEmission.renormalise(normEmission.DIFF_NUM_FLUX, normParticle, 1.);
  Volume2Density = cube(normHD.normQuantity(normHD.LEN))
                       .normalizeTo(1. / normEmission.normQuantity(normEmission.NUM_DENS));
}

#include "Projection.H"

// void ProjectionReduction::reduce(const size_t& iDirection, const EmissivityContainer& emissivity)
// {
//   // Restructure data (to match the projection format)
//   NumMatrix<double, 5> restructuredData;
//   restructuredData.resize(Index::set(0, 0, 0, 0, 0), Index::set(grid.mx[0], grid.mx[1],
//   grid.mx[2],
//                                                                 nEnergies - 1, nChannels - 1));
//   restructuredData.clear();

//   for (int iz = 0; iz <= grid.mx[2]; ++iz) {
//     for (int iy = 0; iy <= grid.mx[1]; ++iy) {
//       for (int ix = 0; ix <= grid.mx[0]; ++ix) {
//         for (int iChannel = 0; iChannel < nChannels; ++iChannel) {
//           for (int iEnergy = 0; iEnergy < nEnergies; ++iEnergy) {
//             restructuredData(ix, iy, iz, iEnergy, iChannel) +=
//                 emissivity(iChannel, iEnergy, ix, iy, iz);
//           }
//         }
//       }
//     }
//   }

//   const double dx = min(x1Grid.get_del(0), x2Grid.get_del(0));
//   double sum = 0;
//   for (int i = 0; i < restructuredData.getSize(); ++i) {
//     sum += restructuredData.data()[i];
//   }

//   static size_t iDirectionInUse = std::numeric_limits<size_t>::max();
//   static ProjectionPlane projectionPlane =
//       ProjectionPlane::newFromDirection(x1Grid, x2Grid, directions[iDirection], up, offset);
//   if (iDirection != iDirectionInUse) {
//   }

//   Projection projection(projectionPlane, nEnergies);

//   NumMatrix<double, 4> projectedData = projection.project(grid, restructuredData, dx);

//   for (int iChannel = 0; iChannel < nChannels; ++iChannel) {
//     for (int iEnergy = 0; iEnergy < nEnergies; ++iEnergy) {
//       for (int ix2 = 0; ix2 <= x2Grid.get_mx(); ++ix2) {
//         for (int ix1 = 0; ix1 <= x1Grid.get_mx(); ++ix1) {
//           data[iChannel](ix1, ix2, iEnergy, iDirection) +=
//               projectedData(ix1, ix2, iEnergy, iChannel) * Volume2Density * FluxRenormFac;
//         }
//       }
//     }
//   }
// }

void ProjectionReduction::reduce(const size_t& iDirection, const size_t& iChannel,
                                 const size_t& iEnergy, const size_t& ix, const size_t& iy,
                                 const size_t& iz, const double& emissivity) {
  static size_t iDirectionInUse = std::numeric_limits<size_t>::max();
  static std::unique_ptr<Projection> projection = nullptr;
  
  if ((iDirectionInUse != iDirection) || (projection == nullptr)) {
    iDirectionInUse = iDirection;
    ProjectionPlane projectionPlane = ProjectionPlane::newFromDirection(x1Grid, x2Grid, directions[iDirection], up, offset);
    projection = std::unique_ptr<Projection>(new Projection(projectionPlane, nEnergies));
  }

  double dataPoint = emissivity * Volume2Density * FluxRenormFac;
  std::array<size_t, 5> dataIdx{ix, iy, iz, iEnergy, iDirection};
  projection->project(grid, dataIdx, dataPoint, data[iChannel]);
}

std::string ProjectionReduction::type() const { return "projection"; }

void ProjectionReduction::mpiMerge() { _mpiMerge(data); }

void ProjectionReduction::writeToH5(Hdf5Stream& h5, hid_t group) const {
  _writeToH5(h5, group, data);
  double dx[2];
  dx[0] = x1Grid.get_del(0);
  dx[1] = x2Grid.get_del(0);
  h5.AddGlobalAttr("dx", dx, group, 2);

  double origin[2];
  origin[0] = x1Grid.get_xL(0);
  origin[1] = x2Grid.get_xL(0);
  h5.AddGlobalAttr("origin", origin, group, 2);

  int Nx[2];
  Nx[0] = x1Grid.get_Nx();
  Nx[1] = x2Grid.get_Nx();
  h5.AddGlobalAttr("Nx", Nx, group, 2);
}
