#ifndef EMISSIVITY_CONTAINER_H
#define EMISSIVITY_CONTAINER_H

#include <memory>
#include "Hdf5File_cbase.H"
#include "grid.H"
#include "matrix.H"

#include <string>
#include "normalisation.H"

#include "EmissionDirection.H"

/**
 * @brief Reduces the computed emissivities (e.g. to a single spectrum or projection)
 * and stores the result internally for a given amount of directions.
 * The class also handles the mpi merging and the h5 writing
 * */
class EmissivityReduction {
 public:
  /**
   * @brief Reduces the emission container and stores the result internally
   */
  virtual void reduce(const size_t& iDirection, const size_t& iChannel, const size_t& iEnergy,
                      const size_t& ix, const size_t& iy, const size_t& iz,
                      const double& emissivity) = 0;

  /**
   * @brief Merges the internal data of all mpi ranks
   */
  virtual void mpiMerge() = 0;

  /**
   * @brief Writes the internal storage to a given hdf5 group
   * Writes also additional data for the postprocessing of the respective subclass
   */
  virtual void writeToH5(Hdf5Stream&, hid_t group) const = 0;

  /**
   * @brief Returns the reduction type
   * This is stored as hdf5 attribute for postprocessing identification
   */
  virtual std::string type() const = 0;

 protected:
  template <class T, int N>
  /** @brief Internal reduction type */
  using ReductionT = NumMatrix<T, N>;

  template <class T, int N>
  /**
   * @brief Internal reduction storage type
   * The reduction of an emissivity map is stored as vector element
   */
  using DataT = std::vector<ReductionT<T, N> >;

  template <class T, int N>
  /**
   * @brief Basic merging strategy for a list of NumMatrices
   * Intended for subclass usage
   */
  static void _mpiMerge(DataT<T, N>& data);

  template <class T, int N>
  /**
   * @brief Basic hdf5 writing strategy for a list of NumMatrices
   * Intended for subclass usage
   */
  void _writeToH5(Hdf5Stream&, hid_t group, const DataT<T, N>& data) const;
};

/**
 * @brief Reduces the emissivities to a single emission spectrum
 */
class SpectrumReduction : public EmissivityReduction {
 private:
  /**
   * indices in order [iChannel](iEnergy, iDirection)
   */
  DataT<double, 2> data;

  const size_t nEnergies, nDirections, nChannels;
  const Grid& grid;

  /** Corrects for the fact that densities are used for the integrations,
   * which might be based on a different volume element with respect to the one implied by the
   * length scale. */
  REAL Volume2Density;

  /** Needed for the normalisation change: ParticleNorm -> EmissionNorm */
  REAL FluxRenormFac;

 public:
  SpectrumReduction(const Grid& grid, const size_t nEnergies, const size_t nDirections,
                    const std::vector<string>& channelNames, const normalisation& normHD,
                    const normalisation& normParticle, const normalisation& normEmission);

  virtual void reduce(const size_t& iDirection, const size_t& iChannel, const size_t& iEnergy,
                      const size_t& ix, const size_t& iy, const size_t& iz,
                      const double& emissivity) override {
    data[iChannel](iEnergy, iDirection) +=
        emissivity * grid.get_CellVolume(ix, iy, iz) * Volume2Density * FluxRenormFac;
  }

  virtual std::string type() const override;

  virtual void mpiMerge() override;
  virtual void writeToH5(Hdf5Stream&, hid_t group) const override;
};

#include "Grid1D.H"

/**
 * @brief Reduces the emissivities to a projection
 */
class ProjectionReduction : public EmissivityReduction {
 private:
  /**
   * indices in order [iChannel](ix, iy, iEnergy, iDirection)
   */
  DataT<double, 4> data;

  const Grid1D x1Grid, x2Grid;
  const size_t nEnergies, nChannels;
  const DirectionList& directions;
  const Grid& grid;

  const Direction up;
  const vector3D offset; 

  /** Corrects for the fact that densities are used for the integrations,
   * which might be based on a different volume element with respect to the one implied by the
   * length scale. */
  REAL Volume2Density;

  /** Needed for the normalisation change: ParticleNorm -> EmissionNorm */
  REAL FluxRenormFac;

 public:
  ProjectionReduction(const Grid& grid, const Grid1D& x1Grid, const Grid1D& x2Grid,
                      const size_t nEnergies, const DirectionList& directions,
                      const std::vector<string>& channelNames, const Direction& up,
                      const vector3D& offset, const normalisation& normHD,
                      const normalisation& normParticle, const normalisation& normEmission);

  virtual void reduce(const size_t& iDirection, const size_t& iChannel, const size_t& iEnergy,
                      const size_t& ix, const size_t& iy, const size_t& iz,
                      const double& emissivity) override;

  virtual std::string type() const override;

  virtual void mpiMerge() override;
  virtual void writeToH5(Hdf5Stream&, hid_t group) const override;
};

#endif