#include <stdexcept>
#include "Projection.H"

Plane::Plane(const WorldVector& basisVector1, const WorldVector& basisVector2,
             const AxisRange& range1, const AxisRange& range2,
             const WorldVector& offsetPlaneOrigin) {
  this->basisVector1 = basisVector1.normalised();
  this->basisVector2 = basisVector2.normalised();
  this->normalVector = basisVector1.cross(basisVector2).normalised();
  this->range1 = range1;
  this->range2 = range2;
  this->offsetPlaneOrigin = offsetPlaneOrigin;
}

double Plane::intersectionDistance(const WorldVector& origin, const WorldVector& direction) const {
  const double normalDistance = (origin - offsetPlaneOrigin).inner(normalVector);
  return -normalDistance / direction.inner(normalVector);
}

Plane::PlaneVector Plane::intersect(const WorldVector& origin, const WorldVector& direction) const {
  const double intersectionDistance = this->intersectionDistance(origin, direction);
  const auto intersection = origin - offsetPlaneOrigin + intersectionDistance * direction;

  PlaneVector planeVector;
  planeVector[0] = basisVector1.inner(intersection);
  planeVector[1] = basisVector2.inner(intersection);

  return planeVector;
}
bool Plane::contains(const PlaneVector& planeVector) const {
  return (planeVector[0] >= range1[0]) && (planeVector[0] <= range1[1]) &&
         (planeVector[1] >= range2[0]) && (planeVector[1] <= range2[1]);
}

bool Plane::intersectedBy(const WorldVector& origin, const WorldVector& direction) const {
  const auto intersection = intersect(origin, direction);
  return contains(intersection);
}

Plane::WorldVector Plane::convert(const PlaneVector& planeVector) const {
  return planeVector[0] * basisVector1 + planeVector[2] * basisVector2 + offsetPlaneOrigin;
}

Box::Box(const vector3D& axisVector1, const vector3D& axisVector2, const vector3D& axisVector3,
         const AxisRange& axisRange1, const AxisRange& axisRange2, const AxisRange& axisRange3) {
  // xy Low
  sides.push_back(Plane(axisVector1, axisVector2, axisRange1, axisRange2, {0, 0, axisRange3[0]}));
  // xz Low
  sides.push_back(Plane(axisVector1, axisVector3, axisRange1, axisRange3, {0, axisRange2[0], 0}));
  // yz Low
  sides.push_back(Plane(axisVector2, axisVector3, axisRange2, axisRange3, {axisRange1[0], 0, 0}));
  // xy High
  sides.push_back(Plane(axisVector1, axisVector2, axisRange1, axisRange2, {0, 0, axisRange3[1]}));
  // xz High
  sides.push_back(Plane(axisVector1, axisVector3, axisRange1, axisRange3, {0, axisRange2[1], 0}));
  // yz High
  sides.push_back(Plane(axisVector2, axisVector3, axisRange2, axisRange3, {axisRange1[1], 0, 0}));
}

const std::vector<Plane> Box::getSides() const { return sides; }

LineOfSight::LineOfSight(const vector3D& origin_, const vector3D& direction_)
    : origin(origin_), direction(direction_) {}

LineOfSight::AxisRange LineOfSight::rangeInsideBox(const Box& box) const {
  const auto& sides = box.getSides();
  static std::vector<double> distances(6);
  distances.clear();
  for (const auto& side : sides) {
    if (side.intersectedBy(origin, direction)) {
      distances.push_back(side.intersectionDistance(origin, direction));
    }
  }
  if (distances.size() < 2) {
    return AxisRange{0, 0};
  } else if (distances.size() == 1) {
    throw std::runtime_error("Line of sight: hitting just one box side. This should not occur.");
  }
  auto range = std::minmax_element(distances.begin(), distances.end());
  return AxisRange{*range.first, *range.second};
}

GridBox::GridBox(const Grid& _grid)
    : Box({1, 0, 0}, {0, 1, 0}, {0, 0, 1}, {_grid.getCen(0, 0), _grid.getCen(0, _grid.mx[0])},
          {_grid.getCen(1, 0), _grid.getCen(1, _grid.mx[1])},
          {_grid.getCen(2, 0), _grid.getCen(2, _grid.mx[2])}),
      grid(_grid) {}

GridBox::GridIndex GridBox::getIndex(const vector3D& position) const {
  GridIndex index;
  for (int dir = 0; dir < 3; ++dir) {
    index[dir] = grid.get_CellIndex(dir, position[dir]);
  }
  return index;
}

vector3D GridBox::getPosition(const GridIndex& index) const {
  vector3D position;
  for (int dir = 0; dir < 3; ++dir) {
    position[dir] = grid.getCen(dir, index[dir]);
  }
  return position;
}

GridBox::NeighbourList GridBox::getNeighbours(const vector3D& position) const {
  const auto containingCell = getIndex(position);
  const auto containingCellPosition = getPosition(containingCell);
  std::array<int, 3> neighbourCellIndex;

  for (int dir = 0; dir < 3; ++dir) {
    auto const neighbourShift = position[dir] < containingCellPosition[dir] ? -1 : +1;

    neighbourCellIndex[dir] = containingCell[dir] + neighbourShift;
    if (neighbourCellIndex[dir] < 0 || neighbourCellIndex[dir] > grid.mx[dir]) {
      neighbourCellIndex[dir] = containingCell[dir] - neighbourShift;
    }
  }

  NeighbourList neighbours(8);
  int n = 0;
  std::generate(neighbours.begin(), neighbours.end(),
                [&containingCell, &neighbourCellIndex, &n]() mutable {
                  GridIndex neighbour;

#define CHECK_BIT(var, pos) ((var) & (1 << (pos)))

                  if (CHECK_BIT(n, 0)) {
                    neighbour[0] = neighbourCellIndex[0];
                  } else {
                    neighbour[0] = containingCell[0];
                  }
                  if (CHECK_BIT(n, 1)) {
                    neighbour[1] = neighbourCellIndex[1];
                  } else {
                    neighbour[1] = containingCell[1];
                  }
                  if (CHECK_BIT(n, 2)) {
                    neighbour[2] = neighbourCellIndex[2];
                  } else {
                    neighbour[2] = containingCell[2];
                  }
                  ++n;
                  return neighbour;
                });

  return neighbours;
}

InterpolationBox::InterpolationBox(const Grid& grid) : GridBox(grid) {}

TrilinearBox::TrilinearBox(const Grid& grid) : InterpolationBox(grid) {}

typename TrilinearBox::InterpolationList TrilinearBox::getInterpolationList(
    const vector3D& position) const {
  NeighbourList neighbours = getNeighbours(position);
  WeightList weights(neighbours.size());
  double weightSum = 0;

  // The weights are given by the opposite volume
  // Compute volumes and insert them to weights in the reversed order
  std::transform(
      neighbours.begin(), neighbours.end(), weights.rbegin(), [&](const GridIndex& neighbour) {
        const auto neighbourPos = getPosition(neighbour);
        const auto diff = position - neighbourPos;

        const double volume =
            std::abs(std::accumulate(diff.begin(), diff.end(), 1., std::multiplies<double>()));
        weightSum += volume;
        return volume;
      });

  // Normalise
  std::transform(weights.begin(), weights.end(), weights.begin(),
                 [&](const double& x) { return x / weightSum; });

  return {neighbours, weights};
}

PiecewiseConstantBox::PiecewiseConstantBox(const Grid& grid) : InterpolationBox(grid) {}

typename PiecewiseConstantBox::InterpolationList PiecewiseConstantBox::getInterpolationList(
    const vector3D& position) const {
  return {{getIndex(position)}, {1.}};
}

#include <functional>
template <class T, int N>
NumMatrix<T, N - 3> LineOfSight::integrate(const InterpolationBox& box, const NumMatrix<T, N>& data,
                                           const double& _dx) const {
  constexpr int rank = N - 3;
  NumMatrix<T, rank> value(&data.getLow()[3], &data.getHigh()[3]);
  value.clear();

  const auto range = rangeInsideBox(box);
  if (range[0] == range[1]) {
    return value;
  }

  auto addIntegrand = [&](const double x, const double weight) {
    const auto position = origin + x * direction;
    const auto list = box.getInterpolationList(position);

    std::array<int, rank> index;
    std::function<void(const int&)> rankAddLoop = [&](const int rankCounter) {
      if (rankCounter > 0) {
        const size_t dim = rankCounter - 1;
        const auto lo = data.getLow()[3 + dim];
        const auto hi = data.getHigh()[3 + dim];
        for (int iter = lo; iter <= hi; ++iter) {
          index[dim] = iter;
          rankAddLoop(rankCounter - 1);
        }
      } else {
        { value[index.data()] += weight * box.interpolate(data, list, index); };
      }
    };

    rankAddLoop(rank);
  };

  // Apply simpson rule
  const size_t subdivisions = ceil((range[1] - range[0]) / _dx);
  const auto dx = (range[1] - range[0]) / subdivisions;
  const auto dxHalf = 0.5 * dx;
  constexpr double oneThird = 1. / 3.;
  constexpr double twoThird = 2. / 3.;
  addIntegrand(range[0], 1. / 6.);
  addIntegrand(range[1], -1. / 6.);
  for (size_t iStep = 0; iStep < subdivisions; ++iStep) {
    const auto xR = range[0] + iStep * dx;
    const auto xC = xR - dxHalf;
    addIntegrand(xC, twoThird);
    addIntegrand(xR, oneThird);
  }

  value *= dx;
  return value;
}

ProjectionPlane::ProjectionPlane(const WorldVector& basisVector1, const WorldVector& basisVector2,
                                 const Grid1D& x1Grid, const Grid1D& x2Grid,
                                 const WorldVector& offsetPlaneOrigin)
    : Plane(basisVector1, basisVector2, {x1Grid.get_xL(0), x1Grid.get_xL(x1Grid.get_Nx())},
            {x2Grid.get_xL(0), x2Grid.get_xL(x2Grid.get_Nx())}, offsetPlaneOrigin),
      x1Grid(x1Grid),
      x2Grid(x2Grid) {}

ProjectionPlane ProjectionPlane::newFromDirection(const Grid1D& x1Grid, const Grid1D& x2Grid,
                                                  const WorldVector& direction,
                                                  const WorldVector& up,
                                                  const WorldVector& offset) {
  assert(abs(up.norm() - 1) < 1e-8);
  assert(abs(direction.norm() - 1) < 1e-8);

  const std::pair<WorldVector, WorldVector> basisPair = computeBasis(direction, up);
  return ProjectionPlane(basisPair.first, basisPair.second, x1Grid, x2Grid, offset);
}

std::pair<Plane::WorldVector, Plane::WorldVector> ProjectionPlane::computeBasis(
    const WorldVector& direction, const WorldVector& up) {
  const auto basisVector1 = up.cross(direction).normalised();
  const auto basisVector2 = direction.cross(basisVector1).normalised();
  return {basisVector1, basisVector2};
}

ProjectionPlane::PlaneIndices ProjectionPlane::getPlaneIndices(
    const PlaneVector& planeVector) const {
  assert(contains(planeVector));
  return {x1Grid.get_ipos(planeVector[0]), x2Grid.get_ipos(planeVector[1])};
}

Plane::PlaneVector Plane::project(const WorldVector& point) const {
  PlaneVector planeVector;
  planeVector[0] = basisVector1.inner(point - offsetPlaneOrigin);
  planeVector[1] = basisVector2.inner(point - offsetPlaneOrigin);
  return planeVector;
}

constexpr std::array<int, 3> Projection::subdivisions;

Projection::Projection(const ProjectionPlane& _projectionPlane, const long& nEnergy)
    : projectionPlane(_projectionPlane) {
  this->nEnergy = nEnergy;

  subcellFraction = 1. / (subdivisions[0] * subdivisions[1] * subdivisions[2]);
}


template<size_t N>
void Projection::recomputeSubcellProjections(const Grid& grid, const std::array<size_t, N>& dataIndex) {
  static_assert(N >= 3);
  bool positionChanged = (dataIndex[0] != gridPosition3D[0]) || (dataIndex[1] != gridPosition3D[1]) ||
                         (dataIndex[2] != gridPosition3D[2]);

  if (positionChanged) {
    vector3D cellCenter;
    vector3D dxSub;
    for (int dir = 0; dir < 3; ++dir) {
      dxSub[dir] = (grid.getEdgR(dir, dataIndex[dir]) - grid.getEdgL(dir, dataIndex[dir])) /
                   (1 + subdivisions[dir]);
    }

    for (int izSub = 0; izSub < subdivisions[2]; ++izSub) {
      cellCenter[2] = grid.getEdgL(2, dataIndex[2]) + dxSub[2] * izSub;
      for (int iySub = 0; iySub < subdivisions[1]; ++iySub) {
        cellCenter[1] = grid.getEdgL(1, dataIndex[1]) + dxSub[1] * iySub;
        for (int ixSub = 0; ixSub < subdivisions[0]; ++ixSub) {
          cellCenter[0] = grid.getEdgL(0, dataIndex[0]) + dxSub[0] * ixSub;
          subcellProjections[ixSub][iySub][izSub] = projectionPlane.project(cellCenter);
        }
      }
    }

    gridPosition3D[0] = dataIndex[0];
    gridPosition3D[1] = dataIndex[1];
    gridPosition3D[2] = dataIndex[2];
  }
}

template <class T, size_t N>
void Projection::project(const Grid& grid, const std::array<size_t, N>& dataIndex,
                         const T& pointData, NumMatrix<T, N - 1>& projectedData) {
  const size_t ix = dataIndex[0];
  const size_t iy = dataIndex[1];
  const size_t iz = dataIndex[2];

  recomputeSubcellProjections(grid, dataIndex);

  std::array<int, N - 1> projectionIndex;
  for (int iData = 3; iData < N; ++iData) {
    projectionIndex[iData - 1] = dataIndex[iData];
  }

  for (int izSub = 0; izSub < subdivisions[2]; ++izSub) {
    for (int iySub = 0; iySub < subdivisions[1]; ++iySub) {
      for (int ixSub = 0; ixSub < subdivisions[0]; ++ixSub) {
        auto projectedPosition = subcellProjections[ixSub][iySub][izSub];

        if (!projectionPlane.contains(projectedPosition)) {
          continue;
        }

        ProjectionPlane::PlaneIndices planeIndices =
            projectionPlane.getPlaneIndices(projectedPosition);

        projectionIndex[0] = planeIndices.first;
        projectionIndex[1] = planeIndices.second;

        const double subCellVolume = subcellFraction * grid.get_CellVolume(ix, iy, iz);
        const double pixelArea = projectionPlane.x1Grid.get_del(projectionIndex[0]) *
                                 projectionPlane.x2Grid.get_del(projectionIndex[1]);
        const double volumeAreaRatio = subCellVolume / pixelArea;

        projectedData[projectionIndex.data()] += volumeAreaRatio * pointData;
      }
    }
  }
}

// ----------------------------------------------------------------------------------
// Explicit instantiations

template NumMatrix<double, 1> LineOfSight::integrate<double, 4>(const InterpolationBox& box,
                                                                const NumMatrix<double, 4>&,
                                                                const double&) const;

template NumMatrix<double, 2> LineOfSight::integrate<double, 5>(const InterpolationBox& box,
                                                                const NumMatrix<double, 5>&,
                                                                const double&) const;

template void Projection::project<double, 5>(const Grid& grid, const std::array<size_t, 5>& dataIndex,
                                  const double& pointData, NumMatrix<double, 4>& projectedData);
