#ifndef PROJECTION_H
#define PROJECTION_H

#include "Grid1D.H"
#include "marray.H"
#include "matrix.H"

class Plane {
 public:
  /** Vector in the basis of the world coordinate system */
  using WorldVector = vector3D;

  /** Vector in the basis of the plane coordinate system */
  using PlaneVector = marray<double, 2>;
  using AxisRange = marray<double, 2>;

  Plane(const WorldVector& basisVector1, const WorldVector& basisVector2, const AxisRange& range1,
        const AxisRange& range2, const WorldVector& offsetPlaneOrigin = {0, 0, 0});

  double intersectionDistance(const WorldVector& origin, const WorldVector& direction) const;
  PlaneVector intersect(const WorldVector& origin, const WorldVector& direction) const;
  bool contains(const PlaneVector&) const;
  bool intersectedBy(const WorldVector& origin, const WorldVector& direction) const;

  WorldVector convert(const PlaneVector&) const;
  PlaneVector project(const WorldVector&) const;

  std::pair<WorldVector, WorldVector> getBasis() const {
    return std::make_pair(basisVector1, basisVector2);
  }

 protected:
  Plane(){};

 private:
  WorldVector basisVector1, basisVector2;
  /** Implicitly computed */
  WorldVector normalVector;
  AxisRange range1, range2;
  vector3D offsetPlaneOrigin;
};

#include <vector>

/** Implements the sides of a box */
class Box {
 public:
  using AxisRange = marray<double, 2>;
  Box(const vector3D& axisVector1, const vector3D& axisVector2, const vector3D& axisVector3,
      const AxisRange& axisRange1, const AxisRange& axisRange2, const AxisRange& axisRange3);

  const std::vector<Plane> getSides() const;

 private:
  std::vector<Plane> sides;
};

#include "grid.H"

/** Implements the cell identification in a gridded box */
class GridBox : public Box {
 public:
  GridBox(const Grid&);

  using GridIndex = marray<std::size_t, 3>;
  GridIndex getIndex(const vector3D&) const;
  vector3D getPosition(const GridIndex&) const;

  using NeighbourList = std::vector<GridIndex>;
  NeighbourList getNeighbours(const vector3D&) const;

 private:
  const Grid& grid;
};

#include <type_traits>

/* Implements the tri-linear interpolation of tensorial data on a 3d grid */
class InterpolationBox : public GridBox {
 public:
  InterpolationBox(const Grid&);

  using WeightList = std::vector<double>;
  using InterpolationList = std::pair<NeighbourList, WeightList>;

  virtual InterpolationList getInterpolationList(const vector3D&) const = 0;

  template <typename T, int N, typename... Args>
  inline T interpolate(const NumMatrix<T, N>& data, const vector3D& position, Args&&... idx) const {
    const auto list = getInterpolationList(position);
    return interpolate(data, list, std::forward<Args>(idx)...);
  }

  template <typename T, int N, typename... Args>
  inline T interpolate(const NumMatrix<T, N>& data, const InterpolationList& interpolationList,
                       Args&&... idx) const {
    static std::array<int, N - 3> tensorIndex;
    tensorIndex = {std::forward<Args>(idx)...};
    return interpolate(data, interpolationList, tensorIndex);
  }

  template <typename T, int N>
  inline T interpolate(const NumMatrix<T, N>& data, const InterpolationList& interpolationList,
                       const std::array<int, N - 3> tensorIdx) const {
    static std::array<int, N> index;
    std::copy(tensorIdx.begin(), tensorIdx.end(), index.begin() + 3);
    const auto& neighbours = interpolationList.first;
    const auto& weights = interpolationList.second;
    return std::inner_product(neighbours.begin(), neighbours.end(), weights.begin(), 0.,
                              std::plus<T>(), [&](const GridIndex& neighbour, const T& weight) {
                                if (weight == 0.) return 0.;
                                std::copy(neighbour.begin(), neighbour.end(), index.begin());
                                return data[index.data()] * T(weight);
                              });
  }

  // requires c++14
  // template <typename T, typename U, int N>
  // T interpolate(const NumMatrix<T, N>& data, const InterpolationList& interpolationList,
  //               std::array<U, N - 3> idx) const {
  //   return interpolate(data, interpolationList, idx, std::make_index_sequence<N - 3>{});
  // }

  // template <typename T, int N, typename... Args>
  // T interpolate(const NumMatrix<T, N>& data, const InterpolationList& interpolationList,
  //               Args&&... idx) const {
  //   const auto& [neighbours, weights] = interpolationList;
  //   return std::inner_product(neighbours.begin(), neighbours.end(), weights.begin(), 0.,
  //                             std::plus<T>(), [&](const GridIndex& neighbour, const T& weight) {
  //                               if (weight == 0.) return 0.;
  //                               return data(neighbour[0], neighbour[1], neighbour[2],
  //                                           std::forward<Args>(idx)...) *
  //                                      T(weight);
  //                             });
  // }

 private:
  // requires c++14
  // template <typename T, int N, typename U, std::size_t... indices>
  // T interpolate(const NumMatrix<T, N>& data, const InterpolationList& interpolationList,
  //               std::array<U, N - 3> idx, std::index_sequence<indices...>) const {
  //   return interpolate(data, interpolationList, idx[indices]...);
  // }
};

class TrilinearBox : public InterpolationBox {
 public:
  TrilinearBox(const Grid&);
  InterpolationList getInterpolationList(const vector3D&) const final;
};

class PiecewiseConstantBox : public InterpolationBox {
 public:
  PiecewiseConstantBox(const Grid&);
  InterpolationList getInterpolationList(const vector3D&) const final;
};

class LineOfSight {
 public:
  using AxisRange = marray<double, 2>;

  const vector3D origin, direction;
  LineOfSight(const vector3D& origin, const vector3D& direction);

  AxisRange rangeInsideBox(const Box& box) const;

  template <class T, int N>
  NumMatrix<T, N - 3> integrate(const InterpolationBox& box, const NumMatrix<T, N>& data,
                                const double& dx) const;
};

class ProjectionPlane : public Plane {
 public:
  ProjectionPlane(const WorldVector& basisVector1, const WorldVector& basisVector2,
                  const Grid1D& x1Grid, const Grid1D& x2Grid,
                  const WorldVector& offsetPlaneOrigin = {0, 0, 0});
  static ProjectionPlane newFromDirection(const Grid1D& x1Grid, const Grid1D& x2Grid,
                                          const WorldVector& direction, const WorldVector& up,
                                          const WorldVector& offset = {0, 0, 0});

  const Grid1D x1Grid, x2Grid;

  using PlaneIndices = std::pair<int, int>;
  PlaneIndices getPlaneIndices(const PlaneVector&) const;

 private:
  static std::pair<WorldVector, WorldVector> computeBasis(const WorldVector& direction,
                                                          const WorldVector& up);
};

class Projection {
 private:
  ProjectionPlane projectionPlane;
  long nEnergy;

  // Subdivision for the integration
  static constexpr std::array<int, 3> subdivisions = {3, 3, 3};
  double subcellFraction;


  template<size_t N>
  void recomputeSubcellProjections(const Grid&, const std::array<size_t, N>&);

  std::array<
      std::array<std::array<ProjectionPlane::PlaneVector, subdivisions[2]>, subdivisions[1]>,
      subdivisions[0]>
      subcellProjections;

  std::array<int, 3> gridPosition3D = {0, 0, std::numeric_limits<int>::max()};

 public:
  Projection(const ProjectionPlane& projectionPlane, const long& nEnergy);

  template <class T, size_t N>
  void project(const Grid& grid, const std::array<size_t, N>& dataIndex, const T& pointData,
                         NumMatrix<T, N - 1>& projectedData);
};

#endif