#include <stdlib.h>
#include <fstream>
#include <matrix.H>

#define ISODD(i) ((i)&1)     

#include <sys/types.h>
#include <sys/stat.h>   
#include <utime.h>

#include <fenv.h>
#include <signal.h>

#include "gridgen.H"
#include "EmissionHandler.H"
#ifdef parallel
#include "mpi.h"
#include "MpiErrorHandler.H"
#endif

cronos_ostream ccout(std::cout, 0);
cronos_ostream ccerr(std::cerr, 0);

using namespace std;

int main(int argc, char *argv[])
{
#ifdef parallel
	int ntasks;
	MPI_Init(&argc, &argv);
	MPI_Errhandler errhandler;
	MPI_Errhandler_create(& MPI_Errorhandler_cronos, & errhandler);
	MPI_Comm_set_errhandler(MPI_COMM_WORLD, errhandler);
	MPI_Comm_size(MPI_COMM_WORLD, &ntasks);
	cout << " Number of tasks: " << ntasks << endl;

	ccout.checkRank();
	ccerr.checkRank();
#endif
	/*
	  Additional arguments:
	  (a) Programm name, reload_step
	  (b) data directory, Program name, reload_step
	*/

	int reload_step;
	if (argc == 4) {
		setenv("poub",argv[1],1);
		setenv("pname",argv[2],1);
		reload_step = atoi(argv[3]);
	} else {
		cerr << "Wrong arguments" << endl;
		exit(1.);
	}

	feclearexcept(FE_ALL_EXCEPT);  		// Clear all exceptions
	signal(SIGFPE, SIG_IGN);       		// Ignore Floating point errors (zero division)
	signal(SIGSEGV, stacktraceHandler); // install the stacktracer for SEGFAULTS
	signal(SIGABRT, stacktraceHandler); // install the stacktracer for aborts (e.g. by failed asserts)


	ProjParameters.read_file();

	string emissionCat = string(argv[1]) + "/" + string(argv[2]) + ".emission";
	ParameterFileReader emissionParameters(emissionCat);
	emissionParameters.read_file();

	// Overwrite proj settings with the emission settings
	for (const auto & pair: emissionParameters.get_all()) {
		ProjParameters.set(pair.first, pair.second);
	}

	Data gdata;
	Environment solver(gdata);

	// Init as usual
	gdata.time = 0.;
	try {
		solver.RKSolver->init(gdata, *solver.gfunc, *solver.Problem);
	} catch (CException exep) {
		solver.Abort(gdata, exep);
	}

	// reload data from hdf5 float file
	solver.LoadData_flt(gdata, reload_step);

	// Main task
	EmissionHandler emissionHandler(emissionParameters);
	solver.Problem->reset_User(gdata);

	try {
		emissionHandler.work(gdata, *solver.Problem, solver.Problem->particleTransport);
	} catch (CException exep) {
		solver.Abort(gdata, exep);
	}

	emissionHandler.dataout(gdata);


#ifdef parallel
	MPI_Finalize();
#endif

	return 0;
}










