#ifndef M_ARRAY_H
#define M_ARRAY_H


#include <algorithm>
#include <cmath>
#include <numeric>
#include <memory>
#include <iostream>

/**
 * @brief std::array wrapper that allows basic numerical computations
 */
template <class T, std::size_t N>
class marray : public std::array<T, N> {
 public:
  template <typename Other>
  marray(Other&& other) : std::array<T, N>({std::forward<std::array<T, N> >(other)}) {}

  template <typename... Args>
  marray(Args&&... args) : std::array<T, N>({T(std::forward<Args>(args))...}) {}

  inline friend marray operator*(const T& scalar, const marray& array) {
    marray tmp;
    std::transform(array.begin(), array.end(), tmp.begin(),
                   [&](const T element) { return scalar * element; });
    return tmp;
  }

  inline friend marray operator*(const marray& array, const T& scalar) { return scalar * array; }

  inline friend marray operator/(const marray& array, const T& scalar) {
    marray tmp;
    std::transform(array.begin(), array.end(), tmp.begin(),
                   [&](const T element) { return element / scalar; });
    return tmp;
  }

  inline friend marray operator+(const marray& a, const marray& b) {
    marray tmp(b);
    std::transform(a.begin(), a.end(), tmp.begin(), tmp.begin(), std::plus<T>());
    return tmp;
  }

  inline friend marray operator-(const marray& a, const marray& b) {
    marray tmp(b);
    std::transform(a.begin(), a.end(), tmp.begin(), tmp.begin(), std::minus<T>());
    return tmp;
  }

  inline friend marray operator*(const marray& a, const marray& b) {
    marray tmp(b);
    std::transform(a.begin(), a.end(), tmp.begin(), tmp.begin(), std::multiplies<T>());
    return tmp;
  }

  inline friend marray operator/(const marray& a, const marray& b) {
    marray tmp(b);
    std::transform(a.begin(), a.end(), tmp.begin(), tmp.begin(), std::divides<T>());
    return tmp;
  }

  inline marray& operator/=(const T& scalar) {
    std::transform(this->begin(), this->end(), this->begin(),
                   [&](const T element) { return element / scalar; });
    return *this;
  }

  inline friend bool operator==(const marray& a, const marray& b) {
    return std::equal(a.begin(), a.end(), b.begin());
  }

  inline T norm() const {
    T quadsum = std::accumulate(this->begin(), this->end(), 0.,
                                [](T sum, const T& element) { return sum + element * element; });
    return sqrt(quadsum);
  }

  inline marray normalised() const { return *this / this->norm(); }

  inline marray cross(const marray& v2) const {
    static_assert(N == 3, "The cross product is only available for 3D arrays");
    const marray& v1 = *this;
    marray tmp;
    tmp[0] = v1[1] * v2[2] - v1[2] * v2[1];
    tmp[1] = v1[2] * v2[0] - v1[0] * v2[2];
    tmp[2] = v1[0] * v2[1] - v1[1] * v2[0];
    return tmp;
  }

  inline T inner(const marray& v2) const {
    return std::inner_product(this->begin(), this->end(), v2.begin(), 0.);
  }
};

template <class T, std::size_t N>
std::ostream& operator<<(std::ostream& os, const marray<T, N>& array) {
  os << "{ ";
  for (const auto& x : array) {
    os << x << " ";
  }
  os << "}";
  return os;
}

using vector3D = marray<double, 3>;
using vector2D = marray<double, 2>;

using index3D = marray<long, 3>;
using index2D = marray<long, 2>;

#endif
