

#include "Projection.H"
#include "catch.hpp"

TEST_CASE("2D Diagonal Projection", "[ProjectionPlane]") {
  const vector3D direction = {1., 0., 1.};
  ProjectionPlane plane(direction, {0., 0., 1.});

  SECTION("basis") {
    const auto basis = plane.getBasis();

    SECTION("is normalised") {
      REQUIRE(basis.first.norm() == Approx(1.));
      REQUIRE(basis.second.norm() == Approx(1.));
    }

    SECTION("is orthogonal") {
      REQUIRE(basis.first.inner(direction) == Approx(0.));
      REQUIRE(basis.second.inner(direction) == Approx(0.));
      REQUIRE(basis.first.inner(basis.second) == Approx(0.));
    }

    SECTION("orientation") {
      REQUIRE(basis.first == vector3D{0, 1, 0});
      REQUIRE(basis.second == vector3D{-1, 0, 1}.normalised());
    }
  }

  SECTION("y axis") {
    vector2D proj = plane.project({0, 1, 0});
    REQUIRE(proj == vector2D{1, 0});
  }

  SECTION("z axis") { vector2D proj = plane.project({0, 0, 1}); }
}

TEST_CASE("3D Diagonal Projection", "[ProjectionPlane]") {
  const vector3D direction = {1., 1., 1.};
  ProjectionPlane plane(direction, {0., 0., 1.});

  SECTION("basis") {
    const auto basis = plane.getBasis();

    SECTION("is normalised") {
      REQUIRE(basis.first.norm() == Approx(1.));
      REQUIRE(basis.second.norm() == Approx(1.));
    }

    SECTION("is orthogonal") {
      REQUIRE(basis.first.inner(direction) == Approx(0.));
      REQUIRE(basis.second.inner(direction) == Approx(0.));
      REQUIRE(basis.first.inner(basis.second) == Approx(0.));
    }

    SECTION("orientation") {
      REQUIRE(basis.first == vector3D{-1, 1, 0}.normalised());
      REQUIRE(basis.second == vector3D{-1, -1, 2}.normalised());
    }
  }

  SECTION("x unit vector") {
    vector2D proj = plane.project({1, 0, 0});
    REQUIRE(proj == vector2D{-1. / sqrt(2.), -1. / sqrt(6.)});
  }

  SECTION("y unit vector") {
    vector2D proj = plane.project({0, 1, 0});
    REQUIRE(proj == vector2D{1. / sqrt(2.), -1. / sqrt(6.)});
  }

  SECTION("z unit vector") {
    vector2D proj = plane.project({0, 0, 1});
    REQUIRE(proj == vector2D{0., 2. / sqrt(6.)});
  }
}

TEST_CASE("intersection", "[Plane]") {
  Plane plane({1, 0, 0}, {0, 1, 0}, {0, 1}, {0, 1}, {0, 0, 1});

  SECTION("Straight") {
    auto intersection = plane.intersect({0.5, 0.5, 0}, {0, 0, 1});
    REQUIRE(plane.contains(intersection) == true);

    intersection = plane.intersect({-0.5, 0.5, 0}, {0, 0, 1});
    REQUIRE(plane.contains(intersection) == false);
  }

  SECTION("Inclined") {
    auto intersection = plane.intersect({0.5, 0.5, 0}, {0.1, 0.1, 0.7});
    REQUIRE(plane.contains(intersection) == true);

    intersection = plane.intersect({-0.5, 0.5, 0}, {0.1, 0.1, 0.7});
    REQUIRE(plane.contains(intersection) == false);
  }

  SECTION("parallel") {
    auto intersection = plane.intersect({0.5, 0.5, 0}, {1, 0, 0});
    REQUIRE(plane.contains(intersection) == false);
  }

  SECTION("On plane") {
    auto intersection = plane.intersect({0.5, 0.5, 1}, {0, 0, 1});
    REQUIRE(plane.contains(intersection) == true);

    intersection = plane.intersect({0.5, 0.5, 1}, {1, 0, 0});
    REQUIRE(plane.contains(intersection) == false);
  }
}

TEST_CASE("box intersection", "[LineOfSight]") {
  Box box({1, 0, 0}, {0, 1, 0}, {0, 0, 1}, {0, 1}, {0, 1}, {0, 1});

  SECTION("Straight") {
    LineOfSight los({0, 0, 2}, {0, 0, -1});
    auto range = los.rangeInsideBox(box);
    REQUIRE(range == decltype(range){1, 2});
  }

  SECTION("Diagonal over corner") {
    LineOfSight los({2, 2, 2}, vector3D({-1, -1, -1}).normalised());
    auto range = los.rangeInsideBox(box);
    REQUIRE(range[0] == Approx(sqrt(3.)));
    REQUIRE(range[1] == Approx(2. * sqrt(3.)));
  }

  SECTION("Inside") {
    LineOfSight los({0.5, 0.5, 0.5}, {0, 0, 0.5});
    auto range = los.rangeInsideBox(box);
    REQUIRE(range[0] == Approx(-1.));
    REQUIRE(range[1] == Approx(1.));
  }

  SECTION("missing box") {
    LineOfSight los({3, 3, 3}, {0, 0, 0.5});
    auto range = los.rangeInsideBox(box);
    REQUIRE(range[0] == 0);
    REQUIRE(range[1] == 0);
  }
}

TEST_CASE("Standard cartesian 2x2x2", "[GridBox]") {
  double xb[3] = {0, 0, 0};
  double xe[3] = {2, 2, 2};
  int Nx[3] = {4, 4, 4};
  Grid grid = Grid::newInDomainGrid(xb, xe, Nx);
  GridBox box(grid);

  SECTION("center") {
    vector3D position = {0.5, 0.5, 0.5};
    auto neighbours = box.getNeighbours(position);

    REQUIRE(neighbours[7] == GridBox::GridIndex{0, 0, 0});
    REQUIRE(neighbours[6] == GridBox::GridIndex{1, 0, 0});
    REQUIRE(neighbours[5] == GridBox::GridIndex{0, 1, 0});
    REQUIRE(neighbours[4] == GridBox::GridIndex{1, 1, 0});
    REQUIRE(neighbours[3] == GridBox::GridIndex{0, 0, 1});
    REQUIRE(neighbours[2] == GridBox::GridIndex{1, 0, 1});
    REQUIRE(neighbours[1] == GridBox::GridIndex{0, 1, 1});
    REQUIRE(neighbours[0] == GridBox::GridIndex{1, 1, 1});
  }
}

TEST_CASE("standard cartesian 2x2x2", "[TrilinearBox]") {
  double xb[3] = {0, 0, 0};
  double xe[3] = {1, 1, 1};
  int Nx[3] = {2, 2, 2};
  Grid grid = Grid::newInDomainGrid(xb, xe, Nx);

  NumMatrix<double, 4> data(Index::set(0, 0, 0, 0), Index::set(1, 1, 1, 1));
  data.clear();
  data(0, 0, 0, 0) = 3.;

  SECTION("Trilinear") {
    TrilinearBox box(grid);

    SECTION("weights center") {
      vector3D position = {0.5, 0.5, 0.5};
      const auto interpolationList = box.getInterpolationList(position);
      const auto& neighbours = interpolationList.first;
      const auto& weights = interpolationList.second;

      for (const auto& weight : weights) {
        REQUIRE(weight == Approx(1. / 8.));
      }
    }

    SECTION("weights corner") {
      vector3D position = {0.75, 0.75, 0.75};
      const auto interpolationList = box.getInterpolationList(position);
      const auto& neighbours = interpolationList.first;
      const auto& weights = interpolationList.second;

      REQUIRE(weights[0] == Approx(1.));
      REQUIRE(weights[1] == Approx(0.));
      REQUIRE(weights[2] == Approx(0.));
      REQUIRE(weights[3] == Approx(0.));
      REQUIRE(weights[4] == Approx(0.));
      REQUIRE(weights[5] == Approx(0.));
      REQUIRE(weights[6] == Approx(0.));
      REQUIRE(weights[7] == Approx(0.));
    }

    SECTION("reconstruction corner") {
      vector3D position = {0.25, 0.25, 0.25};
      REQUIRE(box.interpolate(data, position, 0) == Approx(3.));
      REQUIRE(box.interpolate(data, position, 1) == Approx(0.));
    }

    SECTION("reconstruction center") {
      vector3D position = {0.5, 0.5, 0.5};
      REQUIRE(box.interpolate(data, position, 0) == Approx(3. / 8.));
      REQUIRE(box.interpolate(data, position, 1) == Approx(0.));
    }
  }

  SECTION("Piecewise Constant") {
    PiecewiseConstantBox box(grid);

    SECTION("weights center") {
      vector3D position = {0.5, 0.5, 0.5};
      const auto interpolationList = box.getInterpolationList(position);
      const auto& neighbours = interpolationList.first;
      const auto& weights = interpolationList.second;

      REQUIRE(neighbours.size() == 1);
      REQUIRE(weights.size() == 1);
      REQUIRE(weights[0] == Approx(1.));
    }

    SECTION("weights corner") {
      vector3D position = {0.75, 0.75, 0.75};
      const auto interpolationList = box.getInterpolationList(position);
      const auto& neighbours = interpolationList.first;
      const auto& weights = interpolationList.second;

      REQUIRE(neighbours.size() == 1);
      REQUIRE(weights.size() == 1);
      REQUIRE(weights[0] == Approx(1.));
    }

    SECTION("reconstruction corner") {
      vector3D position = {0.25, 0.25, 0.25};
      REQUIRE(box.interpolate(data, position, 0) == Approx(3.));
      REQUIRE(box.interpolate(data, position, 1) == Approx(0.));
    }

    SECTION("reconstruction center") {
      vector3D position = {0.5, 0.5, 0.5};
      REQUIRE(box.interpolate(data, position, 0) == Approx(0.));
      REQUIRE(box.interpolate(data, position, 1) == Approx(0.));
    }
  }
}

TEST_CASE("standard cartesian 2x2x2 integration", "[LineOfSight]") {
  double xb[3] = {0, 0, 0};
  double xe[3] = {1, 1, 1};
  int Nx[3] = {2, 2, 2};
  Grid grid = Grid::newInDomainGrid(xb, xe, Nx);

  NumMatrix<double, 4> data(Index::set(0, 0, 0, 0), Index::set(1, 1, 1, 1));
  data.clear();

  data(0, 0, 0, 0) = 1.;
  data(1, 0, 0, 0) = 2.;
  data(0, 1, 0, 0) = 4.;
  data(1, 1, 0, 0) = 8.;
  data(0, 0, 1, 0) = 16.;
  data(1, 0, 1, 0) = 32.;
  data(0, 1, 1, 0) = 64.;
  data(1, 1, 1, 0) = 128.;
  data(0, 0, 0, 1) = 256.;
  data(1, 0, 0, 1) = 512.;
  data(0, 1, 0, 1) = 1024.;
  data(1, 1, 0, 1) = 2048.;
  data(0, 0, 1, 1) = 4096.;
  data(1, 0, 1, 1) = 8192.;
  data(0, 1, 1, 1) = 16384.;
  data(1, 1, 1, 1) = 32768.;

  SECTION("corner") {
    LineOfSight los({0.25, -1, 0.25}, {0, 1, 0});
    SECTION("Trilinear") {
      TrilinearBox box(grid);
      auto integral = los.integrate(box, data, 1e-4);
      REQUIRE(integral(0) == Approx(1.25).margin(0.01));
      REQUIRE(integral(1) == Approx((256. + 1024.) / 4.).margin(1));
    }

    SECTION("constant") {
      PiecewiseConstantBox box(grid);
      auto integral = los.integrate(box, data, 1e-4);
      // High error due to first order interpolation
      REQUIRE(integral(0) == Approx(1.25).margin(0.1));
      REQUIRE(integral(1) == Approx(320).margin(0.1));
    }
  }
}

TEST_CASE("Projections project data 3D -> 2D", "[Projection]") {
  // lets consider a 2x2x2 cube

  // LinearGrid x1GridCube(0, 1, Nx, 0);
  // LinearGrid x2GridCube(0, 1, Nx, 0);
  // LinearGrid x3GridCube(0, 1, Nx, 0);

  const double xb[3] = {-5, -5, -5};
  const double xe[3] = {5, 5, 5};
  const int Nx = 5;
  const double dx = (xe[0] - xb[0]) / Nx;
  const int Nxs[3] = {Nx, Nx, Nx};

  Grid grid = Grid::newInDomainGrid(xb, xe, Nxs);

  // Define some data
  NumMatrix<double, 4> data(Index::set(0, 0, 0, 0), Index::set(Nx, Nx, Nx, 0));
  data(1, 1, 1, 0) = 1;
  data(3, 3, 3, 0) = 2;
  data(3, 1, 1, 0) = 4;
  data(1, 3, 1, 0) = 8;
  data(1, 1, 3, 0) = 16;

  data(3, 1, 3, 0) = 32;
  data(3, 3, 1, 0) = 64;
  data(1, 3, 3, 0) = 128;

  LinearGrid x1GridProj(-7, 7, 7, 0);
  LinearGrid x2GridProj(-7, 7, 7, 0);

  SECTION("+x direction") {
    vector3D direction = {1, 0, 0};
    vector3D up = {0, 0, 1};
    ProjectionPlane plane(direction, up);

    Projection proj(plane, x1GridProj, x2GridProj, 1);
    NumMatrix<double, 3> projectedData = proj.project(grid, data, 1);

    REQUIRE(projectedData(2, 2, 0) == Approx((1 + 4) * dx).margin(0.01));
    REQUIRE(projectedData(2, 4, 0) == Approx((16 + 32) * dx).margin(0.01));
    REQUIRE(projectedData(4, 4, 0) == Approx((128 + 2) * dx).margin(0.1));
    REQUIRE(projectedData(4, 2, 0) == Approx((8 + 64) * dx).margin(0.1));
  }
}

TEST_CASE("Tensor 2 projection") {
  // lets consider a 2x2x2 cube

  const double xb[3] = {0, 0, 0};
  const double xe[3] = {2, 2, 2};
  const int Nx = 2;
  const double dx = (xe[0] - xb[0]) / Nx;
  const int Nxs[3] = {Nx, Nx, Nx};

  Grid grid = Grid::newInDomainGrid(xb, xe, Nxs);

  // Define some data
  NumMatrix<double, 5> data(Index::set(0, 0, 0, 0), Index::set(Nx, Nx, Nx, 1, 1));

  for (int iz = 0; iz < Nx; ++iz) {
    for (int iy = 0; iy < Nx; ++iy) {
      for (int ix = 0; ix < Nx; ++ix) {
        data(ix, iy, iz, 0, 0) = 1.;
        data(ix, iy, iz, 0, 1) = 2.;
        data(ix, iy, iz, 1, 0) = 4.;
        data(ix, iy, iz, 1, 1) = 8.;
      }
    }
  }

  LinearGrid x1GridProj(0, 2, 1, 0);
  LinearGrid x2GridProj(0, 2, 1, 0);

  SECTION("+x direction") {
    vector3D direction = {1, 0, 0};
    vector3D up = {0, 0, 1};
    ProjectionPlane plane(direction, up);

    Projection proj(plane, x1GridProj, x2GridProj, 1);
    NumMatrix<double, 4> projectedData = proj.project(grid, data, 1);

    REQUIRE(projectedData(0, 0, 0, 0) == Approx(1.));
    REQUIRE(projectedData(0, 0, 0, 1) == Approx(2.));
    REQUIRE(projectedData(0, 0, 1, 0) == Approx(4.));
    REQUIRE(projectedData(0, 0, 1, 1) == Approx(8.));
  }
}