#ifndef CRONOS_PROBLEM_H
#define CRONOS_PROBLEM_H = 1

#include <string>
#include <map>
#include "matrix.H"
#include "util.H"
#include "data.H"
#include "constants.H"
#include "movie.H"

#include "vector.H"
#include "fields_local.H"
#include "normalisation.H"
#include "units.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)
#include "ParticleTransport.H"
#endif

#include "physical_constants.H"

using namespace std;

/*
  Base class for specific numerical problem. Derived classes have at
  least to supply the routine:

  init_fields(q, i, j, k)

  for initialisation of all relevant quantities and a corresponding
  constructor.

  Routines:

  bc_User(gdata, omb, dir, top, q, rim)
  bc_periodic(gdata, Send, omb, q, dir, top)
  bc_Flux(gdata, HXpmWE, HXpmSN, HXpmBT);
  bc_Flux(gdata, HXpmWE, direction);
  bc_emf(gdata, Ex, Ey, Ez);
  src_User(gdata, nom, nom_user)
  computePhystest(gdata)
  writePhystest(gdata, outfile)
  computeFluct(gdata, EkinFluct, EmagFluct)
  get_NameMovie(gdata, q)
  set_MovieOut(gdata, q)

  are optional and are used as to do nothing, when not redeclared in
  derived class.
*/

class ProblemType {
public:
	ProblemType(const Data &gdata);
	virtual void init_fields(Data &, int [], int []) = 0;
	// virtual void init_fields(Data &, int q, int [], int []);
	// virtual REAL init_fields(Data &, int q, REAL xPos, REAL yPos, REAL zPos);
	void init_fields(Data &);
	void init_fields(Data &, int);
	virtual void init_problem(Data &) {};
	virtual string get_Name();   
	int get_Type();
	virtual void bc_User(Data &, NumMatrix<REAL,3> &, int, int, int, int);
	virtual void bc_periodic(Data &,
	                         NumMatrix<double,3> &, NumMatrix<double,3> &,
	                         int, int, bool){};
	virtual void bc_Flux(Data &, NumMatrix<REAL,3> [],
	                     NumMatrix<REAL,3> [], NumMatrix<REAL,3> []) {}
	virtual void bc_Flux(Data &, NumMatrix<REAL,3> [], int) {}
	virtual void bc_emf(Data &, NumMatrix<double,3> &,
	                    NumMatrix<double,3> &, NumMatrix<double,3> &) {}
	virtual REAL c2_iso(Data &, const REAL &, const REAL &, const REAL &);
	virtual void src_User(Data &, NumMatrix<REAL,3> [], NumMatrix<REAL,3> []) {}
	virtual void src_Add(Data &, bool [N_OMINT]) {}
	virtual void reset_User(Data &) {}
	virtual bool isInternalBoundary(Data&, int, int, int) {return false;}
	virtual bool setInternalBoundary(Data& gdata, int ix, int iy, int iz) {return isInternalBoundary(gdata, ix, iy, iz);}
	virtual REAL eta(Data &, const REAL &, const REAL &, const REAL &);
	virtual REAL nu(Data &, const REAL &, const REAL &, const REAL &);
	virtual void computePhystest(Data &){}
	virtual void writePhystestInfo(Data &){}
	virtual void writePhystest(Data &, ofstream &){}
	virtual void computeFluct(Data &, double &, double &);
	virtual int checkConvergence(Data &);
	virtual void writeMovie(const Data &gdata, Movie & mov) const ;
	virtual bool force_max(const int &);
	virtual REAL max_Val(const int &);
	virtual bool force_min(const int &);
	virtual REAL min_Val(const int &);
	virtual REAL grid_user_x(REAL ratio);
	virtual REAL grid_user_y(REAL ratio);
	virtual REAL grid_user_z(REAL ratio);
	virtual void ReadFromH5(Hdf5iStream &) {}
	virtual void WriteToH5(Hdf5Stream &) {}
	// Allow a change of the User fields between conservative and
	// characteristic form
	virtual void get_Cons(Data &, cronos::vector<REAL> &Pos,
	                      phys_fields_1D &physField, 
	                      phys_fields_1D &physFieldsUser,
	                      int dir, REAL shift);
	// Routines replaced by TransPrim2Cons et al...
	// virtual void TransConsCharUser(Data &) {}
	// virtual void TransCharConsUser(Data &) {}

	virtual void TransPrim2Cons(Data &) {}
	virtual void TransCons2Prim(Data &) {}


	virtual void get_PhysFluxUser(Data &gdata, cronos::vector<REAL> &iPos,
	                              phys_fields_1D &pf, phys_fields_1D &pfUser,
	                              int dir, REAL shift) {}
	virtual void get_PhysFlux(Data &gdata, cronos::vector<REAL> &iPos,
			phys_fields_1D &pf, int dir, REAL shift) {}

	void name_User(Data &);
	/**Routine to access some internal variables from anywhere (e.g. from an Euler solver)*/
	double get_internalValue(int){return 0.;}
	int get_q(NumMatrix<REAL,3> &, string fluidName="fluid00");
	void set_q(int q, NumMatrix<REAL,3> &);
	void set_Info(bool);
	void set_AsciiOut(bool);
	void set_InfoData(Data &);
	bool checkout(int);
	bool get_Info();
	bool get_AsciiOut();
	virtual ~ProblemType(){}

	normalisation * TrafoNorm;

  /** mass fractions of hydrogen */
  virtual double getMassFractionEP(const Data &gdata, const int ix, const int iy,
                                   const int iz) const {
    return 0;
  }

  /** mass fractions of hydrogen */
  virtual double getMassFractionX(const Data &gdata, const int ix, const int iy,
                                  const int iz) const {
    return 1;
  }

  /** mass fraction helium */
  virtual double getMassFractionY(const Data &gdata, const int ix, const int iy,
                                  const int iz) const {
    return 0;
  }

  /** Default mean particle mass for pure, fully-ionised electron-proton plasma */
  virtual Quantity getMeanParticleMass(const Data &gdata, const int ix, const int iy,
                                       const int iz) const {
    const double massFractionX = getMassFractionX(gdata, ix, iy, iz);
    const double massFractionY = getMassFractionY(gdata, ix, iy, iz);
    const double massFractionEP = getMassFractionEP(gdata, ix, iy, iz);
    const double massFractionZ = 1. - massFractionX - massFractionY - massFractionEP;

    const Quantity meanParticleMass =
        CRONOS_CONSTANTS::AtomicMassUnit /
        (2. * massFractionX + 0.75 * massFractionY + 0.5 * massFractionZ + 2000 * massFractionEP);
    return meanParticleMass;
  }

	REAL gamma;
#if (EOS == POLYTROPIC)
	REAL polytropic_K;
#endif
	REAL rho0;
	bool mag, info, asciiOut, infodata[9];
	int q_rho, q_sx, q_sy, q_sz, q_Bx, q_By, q_Bz, q_Eges, q_Eadd;
	int q_Ax, q_Ay, q_Az;
	int n_omInt, n_omIntAll;

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	ParticleTransport particleTransport;
	virtual void init_particleTransport(Data &) = 0;
#endif

#ifdef USE_VALENCIA
	Valencia * valencia;
#endif
	virtual double getOrbitalPhase() const {return 0.;}

protected:
	REAL nu0, eta0;
	REAL cs2;
	string name;
	int type;
#if(FLUID_TYPE==CRONOS_MULTIFLUID)
	int numFluids;
#endif
	map<std::string, int> fieldNames;
};  

#endif
