#include <iostream>
#include "BoundaryHandler1D.H"
#include <stdlib.h>
#include "math.h"

using namespace std;


BoundaryHandler1D::BoundaryHandler1D(int lowerBoundaryType, int upperBoundaryType) {
	// Types of boundary conditions
	// (0) -> zero	
	// (1) -> periodic
	// (2) -> linear extrapolation
	// (3) -> linear extrapolation + outflow
	// (11) -> log extrapolation (untested)

	bcType.resize(2);
	bcType[0] = lowerBoundaryType;
	bcType[1] = upperBoundaryType;
}

int BoundaryHandler1D::get_bcType(int dir) const {
	return bcType[dir];
}

void BoundaryHandler1D::do_BCsCellAve(NumMatrix<double,1> &spec, const Grid1D &grid, std::function<double(double)> & advectionFunction) const {
	//! Boundary conditions for momentum direction

	for(int top = 0; top <= 1; ++top) {
		switch(bcType[top]) {
				case 0 : bcZero(spec, grid, top); break;
				case 1 : bcPeriodic(spec, grid, top); break;
				case 2 : bcLinearExtrapolation(spec, grid, top); break;
				case 3 : bcLinearExtrapolationOutflow(spec, grid, advectionFunction, top); break;
				case 11 : bcLogExtrapolation(spec, grid, top); break;
				default:
					throw std::logic_error("The boundary handler doesn't know what to do " + to_string(top) + " " + to_string(bcType[top]));
					break;
		}
	}
}

void BoundaryHandler1D::do_BCsSecond(NumMatrix<double,1> &secondOrderParams, const Grid1D &grid, std::function<double(double)> & advectionFunction) const {
	//! Boundary conditions for momentum direction

	for(int top = 0; top <= 1; ++top) {
		switch(bcType[top]) {
				case 0 : bcSecondZero(secondOrderParams, grid, top); break;
				case 1 : bcSecondPeriodic(secondOrderParams, grid, top); break;
				case 2 : bcSecondLinearExtrapolation(secondOrderParams, grid, top); break;
				case 3 : bcSecondLinearExtrapolationOutflow(secondOrderParams, grid, advectionFunction, top); break;
				case 11 : bcSecondLogExtrapolation(secondOrderParams, grid, top); break;
				default:
					throw std::logic_error("The boundary handler doesn't know what to do " + to_string(top) + " " + to_string(bcType[top]));
					break;
		}
	}
}


void BoundaryHandler1D::bcPeriodic(NumMatrix<double,1> &spec, const Grid1D &grid, int top) const {
	assert(top == 0 || top == 1);
	const int mx = grid.get_mx();
	const int Nx = grid.get_Nx();
	const int rim = grid.get_rim();

	if (top == 0) {
		for (int ipos = -1; ipos >= -rim; --ipos) {
			spec(ipos) = spec(Nx + ipos);
		}
	} else {
		for (int ipos = 0; ipos < rim; ++ipos) {
			spec(Nx + ipos) = spec(ipos);
		}
	}
}


void BoundaryHandler1D::bcZero(NumMatrix<double,1> &spec, const Grid1D &grid, int top) const {
	assert(top == 0 || top == 1);
	const int mx = grid.get_mx();
	const int rim = grid.get_rim();

	if (top == 0) {
		for (int ipos = -1; ipos >= -rim; --ipos) {
			spec(ipos) = 0.;
		}
	} else {
		for (int ipos = mx + 1; ipos <= mx + rim; ++ipos) {
			spec(ipos) = 0.;
		}
	}
}

void BoundaryHandler1D::bcLinearExtrapolation(NumMatrix<double,1> &spec, const Grid1D &grid, int top) const {
	assert(top == 0 || top == 1);
	const int mx = grid.get_mx();
	const int rim = grid.get_rim();

	if (top == 0) {
		for (int ipos = -1; ipos >= -rim; --ipos) {
			const double fac = ((grid.get_del(ipos) + grid.get_del(ipos + 1)) /
						  (grid.get_del(ipos + 1) + grid.get_del(ipos + 2)));
			spec(ipos) = spec(ipos + 1) + fac * (spec(ipos + 1) - spec(ipos + 2));
			
			// Ensure positivity
			spec(ipos) = max(spec(ipos), 0.);
		}
	} else {
		for (int ipos = mx + 1; ipos <= mx + rim; ++ipos) {
			const double fac = ((grid.get_del(ipos) + grid.get_del(ipos - 1)) /
						  (grid.get_del(ipos - 1) + grid.get_del(ipos - 2)));
			spec(ipos) = spec(ipos -  1) + fac * (spec(ipos - 1) - spec(ipos - 2));
			
			// Ensure positivity
			spec(ipos) = max(spec(ipos), 0.);
		}
	}
}

void BoundaryHandler1D::bcLogExtrapolation(NumMatrix<double,1> &spec, const Grid1D &grid, int top) const {
	// (this was not tested)
	assert(top == 0 || top == 1);
	const int mx = grid.get_mx();
	const int rim = grid.get_rim();

	if (top == 0) {
		double xR = 0;
		double xRR = log(grid.get_xCen(1));

		double specR = 0;
		double specRR = log(spec(1));

		for (int ipos = -1; ipos >= -rim; --ipos) {
			xR = log(grid.get_xCen(ipos + 1));
			specR = log(spec(ipos + 1));
			const double s = (specRR - specR)/(xRR - xR);

			spec(ipos) = spec(ipos + 1) * pow(grid.get_xCen(ipos)/grid.get_xCen(ipos + 1), s);

			xRR = xR;
			specRR = specR;
		}
	} else {
		throw logic_error("Not implemented yet");
	}
}

void BoundaryHandler1D::bcLinearExtrapolationOutflow(NumMatrix<double,1> &spec, const Grid1D &grid, std::function<double(double)> & advectionFunction, int top) const {
	assert(top == 0 || top == 1);
	if (top == 0) {
		if (advectionFunction(grid.get_xL(0)) < 0) {
			bcLinearExtrapolation(spec, grid, top);
		} else {
			bcZero(spec, grid, top);
		}
	} else {
		if (advectionFunction(grid.get_xL(grid.get_mx() + 1)) > 0) {
			bcLinearExtrapolation(spec, grid, top);
		} else {
			bcZero(spec, grid, top);
		}
	}
}



void BoundaryHandler1D::bcSecondPeriodic(NumMatrix<double,1> &secondOrderParams, const Grid1D &grid, int top) const {
}


void BoundaryHandler1D::bcSecondZero(NumMatrix<double,1> &secondOrderParams, const Grid1D &grid, int top) const {
	assert(top == 0 || top == 1);
	const int mx = grid.get_mx();
	const int rim = grid.get_rim();

	if (top == 0) {
		for (int ipos = -1; ipos >= -rim; --ipos) {
			secondOrderParams(ipos) = 0.;
		}
	} else {
		for (int ipos = mx + 1; ipos <= mx + rim; ++ipos) {
			secondOrderParams(ipos) = 0.;
		}
	}
}

void BoundaryHandler1D::bcSecondLinearExtrapolation(NumMatrix<double,1> &secondOrderParams, const Grid1D &grid, int top) const {
}

void BoundaryHandler1D::bcSecondLogExtrapolation(NumMatrix<double,1> &secondOrderParams, const Grid1D &grid, int top) const {
}

void BoundaryHandler1D::bcSecondLinearExtrapolationOutflow(NumMatrix<double,1> &secondOrderParams, const Grid1D &grid, std::function<double(double)> & advectionFunction, int top) const {
	bcSecondZero(secondOrderParams, grid, top);
}
