#include "constants.H"
#if (PARTICLE_TRANSPORT == CRONOS_ON)

#include "DiffusionConvection1D.H"
#define SLOW_COMMUNICATION CRONOS_OFF

DiffusionConvection1D::DiffusionConvection1D(
		Data & gdata,
		ProblemType & problem
		):
		particleTransport(problem.particleTransport),
		boundaryHandler1D(0, 0),
		momentumLossFunction(problem.particleTransport),
		semilagra(SemilagrangeSolver::TimeSolver::Second, SemilagrangeSolver::SpaceSolver::Linear)
{
	ccout << " Setting up Convection/Diffusion Solver " << endl;

	lo = 0;
	hi = 0;

	nParticleSpecies = value("nParticleSpecies");
	timerParticle = new CronosTimer[nParticleSpecies];

	ccout << " Adding Semilagrange Solver " << endl;
	if (value_exists("SemilagrangeLimiter")) {
		ccout << "Using SemilagrangeLimiter: " << svalue("SemilagrangeLimiter") << endl;
		semilagra.set_slopeLimiter(svalue("SemilagrangeLimiter"));
	} else {
		ccout << "Using SemilagrangeLimiter: minmod" << endl;
		semilagra.set_slopeLimiter("minmod");
	}
	semilagra.set_advectionFunction(momentumLossFunction);

#ifndef NDEBUG
	semilagra.disableWarnings();
#endif

	initialized = false;
}


DiffusionConvection1D::~DiffusionConvection1D()
{
	delete [] timerParticle;
}

void DiffusionConvection1D::init(
		Data & gdata,
		gridFunc & gfunc,
		ProblemType & problem
	)
{
	//
	//	Initializing buffers
	//

	for (int iParticle = 0; iParticle < nParticleSpecies; ++iParticle) {
		Particle & particle = problem.particleTransport.get_particle(iParticle);

		int rim = 1;
		grids.emplace_back(particle.get_momentumEdge(particle.getLo()), particle.get_momentumEdge(particle.getHi()), particle.getNBins(), rim, 2);
		Grid1D & grid = grids[iParticle];
//		NumMatrix<double,1> edges(grid.get_xL());
//		edges(edges.getLow(0)) = max(1., edges(edges.getLow(0))); /** Make sure that gamma is not lower than 1 */

		/** There was something fishy with the above log grid generation - do it manually for now */
		NumMatrix<double,1> edges(Index::set(-rim), Index::set(particle.getHi()+1+rim));
		for (int ip = 0; ip <= particle.getHi()+1; ip++) {
			edges(ip) = particle.get_momentumEdge(ip);
		}
		edges(-rim) = 1.;
		edges(particle.getHi()+1+rim) =  edges(particle.getHi() + 1) * ( edges(particle.getHi() + 1) / edges(particle.getHi()) );

		grid.set_grid(edges, rim, particle.getNBins());

//		int lo = particle.getLo();
//		int hi = particle.getHi();
//
//		// To include extrapolated cells
//		lo -= 1;
//		hi += 1;
//
//		// Buffers for the parameters of the second order characteristics in momentum space
//		momentumLoss_c1_int.emplace_back(Index::set(lo), Index::set(hi+1));
//		momentumLoss_c2.emplace_back(Index::set(lo), Index::set(hi+1));
//		momentumLoss_c2_old.emplace_back(Index::set(lo), Index::set(hi+1));
//		momentumLoss_c2_old_save.emplace_back(Index::set(lo, -B,-B,-B), Index::set(hi+1, gdata.mx[0]+B,gdata.mx[1]+B,gdata.mx[2]+B));
	}

	initialized = true;
}

void DiffusionConvection1D::restart(
		Data & gdata,
		gridFunc & gfunc,
		ProblemType & problem
	)
{
	return init(gdata, gfunc, problem);
}


void DiffusionConvection1D::compute_characteristica(
		Data & gdata,
		ParticleTransport & particleTransport,
		const NumMatrix<REAL,1> & momentumEdges,
		int iParticle,
		int ix, int iy, int iz
		)
{
	assert(initialized);
	// Compute parameters for the second order characteristic curves
	for (int iMom=lo; iMom <= hi + 1; iMom++) {
		momentumLoss_c1_int[iParticle](iMom)	= particleTransport.get_momentumLoss_c1_int(momentumEdges(iMom));
		momentumLoss_c2[iParticle](iMom) 		= particleTransport.get_momentumLoss_c2(momentumEdges(iMom));
		momentumLoss_c2_old[iParticle](iMom) 	= momentumLoss_c2_old_save[iParticle](iMom, ix, iy, iz);
		momentumLoss_c2_old_save[iParticle]
							(iMom, ix, iy, iz) 	= momentumLoss_c2[iParticle](iMom);
	}
}



REAL DiffusionConvection1D::singlestep(
		Data & gdata,
		gridFunc & gfunc,
		ProblemType & problem
		)
{
	assert(initialized);
	ccout << "......................................................" << endl;

	NumMatrix<REAL,1> momentumEdges;
	NumMatrix<REAL,1> momentumCenters;
	NumMatrix<REAL,1> velocities;			// equivalent to p_dot
	NumMatrix<REAL,1> momentumSpectrum;
	NumMatrix<REAL,1> momentumCenters_propagated;

	// Bins for which the injection boundary condition
	// will be applied - hardcoded for the time being
	int d[1] = {1};
	NumMatrix<int, 1> injectionBins(d);
	injectionBins(0) = 0;


	// ----------------------------------------------------------------
	// Starting the clock:
	// ----------------------------------------------------------------
	timerFullStep.reset();
	timerFullStep.start();

	timerBoundaries.reset();

	// ----------------------------------------------------------------
	// Resetting some debug fields:
	// ----------------------------------------------------------------
	gdata.negativeParticles.clear();

	// ----------------------------------------------------------------
	//  Change to Conserved Quantities
	// ----------------------------------------------------------------
	problem.particleTransport.TransPrim2Cons(gdata);

	// ----------------------------------------------------------------
	//  Apply SemiLagrangian scheme to the
	//	transport equation of every particle species
	// ----------------------------------------------------------------
	for (int iParticle=0; iParticle < nParticleSpecies; ++iParticle) {
		timerParticle[iParticle].reset();
		timerParticle[iParticle].start();

		// Setting particle of interest
		Particle & particle = problem.particleTransport.set_currentParticle(iParticle);
		Grid1D & grid = grids[iParticle];

		ccout << "  SLSTEP for Particle " << iParticle << ": " << particle.get_particleName() << endl;

		momentumSpectrum.resize(grid.getLow(), grid.getHigh());
		momentumSpectrum.clear();
		assert(grid.getLo() <= particle.getLo());
		assert(grid.getHi() >= particle.getHi());

		semilagra.resize(grid);
		semilagra.reset_reachedSubcycleDepth();

#if (SLOW_COMMUNICATION == CRONOS_OFF)
		for (int ix = 0; ix <= gdata.mx[0]; ix++){
		for (int iy = 0; iy <= gdata.mx[1]; iy++){
		for (int iz = 0; iz <= gdata.mx[2]; iz++){
#else
		for (int ix = -1; ix <= gdata.mx[0]+1; ix++){
		for (int iy = -1; iy <= gdata.mx[1]+1; iy++){
		for (int iz = -1; iz <= gdata.mx[2]+1; iz++){
#endif
			problem.particleTransport.compute_constants(gdata, ix, iy, iz);

			// Get the particle spectrum from the cronos user fields
			for (int ip = 0; ip <= grid.get_mx(); ip++) {
				momentumSpectrum(ip) = gdata.om_user[ip + particle.getOmb()](ix, iy, iz);
			}

			// Inject Particles
			int injection = problem.particleTransport.injectParticleSpectrum(
											gdata, ix, iy, iz,
											grid.get_xL(), momentumSpectrum
											);
			gdata.injection(ix, iy, iz) = injection;

			// Solve the momentum advection
			if ( injection <=0 ) {			
				//const double maxSpec = momentumSpectrum.max_norm();
				bool containsParticles = false;

				for (int ip = 0; ip <= grid.get_mx(); ip++) {
					if (momentumSpectrum(ip) > 0) {
						containsParticles = true;
						break;
					}
				}

				if ( containsParticles ) {
					problem.particleTransport.compute_transportConstants(gdata, ix, iy, iz);

					// Estimate the relevant momentum range to save computation time
					// this criteria should be revisited at some point, since they assume losses to be dominant
					int ipMax = grid.get_mx();
					for (int ip = ipMax; ip > 0; ip--) {
						if ( momentumSpectrum(ip) > 1e-20 * momentumSpectrum(ip - 1) ) {
							break;
						} else {
							ipMax = ip;
						}
					}

					semilagra.set_computationLimits(0, ipMax);

					semilagra.singlestep(grid, boundaryHandler1D, momentumSpectrum, gdata.dt);

				}
			}

			// write the particle spectrum to the cronos user fields
			for (int ip = 0; ip <= grid.get_mx(); ip++) {
//						gdata.om_user[ip + particle.getOmb()](ix, iy, iz) = momentumSpectrum(ip);
				// Todo This is just a hack -> The scheme should not produce negative numbers
				// gdata.om_user[ip + particle.getOmb()](ix, iy, iz) = max(momentumSpectrum(ip), 0.);
				if(momentumSpectrum(ip) < 0) {
					gdata.negativeParticles(ix, iy, iz) += 1;
				}
				
				if(!std::isfinite(momentumSpectrum(ip))) {
//					cout << "Particle containing non-finite value: setting zero at " << ip << " " << momentumSpectrum(ip) << std::endl;
					gdata.om_user[ip + particle.getOmb()](ix, iy, iz) = 0;
					continue;
				}
				gdata.om_user[ip + particle.getOmb()](ix, iy, iz) = momentumSpectrum(ip);
			}
		}
		}
		}
		// ----------------------------------------------------------------
		//   Performance estimate for single particle
		// ----------------------------------------------------------------
		timerParticle[iParticle].stop();
	}

	// ----------------------------------------------------------------
	//  Change to primitive quantities
	// ---------------------------------------------------------------
	 problem.particleTransport.TransCons2Prim(gdata);


	// ----------------------------------------------------------------
	//   Synchronisation (in order to get the timing right)
	// ----------------------------------------------------------------
	double timeSLStep = timerFullStep.dtime();
	double maxSLStep = timeSLStep;
	double meanSLStep = timeSLStep;
#ifdef parallel
	MPI_Allreduce(&timeSLStep, &maxSLStep, 1, MPI_DOUBLE, MPI_MAX, gdata.comm3d);
	MPI_Allreduce(&timeSLStep, &meanSLStep, 1, MPI_DOUBLE, MPI_SUM, gdata.comm3d);
	meanSLStep /= gdata.ntasks;
#endif

	// ----------------------------------------------------------------
	//   Apply Boundary Conditions for particle fields
	// ----------------------------------------------------------------
	timerBoundaries.start();
#if (SLOW_COMMUNICATION == CRONOS_OFF)
	//gfunc.boundary(gdata, problem);
	gfunc.BoundaryHandler.boundary(gdata, problem, gdata.om_user, N_OMINT_PARTICLE);
#endif
	timerBoundaries.stop();
	timerFullStep.stop();

	ccout << "   Reached Subcycle Depth " << semilagra.get_reachedSubcycleDepth() << endl;

	if(problem.get_Info() && problem.checkout(5)) {
		for (int iParticle=0; iParticle < nParticleSpecies; ++iParticle) {
		ccout << "   Time needed for particle    " << iParticle << ": " << timerParticle[iParticle].time() << " [" << problem.particleTransport.get_particle(iParticle).get_particleName() << "]" << endl;
		}
		ccout << "               for SL step : max = " << maxSLStep << " mean = " << meanSLStep << endl;
		ccout << "               for SL boundaries: " << timerBoundaries.time()  << endl;
		ccout << "               for full SL step : " << timerFullStep.time() << endl;
	}

	return 1;
}


void DiffusionConvection1D::do_linear_extrapolation(
		const NumMatrix<REAL,1> & cellCenters,
		NumMatrix<REAL,1> & values
		)
{
//	values(lo) = values(lo+1) + (values(lo+2) - values(lo+1))/(cellCenters(lo+2) - cellCenters(lo+1)) * (cellCenters(lo) - cellCenters(lo+1));
//	values(hi) = values(hi-1) + (values(hi-1) - values(hi-2))/(cellCenters(hi-1) - cellCenters(hi-2)) * (cellCenters(hi) - cellCenters(hi-1));
	values(lo) = 0.;
	values(hi) = 0.;
}

void DiffusionConvection1D::do_powerlaw_extrapolation(
		const NumMatrix<REAL,1> & cellCenters,
		NumMatrix<REAL,1> & values
		)
{
//	values(lo) = values(lo+1) + (values(lo+2) - values(lo+1))/(cellCenters(lo+2) - cellCenters(lo+1)) * (cellCenters(lo) - cellCenters(lo+1));
//	values(hi) = values(hi-1) + (values(hi-1) - values(hi-2))/(cellCenters(hi-1) - cellCenters(hi-2)) * (cellCenters(hi) - cellCenters(hi-1));

	values(lo) = values(lo+1) * pow(values(lo+1)/values(lo+2),
			log(cellCenters(lo)/cellCenters(lo+1))/log(cellCenters(lo+1)/cellCenters(lo+2)));
	values(hi) = values(hi-1) * pow(values(hi-1)/values(hi-2),
			log(cellCenters(hi)/cellCenters(hi-1))/log(cellCenters(hi-1)/cellCenters(hi-2)));
	cout << values(lo+1)/values(lo+2) << " " << values(lo) << " - " << values(hi-1)/values(hi-2) << " " << values(hi) << endl;
}



void DiffusionConvection1D::addConstants_toH5(
		Data &gdata,
		Hdf5Stream &h5out
		)
{
	h5out.AddGlobalAttr("nParticles", nParticleSpecies);

	hid_t group = h5out.AddGroup("Data/Particles");

	for (int iParticle=0; iParticle < nParticleSpecies; ++iParticle) {
		Particle & particle = particleTransport.get_particle(iParticle);

		NumMatrix<float, 1> momentumCenters_flt = convertMatrix<float>(particle.get_momentumCenter());
		NumMatrix<float, 1> momentumEdges_flt = convertMatrix<float>(particle.get_momentumEdge());

		h5out.AddGlobalAttr("particle" + ToString(iParticle, 2) + "_bins", particle.getNBins());
		h5out.AddGlobalAttr("particle" + ToString(iParticle, 2) + "_bins", particle.getNBins(), group);

		h5out.Write1DMatrix("momentumCenters_" + ToString(iParticle, 2), momentumCenters_flt, group);
		h5out.Write1DMatrix("momentumEdges_" + ToString(iParticle, 2), momentumEdges_flt, group);
		h5out.AddGlobalAttr("particle" + ToString(iParticle, 2) + "_type", particle.get_particleType(), group);
		h5out.AddGlobalAttr("particle" + ToString(iParticle, 2) + "_name", particle.get_particleName(), group);
	}

	// Now close the new group
	h5out.CloseGroup(group);

}


MomentumLossFunction::MomentumLossFunction(
		ParticleTransport & _particleTransport
		) :
	particleTransport(_particleTransport)
{
}


REAL MomentumLossFunction::_get_value()
{
	REAL loss;
	particleTransport.get_momentumLoss(x, loss);
	return loss;
}

#endif
