//
// Created by dave on 06.03.20.
//

#ifndef CRONOSGAMMARAYBINARIES_EMISSIVITYSPECTRUMINTEGRATOR_H
#define CRONOSGAMMARAYBINARIES_EMISSIVITYSPECTRUMINTEGRATOR_H

#include <gsl/gsl_integration.h>
#include <functional>
#include "PowerlawReconstruction.H"
#include "matrix.H"

class EmissivitySpectrumIntegrator {
 public:
  struct integrationLimits {
    int gammaLoIdx; /** lowest cell edge above which we still have particles */
    int gammaHiIdx; /** highest cell edge below which we still have particles */
    double gammaLo;
    double gammaHi;
  };

  virtual ~EmissivitySpectrumIntegrator() {}

  virtual double integrate(const LogarithmicGrid& grid,
                           const NumMatrix<REAL, 1>& values,
                           const integrationLimits& limits,
                           const std::function<double(double)>& emissivity) = 0;
};

class MidpointIntegrator : public EmissivitySpectrumIntegrator {
 public:
  double integrate(const LogarithmicGrid& grid,
                   const NumMatrix<REAL, 1>& values,
                   const integrationLimits& limits,
                   const std::function<double(double)>& emissivity)
  {
    double photonEmissivity = 0;

    for (int ip = limits.gammaLoIdx; ip < limits.gammaHiIdx; ++ip) {
      photonEmissivity +=
          emissivity(grid.get_xCen(ip)) * values(ip) * grid.get_del(ip);
    }

    return photonEmissivity;
  }
};

class GaussPowerlawIntegrator : public EmissivitySpectrumIntegrator {
 public:
  GaussPowerlawIntegrator()
  {
    // For the particle gamma Integration
    const int nGauss = value_exists("InverseComptonIntegrationNGauss")
                           ? value("InverseComptonIntegrationNGauss")
                           : 10;

    fixedWorkspace = gsl_integration_fixed_alloc(gsl_integration_fixed_legendre,
                                                 nGauss, 0, 1, 0, 0);
  }

  ~GaussPowerlawIntegrator() { gsl_integration_fixed_free(fixedWorkspace); }

  double integrate(const LogarithmicGrid& grid,
                   const NumMatrix<REAL, 1>& values,
                   const integrationLimits& limits,
                   const std::function<double(double)>& emissivity)
  {
    powerlawReconstruction.setData(grid, values, limits.gammaLoIdx,
                                   limits.gammaHiIdx);
    return gaussFixed(powerlawReconstruction, emissivity, limits);
  }

 private:
  PowerlawReconstruction powerlawReconstruction;

  /** Applies a gaussian quadrature to integrate the
   * log transformed diff emissivity */
  double gaussFixed(PowerlawReconstruction& reconstruction,
                    const std::function<double(double)>& diffEmissivity,
                    const integrationLimits& limits)
  {
    // log transformed integrand (y = log(x))
    // int_a^b f(x) dx -> int_log(a)^log(b) f(exp(y)) x dy
    // plus mapping of log(a) -> 0, log(b) -> 1 for gauss integrator (as defined
    // by the workspace)
    static auto integrand = [&](double z) -> double {
      double particleLorentz_COM =
          powf(limits.gammaHi / limits.gammaLo, z) * limits.gammaLo;
      return diffEmissivity(particleLorentz_COM) * particleLorentz_COM *
             reconstruction(particleLorentz_COM);
    };

    const double result = gaussLegendre(integrand);
    return result * log(limits.gammaHi / limits.gammaLo);
  }

  gsl_integration_fixed_workspace* fixedWorkspace;

  template <typename T>
  double gaussLegendre(T& integrand) const
  {
    static auto evaluator = [](double x, void* fun) {
      return (*static_cast<T*>(fun))(x);
    };
    static gsl_function F = {.function = evaluator};
    F.params = &integrand;
    double result;
    gsl_integration_fixed(&F, &result, fixedWorkspace);
    return result;
  }
};

#endif  // CRONOSGAMMARAYBINARIES_EMISSIVITYSPECTRUMINTEGRATOR_H
