#include "Grid1D.H"
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>

using namespace std;

Grid1D::Grid1D(double xb, double xe, int Nx, int rim, int type) {
	switch(type) { // Add other grids 
		case 0:	*this = LinearGrid(xb, xe, Nx, rim);
						return;
		case 2:	*this = LogarithmicGrid(xb, xe, Nx, rim);
						return;
	}

	is_set = false;
	this->Nx = Nx;
	mx = Nx-1;
	this->rim = rim; // Number of boundary cells
	this->xb = xb;
	this->xe = xe;
	this->type = type; // Type of grid
	compute_depedentVars();
	build_grid();
}


// Grid1D::Grid1D(const Grid1D& grid) {
// 	is_set = false;
// 	Nx = grid.Nx;
// 	mx = grid.mx;
// 	rim = grid.rim;
// 	xb = grid.xb;
// 	xe = grid.xe;
// 	type = grid.type;
// 	compute_depedentVars();
// 	build_grid();
// }

Grid1D::Grid1D() {
	is_set = false;
	Nx = 0;
	mx = 0;
	rim = 0;
	xb = 0.;
	xe = 0.;
	type = -99;
}

void Grid1D::compute_depedentVars() {
	ipos_get = 0;
	Len = xe - xb;

	del = Len/(1.*Nx);
	idel = 1./del;
}


void Grid1D::set_grid(NumMatrix<double,1> Edges, int rim, int Nx) {
	this->Nx = Nx;
	this->rim = rim;
	mx = Nx-1;

	lo = Index::set(-rim);
	hi = Index::set(mx + rim);
	cellCentres.resize(lo, hi);
	cellWidths.resize(lo, hi);
	cellEdges.resize(lo, hi + 1);

	assert(Edges.getHigh(0) >= cellCentres.getHigh(0) &&
	       Edges.getLow(0)  <= cellCentres.getLow(0));

	cellEdges = Edges;

	// Compute cell centres & widhts:
	for(int ipos=-rim; ipos<=mx+rim; ++ipos) {
		cellCentres(ipos) = 0.5*(cellEdges(ipos) + cellEdges(ipos+1));
		cellWidths(ipos) = cellEdges(ipos+1) - cellEdges(ipos);
	}

	is_set = true;
}


void Grid1D::build_grid() {
	lo = Index::set(-rim);
	hi = Index::set(mx + rim);
	cellCentres.resize(lo, hi);
	cellWidths.resize(lo, hi);
	cellEdges.resize(lo, hi + 1);

	if(type == 0) { // Linear grid
		assert(false); // This case should be caught in the constructor
		return;

	} else if(type == 1) { // Sinusoidal grid

		// Loop over all cell edges:
		for(int ipos=-rim; ipos<=mx+rim+1; ++ipos) {
			cellEdges(ipos) = xb + del*(ipos + 2.*sin(2.*M_PI*ipos/(1.*Nx)));
		}

	} else if(type == 2) { // Log grid
		assert(false); // This case should be caught in the constructor
		return;
	}

	// std::cout << " mx: " << mx << std::endl;
	// exit(2);

	// Compute cell centres & widhts:
	for(int ipos=-rim; ipos<=mx+rim; ++ipos) {
		cellCentres(ipos) = 0.5*(cellEdges(ipos) + cellEdges(ipos+1));
		cellWidths(ipos) = cellEdges(ipos+1) - cellEdges(ipos);
	}

	is_set = true;
}

double Grid1D::get_del(int ipos) const {
	// return del;
	assert(is_set);
	return cellWidths(ipos);
}

double Grid1D::get_xb() const {
	assert(is_set);
	return xb;
}

double Grid1D::get_xe() const {
	assert(is_set);
	return xe;
}

double Grid1D::get_Len() const {
	assert(is_set);
	return Len;
}

double Grid1D::get_xCen(int ipos) const {
	// return (ipos+0.5)*del + xb;
	assert(is_set);
	return cellCentres(ipos);
}

double Grid1D::get_xL(int ipos) const {
	// return ipos*del + xb;
	assert(is_set);
	return cellEdges(ipos);
}

// int Grid1D::get_ipos(const double pos) const {
// 	assert(is_set);
// 	bool goLeft(false);
// 	if(pos < get_xL(ipos_get)) {
// 		goLeft = true;
// 	}

// 	if(goLeft) {
// 		while(pos < get_xL(ipos_get)) {
// 			ipos_get--;
// 		}
// 	} else {
// 		while(pos > get_xL(ipos_get)) {
// 			ipos_get++;
// 		}
// 		ipos_get--;
// 	}
// 	return ipos_get;
// }

int Grid1D::get_mx() const {
	assert(is_set);
	return mx;
}

int Grid1D::get_Nx() const {
	assert(is_set);
	return Nx;
}

int Grid1D::get_rim() const {
	assert(is_set);
	return rim;
}

const Index & Grid1D::getLow() const {
	assert(is_set);
	return lo;
}

const Index & Grid1D::getHigh() const {
	assert(is_set);
	return hi;
}

int Grid1D::getLo() const {
	assert(is_set);
	return lo(0);
}

int Grid1D::getHi() const {
	assert(is_set);
	return hi(0);
}

const NumMatrix<double,1> & Grid1D::get_xCen() const {
	assert(is_set);
	return cellCentres;
}

const NumMatrix<double,1> & Grid1D::get_del() const {
	assert(is_set);
	return cellWidths;
}

const NumMatrix<double,1> & Grid1D::get_xL() const {
	assert(is_set);
	return cellEdges;
}

int Grid1D::get_ipos(const double pos) const {
	assert(is_set);

  long lower = getLo();
  long upper = getHi() + 1;
  long mid = 0;

  while (upper - lower > 1) {
    mid = (upper + lower) / 2;
		const double midPos = cellEdges(mid);
    if (pos < midPos) {
      upper = mid;
    } else {
      lower = mid;
    }
  }
	return lower;
 }

std::unique_ptr<Grid1D> Grid1D::clone() const {
	return std::unique_ptr<Grid1D>(new Grid1D(*this));
}

LogarithmicGrid::LogarithmicGrid(double xb, double xe, int Nx, int rim)
    : Grid1D() {
  fx = pow(xe / xb, 1. / Nx);
  fxLog = log(fx);

  Index lo = Index::set(-rim);
  Index hi = Index::set(Nx - 1 + rim);
  Index hiEdge = Index::set(Nx + rim);

  NumMatrix<double, 1> cellEdgesLocal(lo, hiEdge);
  cellEdgesLog.resize(lo, hiEdge);
  cellCentersLog.resize(lo, hi);

  const double xbLog = log(xb);
  for (int ip = -rim; ip <= Nx + rim; ip++) {
    cellEdgesLocal(ip) = xb * pow(fx, ip);
    cellEdgesLog(ip) = xbLog + ip * fxLog;
  }

  set_grid(cellEdgesLocal, rim, Nx);

  for (int ip = -rim; ip <= Nx + rim - 1; ip++) {
    cellCentersLog(ip) = log(get_xCen(ip));
  }
}

LogarithmicGrid::LogarithmicGrid() : Grid1D() {}

#define EXACT true
int LogarithmicGrid::get_ipos(const double pos) const {
  const int idx = int(logf(pos / get_xL(0)) / fxLog);

#if (EXACT == true)
	if (pos < get_xL(idx)) return idx - 1;
	else if (pos > get_xL(idx + 1)) return idx + 1;
	else return idx;
#else
	return idx;
#endif
}
#undef EXACT

double LogarithmicGrid::get_xC_log(const int ix) const {
  return cellCentersLog(ix);
};
double LogarithmicGrid::get_xL_log(const int ix) const {
  return cellEdgesLog(ix);
};

std::unique_ptr<Grid1D> LogarithmicGrid::clone() const {
	return std::unique_ptr<LogarithmicGrid>(new LogarithmicGrid(*this));
}


LinearGrid::LinearGrid(double xb, double xe, int Nx, int rim)
    : Grid1D() {
	const double length = xe - xb;
	dx = length / double(Nx);

  Index lo = Index::set(-rim);
  Index hi = Index::set(Nx - 1 + rim);
  Index hiEdge = Index::set(Nx + rim);

  NumMatrix<double, 1> cellEdgesLocal(lo, hiEdge);
  for (int ip = -rim; ip <= Nx + rim; ip++) {
    cellEdgesLocal(ip) = ip * dx + xb;
  }

  set_grid(cellEdgesLocal, rim, Nx);
}

LinearGrid::LinearGrid() : Grid1D() {}

#define EXACT true
int LinearGrid::get_ipos(const double pos) const {
  const int idx = int((pos - get_xL(0)) / dx);

#if (EXACT == true)
	if (pos < get_xL(idx)) return idx - 1;
	else if (pos > get_xL(idx + 1)) return idx + 1;
	else return idx;
#else
	return idx;
#endif
}
#undef EXACT

std::unique_ptr<Grid1D> LinearGrid::clone() const {
	return std::unique_ptr<LinearGrid>(new LinearGrid(*this));
}