#include "Interpolation1D.H"

using namespace std;

/*------------------------------------------------------------
  1D Interpolations
  ------------------------------------------------------------*/
Interpolation1D::Interpolation1D(
		const NumMatrix<REAL, 1> &_supports,
		const NumMatrix<REAL, 1> &_values,
		const int extrapolationType,
		bool differentSize
):
		supports(_supports),
		values(_values)
{
	if (!differentSize)
	if (supports.getSize() != values.getSize()) {
		throw std::invalid_argument("Interpolation1D: supports and values size do not match");
	}

	lo = supports.getLow(0);
	hi = supports.getHigh(0);
	xLow = supports(lo);
	xHigh = supports(hi);

	loVal = values.getLow(0);
	hiVal = values.getHigh(0);

	set_extrapolation(extrapolationType);
}

void Interpolation1D::set_extrapolation(const int extrapolationType) {
	switch(extrapolationType) {
		case -1  :
			// Constant
			compute_extrapolation = &Interpolation1D::extrapolate_error;
			break;

		case 0  :
			// Constant
			compute_extrapolation = &Interpolation1D::extrapolate_constant;
			break;

		case 2  :
			// Periodic continuous
			compute_extrapolation = &Interpolation1D::extrapolate_periodicContinuous;
			break;

		case 3  :
			// Powerlaw extrapolation
			compute_extrapolation = &Interpolation1D::extrapolate_powerlaw;
			break;

		default :
			cerr << " No such extrapolation type " << extrapolationType << endl;
			throw std::invalid_argument("No such extrapolation type");
	}
}

REAL Interpolation1D::operator()(
		REAL x
) {
	if (x < xLow || x > xHigh) {
		return (this->*compute_extrapolation)(x);
	} else {
		return compute_interpolation(x);
	}
}

REAL Interpolation1D::extrapolate_error(
		REAL x
) {
	throw std::runtime_error("No extrapolation available");
}

REAL Interpolation1D::extrapolate_constant(
		REAL x
) {
	if (x < xLow) {
		return values(loVal);
	} else {
		return values(hiVal);
	}
}

REAL Interpolation1D::extrapolate_periodicContinuous(
		REAL x
) {
	/*
	 * This function extrapolates continuously and "periodically"
	 * The signal is copied to the left or to the right (periodically), but is offsetted to remain continuous
	 */

	REAL xDiff = xHigh - xLow;
	REAL valueDiff = values(hi) - values(lo);

	REAL xRel = (x - xLow) / xDiff;

	if(valueDiff < 0) {
		valueDiff *= -1.;
	}

	int shifts = floor(xRel);
	REAL xMid = x - shifts * xDiff;

	assert(xMid >= xLow && xMid <= xHigh);
	return shifts * valueDiff + compute_interpolation(xMid);
}

/*------------------------------------------------------------
  Divided-Differences Interpolation Scheme
  ------------------------------------------------------------*/

void DividedDifferences::calc_divdiff(	const NumMatrix<REAL, 1>& support,
										const NumMatrix<REAL, 1>& values,
										NumMatrix<REAL, 1>& coeff)
{
	coeff.resize(values.getDims());
	for (int i=coeff.getLow(0); i <= coeff.getHigh(0); i++){
		coeff(i) = values(i);
	}

	for (int step=1; step < coeff.getSize(); step++) {
		for (int i=coeff.getHigh(0); i >= coeff.getLow(0) + step; i--){
			coeff(i) = (coeff(i) - coeff(i-1)) / (supports(i) - supports(i-step));
		}
	}
}


REAL DividedDifferences::compute_interpolation(const REAL x)
{
	return horner_scheme(coeff, supports, x);
}


REAL DividedDifferences::horner_scheme(
		const NumMatrix<REAL, 1> &_coeff,
		const NumMatrix<REAL, 1> &_supports,
		const REAL x)
const {
	REAL val = _coeff(_coeff.getHigh(0));
	for (int i=_coeff.getHigh(0) - 1; i >= _coeff.getLow(0); i--){
		val *= x - _supports(i);
		val += _coeff(i);
	}
	return val;
}


/*------------------------------------------------------------
  Piecewise Linear Interpolation Scheme
  ------------------------------------------------------------*/

REAL PiecewiseLinear::compute_interpolation(const REAL x)
{
	long lower = supports.getLow(0);
	long upper = supports.getHigh(0);
	long mid = 0;

	while (upper - lower > 1) {
		mid = long(0.5 * (upper + lower));
		if (supports(mid) > x) {
			upper = mid;
		} else {
			lower = mid;
		}
	}

	return values(lower) + (x - supports(lower)) / (supports(upper) - supports(lower)) * (values(upper) - values(lower));
}

/*------------------------------------------------------------
  Piecewise Constant Interpolation Scheme
  ------------------------------------------------------------*/

REAL PiecewiseConstant::compute_interpolation(const REAL x)
{
	long lower = supports.getLow(0);
	long upper = supports.getHigh(0);
	long mid = 0;

	while (upper - lower > 1) {
		mid = long(0.5 * (upper + lower));
		if (supports(mid) > x) {
			upper = mid;
		} else {
			lower = mid;
		}
	}

	return values(lower);


	if (x == supports(supports.getHigh(0))) {
		return values(lower);
	} else if (x == supports(upper)) {
		return values(upper);
	} else {
		return values(lower);
	}
}


/*------------------------------------------------------------
  Piecewise Powerlaw Interpolation Scheme
  ------------------------------------------------------------*/

REAL PiecewisePowerlaw::compute_interpolation(const REAL x)
{
	long lower = supports.getLow(0);
	long upper = supports.getHigh(0);
	long mid = 0;

	while (upper - lower > 1) {
		mid = long(0.5 * (upper + lower));
		if (supports(mid) > x) {
			upper = mid;
		} else {
			lower = mid;
		}
	}

	if (x == supports(lower)) {
		return values(lower);
	} else if (x == supports(upper)) {
		return values(upper);
	} else if (values(upper) == 0) {
		return 0;
	} else if (values(lower) == 0) {
		return 0;
	} else {
		const REAL s = log(values(upper)/values(lower)) / log(supports(upper)/supports(lower));
		return  values(lower) * pow(x / supports(lower), s);
	}
}

REAL Interpolation1D::extrapolate_powerlaw(
		REAL x
) {
	if (x < xLow) {
		static const REAL s = log(values(lo+1)/values(lo)) / log(supports(lo+1)/supports(lo));
		return values(lo) * pow(x / supports(lo), s);
	} else {
		static const REAL s = log(values(hi)/values(hi-1)) / log(supports(hi)/supports(hi-1));
		return values(hi) * pow(x / supports(hi), s);
	}
}

/*------------------------------------------------------------
  Piecewise Parabolic Interpolation Scheme
  ------------------------------------------------------------*/

REAL PiecewiseParabolic::compute_interpolation(const REAL x)
{
	long lower = supports.getLow(0);
	long upper = supports.getHigh(0);
	long mid = 0;

	while (upper - lower > 1) {
		mid = long(0.5 * (upper + lower));
		if (supports(mid) > x) {
			upper = mid;
		} else {
			lower = mid;
		}
	}

	return values(lower)
		   + (x - supports(lower)) / (supports(upper) - supports(lower)) * (values(upper) - values(lower))
		   +  (x - supports(lower)) * (x - supports(upper)) * parabCoeffs(lower);
}
