#ifndef INTERPOLATION_1D_H
#define INTERPOLATION_1D_H

#ifdef MODULE_TEST
#define REAL double
#else
#include "constants.H"
#endif

#include "matrix.H"
#include <cassert>
#include <stdexcept>

/*------------------------------------------------------------
  Onedimensional Interpolations for the Semilagrange Solver
  ------------------------------------------------------------*/

class Interpolation1D {
public:
	Interpolation1D(
			const NumMatrix<REAL, 1> &_supports,
			const NumMatrix<REAL, 1> &_values,
			const int extrapolationType = -1,
			bool differentSize = false
	);

	~Interpolation1D() {};

	REAL operator()(
			REAL x
	);

protected:
	virtual REAL compute_interpolation(
			REAL
	) = 0;

	const NumMatrix<REAL, 1>& supports;
	const NumMatrix<REAL, 1>& values;

private:
	int lo, hi;
	REAL xLow, xHigh;

	int loVal, hiVal;

	virtual void set_extrapolation(const int extrapolationType);

	REAL (Interpolation1D::*compute_extrapolation)(REAL);
	REAL extrapolate_error(REAL x);
	REAL extrapolate_constant(REAL x);
	REAL extrapolate_periodicContinuous(REAL x);
	REAL extrapolate_powerlaw(REAL x);


};


class DividedDifferences: public Interpolation1D {
public:
	DividedDifferences(
			const NumMatrix<REAL, 1> &supports,
			const NumMatrix<REAL, 1> &values,
			const int extrapolationType = -1
	):
			Interpolation1D(supports, values, extrapolationType)
	{
		calc_divdiff(supports, values, coeff);
	};

	const NumMatrix<REAL, 1>& get_coeff(
			void
	) const {return coeff;};

private:
	void calc_divdiff(
			const NumMatrix<REAL, 1>& support,
			const NumMatrix<REAL, 1>& values,
			NumMatrix<REAL, 1>& coeff
	);

	REAL compute_interpolation(
			REAL x
	);

	REAL horner_scheme(
			const NumMatrix<REAL, 1>& coeff,
			const NumMatrix<REAL, 1>& support,
			const REAL x
	) const;

	NumMatrix<REAL, 1> coeff;

};

class PiecewiseConstant: public Interpolation1D {
public:
	PiecewiseConstant(
			const NumMatrix<REAL, 1> &limits,
			const NumMatrix<REAL, 1> &values,
			const int extrapolationType = -1
	):
			Interpolation1D(limits, values, extrapolationType, true)
	{};

private:
	REAL compute_interpolation(
			REAL x
	);
};

class PiecewiseLinear: public Interpolation1D {
public:
	PiecewiseLinear(
			const NumMatrix<REAL, 1> &supports,
			const NumMatrix<REAL, 1> &values,
			const int extrapolationType = -1
	):
			Interpolation1D(supports, values, extrapolationType)
	{};

private:
	REAL compute_interpolation(
			REAL x
	);
};

class PiecewisePowerlaw: public Interpolation1D {
public:
	PiecewisePowerlaw(
			const NumMatrix<REAL, 1> &supports,
			const NumMatrix<REAL, 1> &values,
			const int extrapolationType = -1
	):
			Interpolation1D(supports, values, extrapolationType)
	{};

private:
	REAL compute_interpolation(
			REAL x
	);

};

class PiecewiseParabolic: public Interpolation1D {
public:
	PiecewiseParabolic(
			const NumMatrix<REAL, 1> &supports,
			const NumMatrix<REAL, 1> &values,
			const NumMatrix<REAL, 1> &parabCoeffs,
			const int extrapolationType = -1
	):
			Interpolation1D(supports, values, extrapolationType),
			parabCoeffs(parabCoeffs)
	{};

private:
	REAL compute_interpolation(
			REAL x
	);

	const NumMatrix<REAL, 1>& parabCoeffs;
};


#endif
