#include "Interpolation1D.H"
#include "InverseCompton.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)

#include <fstream>
#include <functional>
#include <typeinfo>

using namespace std;
using namespace CRONOS_CONSTANTS;

std::shared_ptr<InverseComptonEmissionIntegrals> InverseComptonEmissionIntegrals::getInstance(
    std::shared_ptr<RadiationField> radiationField) {
  if (auto radField = std::dynamic_pointer_cast<RadiationFieldFactorized>(radiationField)) {
    return std::make_shared<InverseComptonEmissionIntegralsFactorized>(radField);
  } else {
    throw CException("InverseComptonEmissionIntegrals not implemented for this radiation field");
  }
}

InverseComptonEmissionIntegralsFactorized::InverseComptonEmissionIntegralsFactorized(
    std::shared_ptr<RadiationFieldFactorized> radiationField) {
  factorNumberDensity = radiationField->get_iRadiationFieldFactor_NumberDensity();
  factorEnergy = radiationField->get_iRadiationFieldFactor_Energy();
  factorDirection = radiationField->get_iRadiationFieldFactor_Direction();

  energyIntegrator = IRFFEnergyIntegrator::getIntegrator(factorEnergy);
  directionIntegrator = IRFFDirectionIntegrator::getIntegrator(factorDirection);
}

double InverseComptonEmissionIntegralsFactorized::gEmission(Data& gdata, const int ix, const int iy,
                                                            const int iz,
                                                            const REAL scatteredPhotonEnergy,
                                                            const REAL observerDirection[3],
                                                            const REAL particleLorentz_COM) {
  return factorNumberDensity->numberDensity(gdata, ix, iy, iz) *
         directionIntegrator->integrate(gdata, ix, iy, iz, scatteredPhotonEnergy, observerDirection,
                                        particleLorentz_COM, energyIntegrator);
}

std::shared_ptr<IRFFEnergyIntegrator> IRFFEnergyIntegrator::getIntegrator(
    std::shared_ptr<IRadiationFieldFactor_Energy> energyFactor) {
  if (auto mono = std::dynamic_pointer_cast<IRadiationFieldFactor_Monochromatic>(energyFactor)) {
    return std::make_shared<IRFFEnergyIntegratorMonochromatic>(mono);
  } else {
    throw CException("IC Energy integrator not implemented for " +
                     string(typeid(energyFactor).name()));
  }
}

void IRFFEnergyIntegratorMonochromatic::compute_constants(Data& gdata, const int ix, const int iy,
                                                          const int iz,
                                                          const REAL particleRestEnergy) {
  energyNorm = 1. / particleRestEnergy;

  targetPhotonEnergyNorm = factorMonochromatic->get_photonEnergy(gdata, ix, iy, iz) * energyNorm;

  lorentzTrafo.set_boostVector(
      gdata.om[gdata.fluid.get_q_sx()](ix, iy, iz), gdata.om[gdata.fluid.get_q_sy()](ix, iy, iz),
      gdata.om[gdata.fluid.get_q_sz()](ix, iy, iz), gdata.lorentz(ix, iy, iz));
}

double IRFFEnergyIntegratorMonochromatic::integrate(Data& gdata, const int ix, const int iy,
                                                    const int iz, const REAL scatteredPhotonEnergy,
                                                    const REAL observerDirection[3],
                                                    const double beamDirection[3],
                                                    const REAL particleLorentz_COM) {
  const double scatteredPhotonEnergyNorm = scatteredPhotonEnergy * energyNorm;
  const double dopplerBeamInv = lorentzTrafo.get_dopplerFactor_inv(beamDirection);
  const double dopplerObserverInv = lorentzTrafo.get_dopplerFactor_inv(observerDirection);

  const double scatteredPhotonEnergyNorm_COM = dopplerObserverInv * scatteredPhotonEnergyNorm;
  const double targetPhotonEnergyNorm_COM = dopplerBeamInv * targetPhotonEnergyNorm;

  if (scatteredPhotonEnergyNorm_COM < targetPhotonEnergyNorm_COM) {
    return 0.;
  }

  double muBeamObserver = 0;
  for (int iDir = 0; iDir < 3; ++iDir) {
    muBeamObserver += beamDirection[iDir] * observerDirection[iDir];
  }
  const double muBeamObserver_COM =
      1. + (muBeamObserver - 1.) / (dopplerBeamInv * dopplerObserverInv);

  const double bBeamObserver_COM =
      2. * (1. - muBeamObserver_COM) * targetPhotonEnergyNorm_COM * particleLorentz_COM;
  const double scatteredPhotonEnergyNormMax_COM =
      bBeamObserver_COM / (1. + bBeamObserver_COM) * particleLorentz_COM;

  // delta integration
  const double fEnergy = 1.;
  if (scatteredPhotonEnergyNorm_COM < scatteredPhotonEnergyNormMax_COM) {
    return 1. / dopplerObserverInv * fEnergy / targetPhotonEnergyNorm *
           InverseComptonFunctions::fGeneric(scatteredPhotonEnergyNorm_COM,
                                             targetPhotonEnergyNorm_COM, particleLorentz_COM,
                                             muBeamObserver_COM);
  } else {
    return 0.;
  }
}

std::shared_ptr<IRFFDirectionIntegrator> IRFFDirectionIntegrator::getIntegrator(
    std::shared_ptr<IRadiationFieldFactor_Direction> directionFactor) {
  if (auto beam = std::dynamic_pointer_cast<IRadiationFieldFactor_Beamed>(directionFactor)) {
    return std::make_shared<IRFFDirectionIntegratorBeamed>(beam);
  } else if (auto spherical =
                 std::dynamic_pointer_cast<IRFFD_SphericalCoordinates>(directionFactor)) {
    return std::make_shared<IRFFDirectionIntegratorSphericalCoords>(spherical);
  } else {
    throw CException("IC Direction integrator not implemented for " +
                     string(typeid(directionFactor).name()));
  }
}

double IRFFDirectionIntegratorBeamed::integrate(
    Data& gdata, const int ix, const int iy, const int iz, const REAL scatteredPhotonEnergy,
    const double observerDirection[3], const REAL particleLorentz_COM,
    shared_ptr<IRFFEnergyIntegrator>& energyIntegrator) {
  double beamDirection[3];
  factorBeamed->get_beamDirection(gdata, ix, iy, iz, beamDirection);

  // delta integration
  return energyIntegrator->integrate(gdata, ix, iy, iz, scatteredPhotonEnergy, observerDirection,
                                     beamDirection, particleLorentz_COM);
}

IRFFDirectionIntegratorSphericalCoords::IRFFDirectionIntegratorSphericalCoords(
    std::shared_ptr<IRFFD_SphericalCoordinates> irffd) {
  factorSpherical = irffd;
  gaussLegendreTableAzimuth = gsl_integration_glfixed_table_alloc(nPointsAzimuth);
  gaussLegendreTableDeclination = gsl_integration_glfixed_table_alloc(nPointsDeclination);
}

IRFFDirectionIntegratorSphericalCoords::~IRFFDirectionIntegratorSphericalCoords() {
  gsl_integration_glfixed_table_free(gaussLegendreTableAzimuth);
  gsl_integration_glfixed_table_free(gaussLegendreTableDeclination);
}

double IRFFDirectionIntegratorSphericalCoords::integrate(
    Data& gdata, const int ix, const int iy, const int iz, const REAL scatteredPhotonEnergy,
    const double observerDirection[3], const REAL particleLorentz_COM,
    shared_ptr<IRFFEnergyIntegrator>& energyIntegrator) {
  double muSeed;
  double beamDirection[3];
  double pos[3] = {gdata.getCen_x(ix), gdata.getCen_y(iy), gdata.getCen_z(iz)};

  auto integrandAzimuth = [&](double phiSeed) -> double {
    factorSpherical->get_dir(gdata, pos, muSeed, phiSeed, beamDirection);
    return factorSpherical->fDirection(gdata, pos, muSeed, phiSeed) *
           energyIntegrator->integrate(gdata, ix, iy, iz, scatteredPhotonEnergy, observerDirection,
                                       beamDirection, particleLorentz_COM);
  };

  const double phiMin = factorSpherical->get_phiMin(gdata, pos);
  const double phiMax = factorSpherical->get_phiMax(gdata, pos);

  auto integralAzimuthEqui = [&](double _muSeed) -> double {
    muSeed = _muSeed;
    double sum = phiMin;
    const double dPhi = (phiMax - phiMin) / nPointsAzimuth;
    double phi = 1e-2;
    for (int i = 0; i < nPointsAzimuth; ++i) {
      sum += integrandAzimuth(phi);
      phi += dPhi;
    }
    return sum * dPhi;
  };

  // auto evaluatorAzimuth = [](double x, void* arg) -> double {
  //   return (*static_cast<decltype(integrandAzimuth)*>(arg))(x);
  // };

  // gslIntegrandAzitmuth.function = evaluatorAzimuth;
  // gslIntegrandAzitmuth.params = &integrandAzimuth;

  // auto integralAzimuth = [&](double _muSeed) -> double {
  //   muSeed = _muSeed;
  //   return gsl_integration_glfixed(&gslIntegrandAzitmuth, phiMin, phiMax,
  //   gaussLegendreTableAzimuth);
  // };

  const double muSeedMin = factorSpherical->get_muMin(gdata, pos);
  const double muSeedMax = factorSpherical->get_muMax(gdata, pos);

  if (muSeedMax - muSeedMin < 1e-4 ) {
    factorSpherical->get_dir(gdata, pos, muSeedMin, 0, beamDirection);
    return energyIntegrator->integrate(gdata, ix, iy, iz, scatteredPhotonEnergy, observerDirection,
                                       beamDirection, particleLorentz_COM);
  }

  auto integrandDeclination = [&](double _muSeed) -> double {
    // return integralAzimuth(_muSeed);
    return integralAzimuthEqui(_muSeed);
  };

  auto evaluatorDeclination = [](double x, void* arg) -> double {
    return (*static_cast<decltype(integrandDeclination)*>(arg))(x);
  };

  gslIntegrandDeclination.function = evaluatorDeclination;
  gslIntegrandDeclination.params = &integrandDeclination;

  const double result = gsl_integration_glfixed(&gslIntegrandDeclination, muSeedMin, muSeedMax,
                                                gaussLegendreTableAzimuth);

  assert(!std::isnan(result));
  return result;
}

void InverseComptonEmissionIntegralsFactorized::compute_constants(Data& gdata, const int ix,
                                                                  const int iy, const int iz,
                                                                  const REAL particleRestEnergy) {
  energyIntegrator->compute_constants(gdata, ix, iy, iz, particleRestEnergy);
  directionIntegrator->compute_constants(gdata, ix, iy, iz, particleRestEnergy);

  energyNorm = 1. / particleRestEnergy;

  lorentzTrafo.set_boostVector(
      gdata.om[gdata.fluid.get_q_sx()](ix, iy, iz), gdata.om[gdata.fluid.get_q_sy()](ix, iy, iz),
      gdata.om[gdata.fluid.get_q_sz()](ix, iy, iz), gdata.lorentz(ix, iy, iz));
}

double InverseComptonEmissionIntegralsFactorized::particleLorentzMin(
    Data& gdata, const int ix, const int iy, const int iz, const REAL scatteredPhotonEnergy,
    const REAL observerDirection[3]) const {
  const REAL scatteredPhotonEnergyNorm = scatteredPhotonEnergy * energyNorm;
  const REAL dopplerObserverInv = lorentzTrafo.get_dopplerFactor_inv(observerDirection);
  const REAL scatteredPhotonEnergyNorm_COM = dopplerObserverInv * scatteredPhotonEnergyNorm;
  const REAL targetPhotonEnergyMaxNorm = factorEnergy->getEMax(gdata, ix, iy, iz) * energyNorm;
  const REAL muBeamObserverMin = factorDirection->getMuMin(gdata, ix, iy, iz, observerDirection);

  // the following quantity is lorentz invariant
  // scatteredPhotonEnergyNorm_COM * targetPhotonEnergyMaxNorm_COM * (1 - muBeamObserverMin_COM) =
  // scatteredPhotonEnergyNorm * targetPhotonEnergyMaxNorm * (1 - muBeamObserverMin)
  // hence we can use the latter to avoid the computation of the doppler factors corresponding to
  // beamDirection thus making the minimum particleLorentzFactor independent of the beamDirection
  // (which would be hard to find in general)

  return 0.5 * scatteredPhotonEnergyNorm_COM *
         (1. + sqrt(1. + 2. / (scatteredPhotonEnergyNorm * targetPhotonEnergyMaxNorm *
                               (1 - muBeamObserverMin))));
}

std::shared_ptr<InverseComptonLossIntegrals> InverseComptonLossIntegrals::getInstance(
    std::shared_ptr<RadiationField> radiationField) {
  if (auto factorized = std::dynamic_pointer_cast<RadiationFieldFactorized>(radiationField)) {
    auto rffn = factorized->get_iRadiationFieldFactor_NumberDensity();
    auto rffe = factorized->get_iRadiationFieldFactor_Energy();
    auto rffd = factorized->get_iRadiationFieldFactor_Direction();

    if (auto beamed = std::dynamic_pointer_cast<IRadiationFieldFactor_Beamed>(rffd)) {
      if (auto mono = std::dynamic_pointer_cast<IRadiationFieldFactor_Monochromatic>(rffe)) {
        return std::make_shared<InverseComptonLossIntegralsMonochromaticBeamed>(rffn, mono, beamed);
      }
    }
  }
  throw CException("IC Loss integrals not implemented for " +
                   string(typeid(radiationField).name()));
}

InverseComptonLossIntegralsMonochromaticBeamed::InverseComptonLossIntegralsMonochromaticBeamed(
    std::shared_ptr<IRadiationFieldFactor_NumberDensity> factorNumberDensity,
    std::shared_ptr<IRadiationFieldFactor_Monochromatic> factorMonochromatic,
    std::shared_ptr<IRadiationFieldFactor_Beamed> factorBeamed) {
  this->factorNumberDensity = factorNumberDensity;
  this->factorMonochromatic = factorMonochromatic;
  this->factorBeamed = factorBeamed;
}

void InverseComptonLossIntegralsMonochromaticBeamed::compute_constants(Data& gdata, const int ix,
                                                                       const int iy, const int iz,
                                                                       REAL particleRestEnergy) {
  energyNorm = 1. / particleRestEnergy;
  targetPhotonEnergyNorm = factorMonochromatic->get_photonEnergy(gdata, ix, iy, iz) * energyNorm;
  factorBeamed->get_beamDirection(gdata, ix, iy, iz, beamDirection);

  lorentzTrafo.set_boostVector(
      gdata.om[gdata.fluid.get_q_sx()](ix, iy, iz), gdata.om[gdata.fluid.get_q_sy()](ix, iy, iz),
      gdata.om[gdata.fluid.get_q_sz()](ix, iy, iz), gdata.lorentz(ix, iy, iz));

  dopplerBeamInv = lorentzTrafo.get_dopplerFactor_inv(beamDirection);
}

REAL InverseComptonLossIntegralsMonochromaticBeamed::FKleinNishina(Data& gdata, const int ix,
                                                                   const int iy, const int iz,
                                                                   REAL particleLorentz_COM) {
  REAL targetPhotonEnergyNorm_COM = dopplerBeamInv * targetPhotonEnergyNorm;
  return sqr(dopplerBeamInv) * factorNumberDensity->numberDensity(gdata, ix, iy, iz) *
         targetPhotonEnergyNorm *
         InverseComptonFunctions::fKleinNishinaIsotrope(targetPhotonEnergyNorm_COM,
                                                        particleLorentz_COM);
}

/*
 *
 * INVERSE COMPTON FUNCTIONS
 *
 */

/**
 * @copydetails Implementation taken from [Moderski 2005] (A3)
 */
REAL InverseComptonFunctions::fGeneric(REAL scatteredPhotonEnergy, REAL targetPhotonEnergy,
                                       REAL particleLorentz, REAL cosTheta) {
  REAL bTheta = 2. * (1. - cosTheta) * targetPhotonEnergy * particleLorentz;
  REAL wRatio = scatteredPhotonEnergy / particleLorentz;
  REAL wRed = (1. - wRatio);

  return 1. + 0.5 * sqr(wRatio) / wRed - 2. * wRatio / (bTheta * wRed) +
         2. * sqr(wRatio / (bTheta * wRed));
}

/**
 * @copydetails Implementation taken from [Moderski 2005] (B4)
 */
REAL InverseComptonFunctions::fIsotrope(REAL scatteredPhotonEnergy, REAL targetPhotonEnergy,
                                        REAL particleLorentz) {
  REAL bTilde = 4. * targetPhotonEnergy * particleLorentz;
  REAL wRatio = scatteredPhotonEnergy / particleLorentz;
  REAL fac = bTilde * (1. - wRatio);

  return 1. + wRatio / sqr(fac) *
                  (+0.5 * wRatio * bTilde * fac + fac - 2. * wRatio - 0.5 * sqr(wRatio) * bTilde -
                   2. * fac * log(fac / wRatio));
}

/**
 * @copydetails Implementation taken from [Moderski 2005] (C5)
 */

REAL InverseComptonFunctions::fKleinNishinaIsotrope(REAL targetPhotonEnergy, REAL particleLorentz) {
  REAL b = 4. * targetPhotonEnergy * particleLorentz;

  //	return 1.;

  if (b < 1e-3) {
    return 1.;
  } else {
    return 9. / cube(b) *
           ((0.5 * b + 6. + 6. / b) * log(1. + b) -
            (11. / 12. * cube(b) + 6. * sqr(b) + 9. * b + 4.) / sqr(1. + b) - 2. +
            2. * gsl_sf_dilog(-b)
            //		  + 2. * InverseComptonFunctions::DiLog(-b)
           );
  }
}

REAL InverseComptonFunctions::DiLog(REAL x) {
  // Taken from Klaus
  // see following for details
  //		arXiv:hep-ph/0410259
  //		10.1016/j.cpc.2004.12.009
  double value = 0;
  if (x >= -1.) {
    REAL power = 1.;
    for (int i = 1; i <= 100; i++) {
      power *= x;  // x^i
      value += power / sqr(i);
    }
  } else if (x < -1.) {
    REAL y = 1. - x;
    REAL power = 1.;
    for (int i = 1; i <= 100; i++) {
      power /= y;  // y^-i
      value += power / sqr(i);
    }
    value += -sqr(M_PI) / 6. - 0.5 * log(y) * (2. * log(-x) - log(y));
  } else {
    std::cout << "Error: x= " << x << std::endl;
  }

  return value;
}

/*
 *
 * 	LOSS PROCESS
 *
 */

InverseComptonProcess::InverseComptonProcess(normalisation* normHD, normalisation* normParticle,
                                             std::shared_ptr<RadiationField> targetRadiationField)
    : RadiativeLossProcess(normHD, normParticle) {
  // inverseComptonIntegrals.set_radiationField(targetRadiationField);

  Quantity inverseComptonFac = 4. / 3. * ThomsonCrossSection * SpeedOfLight;
  inverseComptonFac_num =
      inverseComptonFac.normalizeTo(1. / normParticle->normQuantity(normParticle->TIME) /
                                    (normParticle->normQuantity(normParticle->NUM_DENS)));

  // inverseComptonEmission =
  //     new InverseComptonEmission(normHD, normParticle, &inverseComptonIntegrals);

  inverseComptonLossIntegrals = InverseComptonLossIntegrals::getInstance(targetRadiationField);
  setEmissionTarget(targetRadiationField);
}

std::string InverseComptonProcess::get_type() const { return "InverseCompton"; }

void InverseComptonProcess::set_particle(Particle& particle) {
  LossProcess::set_particle(particle);
  inverseComptonEmission->set_particle(particle);
}

void InverseComptonProcess::compute_constants(Data& gdata, const int ix, const int iy,
                                              const int iz) {
  LossProcess::compute_constants(gdata, ix, iy, iz);
  inverseComptonLossIntegrals->compute_constants(gdata, ix, iy, iz,
                                                 particle->particleRestEnergy_num);

  // For the hack
  gdata_link = &gdata;
  this->ix = ix;
  this->iy = iy;
  this->iz = iz;

  // Debug output
  gdata.ic_fac(ix, iy, iz) = inverseComptonFac_num * inverseComptonLossIntegrals->FKleinNishina(
                                                         *gdata_link, ix, iy, iz, 1.);
}

REAL InverseComptonProcess::get_momentumLoss_c2(REAL particleLorentz_COM) {
  return inverseComptonFac_num *
         inverseComptonLossIntegrals->FKleinNishina(*gdata_link, ix, iy, iz, particleLorentz_COM);
}

void InverseComptonProcess::get_momentumLoss(REAL particleLorentz_COM, REAL& loss) {
  loss -= inverseComptonFac_num * sqr(particleLorentz_COM) *
          inverseComptonLossIntegrals->FKleinNishina(*gdata_link, ix, iy, iz, particleLorentz_COM);
}

void InverseComptonProcess::setEmissionTarget(
    std::shared_ptr<RadiationField> targetRadiationField) {
  inverseComptonEmission =
      std::make_shared<InverseComptonEmission>(normHD, normParticle, targetRadiationField);
  this->emissionField = inverseComptonEmission.get();
}

/*
 *
 * IC EMISSION
 *
 */

InverseComptonEmission::InverseComptonEmission(normalisation* normHD, normalisation* normParticle,
                                               std::shared_ptr<RadiationField> targetRadiationField)
    : EmissionField(normHD, normParticle) {
  inverseComptonEmissionIntegrals =
      InverseComptonEmissionIntegrals::getInstance(targetRadiationField);

  Quantity inverseComptonEmissionFac = 3. / (16. * pi) * ThomsonCrossSection * SpeedOfLight;
  inverseComptonEmissionFac_num =
      inverseComptonEmissionFac.normalizeTo(1. / normParticle->normQuantity(normParticle->TIME) /
                                            normParticle->normQuantity(normParticle->NUM_DENS));

  // For the particle gamma Integration
  gslIntegrationWorkspace = gsl_integration_workspace_alloc(nIntegrationMax);

  const int nGauss = value_exists("InverseComptonIntegrationNGauss")
                         ? value("InverseComptonIntegrationNGauss")
                         : 10;

  fixedWorkspace = gsl_integration_fixed_alloc(gsl_integration_fixed_legendre, nGauss, 0, 1, 0, 0);

  const IntegrationType type = value_exists("InverseComptonIntegrationType")
                                   ? catParameterMap[svalue("InverseComptonIntegrationType")]
                                   : IntegrationType::GaussPowerlaw;

  setIntegrationType(type);
}

InverseComptonEmission::~InverseComptonEmission() {
  gsl_integration_workspace_free(gslIntegrationWorkspace);
  gsl_integration_fixed_free(fixedWorkspace);
}

void InverseComptonEmission::set_particle(Particle& particle) { this->particle = &particle; }

inline REAL InverseComptonEmission::diffEmissivity(Data& gdata, const int ix, const int iy,
                                                   const int iz, const REAL photonEnergy,
                                                   const REAL photonDirection[3],
                                                   const REAL particleLorentz_COM) {
  return inverseComptonEmissionIntegrals->gEmission(gdata, ix, iy, iz, photonEnergy,
                                                    photonDirection, particleLorentz_COM) /
         sqr(particleLorentz_COM);
}

inline REAL InverseComptonEmission::diffEmissivityLogTransformed(Data& gdata, const int ix,
                                                                 const int iy, const int iz,
                                                                 const REAL photonEnergy,
                                                                 const REAL photonDirection[3],
                                                                 const REAL particleLorentz_COM) {
  return inverseComptonEmissionIntegrals->gEmission(gdata, ix, iy, iz, photonEnergy,
                                                    photonDirection, particleLorentz_COM) /
         particleLorentz_COM;
}

template <typename T>
double InverseComptonEmission::gaussLegendre(T& integrand) const {
  static auto evaluator = [](double x, void* fun) { return (*static_cast<T*>(fun))(x); };
  static gsl_function F = {.function = evaluator};
  F.params = &integrand;
  double result;
  gsl_integration_fixed(&F, &result, fixedWorkspace);
  return result;
}

template <typename T>
double InverseComptonEmission::qag(T& integrand, const double tolerance) {
  static auto evaluator = [](double x, void* fun) { return (*static_cast<T*>(fun))(x); };
  static gsl_function F = {.function = evaluator};
  F.params = &integrand;
  double result, error;

  gsl_error_handler_t* oldHandler = gsl_set_error_handler_off();
  int status = gsl_integration_qag(&F, 0, 1, 0, tolerance, nIntegrationMax, 1,
                                   gslIntegrationWorkspace, &result, &error);
  gsl_set_error_handler(oldHandler);

  if (status > 0) {
    if (status == GSL_EMAXITER) {
    } else {
      std::string error = gsl_strerror(status);
      throw CException(error);
    }
  }

  return result;
}

InverseComptonEmission::integrationLimits InverseComptonEmission::getIntegrationLimits(
    Data& gdata, const int ix, const int iy, const int iz, const REAL photonEnergy,
    const REAL photonDirection[3]) const {
  integrationLimits limits;

  limits.gammaHi = particle->get_momentumEdge(particle->getHi() + 1);
  const double gammaLoMin = inverseComptonEmissionIntegrals->particleLorentzMin(
      gdata, ix, iy, iz, photonEnergy, photonDirection);
  limits.gammaLo = gammaLoMin;

  if (gammaLoMin > limits.gammaHi) return limits;

  limits.gammaLoIdx = particle->getHi();
  limits.gammaHiIdx = 0;
  for (int iMom = particle->getLo(); iMom <= particle->getHi(); ++iMom) {
    if ((*particle)(iMom, ix, iy, iz) > 0) {
      limits.gammaHiIdx = max(iMom, limits.gammaHiIdx);

      if (gammaLoMin < particle->get_momentumEdge(iMom + 1)) {
        limits.gammaLoIdx = min(iMom, limits.gammaLoIdx);
      }
    }
  }

  limits.gammaHi = particle->get_momentumEdge(limits.gammaHiIdx + 1);
  limits.gammaLo = max(particle->get_momentumEdge(limits.gammaLoIdx), gammaLoMin);

  return limits;
}

REAL InverseComptonEmission::intBinnedCenter(Data& gdata, const int ix, const int iy, const int iz,
                                             const REAL photonEnergy, const REAL photonDirection[3],
                                             const integrationLimits& limits) {
  REAL photonEmissivity = 0;
  double gammaLoMin = inverseComptonEmissionIntegrals->particleLorentzMin(
      gdata, ix, iy, iz, photonEnergy, photonDirection);

  for (int iMom = particle->getLo(); iMom <= particle->getHi(); ++iMom) {
    const REAL particleLorentz_COM = particle->get_momentumCenter(iMom);
    if (particleLorentz_COM < gammaLoMin) {
      continue;
    }

    photonEmissivity +=
        diffEmissivity(gdata, ix, iy, iz, photonEnergy, photonDirection, particleLorentz_COM) *
        (*particle)(iMom, ix, iy, iz) * particle->get_momentumWidth(iMom);
  }

  return photonEmissivity;
}

template <typename T>
REAL InverseComptonEmission::integrateLogTrafoFixed(Data& gdata, const int ix, const int iy,
                                                    const int iz, const REAL photonEnergy,
                                                    const REAL photonDirection[3],
                                                    const integrationLimits& limits,
                                                    T& reconstruction) {
  // static int counter = 0;
  // if (counter < 10 && limits.gammaHi > 100 * limits.gammaLo) {
  //   ofstream myfile;
  //   myfile.open("/tmp/tmp.UT7S6anv2T/data_" + ToString(counter));
  //   for (double x = limits.gammaLo; x <= limits.gammaHi; x *= 1.1) {
  //     myfile << x << " " << reconstruction(x) << "\n";
  //   }
  //   myfile.close();
  //   ++counter;
  // }

  static auto integrand = [&](double z) -> double {
    double particleLorentz_COM = powf(limits.gammaHi / limits.gammaLo, z) * limits.gammaLo;
    return diffEmissivityLogTransformed(gdata, ix, iy, iz, photonEnergy, photonDirection,
                                        particleLorentz_COM) *
           reconstruction(particleLorentz_COM);
  };

  const double result = gaussLegendre(integrand);
  return result * log(limits.gammaHi / limits.gammaLo);
}

template <typename T>
REAL InverseComptonEmission::integrateLogTrafoAdaptive(Data& gdata, const int ix, const int iy,
                                                       const int iz, const REAL photonEnergy,
                                                       const REAL photonDirection[3],
                                                       const integrationLimits& limits,
                                                       T& reconstruction, const double tolerance) {
  static auto integrand = [&](double z) -> double {
    double particleLorentz_COM = powf(limits.gammaHi / limits.gammaLo, z) * limits.gammaLo;
    return diffEmissivityLogTransformed(gdata, ix, iy, iz, photonEnergy, photonDirection,
                                        particleLorentz_COM) *
           reconstruction(particleLorentz_COM);
  };

  const double result = qag(integrand, tolerance);
  return result * log(limits.gammaHi / limits.gammaLo);
}

REAL InverseComptonEmission::intAdaptivePowerlaw(Data& gdata, const int ix, const int iy,
                                                 const int iz, const REAL photonEnergy,
                                                 const REAL photonDirection[3],
                                                 const integrationLimits& limits) {
  powerlawReconstruction.setData(particle->grid, particleSpectrumBuffer, limits.gammaLoIdx,
                                 limits.gammaHiIdx);

  REAL photonEmissivity = integrateLogTrafoAdaptive(
      gdata, ix, iy, iz, photonEnergy, photonDirection, limits, powerlawReconstruction, 1e-2);
  return photonEmissivity;
}

REAL InverseComptonEmission::intGlobalPowerlaw(Data& gdata, const int ix, const int iy,
                                               const int iz, const REAL photonEnergy,
                                               const REAL photonDirection[3],
                                               const integrationLimits& limits) {
  powerlawReconstruction.setData(particle->grid, particleSpectrumBuffer, limits.gammaLoIdx,
                                 limits.gammaHiIdx);

  REAL photonEmissivity = integrateLogTrafoFixed(gdata, ix, iy, iz, photonEnergy, photonDirection,
                                                 limits, powerlawReconstruction);
  return photonEmissivity;
}

REAL InverseComptonEmission::intGlobalPS(Data& gdata, const int ix, const int iy, const int iz,
                                         const REAL photonEnergy, const REAL photonDirection[3],
                                         const integrationLimits& limits) {
  parabolicSplineReconstruction.set_grid(particle->grid);
  parabolicSplineReconstruction.get_reconst_PSM(particleSpectrumBuffer);

  REAL photonEmissivity = integrateLogTrafoFixed(gdata, ix, iy, iz, photonEnergy, photonDirection,
                                                 limits, parabolicSplineReconstruction);
  return photonEmissivity;
}

REAL InverseComptonEmission::intGlobalConst(Data& gdata, const int ix, const int iy, const int iz,
                                            const REAL photonEnergy, const REAL photonDirection[3],
                                            const integrationLimits& limits) {
  auto reconst = [&](double x) { return particleSpectrumBuffer(particle->grid.get_ipos(x)); };

  REAL photonEmissivity =
      integrateLogTrafoFixed(gdata, ix, iy, iz, photonEnergy, photonDirection, limits, reconst);
  return photonEmissivity;
}

void InverseComptonEmission::setIntegrationType(const IntegrationType type) {
  switch (type) {
    case IntegrationType::BinnedCenter:
      integral = &InverseComptonEmission::intBinnedCenter;
      ccout << "  - ic: using binned integration (midpoint rule)" << endl;
      break;
      ;
    case IntegrationType::GaussConstant:
      integral = &InverseComptonEmission::intGlobalConst;
      ccout << "  - ic: using fixed integration (piecewise constant reconst)" << endl;
      break;
      ;
    case IntegrationType::GaussPowerlaw:
      integral = &InverseComptonEmission::intGlobalPowerlaw;
      ccout << "  - ic: using fixed integration (piecewise powerlaw reconst)" << endl;
      break;
      ;
    case IntegrationType::GaussParabolicSpline:
      integral = &InverseComptonEmission::intGlobalPS;
      ccout << "  - ic: using fixed integration (parabolic spline reconst)" << endl;
      break;
      ;
    case IntegrationType::AdaptivePowerlaw:
      integral = &InverseComptonEmission::intAdaptivePowerlaw;
      ccout << "  - ic: using adaptive integration (piecewise powerlaw reconst)" << endl;
      break;
      ;
  }
}

REAL InverseComptonEmission::get_photonEmissivity_LAB(Data& gdata, const int ix, const int iy,
                                                      const int iz, const REAL photonEnergy,
                                                      const REAL photonDirection[3]) {
  assert(particle != NULL);
  assert(integral != nullptr);
  REAL photonEmissivity = 0;

  try {
    inverseComptonEmissionIntegrals->compute_constants(gdata, ix, iy, iz,
                                                     particle->particleRestEnergy_num);
    integrationLimits limits = getIntegrationLimits(gdata, ix, iy, iz, photonEnergy, photonDirection);
    if (limits.gammaLo > limits.gammaHi) return 0;
    particle->getSpectrum(ix, iy, iz, particleSpectrumBuffer);
    photonEmissivity = (this->*integral)(gdata, ix, iy, iz, photonEnergy, photonDirection, limits);
  } catch (std::exception& e) {
    std::stringstream ss;
    ss << "Exception occured in InverseComptonEmission: " << e.what();
    std::cerr << ss.str() << std::endl;
  }

  photonEmissivity *= inverseComptonEmissionFac_num / particle->particleRestEnergy_num;
  return photonEmissivity;
}

#endif
