#ifndef CRONOS_INVERSE_COMPTON_H
#define CRONOS_INVERSE_COMPTON_H

#include <gsl/gsl_integration.h>
#include <gsl/gsl_sf.h>
#include <functional>
#include "ParabolicSpline.H"
#include "PowerlawReconstruction.H"
#include "RadiationField.H"
#include "RadiativeLossProcess.H"

/**
 * @brief Inverse Compton Integral - Interface class
 */
class InverseComptonEmissionIntegrals {
 public:
  static std::shared_ptr<InverseComptonEmissionIntegrals> getInstance(
      std::shared_ptr<RadiationField> radiationField);

  virtual ~InverseComptonEmissionIntegrals(){};

  virtual void compute_constants(Data &gdata, const int ix, const int iy, const int iz,
                                 const REAL particleRestEnergy){};

  virtual REAL gEmission(Data &gdata, const int ix, const int iy, const int iz,
                         const REAL scatteredPhotonEnergy, const REAL observerDirection[3],
                         const REAL particleLorentz_COM) = 0;

  virtual REAL particleLorentzMin(Data &gdata, const int ix, const int iy, const int iz,
                                  const REAL photonEnergy, const REAL photonDirection[3]) const = 0;
};

/**
 * @brief Inverse Compton Integral - Interface class
 */
class InverseComptonLossIntegrals {
 public:
  static std::shared_ptr<InverseComptonLossIntegrals> getInstance(
      std::shared_ptr<RadiationField> radiationField);

  virtual ~InverseComptonLossIntegrals(){};

  virtual void compute_constants(Data &gdata, const int ix, const int iy, const int iz,
                                 const REAL particleRestEnergy){};

  virtual REAL FKleinNishina(Data &gdata, const int ix, const int iy, const int iz,
                             const REAL particleLorentz_COM) = 0;
};

class InverseComptonLossIntegralsMonochromaticBeamed : public InverseComptonLossIntegrals {
 public:
  InverseComptonLossIntegralsMonochromaticBeamed(
      std::shared_ptr<IRadiationFieldFactor_NumberDensity> factorNumberDensity,
      std::shared_ptr<IRadiationFieldFactor_Monochromatic> factorMonochromatic,
      std::shared_ptr<IRadiationFieldFactor_Beamed> factorBeamed);

  virtual ~InverseComptonLossIntegralsMonochromaticBeamed(){};

  virtual void compute_constants(Data &gdata, const int ix, const int iy, const int iz,
                                 const REAL particleRestEnergy);

  virtual REAL FKleinNishina(Data &gdata, const int ix, const int iy, const int iz,
                             const REAL particleLorentz_COM) override;

 private:
  double energyNorm;
  double targetPhotonEnergyNorm;
  double dopplerBeamInv;
  double beamDirection[3];
  LorentzTrafo lorentzTrafo;

  std::shared_ptr<IRadiationFieldFactor_NumberDensity> factorNumberDensity;
  std::shared_ptr<IRadiationFieldFactor_Monochromatic> factorMonochromatic;
  std::shared_ptr<IRadiationFieldFactor_Beamed> factorBeamed;
};

class IRFFEnergyIntegrator {
 public:
  static std::shared_ptr<IRFFEnergyIntegrator> getIntegrator(
      std::shared_ptr<IRadiationFieldFactor_Energy> irff);

  virtual ~IRFFEnergyIntegrator(){};

  virtual void compute_constants(Data &gdata, const int ix, const int iy, const int iz,
                                 REAL particleRestEnergy) {}

  virtual double integrate(Data &gdata, const int ix, const int iy, const int iz,
                           const REAL scatteredPhotonEnergy, const REAL observerDirection[3],
                           const double beamDirection[3], const REAL particleLorentz_COM) = 0;
};

class IRFFDirectionIntegrator {
 public:
  static std::shared_ptr<IRFFDirectionIntegrator> getIntegrator(
      std::shared_ptr<IRadiationFieldFactor_Direction> irff);

  virtual ~IRFFDirectionIntegrator(){};

  virtual void compute_constants(Data &gdata, const int ix, const int iy, const int iz,
                                 const REAL particleRestEnergy) {}

  virtual double integrate(Data &gdata, const int ix, const int iy, const int iz,
                           const REAL scatteredPhotonEnergy, const double observerDirection[3],
                           const REAL particleLorentz_COM,
                           shared_ptr<IRFFEnergyIntegrator> &energyIntegrator) = 0;
};

class IRFFEnergyIntegratorMonochromatic : public IRFFEnergyIntegrator {
 public:
  IRFFEnergyIntegratorMonochromatic(std::shared_ptr<IRadiationFieldFactor_Monochromatic> irff) {
    this->factorMonochromatic = irff;
  }
  virtual void compute_constants(Data &gdata, const int ix, const int iy, const int iz,
                                 const REAL particleRestEnergy);

  virtual double integrate(Data &gdata, const int ix, const int iy, const int iz,
                           const REAL scatteredPhotonEnergy, const REAL observerDirection[3],
                           const double beamDirection[3], const REAL particleLorentz_COM) override;

 private:
  std::shared_ptr<IRadiationFieldFactor_Monochromatic> factorMonochromatic;
  LorentzTrafo lorentzTrafo;
  double energyNorm;
  double targetPhotonEnergyNorm;
};

class IRFFDirectionIntegratorBeamed : public IRFFDirectionIntegrator {
 public:
  IRFFDirectionIntegratorBeamed(std::shared_ptr<IRadiationFieldFactor_Beamed> irffd) {
    factorBeamed = irffd;
  }
  virtual double integrate(Data &gdata, const int ix, const int iy, const int iz,
                           const REAL scatteredPhotonEnergy, const double observerDirection[3],
                           const REAL particleLorentz_COM,
                           shared_ptr<IRFFEnergyIntegrator> &energyIntegrator) override;

 private:
  std::shared_ptr<IRadiationFieldFactor_Beamed> factorBeamed;
};

class IRFFDirectionIntegratorSphericalCoords : public IRFFDirectionIntegrator {
 public:
  IRFFDirectionIntegratorSphericalCoords(std::shared_ptr<IRFFD_SphericalCoordinates> irffd);
  virtual ~IRFFDirectionIntegratorSphericalCoords();
  virtual double integrate(Data &gdata, const int ix, const int iy, const int iz,
                           const REAL scatteredPhotonEnergy, const double observerDirection[3],
                           const REAL particleLorentz_COM,
                           shared_ptr<IRFFEnergyIntegrator> &energyIntegrator) override;

 private:
  std::shared_ptr<IRFFD_SphericalCoordinates> factorSpherical;
  gsl_integration_glfixed_table *gaussLegendreTableAzimuth;
  gsl_integration_glfixed_table *gaussLegendreTableDeclination;
  gsl_function gslIntegrandAzitmuth;
  gsl_function gslIntegrandDeclination;

  const int nPointsAzimuth = 5;
  const int nPointsDeclination = 3;
};

/**
 * @brief Implementation of the  Inverse Compton Integrals for beamed monochromatic seed photon
 * fields
 */
class InverseComptonEmissionIntegralsFactorized : public InverseComptonEmissionIntegrals {
 public:
  InverseComptonEmissionIntegralsFactorized(
      std::shared_ptr<RadiationFieldFactorized> radiationField);

  void compute_constants(Data &gdata, const int ix, const int iy, const int iz,
                         const REAL particleRestEnergy);

  REAL gEmission(Data &gdata, const int ix, const int iy, const int iz,
                 const REAL scatteredPhotonEnergy, const REAL observerDirection[3],
                 const REAL particleLorentz_COM);

  REAL particleLorentzMin(Data &gdata, const int ix, const int iy, const int iz,
                          const REAL photonEnergy, const REAL photonDirection[3]) const;

 private:
  std::shared_ptr<IRadiationFieldFactor_NumberDensity> factorNumberDensity;
  std::shared_ptr<IRadiationFieldFactor_Energy> factorEnergy;
  std::shared_ptr<IRadiationFieldFactor_Direction> factorDirection;

  std::shared_ptr<IRFFEnergyIntegrator> energyIntegrator;
  std::shared_ptr<IRFFDirectionIntegrator> directionIntegrator;

  LorentzTrafo lorentzTrafo;

  /** Buffer quantities - set by compute_constants */
  REAL energyNorm;
};

/**
 * @brief Some generic inverse compton functions
 */
namespace InverseComptonFunctions {
REAL fGeneric(REAL scatteredPhotonEnergy, REAL targetPhotonEnergy, REAL particleLorentz,
              REAL cosTheta);

REAL fIsotrope(REAL scatteredPhotonEnergy, REAL targetPhotonEnergy, REAL particleLorentz);

REAL fKleinNishinaIsotrope(REAL targetPhotonEnergy, REAL particleLorentz);

REAL DiLog(REAL x);
}  // namespace InverseComptonFunctions

/**
 * @brief Inverse compton emission field
 * @details This evaluates parts of an aggregated instance of InverseComptonEmissionIntegrals
 */
class InverseComptonEmission : public EmissionField {
 public:
  InverseComptonEmission(normalisation *normHD, normalisation *normParticle,
                         std::shared_ptr<RadiationField> targetRadiationField);

  ~InverseComptonEmission();

  REAL get_photonEmissivity_LAB(Data &gdata, const int ix, const int iy, const int iz,
                                const REAL photonEnergy_LAB, const REAL photonDirection_LAB[3]);

  void set_particle(Particle &particle);

 private:
  Particle *particle;

  std::shared_ptr<InverseComptonEmissionIntegrals> inverseComptonEmissionIntegrals;
  REAL inverseComptonEmissionFac_num;

  /** For the emissivity integration */
  static constexpr long unsigned int nIntegrationMax = 10;
  gsl_integration_workspace *gslIntegrationWorkspace;
  gsl_integration_fixed_workspace *fixedWorkspace;

  PowerlawReconstruction powerlawReconstruction;
  ParabolicSplineReconstruction parabolicSplineReconstruction;
  NumMatrix<REAL, 1> particleSpectrumBuffer;

  /** Defines the differential emissivity to be folded with the particle spectrum */
  REAL diffEmissivity(Data &gdata, const int ix, const int iy, const int iz,
                      const REAL photonEnergy, const REAL photonDirection[3],
                      const REAL particleLorentz_COM);

  /** Defines the differential emissivity to be folded with the particle spectrum */
  REAL diffEmissivityLogTransformed(Data &gdata, const int ix, const int iy, const int iz,
                                    const REAL photonEnergy, const REAL photonDirection[3],
                                    const REAL particleLorentz_COM);

  /** Integrates from 0..1 */
  template <typename T>
  double gaussLegendre(T &integrand) const;

  /** Integrates from 0..1 */
  template <typename T>
  double qag(T &integrand, const double tolerance);

  struct integrationLimits {
    int gammaLoIdx;
    int gammaHiIdx;
    double gammaLo;
    double gammaHi;
  };

  /** Compute Integration Limits */
  integrationLimits getIntegrationLimits(Data &gdata, const int ix, const int iy, const int iz,
                                         const REAL photonEnergy,
                                         const REAL photonDirection[3]) const;

  /** Applies a gaussian quadrature to integrate the
   * log transformed diff emissivity */
  template <typename T>
  REAL integrateLogTrafoFixed(Data &gdata, const int ix, const int iy, const int iz,
                              const REAL photonEnergy, const REAL photonDirection[3],
                              const integrationLimits &limits, T &reconstruction);

  /** Applies an adaptive integration method to integrate the
   * log transformed diff emissivity */
  template <typename T>
  REAL integrateLogTrafoAdaptive(Data &gdata, const int ix, const int iy, const int iz,
                                 const REAL photonEnergy, const REAL photonDirection[3],
                                 const integrationLimits &limits, T &reconstruction,
                                 const double tolerance);

  enum class IntegrationType {
    BinnedCenter,
    GaussConstant,
    GaussPowerlaw,
    GaussParabolicSpline,
    AdaptivePowerlaw
  };
  std::map<std::string, IntegrationType> catParameterMap = {
      {"BinnedConstant", IntegrationType::BinnedCenter},
      {"GaussConstant", IntegrationType::GaussConstant},
      {"GaussPowerlaw", IntegrationType::GaussPowerlaw},
      {"GaussParabolicSpline", IntegrationType::GaussParabolicSpline},
      {"AdaptivePowerlaw", IntegrationType::AdaptivePowerlaw}};
  void setIntegrationType(const IntegrationType);

  /** Integration variants */
  REAL(InverseComptonEmission::*integral)
  (Data &, const int, const int, const int, const REAL, const REAL *, const integrationLimits &);

  /** Basic midpoint rule per energy bin */
  REAL intBinnedCenter(Data &gdata, const int ix, const int iy, const int iz,
                       const REAL photonEnergy, const REAL photonDirection[3],
                       const integrationLimits &limits);

  /** Gaussian quadrature with constant reconstruction */
  REAL intGlobalConst(Data &gdata, const int ix, const int iy, const int iz,
                      const REAL photonEnergy, const REAL photonDirection[3],
                      const integrationLimits &limits);

  /** Gaussian quadrature with powerlaw reconstruction */
  REAL intGlobalPowerlaw(Data &gdata, const int ix, const int iy, const int iz,
                         const REAL photonEnergy, const REAL photonDirection[3],
                         const integrationLimits &limits);

  /** Gaussian quadrature with parabolic spline reconstruction */
  REAL intGlobalPS(Data &gdata, const int ix, const int iy, const int iz, const REAL photonEnergy,
                   const REAL photonDirection[3], const integrationLimits &limits);

  /** Adaptive integration with powerlaw reconstruction */
  REAL intAdaptivePowerlaw(Data &gdata, const int ix, const int iy, const int iz,
                           const REAL photonEnergy, const REAL photonDirection[3],
                           const integrationLimits &limits);
};

/**
 * @brief Inverse compton loss process
 * @details This evaluates parts of an aggregated instance of InverseComptonEmissionIntegrals
 */
class InverseComptonProcess : public RadiativeLossProcess {
 public:
  InverseComptonProcess(normalisation *normHD, normalisation *normParticle,
                        std::shared_ptr<RadiationField> targetPhotonField);

  void set_particle(Particle &particle);

  void compute_constants(Data &gdata, const int ix, const int iy, const int iz);

  void get_momentumLoss(REAL particleLorentz_COM, REAL &loss);

  REAL get_momentumLoss_c2(REAL particleLorentz_COM);

  std::string get_type() const final;

  void setEmissionTarget(std::shared_ptr<RadiationField> targetPhotonField);

 private:
  std::shared_ptr<InverseComptonLossIntegrals> inverseComptonLossIntegrals;
  std::shared_ptr<InverseComptonEmission> inverseComptonEmission;

  REAL inverseComptonFac_num;

  // The hack
  Data *gdata_link;
  int ix, iy, iz;
};

#endif
