#ifndef PSM_H
#define PSM_H

#include "Grid1D.H"
#include "matrix.H"

class ParabolicSplineReconstruction {
 private:
  NumArray<double> valLocL;

  NumArray<double> diag;
  NumArray<double> upperDiag;

  NumArray<double> rhs, rhsFac, rhsFacL;

  NumMatrix<double, 1> constPolA0;
  NumMatrix<double, 1> constPolA1;
  NumMatrix<double, 1> constPolA2;

  const Grid1D *grid;

  bool use_Limiter = true;
  int mx, Nx;
  static constexpr double third = 1. / 3.;

  // Buffers

  // Tridiag Solver buffer
  NumArray<double> gamma;

  // Limiter Buffer
  NumMatrix<double, 1> valLocLExt;

 public:
  ParabolicSplineReconstruction() {}

  void set_grid(const Grid1D &_grid) {
    if (grid != &_grid) {
      grid = &_grid;

      Nx = grid->get_Nx();
      mx = grid->get_mx();

      constPolA0.resize(grid->getLow(), grid->getHigh());
      constPolA1.resize(grid->getLow(), grid->getHigh());
      constPolA2.resize(grid->getLow(), grid->getHigh());

      valLocL.resize(Nx + 1);
      rhs.resize(Nx + 1);
      rhsFac.resize(Nx + 1);
      rhsFacL.resize(Nx + 1);

      diag.resize(Nx + 1);
      upperDiag.resize(Nx);
      NumArray<double> lowerDiag(Nx);

      // Assign solution vector with RHS of Eq. (14) from ZWS06
      for (int ipos = 0; ipos <= mx; ++ipos) {
        lowerDiag(ipos) = 1. / grid->get_del(ipos);
      }
      upperDiag = lowerDiag;

      for (int ipos = 1; ipos <= mx; ++ipos) {
        diag(ipos) =
            2. * (1. / grid->get_del(ipos - 1) + 1. / grid->get_del(ipos));
      }

      // Assign solution vector with RHS of Eq. (14) from ZWS06
      // rhsFac is the factor for rho_i
      // rhsFacR is the factor for rho_i+1
      // rhs_i = rhsFac_i * rho_i + rhsFacR_i * rho_i+1

      for (int ipos = 1; ipos <= mx; ++ipos) {
        rhsFacL(ipos) = 3. / grid->get_del(ipos - 1);
        rhsFac(ipos) = 3. / grid->get_del(ipos);
      }

      // Add two equations, such that the derivative at the boundaries are zero
      // Left boundary
      diag(0) = -4. / grid->get_del(0);
      upperDiag(0) = -2. / grid->get_del(0);
      rhsFac(0) = -6. / grid->get_del(0);
      rhsFacL(0) = 0.;

      // Right boundary
      diag(mx + 1) = 4. / grid->get_del(mx);
      lowerDiag(mx) = 2. / grid->get_del(mx);
      rhsFacL(mx + 1) = 6. / grid->get_del(mx);
      rhsFac(mx + 1) = 0;

      // Normalize: multiply with the inverse of the lower diagonal
      for (int ipos = 0; ipos <= mx; ++ipos) {
        const double fac = 1. / lowerDiag(ipos);
        diag(ipos + 1) *= fac;
        rhsFac(ipos + 1) *= fac;
        rhsFacL(ipos + 1) *= fac;
        if (ipos + 1 <= mx) upperDiag(ipos + 1) *= fac;
      }

      // Tridiag Solver buffer
      gamma.resize(diag.getLength());
      gamma = diag;

      // Limiter Buffer
      valLocLExt.resize(grid->getLow(), grid->getHigh());
    }
  }

  void get_reconst_PSM(const NumMatrix<double, 1> &valAve) {
    //! Compute the polynomial reconstruction

    /*!

    Compute parameters of the polynomial PSM reconstruction from
    first computing the point values at the left face of each cell
    from the cell averages. Here we use the parabolic spline method
    discussed in the references
    Zerroukat, Wood & Staniforth (2006) Int. J. Numer. Meth. Fluids,
    Vol 51, p1297-1318 [ZWS06]
    where additional limiter considerations are taken from
    Zerroukat, Wood & Staniforth (2005) Q.J.R. Meteorol. Soc.
    Vol 131, p2923-2936 [ZWS05]

    \param valAve array holding the cell-averages
    \param constPolA0 resulting polynomial parameter -- according to
    Eq. (10) from ZWS06
    \param constPolA1 same but according to Eq. (11)
    \param constPolA2 same but according to Eq. (12)

    */

    // Assign solution vector with RHS of Eq. (14) from ZWS06
    for (int ipos = 0; ipos <= mx + 1; ++ipos) {
      valLocL(ipos) = 0;
      valLocL(ipos) += rhsFacL(ipos) > 0 ? rhsFacL(ipos) * valAve(ipos - 1) : 0;
      valLocL(ipos) += rhsFac(ipos) > 0 ? rhsFac(ipos) * valAve(ipos) : 0;
    }

    double alpha = upperDiag(mx);
    double beta = 1.;

    // Solver tridiagonal system (for periodic bcs only)
    // solve_SpecialTridagCyclic(diag, upperdiag, valLocL, alpha, beta);
    solve_SpecialTridag(diag, upperDiag, valLocL);

    if (use_Limiter) {
      // Limiter working on Grid scale (see discussion in ZWS06)
      ensure_GridScaleMonotonicity(valAve, valLocL);
    }

    // Compute parameters of the polynomial for each cell in
    // computational domain:
    int rim = grid->get_rim();

    // Hack for periodic bcs:
    // valLocL(mx + 1) = valLocL(0);

    for (int ipos = 0; ipos <= mx; ++ipos) {
      constPolA0(ipos) = valLocL(ipos);
      constPolA1(ipos) =
          -4. * valLocL(ipos) - 2. * valLocL(ipos + 1) + 6. * valAve(ipos);
      constPolA2(ipos) =
          3. * valLocL(ipos) + 3. * valLocL(ipos + 1) - 6. * valAve(ipos);
    }

    if (use_Limiter) {
      for (int ipos = 0; ipos <= mx; ++ipos) {
        valLocLExt(ipos) = valLocL(ipos);
      }
      // Hack for periodic bcs:
      // for (int ipos = -rim; ipos < 0; ++ipos) {
      //   valLocLExt(ipos) = valLocLExt(mx + ipos + 1);
      // }
      // for (int ipos = mx + 1; ipos <= mx + rim; ++ipos) {
      //   valLocLExt(ipos) = valLocLExt(ipos - mx - 1);
      // }
      ensure_SubGridMonotonicity(valAve, valLocLExt, constPolA0, constPolA1,
                                 constPolA2);
    }

    // // Hack for periodic bcs:
    // for (int ipos = -rim; ipos < 0; ++ipos) {
    //   constPolA0(ipos) = constPolA0(mx + ipos + 1);
    //   constPolA1(ipos) = constPolA1(mx + ipos + 1);
    //   constPolA2(ipos) = constPolA2(mx + ipos + 1);
    // }

    // // Hack for periodic bcs:
    // for (int ipos = mx + 1; ipos <= mx + rim; ++ipos) {
    //   constPolA0(ipos) = constPolA0(ipos - mx - 1);
    //   constPolA1(ipos) = constPolA1(ipos - mx - 1);
    //   constPolA2(ipos) = constPolA2(ipos - mx - 1);
    // }
  }

  void solve_SpecialTridagCyclic(const NumArray<double> &diag,
                                 const NumArray<double> &superdiag,
                                 NumArray<double> &rhs, double alpha,
                                 double beta) {
    //! Special solver for periodic tri-diagnoal matrices
    /*!

    Solution of the special normalised tri-diagonal matrix (see
    discussion of the normalisation in appendix A.2. in Zerroukat et
    al. (2006). This normalisation results in the sub-diagonal
    consisting of unit values.

    In this case the solver assumes periodic boundary
    conditions. These can be treated with a modified version of the
    tri-diagonal solver (as explained in 'numerical recipes' under
    'cyclic tridiagonal systems')

    \param diag - a 1D matrix of extent 0...N-1 holding values along the
    diagonal \param superdiag - a 1D matrix of extent 0...N-2 holding values
    on super-diagonal \param rhs - right hand side of system of equations
    (array is overwritten with solution) \param alpha, beta additional
    off-diagonal entries
    */

    int len = diag.getLength();
    NumArray<double> diagVar(diag);  // vector holding modified diag

    // Replace values of first and last entry:
    double gamma = -diag(0);
    diagVar(0) -= gamma;
    diagVar(len - 1) -= alpha * beta / gamma;

    // Copy rhs into field result and solve for result
    solve_SpecialTridag(diagVar, superdiag, rhs);

    // Set up new RHS vector like in num rec (designated as u)
    NumArray<double> new_rhs(len);
    new_rhs.clear();
    new_rhs(0) = gamma;
    new_rhs(len - 1) = alpha;

    // Solve for new result:
    solve_SpecialTridag(diagVar, superdiag, new_rhs);

    // Compute factor
    double fact = (rhs(0) + beta * rhs(len - 1) / gamma) /
                  (1. + new_rhs(0) + beta * new_rhs(len - 1) / gamma);

    // Use this to compute solution vector:
    for (int ipos = 0; ipos < len; ++ipos) {
      rhs(ipos) -= fact * new_rhs(ipos);
    }
  }

  void solve_SpecialTridag(const NumArray<double> &diag,
                           const NumArray<double> &superdiag,
                           NumArray<double> &rhs) {
    //! Special solver for tri-diagnoal matrices
    /*!
      Solution of the special normalised tri-diagonal matrix (see
      discussion of the normalisation in appendix A.2. in Zerroukat et
      al. (2006). This normalisation results in the sub-diagonal
      consisting of unit values.

      \param diag - a 1D matrix of extent 0...N-1 holding values along the
      diagonal \param superdiag - a 1D matrix of extent 0...N-2 holding values
      on super-diagonal \param rhs - right hand side of system of equations
      (array is overwritten with solution)
    */

    int num = diag.getLength();
    double beta = 0.;

    beta = diag(0);
    // work(0) = rhs(0)/beta;
    rhs(0) /= beta;

    // Forward loop
    for (int ix = 1; ix < num; ++ix) {
      gamma(ix) = superdiag(ix - 1) / beta;
      beta = diag(ix) - gamma(ix);
      rhs(ix) = (rhs(ix) - rhs(ix - 1)) / beta;
    }

    // Backwards loop
    for (int ix = num - 2; ix >= 0; ix--) {
      //		work(ix)  = work(ix) - gamma(ix+1)*work(ix+1);
      rhs(ix) -= gamma(ix + 1) * rhs(ix + 1);
    }

    // Overwrite rhs with solution of tridiagonal system
    // rhs = work;
  }

  void ensure_GridScaleMonotonicity(const NumMatrix<double, 1> &valAve,
                                    NumArray<double> &valLocL) {
    //! Correction of grid scale monotonicity for PSM reconstruction

    /*!
          Limit the PSM reconstruction according to description in
          ZWS06. Here local reconstructed values are
          modified in order to fulfil monotonicity constraints.

          \param valAve array holding the cell-averages
          \param valLocL array holding local reconstruction values
         */

    // Correction of grid-scale monotonicity violations For a
    // description see ZWS05
    for (int ipos = 0; ipos <= mx; ++ipos) {
    // for (int ipos = 3; ipos <= mx - 2; ++ipos) {
      bool isSpurious(false);

      // Check according to Eq. (3.1)
      if ((valLocL(ipos) - valAve(ipos - 1)) * (valAve(ipos) - valLocL(ipos)) <
          0.) {
        if (((valAve(ipos - 1) - valAve(ipos - 2)) *
                 (valAve(ipos + 1) - valAve(ipos)) >=
             0.) ||
            ((valAve(ipos - 1) - valAve(ipos - 2)) *
                 (valAve(ipos - 2) - valAve(ipos - 3)) <=
             0.) ||
            ((valAve(ipos + 1) - valAve(ipos)) *
                 (valAve(ipos + 2) - valAve(ipos + 1)) <=
             0.) ||
            ((valLocL(ipos) - valAve(ipos - 1)) *
                 (valAve(ipos - 1) - valAve(ipos - 2)) <=
             0.)) {
          isSpurious = true;
        }
      }
      if (isSpurious) {  // Abbreviated form of Eq. (3.6) + (3.7)
        if ((std::abs(valLocL(ipos) - valAve(ipos)) -
             std::abs(valLocL(ipos) - valAve(ipos - 1))) >= 0.) {
          valLocL(ipos) = valAve(ipos - 1);
        } else {
          valLocL(ipos) = valAve(ipos);
        }
      }
    }
  }

  void ensure_SubGridMonotonicity(const NumMatrix<double, 1> &valAve,
                                  NumMatrix<double, 1> &valLocL,
                                  NumMatrix<double, 1> &constPolA0,
                                  NumMatrix<double, 1> &constPolA1,
                                  NumMatrix<double, 1> &constPolA2) {
    //! Correction of sub grid scale monotonicity for PSM reconstruction

    /*!
      Limit the PSM reconstruction according to description in
      ZWS06. Here local reconstructed values are
      modified in order to fulfil monotonicity constraints.

      \param valAve array holding the cell-averages
      \param valLocL array holding local reconstruction values
      \param constPolA0, constPolA1, constPolA2 arrays holding parameters of
      local polynomials
     */

    for (int ipos = 0; ipos <= mx; ++ipos) {
    // for (int ipos = 2; ipos <= mx - 3; ++ipos) {
      // Need only check for single extremum (here we only use third order)

      // Look for single maximum (use Product of gradient at left
      // and right boundary of cell - if both have different sign
      // there has to be an extremum inside the cell):
      if (constPolA1(ipos) * (constPolA1(ipos) + 2. * constPolA2(ipos)) < 0.) {
        // This local extremum is assumed to be genuine if none of
        // the following conditions are satisfied (see Eqs. (3.10)
        // - (3.13) in ZWS05)
        bool genuine = true;
        if (((valLocL(ipos) - valLocL(ipos - 1)) *
                 (valLocL(ipos + 2) - valLocL(ipos + 1)) >=
             0.) ||
            ((valLocL(ipos - 1) - valLocL(ipos - 2)) *
                 (valLocL(ipos) - valLocL(ipos - 1)) <=
             0.) ||
            ((valLocL(ipos + 2) - valLocL(ipos + 1)) *
                 (valLocL(ipos + 3) - valLocL(ipos + 2)) <=
             0.) ||
            ((valLocL(ipos) - valLocL(ipos - 1)) * constPolA1(ipos) <= 0.)) {
          genuine = false;
        }

        if (!genuine) {
          // if((valAve(ipos) <= std::min(valLocL(ipos), valLocL(ipos+1))) ||
          //    (valAve(ipos) >= std::max(valLocL(ipos), valLocL(ipos+1)))) {
          if ((valAve(ipos) - valLocL(ipos)) *
                  (valAve(ipos) - valLocL(ipos + 1)) >=
              0.) {
            // If non-genuine extremum outside range of
            // cell-face values - use constant reconstruction
            // instead
            constPolA0(ipos) = valAve(ipos);
            constPolA1(ipos) = 0.;
            constPolA2(ipos) = 0.;

          } else {
            // Compute shifted parabolas for cases valAve
            // between valLocL and valL or between valLocR and
            // valR (definitions of the latter see below or in
            // Eqs (20) & (21) from ZWS06

            double valL = third * (2. * valLocL(ipos) + valLocL(ipos + 1));
            double valR = third * (valLocL(ipos) + 2. * valLocL(ipos + 1));

            if (valLocL(ipos + 1) >
                valLocL(ipos)) {  // Case from ZWS06 Eq. (20) & (21)
              if (valAve(ipos) < valL) {
                set_PolConstLeft(valAve, valLocL, constPolA0, constPolA1,
                                 constPolA2, ipos);

              } else if (valAve(ipos) > valR) {
                set_PolConstRight(valAve, valLocL, constPolA0, constPolA1,
                                  constPolA2, ipos);
              }
            } else {
              if (valAve(ipos) < valR) {
                set_PolConstRight(valAve, valLocL, constPolA0, constPolA1,
                                  constPolA2, ipos);

              } else if (valAve(ipos) > valL) {
                set_PolConstLeft(valAve, valLocL, constPolA0, constPolA1,
                                 constPolA2, ipos);
              }
            }
          }
        }
      }
    }
  }

  void set_PolConstLeft(const NumMatrix<double, 1> &valAve,
                        const NumMatrix<double, 1> &valLocL,
                        NumMatrix<double, 1> &constPolA0,
                        NumMatrix<double, 1> &constPolA1,
                        NumMatrix<double, 1> &constPolA2, int ipos) {
    constPolA1(ipos) = 0.;
    constPolA2(ipos) = 3. * (valAve(ipos) - valLocL(ipos));
  }

  void set_PolConstRight(const NumMatrix<double, 1> &valAve,
                         const NumMatrix<double, 1> &valLocL,
                         NumMatrix<double, 1> &constPolA0,
                         NumMatrix<double, 1> &constPolA1,
                         NumMatrix<double, 1> &constPolA2, int ipos) {
    constPolA0(ipos) = -2. * valLocL(ipos + 1) + 3. * valAve(ipos);
    constPolA1(ipos) = 6. * (valLocL(ipos + 1) - valAve(ipos));
    constPolA2(ipos) = 3. * (valAve(ipos) - valLocL(ipos + 1));
  }

  double operator()(double x) const {
    const int idx = grid->get_ipos(x);
    const double xNorm = (x - grid->get_xL(idx)) / grid->get_del(idx);

    return ((constPolA2(idx) * xNorm) + constPolA1(idx)) * xNorm +
           constPolA0(idx);
  }
};
#endif