#include "Particle.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)

using namespace CRONOS_CONSTANTS;


const string Particle::get_particleType() const {
	return particleType;
}
const string Particle::get_particleName() const {
	return particleName;
}
const NumMatrix<REAL,1> & Particle::get_momentumCenter() const {
	return grid.get_xCen();
}
const NumMatrix<REAL,1> & Particle::get_momentumEdge() const {
	return grid.get_xL();
}
const NumMatrix<REAL,1> & Particle::get_momentumWidth() const {
	return grid.get_del();
}
const REAL Particle::get_momentumCenter(int ip) const {
	return grid.get_xCen(ip);
}
const REAL Particle::get_momentumEdge(int ip) const {
	return grid.get_xL(ip);
}
const REAL Particle::get_momentumWidth(int ip) const {
	return grid.get_del(ip);
}
const int Particle::getLo() const {
	return lo;
}
const int Particle::getHi() const {
	return hi;
}
const int Particle::getNBins() const {
	return nBins;
}
const int Particle::getOmb() const {
	return omb;
}
const int Particle::getOme() const {
	return ome;
}


Particle::Particle(
		Data & gdata,
		normalisation * normParticle,
		const int particleID,
		const int omb
		):
	particleID(particleID),
	gdata(gdata)
{

	this->normParticle = normParticle;

	//
	//	Particle Properties
	//

	// Defining particle species
	// rest mass and other parameter will be set here

	if (value_exists("ParticleType_" + ToString(particleID))) {
		particleType = svalue("ParticleType_" + ToString(particleID));
	} else {
		particleType = "defaultParticle";
	}

	if (value_exists("ParticleName_" + ToString(particleID))) {
		particleName = svalue("ParticleName_" + ToString(particleID));
	} else {
		particleName = particleType + "_" + ToString(particleID);
	}

	if (particleType == "proton") {
		particleRestMass = ProtonMass;
		particleCharge = ElementaryCharge;
	}
	else if (particleType == "electron") {
		particleRestMass = ElectronMass;
		particleCharge = (-1) * ElementaryCharge;
	}
	else if (particleType == "defaultParticle") {
		particleRestMass = 1. * normParticle->normQuantity(normParticle->MASS);
		particleCharge = ElementaryCharge;
	}
	else {
		throw CException(" No such particle available ");
	}

	// Compute numerical values
	particleRestMass_num = normParticle->get_num(particleRestMass);
	particleRestEnergy_num = normParticle->get_num(particleRestMass * sqr(SpeedOfLight));
	particleCharge_num = normParticle->get_num(particleCharge);
	particleChargeAbs_num = abs(particleCharge_num);



	//
	//	Injection Stuff
	//

	if (value_exists("injectionRatio_" + ToString(particleID))) {
		injectionRatio = value("injectionRatio_" + ToString(particleID));
	} else {
		injectionRatio = value("injectionRatio");
	}

	// get parameters for the injected spectrum
	injectionSpectrumParticleRatio = value("injectionSpectrumParticleRatio"); //0.5
	injectionSpectrumEnergyRatio = value("injectionSpectrumEnergyRatio"); //0.5
	injectionSpectrumIndex = value("injectionSpectrumIndex"); //1.5
	injectionAccelerationFactor = value_exists("injectionAccelerationFactor") ? value("injectionAccelerationFactor") : 1; //1.0


	//
	//	Data Access
	//

	// 	set index limits for the user fields storing the regarding particle spectrum
	// 	and rename them
	nBins = value("nMomentumBins");
	this->omb = omb;
	ome = omb + nBins - 1;

	int paddingBins = log10(ome - omb) + 1;
	for (int iom = omb; iom <= ome; iom++) {
		gdata.om_user[iom].rename( "particle" + ToString(particleID, 2) + "_" + ToString(iom - omb, paddingBins));
	}

	//
	//	Momentum Gridding
	//
	lo = 0;
	hi = nBins - 1;

	REAL momentumb = 0;
	REAL momentume = 0;

	if (value_exists("momentumb_" + ToString(particleID))) {
		momentumb = value("momentumb_" + ToString(particleID, 2)) / particleRestMass_num;
	} else {
		momentumb = value("momentumb") / particleRestMass_num;
	}

	if (value_exists("momentume_" + ToString(particleID))) {
		momentume = value("momentume_" + ToString(particleID, 2)) / particleRestMass_num;
	} else {
		momentume = value("momentume") / particleRestMass_num;
	}

	grid = LogarithmicGrid(momentumb, momentume, nBins, 3);

	ccout << " ... Particle " << particleID << ": " << particleType << "\n";
  ccout << "      - Charge: " << particleCharge << " " << particleCharge_num << "\n";
  ccout << "      - Mass:   " << particleRestMass << " " << particleRestMass_num << " " << particleRestEnergy_num << "\n";
	ccout << "      - Range:  " << get_momentumCenter(lo) << " - " << get_momentumCenter(hi) << "\n";
	ccout << "      - Fields: " << omb << " - " << ome << endl;

}

Particle::~Particle(
		)
{
}

inline REAL compute_powerlawIntegral (
		REAL lorentz,
		REAL spectralIndex
		)
{
	REAL ret = pow(lorentz, 1. - spectralIndex) / (1. - spectralIndex);
	return ret;
}

void Particle::set_powerlaw(
		REAL powerlawConstant,
		REAL spectralIndex,
		int ix, int iy, int iz
		)
{
	REAL this_integral = 0;
	REAL last_integral = compute_powerlawIntegral(get_momentumEdge(lo), spectralIndex);

	for (int ip=lo; ip <= hi; ip++) {
		   this_integral = compute_powerlawIntegral(get_momentumEdge(ip + 1), spectralIndex);
		   (*this)(ip, ix, iy, iz) = powerlawConstant * (this_integral - last_integral)	/ (get_momentumEdge(ip + 1) - get_momentumEdge(ip));
		   last_integral = this_integral;
	}
}

void Particle::clear()
{
	for (int iom = omb; iom <= ome; iom++) {
		gdata.om_user[iom].clear();
	}
}

REAL Particle::numberDensity(
		int ix, int iy, int iz,
		int method
) const
{
	if (method == 0) {
		REAL sum = 0;
		for (int ip=lo; ip <= hi; ip++) {
			sum += (*this)(ip, ix, iy, iz) * get_momentumWidth(ip);
		}
		return sum;
	} else {
		throw CException("Invalid Method");
	}
}


REAL Particle::energyDensity(
		int ix, int iy, int iz,
		int method
) const
{
	if (method == 0) {
		REAL sum = 0;
		for (int ip=lo; ip <= hi; ip++) {
			sum += (*this)(ip, ix, iy, iz) * get_momentumCenter(ip) * get_momentumWidth(ip);
		}
		sum *= particleRestMass_num;
		return sum;
	} else {
		throw CException("Invalid Method");
	}
}


#endif
