#include "constants.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)
#include "ParticleTransport.H"
#include "GslWrapper.H"

using namespace std;
using namespace CRONOS_CONSTANTS;

class ParticleTransportNormalisation : public normalisation {
public:
	virtual void setup();
};

void ParticleTransportNormalisation::setup() {

	set_definitionFromParent(TEMP);
	set_definitionFromParent(TIME);
	set_definitionFromParent(NUM_DENS);

	set_definition(VEL, SpeedOfLight);
	set_definition(ENERGY, ElectronVolt);
	set_definition(MAG_IND, Tesla);

	finish_setup();
}


ParticleTransport::ParticleTransport(
		const Data & gdata
		)
{
	q_rho = gdata.fluid.get_q_rho();
	q_sx = gdata.fluid.get_q_sx();
	q_sy = gdata.fluid.get_q_sy();
	q_sz = gdata.fluid.get_q_sz();
	q_Bx = gdata.fluid.get_q_Bx();
	q_By = gdata.fluid.get_q_By();
	q_Bz = gdata.fluid.get_q_Bz();
	q_Eges = gdata.fluid.get_q_Eges();
	q_Eadd = gdata.fluid.get_q_Eadd();

	nParticles = 0;
	nLossProcesses = 0;
	iCurrentParticle = -1;
	currentParticle = NULL;

	// Bins for which the injection boundary condition
	// will be applied - hardcoded for the time being
	int d[1] = {1};
	NumMatrix<int, 1> injectionBins(d);
	injectionBins(0) = 0;

	// Todo set standart norms
	normHD = NULL;
	normParticle = NULL;

	lorentz_inv = 1.;
}


ParticleTransport::~ParticleTransport()
{
	delete normParticle;
	for (auto ptr: radiativeLossProcesses) delete ptr;
	for (auto ptr: emissionPostProcesses) delete ptr;
}


void ParticleTransport::set_normHD(
		normalisation * normHD
		)
{
	this->normHD = normHD;

	normParticle = new ParticleTransportNormalisation;
	normParticle->set_parent(normHD);
	normParticle->setup();

	ccout << "Finished setting Particle Normalisation" << endl;
	normParticle->print_norms(ccout);
}


normalisation & ParticleTransport::get_normHD(
		)
{
	assert(normHD != NULL);
	return * normHD;
}

normalisation & ParticleTransport::get_normParticle(
		)
{
	assert(normParticle != NULL);
	return * normParticle;
}

void ParticleTransport::add_particle(
		Particle * particle
		)
{
	particles.push_back(particle);
	nParticles++;
}

Particle & ParticleTransport::get_particle(
		int iParticle
		)
{
	return *particles[iParticle];
}

Particle & ParticleTransport::get_currentParticle(
		)
{
	return * currentParticle;
}

Particle & ParticleTransport::set_currentParticle(
		int iParticle
		)
{
	iCurrentParticle = iParticle;
	currentParticle = particles[iParticle];
	Particle & particle = * currentParticle;

	for (int iLoss = 0; iLoss < nLossProcesses; iLoss++) {
		lossProcesses[iLoss]->set_particle(particle);
	}

	return particle;
}


void ParticleTransport::add_lossProcess(
		LossProcess * lossProcess
		)
{
	lossProcesses.push_back(lossProcess);

	nLossProcesses++;
}

void ParticleTransport::add_lossProcess(
		RadiativeLossProcess * lossProcess
		)
{
	radiativeLossProcesses.push_back(lossProcess);
	add_lossProcess((LossProcess *) lossProcess);
}

/**
 * @brief Adds magnetic field to the particle transport
 * @details The magnetic fields added with this method will be considered for the particle injection.
 * 			These magnetic fields are NOT relevant for the particle transport itself.
 * @param magneticField that should be considered
 */
void ParticleTransport::add_magneticField(
		MagneticField * magneticField
)
{
	magneticFields.push_back(magneticField);
}

/**
 * @brief Finishes the setup of particle transport
 * @details This function is called after Problem::init_particles
 * @param gdata Data object
 */
void ParticleTransport::finish_setup(
		Data & gdata
		)
{
	ccout << "Finishing particle transport setup..." << endl;
	assert(normHD != NULL);
	assert(normParticle != NULL);

	Quantity gammaMaxCoeff = sqrt(
			VacuumPermeability * SpeedOfLight * 1.854 /* F(pi/2, 1/2) *//
			(sqrt(2) * pi * ThomsonCrossSection)
	);
	gammaMaxCoeff_num = gammaMaxCoeff.normalizeTo(
			sqrt( normHD->normQuantity(normHD->MAG_IND) /
			      normParticle->normQuantity(normParticle->CHAR)
			)
	);

	ccout << "Particle transport setup is done!" << endl;
}



void ParticleTransport::compute_constants(
		Data & gdata,
		int ix, int iy, int iz
		)
{
	// This function is called at the very beginning of every momentum singlestep

	//
	//	For Injection & Adiabatic Losses
	//

	gdata.du_cov(ix, iy, iz) = 0.;
	gdata.du_cov(ix, iy, iz) += (gdata.lorentz(ix, iy, iz) - gdata.lorentz_old(ix, iy, iz)) / gdata.dt;
//	cout << gdata.du_cov(ix, iy, iz) << " | " << gdata.lorentz(ix, iy, iz) << " " << gdata.lorentz_old(ix, iy, iz) << " " << gdata.dt << endl;

	gdata.du_cov(ix, iy, iz) += (gdata.om[q_sx](ix + 1, iy, iz) - gdata.om[q_sx](ix - 1, iy, iz)) /
								(gdata.getCen(0, ix + 1) - gdata.getCen(0, ix - 1));
	gdata.du_cov(ix, iy, iz) += (gdata.om[q_sy](ix, iy + 1, iz) - gdata.om[q_sy](ix, iy - 1, iz)) /
								(gdata.getCen(1, iy + 1) - gdata.getCen(1, iy - 1));
	gdata.du_cov(ix, iy, iz) += (gdata.om[q_sz](ix, iy, iz + 1) - gdata.om[q_sz](ix, iy, iz - 1)) /
								(gdata.getCen(2, iz + 1) - gdata.getCen(2, iz - 1));


	//
	//	Auxiliaries
	//


	// To be consistent
//	gdata.du_cov(ix, iy, iz) = -log(momentumLoss_c1_int) * 3. * 0.5 * (gdata.lorentz(ix, iy, iz) + gdata.lorentz_old(ix, iy, iz)) / gdata.dt;


	// debug
//	compute_transportConstants(gdata, ix, iy, iz);

	gdata.magneticField_LAB(ix, iy, iz) = 0; // in hd units
	gdata.magneticField_COM(ix, iy, iz) = 0; // in hd units
	for(auto const& magField: magneticFields) {
		gdata.magneticField_LAB(ix, iy, iz) += magField->get_magneticFieldStrength_LAB(gdata, ix, iy, iz);
		gdata.magneticField_COM(ix, iy, iz) += magField->get_magneticFieldStrength_COM(gdata, ix, iy, iz);
	}

}

/**
 * @brief Gets loss processes ready for the SL-step.
 * @details This function calls the compute_constants() for every loss process.
 * 			The computed constants are needed for the Semilagrange-Step.
 * @param gdata Data object
 * @param ix,iy,iz Location indices
 */
void ParticleTransport::compute_transportConstants(
		Data & gdata,
		int ix, int iy, int iz
		)
{
	for (int iLoss = 0; iLoss < nLossProcesses; iLoss++) {
		lossProcesses[iLoss]->compute_constants(gdata, ix, iy, iz);
	}

	lorentz_inv = 1. / gdata.lorentz(ix, iy, iz);
}


void ParticleTransport::get_momentumLoss(
		REAL momentum,
		REAL & loss
		)
{
	loss = 0;

	for (int iLoss = 0; iLoss < nLossProcesses; iLoss++) {
		lossProcesses[iLoss]->get_momentumLoss(momentum, loss);
	}

	loss *= lorentz_inv;
}

REAL ParticleTransport::get_momentumLoss_c1_int(
		REAL momentum
	)
{
	REAL momentumLoss_c1_int = 1.;
	for (int iLoss = 0; iLoss < nLossProcesses; iLoss++) {
		momentumLoss_c1_int *= lossProcesses[iLoss]->get_momentumLoss_c1_int(momentum);
	}

	return momentumLoss_c1_int;
}

REAL  ParticleTransport::get_momentumLoss_c2(
		REAL momentum
		)
{
	REAL momentumLoss_c2 = 0.;
	for (int iLoss = 0; iLoss < nLossProcesses; iLoss++) {
		momentumLoss_c2 += lossProcesses[iLoss]->get_momentumLoss_c2(momentum);
	}

	return momentumLoss_c2;
}


void ParticleTransport::get_Cons(
		Data & gdata,
		cronos::vector<REAL> &Pos,
		phys_fields_1D &physField,
		phys_fields_1D &physFieldsUser,
		int dir, REAL shift)
{
	// Set / compute conservative variables:
	REAL lorentz;
	for (int iParticle = 0; iParticle < nParticles; ++iParticle) {
		Particle & particle = *particles[iParticle];
		for (int ip = particle.getOmb(); ip <= particle.getOme(); ++ip) {
			for (int i = -2; i <= gdata.mx[dir]+1; ++i){
				physFieldsUser.uCon[ip](i) = physFieldsUser.uPri[ip](i) * physField.uCon[q_rho](i) / physField.uPri[q_rho](i);
			}

//			for (int i = -2; i <= gdata.mx[dir]+1; ++i){
//				lorentz = physField.uCon[q_rho](i) / physField.uPri[q_rho](i);
//				physFieldsUser.uCon[iMom](i) = physFieldsUser.uPri[iMom](i) * lorentz;
//			}
		}

	}

}


void ParticleTransport::TransPrim2Cons(
		Data & gdata
		)
{
	// Change order maybe ??
	for (int iParticle = 0; iParticle < nParticles; ++iParticle) {
		for (int iMom = particles[iParticle]->getLo(); iMom <= particles[iParticle]->getHi(); ++iMom) {
			for (int iz = 0; iz <= gdata.mx[2]; iz++) {
			for (int iy = 0; iy <= gdata.mx[1]; iy++) {
			for (int ix = 0; ix <= gdata.mx[0]; ix++) {
				(*particles[iParticle])(iMom, ix, iy, iz) *= gdata.lorentz(ix, iy, iz);
			}
			}
			}
		}
	}
}

void ParticleTransport::TransCons2Prim(
		Data & gdata
		)
{
	// Change order maybe ??
	for (int iParticle = 0; iParticle < nParticles; ++iParticle) {
		for (int iMom = particles[iParticle]->getLo(); iMom <= particles[iParticle]->getHi(); ++iMom) {
			for (int iz = 0; iz <= gdata.mx[2]; iz++) {
			for (int iy = 0; iy <= gdata.mx[1]; iy++) {
			for (int ix = 0; ix <= gdata.mx[0]; ix++) {
				(*particles[iParticle])(iMom, ix, iy, iz) /= gdata.lorentz(ix, iy, iz);
			}
			}
			}
		}
	}
}


void ParticleTransport::get_PhysFlux(
		Data &gdata,
		cronos::vector<REAL> &iPos,
		phys_fields_1D &pf,
		phys_fields_1D &pfUser,
		int dir,
		REAL shift
		)
{
	for (int iParticle = 0; iParticle < nParticles; ++iParticle) {
		for (int ip = particles[iParticle]->getOmb(); ip <= particles[iParticle]->getOme(); ip++) {
			pfUser.flux[ip].clear();
			for (int i = -1; i <= gdata.mx[dir]+1; ++i){
				pfUser.flux[ip](i) = pf.uPri[q_sx+dir](i) * pfUser.uPri[ip](i) ;
			}
		}
	}
}

std::vector<RadiativeLossProcess *> & ParticleTransport::get_radiativeLossProcesses(
		)
{
	return radiativeLossProcesses;
}

void ParticleTransport::add_emissionPostProcess(
		EmissionPostProcess * EmissionPostProcess
)
{
	emissionPostProcesses.push_back(EmissionPostProcess);
}

std::vector<EmissionPostProcess *> & ParticleTransport::get_emissionPostProcesses(
)
{
	return emissionPostProcesses;
}


int ParticleTransport::injectParticles(
		Data & gdata,
		int ix, int iy, int iz,
		const NumMatrix<REAL, 1> & momentumEdges,
		NumMatrix<REAL, 1> & momentumSpectrum
		)
{
	if (gdata.du_cov(ix, iy, iz) < -100) {
		Particle & particle = *currentParticle;

//		REAL injectionRate = particle.injectionRatio * normParticle->renormalise(normParticle->DENS, normHD, gdata.om[q_rho](ix, iy, iz)) / particle.particleRestMass_num;
		REAL injectionRate = particle.injectionRatio;
		int iBin;
		for (int ip = injectionBins.getLow(0); ip <= injectionBins.getHigh(0); ip++) {

			iBin = injectionBins(ip);
			REAL loss = 1.;
//			get_momentumLoss(particle.get_momentumCenter(iBin), loss);
			momentumSpectrum(ip) += injectionRate / loss * gdata.dt;
		}
		return 1;
	}
//	else
//	if (gdata.du_cov(ix, iy, iz) < 0) {
//		return 2;
//	}

	return 0;
}


//class ImplicitInjectionGammaMin: public ExplicitSlopeProblem {
//public:
//	virtual void set_x(REAL gamma_min_) {
//		gamma_min = std::abs(gamma_min_);
//		gamma_min_s = std::pow(gamma_min, 1. - spectralIndex);
//		lhs = gamma_min_s * (coeff_a + coeff_b * gamma_min);
//	}
//
//	virtual REAL get_value() {
//		return lhs - rhs;
//	}
//
//	virtual REAL get_slope() {
//		return coeff_b * ( lhs / gamma_min + gamma_min_s );
//	}
//
//	void set_parameters(REAL _spectralIndex, REAL _specificEnergy_nt, REAL _gamma_max) {
//		spectralIndex = _spectralIndex;
//		specificEnergy_nt = _specificEnergy_nt;
//		gamma_max = _gamma_max;
//
//		gamma_max_s = std::pow(gamma_max, 1. - spectralIndex);
//
//		coeff_a = specificEnergy_nt * (spectralIndex - 2.);
//		coeff_b = 1. - spectralIndex;
//
//		rhs = gamma_max_s * (coeff_a + coeff_b * gamma_max);
//	}
//
//private:
//	REAL spectralIndex, specificEnergy_nt, gamma_min, gamma_max;
//	REAL gamma_min_s, gamma_max_s, coeff_a, coeff_b, lhs, rhs;
//};


ImplicitInjectionGammaMin::ImplicitInjectionGammaMin() : ExplicitSlopeProblem("ImplicitInjectionGammaMin")
{
}

void ImplicitInjectionGammaMin::set_x(
		REAL _gammaMin)
{
	gammaMin = _gammaMin;
	gammaMinS = std::pow(gammaMin, spectralIndex);
}

REAL ImplicitInjectionGammaMin::get_specific()
{
	double denom = (gammaMaxS*gammaMin - gammaMax*gammaMinS) * (spectralIndex - 2);
	double num =
			(-2. + spectralIndex) * gammaMax * gammaMinS +
			(2. - spectralIndex) * gammaMaxS * gammaMin +
			(-1. + spectralIndex) * gammaMaxS * sqr(gammaMin) +
			(1. - spectralIndex) * sqr(gammaMax) * gammaMinS ;

	return num / denom;
}

REAL ImplicitInjectionGammaMin::get_value()
{
	double denom = (gammaMaxS*gammaMin - gammaMax*gammaMinS) * (spectralIndex - 2);
	double num =
			(-2. + spectralIndex) * gammaMax * gammaMinS +
			(2. - spectralIndex) * gammaMaxS * gammaMin +
			(-1. + spectralIndex) * gammaMaxS * sqr(gammaMin) +
			(1. - spectralIndex) * sqr(gammaMax) * gammaMinS ;

	return nonthermalEnergy_specific - num/denom;
}

REAL ImplicitInjectionGammaMin::get_slope()
{
	double denom = sqr(gammaMaxS*gammaMin - gammaMax*gammaMinS)*(-2 + spectralIndex);
	double num = gammaMaxS*(-1 + spectralIndex) * (
			gammaMax*gammaMin*gammaMinS*(-2 + spectralIndex)
			+ gammaMaxS*sqr(gammaMin)
			- gammaMinS*(-1 + spectralIndex)*sqr(gammaMax)
	);

	return -num/denom;
}

void ImplicitInjectionGammaMin::set_parameters(REAL _spectralIndex, REAL _nonthermalEnergy_specific, REAL _gammaMax)
{
	spectralIndex = _spectralIndex;
	nonthermalEnergy_specific = _nonthermalEnergy_specific;
	gammaMax = _gammaMax;
	gammaMaxS = std::pow(gammaMax, spectralIndex);

	if (spectralIndex == 2)	spectralIndex += 1e-6;
}



inline REAL compute_spectrumParticleIntegral (
		REAL lorentz,
		REAL injectionSpectrumIndex
		)
{
	if (injectionSpectrumIndex == 1.) {
		return log(lorentz);
	} else {
		REAL sred = 1. - injectionSpectrumIndex;
		return pow(lorentz, sred) / sred;
	}
}

inline REAL compute_spectrumParticleIntegral_momentum (
		REAL momentum,
		REAL injectionSpectrumIndex
		)
{
	const REAL lorentz = sqrt(sqr(momentum) + 1.);
	return compute_spectrumParticleIntegral(lorentz,injectionSpectrumIndex);
}

inline REAL compute_spectrumEnergyIntegral (
		REAL lorentz,
		REAL injectionSpectrumIndex
		)
{
	if (injectionSpectrumIndex == 2.) {
		return log(lorentz);
	} else {
		REAL sred = 2. - injectionSpectrumIndex;
		return pow(lorentz, sred) / sred;
	}
}

inline REAL compute_spectrumEnergyIntegral_momentum (
		REAL momentum,
		REAL injectionSpectrumIndex
		)
{
	const REAL lorentz = sqrt(sqr(momentum) + 1.);
	return compute_spectrumEnergyIntegral(lorentz, injectionSpectrumIndex);
}


inline REAL compute_spectrumMomentumIntegral (
		REAL lorentz,
		REAL injectionSpectrumIndex
		)
{
	return compute_spectrumEnergyIntegral(lorentz, injectionSpectrumIndex)
		- compute_spectrumParticleIntegral(lorentz, injectionSpectrumIndex);
}

inline REAL compute_spectrumMomentumIntegral_momentum (
		REAL momentum,
		REAL injectionSpectrumIndex
		)
{
	const REAL lorentz = sqrt(sqr(momentum) + 1.);
	return compute_spectrumEnergyIntegral(lorentz, injectionSpectrumIndex);
}

inline REAL compute_maxwellianIntegral (
    const REAL lorentz,
    const REAL meanLorentz
)
{
  return - std::exp(- lorentz / meanLorentz) *
         (sqr(lorentz) + 2 * lorentz * meanLorentz + 2 * sqr(meanLorentz)) /
         (2 * sqr(meanLorentz));
}


PowerlawParameters ParticleTransport::computePowerlawParameters(
    Data & gdata,
    int ix, int iy, int iz,
    const Particle& particle,
    const double numberDensity,
    const double energyDensityNormalised
)
{
  const double specificEnergyNormalised = energyDensityNormalised / numberDensity;
  gdata.powerlawSpecEnergy(ix, iy, iz) = specificEnergyNormalised;

  PowerlawParameters powerlawParameters;
  powerlawParameters.errorcode = -1;
  powerlawParameters.spectralIndex = particle.injectionSpectrumIndex;

  REAL magneticField_strength = 0; // in hd units
  for(auto const& magField: magneticFields) {
    magneticField_strength += magField->get_magneticFieldStrength_COM(gdata, ix, iy, iz);
  }

  powerlawParameters.gammaMax = gammaMaxCoeff_num * sqrt(abs(particle.particleCharge_num)/(magneticField_strength * particle.injectionAccelerationFactor));
  gdata.gammaMax(ix,iy,iz) = powerlawParameters.gammaMax;

  const double numberIntegralGammaMax = compute_spectrumParticleIntegral(powerlawParameters.gammaMax, particle.injectionSpectrumIndex);
  const double energyIntegralGammaMax = compute_spectrumEnergyIntegral(powerlawParameters.gammaMax, particle.injectionSpectrumIndex);

  // now onto gamma_min
  powerlawParameters.gammaMin = 1;
  if (particle.injectionSpectrumIndex > 2.) {
    powerlawParameters.gammaMin = specificEnergyNormalised * (particle.injectionSpectrumIndex - 2.) / (particle.injectionSpectrumIndex - 1.);
  }

  // Check if enough energy is available
  const double spectrumParticleIntegralMax = numberIntegralGammaMax
                                         - compute_spectrumParticleIntegral(1., particle.injectionSpectrumIndex);

  const double spectrumEnergyIntegralMax = energyIntegralGammaMax
                                         - compute_spectrumEnergyIntegral(1., particle.injectionSpectrumIndex);

  const double spectrumSpecificEnergyMin = spectrumEnergyIntegralMax / spectrumParticleIntegralMax;

  if (spectrumSpecificEnergyMin > specificEnergyNormalised) {
    // Not enough energy available in the fluid
    // reducing particle number

    powerlawParameters.gammaMin = 1;
    powerlawParameters.normalisation = energyDensityNormalised / spectrumEnergyIntegralMax;
    powerlawParameters.errorcode = 2;
    return powerlawParameters;

  } else if (powerlawParameters.gammaMax < specificEnergyNormalised) {
    // Too much energy in the fluid: the spectrum cannot contain more energy (gammaMin = gammaMax)!
    // Reducing the energy

    powerlawParameters.gammaMin = powerlawParameters.gammaMax;
    powerlawParameters.normalisation = numberDensity / spectrumParticleIntegralMax;
    powerlawParameters.errorcode = 3;
    return powerlawParameters;
  } else {

    const double s = particle.injectionSpectrumIndex;

    auto fun = [=](REAL gamma_min) -> REAL {
      const REAL LHS = energyIntegralGammaMax - compute_spectrumEnergyIntegral(gamma_min, s);
      const REAL RHS = specificEnergyNormalised
                       * (numberIntegralGammaMax - compute_spectrumParticleIntegral(gamma_min, s));
      return LHS - RHS;
    };

    static GslRootSolver rootSolver;
    rootSolver.max_iter = 40;
    rootSolver.tolRel = 1e-6;
    rootSolver.fun_f = fun;

    powerlawParameters.gammaMin = rootSolver.solve_f(1., 0.999 * powerlawParameters.gammaMax);

    if (powerlawParameters.gammaMin > powerlawParameters.gammaMax || cronos_isnan(powerlawParameters.gammaMin)) {
      cerr << " gamma_min bigger than gamma_max " << powerlawParameters.gammaMin << " | " << gdata.rank << " " << ix << " " << iy << " " << iz << endl;

      powerlawParameters.errorcode = -10;
      return powerlawParameters;
    } else {
      powerlawParameters.errorcode = 1;
      powerlawParameters.normalisation = numberDensity / (numberIntegralGammaMax - compute_spectrumParticleIntegral(powerlawParameters.gammaMin, s));
    }

    return powerlawParameters;
  }
}

inline double clipGamma(const double gamma, const PowerlawParameters& powerlawParameters) {
  return min(max(gamma, powerlawParameters.gammaMin), powerlawParameters.gammaMax);
}

void injectPowerlaw(const PowerlawParameters& powerlawParameters, const Particle& particle,
                    const NumMatrix<REAL, 1> & momentumEdges, NumMatrix<REAL, 1> & momentumSpectrum) {
  int lo = particle.getLo();
  int hi = particle.getHi();

  // Determine the spectrum Limits
  int injectionLo = lo, injectionHi=hi;

  for (int ip=lo; ip <= hi; ip++) {
    if( powerlawParameters.gammaMin > momentumEdges(ip) ) {
      injectionLo = ip;
    }
    if( powerlawParameters.gammaMax < momentumEdges(ip + 1) ) {
      injectionHi = ip;
      break;
    }
  }

  REAL last_integral = compute_spectrumParticleIntegral(clipGamma(momentumEdges(injectionLo), powerlawParameters),
                                                                 powerlawParameters.spectralIndex);
  REAL this_integral = 0;

  for (int ip = injectionLo; ip <= injectionHi; ip++) {
    this_integral = compute_spectrumParticleIntegral(clipGamma(momentumEdges(ip + 1), powerlawParameters)
        , powerlawParameters.spectralIndex);
    momentumSpectrum(ip) += powerlawParameters.normalisation * (this_integral - last_integral)
                            / (momentumEdges(ip + 1) - momentumEdges(ip));
    last_integral = this_integral;
  }
}

struct MaxwellianParameters {
  double normalistion;
  double gammaMean;
};


MaxwellianParameters computeMaxwellianParameters(
    Data & gdata,
    int ix, int iy, int iz,
    const Particle& particle,
    const double numberDensity,
    const double energyDensityNormalised
)
{
  MaxwellianParameters maxwellianParameters;

  if (numberDensity < 1e-30 * energyDensityNormalised) {
    maxwellianParameters.normalistion = 0.;
    maxwellianParameters.gammaMean = 0.;
    return maxwellianParameters;
  }

  const double specificEnergyNormalised = energyDensityNormalised / numberDensity;

  maxwellianParameters.normalistion = numberDensity;
  maxwellianParameters.gammaMean = specificEnergyNormalised / 3.;

  return maxwellianParameters;
}


void injectMaxwellian(const MaxwellianParameters& parameters, const Particle& particle, const NumMatrix<REAL, 1> & momentumEdges, NumMatrix<REAL, 1> & momentumSpectrum) {
  int lo = particle.getLo();
  int hi = particle.getHi();

  REAL this_integral = 0;
  REAL last_integral = compute_maxwellianIntegral(momentumEdges(lo), parameters.gammaMean);

  for (int ip = lo; ip <= hi; ip++) {
    this_integral = compute_maxwellianIntegral(momentumEdges(ip + 1), parameters.gammaMean);
    momentumSpectrum(ip) +=
        parameters.normalistion * (this_integral - last_integral) / (momentumEdges(ip + 1) - momentumEdges(ip));
    last_integral = this_integral;
  }
}


int ParticleTransport::injectParticleSpectrum(
		Data & gdata,
		int ix, int iy, int iz,
		const NumMatrix<REAL, 1> & momentumEdges,
		NumMatrix<REAL, 1> & momentumSpectrum
		)
{
  gdata.fluidSpecEnergy(ix,iy,iz) = 0;
  gdata.powerlawSpecEnergy(ix,iy,iz) = 0;

  gdata.injectionK(ix,iy,iz) = 0;
  gdata.gammaMin(ix,iy,iz) = 0;
  gdata.gammaMax(ix,iy,iz) = 0;

  gdata.injectionGammaMaxwellian(ix,iy,iz) = 0;

	if (gdata.du_cov(ix, iy, iz) < -10) {
		Particle & particle = *currentParticle;

    // Compute the particle fraction
    REAL particleFraction = gdata.om_user[1](ix, iy, iz) / (gdata.om[q_rho](ix, iy, iz) * gdata.lorentz(ix, iy, iz));
    particleFraction = min(1., max(0., particleFraction));

    // Do nothing if there is no fluid
    if (particleFraction < 1e-8) {
      return -3;
    }

    // Get the available fluid density and energy
		REAL particleDensity = gdata.om[q_rho](ix, iy, iz);
    particleDensity *= particleFraction;
		particleDensity = normParticle->renormalise(normParticle->DENS, normHD, particleDensity) / particle.particleRestMass_num;

		REAL thermalEnergyDensity = gdata.om[q_Eges](ix, iy, iz);
    thermalEnergyDensity *= particleFraction;
		thermalEnergyDensity = normParticle->renormalise(normParticle->E_DENS, normHD, thermalEnergyDensity) / particle.particleRestEnergy_num;

    gdata.fluidSpecEnergy(ix, iy, iz) = thermalEnergyDensity / particleDensity;

    // Apply the wind correction factor
    static const double zetaNPulsarWindCorrection = value_exists("zetaNPulsarWindCorrection") ? value("zetaNPulsarWindCorrection") : 1;
    particleDensity *= zetaNPulsarWindCorrection;

    // Compute the powerlaw parameters
    const double numberPowerlaw = particle.injectionSpectrumParticleRatio * particleDensity;
    const double energyPowerlaw = particle.injectionSpectrumEnergyRatio * thermalEnergyDensity;
		PowerlawParameters powerlawParameters = computePowerlawParameters(gdata, ix, iy, iz, particle, numberPowerlaw, energyPowerlaw);

    gdata.gammaMin(ix,iy,iz) = powerlawParameters.gammaMin;
    gdata.injectionK(ix,iy,iz) =  powerlawParameters.normalisation;

    if (powerlawParameters.errorcode <= 0) {
      return powerlawParameters.errorcode;
    }

		// Set the spectrum
		momentumSpectrum.clear();
    injectPowerlaw(powerlawParameters, particle, momentumEdges, momentumSpectrum);


    // Maxwellian
    const double numberMaxwellian = (1. - particle.injectionSpectrumParticleRatio) * particleDensity;
    const double energyMaxwellian = (1. - particle.injectionSpectrumEnergyRatio) * thermalEnergyDensity;
    MaxwellianParameters maxwellianParameters = computeMaxwellianParameters(gdata, ix, iy, iz, particle, numberMaxwellian, energyMaxwellian);

    gdata.injectionGammaMaxwellian(ix,iy,iz) = maxwellianParameters.gammaMean;
    injectMaxwellian(maxwellianParameters, particle, momentumEdges, momentumSpectrum);


		return powerlawParameters.errorcode;
	}

	return 0;
}

#endif
