#ifndef CRONOS_PARTICLE_TRANSPORT_H
#define CRONOS_PARTICLE_TRANSPORT_H


#include <cmath>
#include <sstream>
#include <vector>

#include "data.H"
#include "normalisation.H"
#include "vector.H"
#include "fields_local.H"
#include "Particle.H"
#include "LossProcess.H"
#include "RadiativeLossProcess.H"
#include "root_finder.H"
#include "MagneticField.H"
#include "EmissionPostProcess.H"


struct PowerlawParameters {
  int errorcode = 1; // 1 == ok
  double normalisation;
  double spectralIndex;
  double gammaMin;
  double gammaMax;
};

class ParticleTransport {
public:
	ParticleTransport(
			const Data & gdata
			);

	~ParticleTransport(
			);

	void set_normHD(
			normalisation * normHD
			);

	normalisation & get_normHD(
			);

	normalisation & get_normParticle(
			);

	void add_particle(
			Particle * particle
			);

	Particle & get_particle(
			int iParticle
			);

	std::vector<Particle *> get_particles() {return particles;}

	Particle & set_currentParticle(
			int iParticle
			);

	Particle & get_currentParticle(
			);

	void add_lossProcess(
			LossProcess * lossProcess
			);

	void add_lossProcess(
			RadiativeLossProcess * lossProcess
			);

	void add_magneticField(
			MagneticField * magneticField
			);

	void get_momentumLoss(
			REAL momentum,
			REAL & loss
			);

	void finish_setup(
			Data & gdata
			);

	REAL get_momentumLoss_c1_int(
			REAL momentum
			);

	REAL get_momentumLoss_c2(
			REAL momentum
			);

	int injectParticles(
			Data & gdata,
			int ix, int iy, int iz,
			const NumMatrix<REAL, 1> & momentumEdges,
			NumMatrix<REAL, 1> & momentumSpectrum
			);

	int injectParticleSpectrum(
			Data & gdata,
			int ix, int iy, int iz,
			const NumMatrix<REAL, 1> & momentumEdges,
			NumMatrix<REAL, 1> & momentumSpectrum
			);

	void compute_constants(
			Data & gdata,
			int ix, int iy, int iz
			);

	void compute_transportConstants(
			Data & gdata,
			int ix, int iy, int iz
			);

	void get_Cons(
			Data & gdata,
			cronos::vector<REAL> &Pos,
			phys_fields_1D &physField,
			phys_fields_1D &physFieldsUser,
			int dir, REAL shift);

	void TransPrim2Cons(
			Data & gdata
			);

	void TransCons2Prim(
			Data & gdata
			);

	void get_PhysFlux(
			Data &gdata,
			cronos::vector<REAL> &iPos,
			phys_fields_1D &pf,
			phys_fields_1D &pfUser,
			int dir,
			REAL shift
			);

	// Emission stuff

	std::vector<RadiativeLossProcess *> & get_radiativeLossProcesses(
			);

	void add_emissionPostProcess(
			EmissionPostProcess * EmissionPostProcess
			);
	std::vector<EmissionPostProcess *> & get_emissionPostProcesses(
			);

protected:

  PowerlawParameters computePowerlawParameters(
      Data & gdata,
      int ix, int iy, int iz,
      const Particle& particle,
      const double numberDensity,
      const double energyDensity
  );

	normalisation * normHD;
	normalisation * normParticle;

	int q_rho, q_sx, q_sy, q_sz, q_Bx, q_By, q_Bz, q_Eges, q_Eadd;

	int nParticles;
	int nLossProcesses;

	NumMatrix<int, 1> injectionBins;

	std::vector<Particle *> particles;
	int iCurrentParticle;
	Particle * currentParticle;

	std::vector<LossProcess *> lossProcesses;
	std::vector<RadiativeLossProcess *> radiativeLossProcesses;
	std::vector<EmissionPostProcess *> emissionPostProcesses;

	std::vector<MagneticField *> magneticFields; /** Magnetic fields for the particle injection */
	REAL gammaMaxCoeff_num;

	// Local constants or buffers
	// set by compute_constants
	REAL lorentz_inv;
};


class ImplicitInjectionGammaMin : public ExplicitSlopeProblem
{
public:
	ImplicitInjectionGammaMin();
	void set_x(REAL _gammaMin);
	REAL get_value();
	REAL get_slope();
	void set_parameters(REAL _spectralIndex, REAL _nonthermalEnergy_specific, REAL _gammaMax);
	REAL get_specific();

private:
	REAL spectralIndex, nonthermalEnergy_specific, gammaMin, gammaMax;
	REAL gammaMinS, gammaMaxS;
};


#endif
