#ifndef POWERLAW_RECONST_H
#define POWERLAW_RECONST_H

#include "matrix.H"
#include "Grid1D.H"
#include <cmath>

#ifndef REAL
#define REAL double
#endif

class PowerlawReconstruction {
public:
	PowerlawReconstruction() {}

	PowerlawReconstruction(const LogarithmicGrid & grid, const NumMatrix<REAL, 1> &values)
	{
		setData(grid, values);
	}

	void setData(const LogarithmicGrid & grid, const NumMatrix<REAL, 1> &values) {
		setData(grid, values, 0, grid.get_mx());
	}

  void setData(const LogarithmicGrid & grid, const NumMatrix<REAL, 1> &values, const int _loComp, const int _hiComp) {
		if (grid_ptr != &grid) {
			grid_ptr = &grid;
			values_ptr = &values;

			amplitudes.resize(grid.getLow(), grid.getHigh());
			exponents.resize(grid.getLow(), grid.getHigh());
			valuesLog.resize(grid.getLow(), grid.getHigh());
			lo = grid.getLo();
			hi = grid.getHi();
		}

		loComp = std::max(_loComp, lo);
		hiComp = std::min(_hiComp, hi);

		xLoComp = grid.get_xL(loComp);
		xHiComp = grid.get_xL(hiComp + 1);

		computeLogs();
		computeExponents();
		computeAmplitudes();
	}

	REAL operator()(const REAL x, const int idx)
	{
		if (amplitudes(idx) == 0) {
			return 0;
		} else {
			return amplitudes(idx) * myPow(x, exponents(idx));
		}
	}

		REAL operator()(const REAL x)
	{
		if (x >= xHiComp) {
			return this->operator()(xHiComp, hiComp);
		}
		if (x <= xLoComp) {
			return this->operator()(xLoComp, loComp);
		}

		int idx = grid_ptr->get_ipos(x);
		idx = std::min(idx, hi);
		idx = std::max(idx, lo);

		return this->operator()(x, idx);
	}

	const NumMatrix<REAL, 1> & getExponents() const {return exponents;};
	const NumMatrix<REAL, 1> & getAmplitudes() const {return amplitudes;};

private:
	const NumMatrix<REAL, 1> * values_ptr;
	const LogarithmicGrid * grid_ptr;

	NumMatrix<REAL, 1> amplitudes, exponents;
	NumMatrix<REAL, 1> valuesLog;
	int lo, hi, loComp, hiComp;
	double xLoComp, xHiComp;

	void computeLogs() {
    const auto & grid = *grid_ptr;
		const auto & values = *values_ptr;

		for(int i=lo; i <= hi; ++i) {
			valuesLog(i) = values(i) > 0 ? myLog(values(i)) : 0;
		}
	}

	void computeExponents() {
		if (loComp != hi)
    if (loComp == lo) {
		  exponents(lo) = computeSR(lo);
    } else {
      exponents(loComp) = computeSLimited(loComp);
    }

		if (hiComp != lo)
    if (hiComp == hi) {
			exponents(hi) = computeSL(hi);
		} else {
      exponents(hiComp) = computeSLimited(hiComp);
    }

		for(int i=loComp + 1; i <= hiComp - 1; ++i) {
      exponents(i) = computeSLimited(i);
		}
	}

	void computeAmplitudes() {
    const auto & grid = *grid_ptr;
		const auto & values = *values_ptr;

		for(int i=loComp; i <= hiComp; ++i) {
			if(values(i) > 0) {
				amplitudes(i) = values(i) * grid.get_del(i) / powerlawIntegralLog(exponents(i), grid.get_xL_log(i), grid.get_xL_log(i+1));
				if (!std::isfinite(amplitudes(i))) {
					std::cerr << "PowerlawReconstruction: non-finite value detected" << std::endl;
					amplitudes(i) = values(i);
					exponents(i) = 0;
				}
			} else {
				amplitudes(i) = 0;
			}
		}
	}

	REAL powerlawIntegral(const REAL s, const REAL a, const REAL b) const {
		const REAL sInt = s+1.;
		if (std::abs(sInt) < 1e-8) {
			return log(b) - log(a);
		} else {
			return (myPow(b, sInt) - myPow(a, sInt)) / sInt;
		}
	}

	REAL powerlawIntegralLog(const REAL s, const REAL aLog, const REAL bLog) const {
		const REAL sInt = s+1.;
		if (std::abs(sInt) < 1e-8) {
			return bLog - aLog;
		} else {
			return (exp(bLog * sInt) - exp(aLog * sInt)) / sInt;
		}
	}


	REAL computeSL(int idx) const {
    const auto & grid = *grid_ptr;
		const auto & values = *values_ptr;

		if(values(idx-1) > 0 && values(idx) > 0 ) {
			return (valuesLog(idx) - valuesLog(idx-1)) / (grid.get_xC_log(idx) - grid.get_xC_log(idx-1));
		} else {
			return 0;
		}
	}

	REAL computeSC(int idx) const {
    const auto & grid = *grid_ptr;
		const auto & values = *values_ptr;

		if(values(idx-1) > 0 && values(idx+1) > 0 ) {
			return (valuesLog(idx+1) - valuesLog(idx-1)) / (grid.get_xC_log(idx+1) - grid.get_xC_log(idx-1));
		} else {
			return 0;
		}
	}

	REAL computeSR(int idx) const {
    const auto & grid = *grid_ptr;
		const auto & values = *values_ptr;

		if(values(idx) > 0 && values(idx+1) > 0 ) {
			return (valuesLog(idx+1) - valuesLog(idx)) / (grid.get_xC_log(idx+1) - grid.get_xC_log(idx));
		} else {
			return 0;
		}
	}

  REAL computeSLimited(int idx) const {
    const auto & grid = *grid_ptr;
		const auto & values = *values_ptr;

		REAL slope = 0;

    if(values(idx-1) > 0 && values(idx) > 0 && values(idx+1) > 0 ) {
      const REAL sL = computeSL(idx);
      const REAL sC = computeSC(idx);
      const REAL sR = computeSR(idx);

			// Van Leer Limiter
			slope = sL * sR > 0 ? sL * sR / sC : 0;

			// Set small numbers to zero to allow optimization
			slope = std::abs(slope) > 1e-12 ? slope : 0;

			// Limit maximum slope
			constexpr const REAL maxSlope = 20;
			const REAL slopeSign = slope >= 0 ? 1. : -1;
			slope = slopeSign * std::min(maxSlope, std::abs(slope));
    } 
		return slope;
  }

	static double myPow(double a, double b) {
    // return powf(a, b);
    // return pow(a, b);
		return pow(a, b);
	}

	static inline double myLog(double x) { return logf(x); }

};

#endif
