#ifndef CRONOS_RADIATION_FIELD_H
#define CRONOS_RADIATION_FIELD_H

#include "data.H"
#include "normalisation.H"
#include "CException.H"
#include "LorentzTrafo.H"
#include "marray.H"

#include <cassert>
#include <vector>


/**
 * @brief Radiation Field Interface class.
 *
 * @details The radiation field provides
 * differential photon number densities in terms of
 *  - dN / (dV dE dOmega), with directional dependece
 *  - dN / (dV dE)
 */
class RadiationField {
public:

	RadiationField() {};

	virtual ~RadiationField() {};

	/** number density: particles per volume */
	virtual REAL numberDensity(
			Data & gdata,
			int ix, int iy, int iz
	);

	virtual REAL numberDensity(
			Data & gdata,
			const REAL pos[3]
	) = 0;

	/** differential number density: particles per volume per (steradian, energy) */
	virtual REAL numberDensity_dEdOmega(
			Data & gdata,
			int ix, int iy, int iz,
			REAL energy,
			const REAL dir[3]
	);

	virtual REAL numberDensity_dEdOmega(
			Data & gdata,
			const REAL pos[3],
			REAL energy,
			const REAL dir[3]
	) = 0;

	/** To check if the RF is defined
	 * 	Default is true
	 * */
	virtual bool isDefined(
			const Data & gdata,
			const REAL pos[3]
	) const;

};


#define templateRadiationField(RadiationFieldType)                                        \
  template <class RadiationFieldType, typename = typename std::enable_if<std::is_base_of< \
                                          RadiationField, RadiationFieldType>::value>::type>


class SphericalSymmetry {
public:
	SphericalSymmetry(
			const vector3D& center
	);

	void get_pos(
			REAL _center [3]
	) const;

	void get_ipos(
			Grid & grid,
			int ipos [3]
	) const;

	virtual void set_center(const vector3D& center);
	const vector3D& get_center() const;

protected:
	index3D ipos;
	vector3D center;
};

class IRadiationFieldFactor_NumberDensity {
public:
	IRadiationFieldFactor_NumberDensity () {};
	virtual ~IRadiationFieldFactor_NumberDensity () {};

	/** @brief Integrated Photon number density at a given position
	 * 	@detail Given in units of 1./Vol (NUM_DENS)
	 */
	REAL numberDensity(
			Data & gdata,
			int ix,int iy, int iz
	);

	virtual REAL numberDensity(
			Data & gdata,
			const REAL pos[3]
	) = 0;
};

class IRadiationFieldFactor_Energy {
public:
	IRadiationFieldFactor_Energy () {};
	virtual ~IRadiationFieldFactor_Energy () {};

	/** @brief Spectral photon density at a given position
	 * 	@detail Given in units 1./Energy (SPECTRAL)
	 */
	virtual double fEnergy(
			Data & gdata,
			int ix,int iy, int iz,
			double energy
	);

	virtual double fEnergy(
			Data & gdata,
			const REAL pos[3],
			REAL energy
	) = 0;

	virtual double getEMax(
			Data & gdata,
			int ix,int iy, int iz
	);

	virtual double getEMax(
			Data & gdata,
			const REAL pos[3]
	) = 0;

	virtual double getEMin(
			Data & gdata,
			int ix,int iy, int iz
	);

	virtual double getEMin(
			Data & gdata,
			const REAL pos[3]
	) = 0;

};

class IRadiationFieldFactor_Direction {
public:
	IRadiationFieldFactor_Direction() {};
	virtual ~IRadiationFieldFactor_Direction() {};

	/** @brief direcitonal photon density at a given position
	 * 	@detail Given in units 1./sr
	 */
	virtual REAL fDirection(
			Data & gdata,
			int ix,int iy, int iz,
			const REAL beamDirection[3]
	);

	virtual REAL fDirection(
			Data & gdata,
			const REAL pos[3],
			const REAL beamDirection[3]
	) = 0;

	virtual double getMuMin(Data & gdata, int ix, int iy, int iz, const double direction[3]);
	virtual double getMuMin(Data & gdata, const double pos[3], const double direction[3]) = 0;
};



using IRFFN = IRadiationFieldFactor_NumberDensity;
using IRFFE = IRadiationFieldFactor_Energy;
using IRFFD = IRadiationFieldFactor_Direction;

/**
 * @brief Radiation Field Interface with assumed factorization
 *
 * @details Radiation Field Implementation
 * Delegates the respective methods to their implementation.
 */
class RadiationFieldFactorized: public RadiationField {
public:
	RadiationFieldFactorized();

	/** number density: particles per volume */
	REAL numberDensity(
			Data & gdata,
			const REAL pos[3]
	);

	/** differential number density: particles per volume per (steradian, energy) */
	REAL numberDensity_dEdOmega(
			Data & gdata,
			const REAL pos[3],
			REAL energy,
			const REAL dir[3]
	);

	void set_iRadiationFieldFactor_NumberDensity(
			std::shared_ptr<IRFFN> iRadiationFieldFactor_NumberDensity
	);
	void set_iRadiationFieldFactor_Energy(
			std::shared_ptr<IRFFE> iRadiationFieldFactor_Energy
	);
	void set_iRadiationFieldFactor_Direction(
			std::shared_ptr<IRFFD> iRadiationFieldFactor_Direction
	);

	std::shared_ptr<IRFFN> get_iRadiationFieldFactor_NumberDensity();
	std::shared_ptr<IRFFE> get_iRadiationFieldFactor_Energy();
	std::shared_ptr<IRFFD> get_iRadiationFieldFactor_Direction();

private:
	std::shared_ptr<IRFFN> iRadiationFieldFactor_NumberDensity;
	std::shared_ptr<IRFFE> iRadiationFieldFactor_Energy;
	std::shared_ptr<IRFFD> iRadiationFieldFactor_Direction;
};


class IRadiationFieldFactor_Monochromatic: public IRadiationFieldFactor_Energy {
public:
	IRadiationFieldFactor_Monochromatic () {};
	~IRadiationFieldFactor_Monochromatic () {};

	REAL fEnergy(
			Data & gdata,
			const REAL pos[3],
			REAL energy
	);

	REAL get_photonEnergy(
			Data & gdata,
			int ix,int iy, int iz
	);

	virtual REAL get_photonEnergy(
			Data & gdata,
			const REAL pos[3]
	) = 0;

	virtual double getEMax(
			Data & gdata,
			const REAL pos[3]
	);

	virtual double getEMin(
			Data & gdata,
			const REAL pos[3]
	);
};

class IRadiationFieldFactor_Beamed: public IRadiationFieldFactor_Direction {
public:
	IRadiationFieldFactor_Beamed () {};
	~IRadiationFieldFactor_Beamed () {};

	REAL fDirection(
			Data & gdata,
			const REAL pos[3],
			const REAL dir[3]
	);

	void get_beamDirection (
			Data & gdata,
			int ix,int iy, int iz,
			REAL beamDirection[3]
	);

	virtual void get_beamDirection (
			Data & gdata,
			const REAL pos[3],
			REAL beamDirection[3]
	) = 0;

	virtual double getMuMin(Data & gdata, const double pos[3], const double direction[3]);
};


#endif

#ifndef CRONOS_RADIATION_FIELD_SPECIAL_H
#include "RadiationField_special.H"
#endif