#ifndef CRONOS_RADIATION_FIELD_H
#define CRONOS_RADIATION_FIELD_H

#include "data.H"
#include "normalisation.H"
#include "CException.H"
#include "LorentzTrafo.H"

#include <cassert>
#include <vector>


/**
 * @brief Base Radiation Field class.
 *
 * @details The radiation field provides
 * differential spectral energy densities in terms of
 *  - dE / (dV dE dOmega), with directional dependece
 *  - dE / (dV dE)
 */
class RadiationField {
public:
	RadiationField(
			normalisation *normHD,
			normalisation *normParticle
	);

	virtual ~RadiationField(
	);

	// spectral energy density: energy per (volume, steradian, energy)

	// All arguments in LAB Frame
	virtual REAL get_dE_dVdEdOmega_LAB(
			const Data &gdata,
			int ix, int iy, int iz,
			const REAL energy_LAB,
			const REAL dir_LAB[3]
	) = 0;

	// All arguments in COM Frame
	virtual REAL get_dE_dVdEdOmega_COM(
			const Data &gdata,
			int ix, int iy, int iz,
			const REAL energy_COM,
			const REAL dir_COM[3]
	) = 0;

	// integrated over solid angle
	// All arguments in LAB Frame
	virtual REAL get_dE_dVdE_LAB(
			const Data &gdata,
			int ix, int iy, int iz,
			const REAL energy_LAB
	) = 0;

	// All arguments in COM Frame
	virtual REAL get_dE_dVdE_COM(
			const Data &gdata,
			int ix, int iy, int iz,
			const REAL energy_COM
	) = 0;

protected:
	normalisation * normHD;
	normalisation * normParticle;
};


////////////////////////////////////////////////////////////////////////////////////////////////
//
//	GENERIC EXTENSIONS
//

/**
 * @brief Radiation Field defined in the LAB frame.
 *
 * @details Assuming the LAB spectral energy densities are
 * provided the COM quantities are defined via lorentz transformations.
 * This class implements these LAB->COM transformation.
 */
class RadiationField_LAB: virtual public RadiationField {
public:
	using RadiationField::RadiationField;

	virtual REAL get_dE_dVdEdOmega_COM(
			const Data &gdata,
			int ix, int iy, int iz,
			const REAL energy_COM,
			const REAL dir_COM[3]
	) override;

protected:
	LorentzTrafo lorentzTrafo {};
};


/**
 * @brief Beamed Radiation Field class.
 *
 * @details The radiation field is assumed to be beamed into a certain direction.
 * Further the energy density per solid angle is ill-defined (delta-distribution).
 */
class BeamedRadiationField: virtual public RadiationField {
public:
	using RadiationField::RadiationField; //Inherit constructors

	virtual REAL get_dE_dVdEdOmega_LAB(
			const Data &gdata,
			int ix, int iy, int iz,
			const REAL energy_LAB,
			const REAL dir_LAB[3]
	) override final;

	virtual REAL get_dE_dVdEdOmega_COM(
			const Data &gdata,
			int ix, int iy, int iz,
			const REAL energy_COM,
			const REAL dir_COM[3]
	) override final;

	virtual void get_beamDirection_LAB(
			const Data &gdata,
			int ix, int iy, int iz,
			REAL dir_LAB[3]
	) = 0;

	virtual void get_beamDirection_COM(
			const Data &gdata,
			int ix, int iy, int iz,
			REAL dir_COM[3]
	) = 0;
};



////////////////////////////////////////////////////////////////////////////////////////////////
//
//	LAB IMPLEMENTATIONS
//

/**
 * @brief Isotrope Radiation Field (isotrope in the LAB frame).
 *
 * @details Assumes isotropy in the LAB frame.
 * The directional dependence in the LAB frame is trivially given.
 * The LAB->COM transformations are inherited.
 */
class IsotropeRadiationField_LAB: virtual public RadiationField_LAB {
public:
	using RadiationField_LAB::RadiationField_LAB; //Inherit constructors

	virtual REAL get_dE_dVdEdOmega_LAB(
			const Data &gdata,
			int ix, int iy, int iz,
			const REAL energy_LAB,
			const REAL dir_LAB[3]
	) override;

	virtual REAL get_dE_dVdE_COM(
			const Data &gdata,
			int ix, int iy, int iz,
			const REAL energy_COM
	) override;

private:
	REAL inverseSolidAngle{1./(4.*pi)};
};


/**
 * @brief Beamed Radiation Field class.
 *
 * @details The radiation field is assumed to be beamed into a certain direction.
 * Further the energy density per solid angle is ill-defined (delta-distribution).
 */
class BeamedRadiationField_LAB: virtual public BeamedRadiationField, virtual public RadiationField_LAB {
public:
//	using RadiationField_LAB::RadiationField_LAB; //Inherit constructors
//	using BeamedRadiationField::get_dE_dVdEdOmega_LAB;
	using BeamedRadiationField::get_dE_dVdEdOmega_COM;
//	using RadiationField_LAB::get_dE_dVdE_LAB;

	virtual REAL get_dE_dVdE_COM(
			const Data &gdata,
			int ix, int iy, int iz,
			const REAL energy_COM
	) override;

	virtual void get_beamDirection_COM(
			const Data &gdata,
			int ix, int iy, int iz,
			REAL dir_COM[3]
	) override;
};


/**
 * @brief Monochromatic Radiation Field (defined in the LAB frame).
 *
 * @details The radiation field is assumed to be monochromatic with a given photon energy.
 * Implements the LAB->COM transformation of the photon energy.
 * The remaining LAB->COM transformations are inherited.
 */
class MonochromaticRadiationField_LAB: virtual public RadiationField_LAB {
public:
	using RadiationField_LAB::RadiationField_LAB; //Inherit constructors

	virtual REAL get_photonEnergy_LAB(
			const Data &gdata,
			int ix, int iy, int iz
	) = 0;

	virtual REAL get_photonEnergy_COM(
			const Data &gdata,
			int ix, int iy, int iz,
			REAL dir_COM[3]
	);
};





#endif
