#include "constants.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)
#include "RadiationField_special.H"

REAL distance(
		const Data & gdata,
		int ix,int iy, int iz,
		const REAL center_num[3],
		int dir = -1
	)
{
	switch(dir) {
#if (GEOM == CYLINDRICAL)
		case 0 :	return gdata.getCen_x(ix);
		case 1 : 	return 0;
		case 2 : 	return gdata.getCen_z(iz) - center_num[2];
#else
		case 0 :	return gdata.getCen_x(ix) - center_num[0];
		case 1 : 	return gdata.getCen_y(iy) - center_num[1];
		case 2 : 	return gdata.getCen_z(iz) - center_num[2];
#endif
		default: 	return sqrt(sqr(distance(gdata, ix, iy, iz, center_num, 0)) +
								sqr(distance(gdata, ix, iy, iz, center_num, 1)) +
								sqr(distance(gdata, ix, iy, iz, center_num, 2)) );
	}
}

REAL distance(
		const Data & gdata,
		int ix,int iy, int iz,
		const vector3D& center_num,
		int dir = -1
	) {
		return distance(gdata, ix, iy, iz, center_num.data(), dir);
	}

REAL distance(
		const Data & gdata,
		const REAL pos[3],
		const REAL center_num[3],
		int dir = -1
	)
{
	switch(dir) {
#if (GEOM == CYLINDRICAL)
		case 0 :	return pos[0];
		case 1 : 	return 0;
		case 2 : 	return pos[2] - center_num[2];
#else
		case 0 :	return pos[0] - center_num[0];
		case 1 : 	return pos[1] - center_num[1];
		case 2 : 	return pos[2] - center_num[2];
#endif
		default: 	return sqrt(sqr(distance(gdata, pos, center_num, 0)) +
								sqr(distance(gdata, pos, center_num, 1)) +
								sqr(distance(gdata, pos, center_num, 2)) );
	}
}

REAL distance(
		const Data & gdata,
		const REAL pos[3],
		const vector3D& center_num,
		int dir = -1
	)
{
		return distance(gdata, pos, center_num.data(), dir);
}



RFFN_Constant::RFFN_Constant (
		REAL _numberDensity
)
{
	theNumberDensity = _numberDensity;
}

REAL RFFN_Constant::numberDensity (
		Data & gdata,
		const REAL pos[3]
)
{
	return theNumberDensity;
}


RFFN_Radial::RFFN_Radial (
		REAL _radialDensity,
		const vector3D& center
	) :
	SphericalSymmetry(center)
{
	radialDensity = _radialDensity;
}

REAL RFFN_Radial::numberDensity (
		Data & gdata,
		const REAL pos[3]
	)
{
	REAL dist = distance(gdata, pos, center);
	return radialDensity / sqr(dist);
}


RFFE_Monochromatic::RFFE_Monochromatic (
		REAL _photonEnergy
	)
{
	photonEnergy = _photonEnergy;
}

REAL RFFE_Monochromatic::get_photonEnergy(
		Data & gdata,
		const REAL pos[3]
	)
{
	return photonEnergy;
}

RFFD_Radial::RFFD_Radial (
		const vector3D& center
	) :
	SphericalSymmetry(center)
{
}

void RFFD_Radial::get_beamDirection (
		Data & gdata,
		const REAL pos[3],
		REAL beamDirection[3]
	)
{
	REAL distance_inv = 1. / distance(gdata, pos, center);
	for (int iDir = 0; iDir < 3; ++iDir) {
		beamDirection[iDir] = distance(gdata, pos, center, iDir) * distance_inv;
	}
}


RFFD_Constant::RFFD_Constant (
		const REAL _beamDirection [3]
)
{
	for (int iDir = 0; iDir < 3; ++iDir)
		beamDirection[iDir] = _beamDirection[iDir];
}

void RFFD_Constant::get_beamDirection (
		Data & gdata,
		const REAL pos[3],
		REAL _beamDirection[3]
)
{
	for (int iDir = 0; iDir < 3; ++iDir) {
		_beamDirection[iDir] = beamDirection[iDir];
	}
}




RFFE_Blackbody::RFFE_Blackbody (
		REAL kbT
)
{
	this->kbT = kbT;
	kbT_inv = 1./kbT;
	REAL norm = 2. * gsl_sf_zeta_int(3) * kbT;
	norm_inv = 1. / norm;
};

RFFE_Blackbody::~RFFE_Blackbody (
		)
{
}

REAL RFFE_Blackbody::fEnergy(
		Data & gdata,
		const REAL pos[3],
		REAL energy
)
{
	REAL x = energy * kbT_inv;
	return sqr(x) / (exp(x) - 1) * norm_inv;
}

double RFFE_Blackbody::getEMax(
		Data & gdata,
		const REAL pos[3]
) {
	// // fEnergy drops to ~1e-8 @ 20 kbT
	// return 20 * kbT;
	return 1e30 * kbT;
}

double RFFE_Blackbody::getEMin(
		Data & gdata,
		const REAL pos[3]
) {
	// // fEnergy drops to ~1e-8 @ 1e-8 kbT
	// return 1e-8 * kbT;
	return 0;
}


RFFN_SphericalExtended::RFFN_SphericalExtended (
		const REAL radialDensity, 
		const vector3D& center,
		const REAL radius
):
	SphericalExtended(radius, center),
		dVec(3)
{
	this->radialDensity = radialDensity;
}

RFFN_SphericalExtended::~RFFN_SphericalExtended (
	)
{
}

REAL RFFN_SphericalExtended::numberDensity(
		Data & gdata,
		const REAL pos[3]
)
{
	REAL dist = distance(gdata, pos, center);
	if (dist > radius) {
		return radialDensity * (1. - sqrt(1. - sqr(radius/dist)));
	} else {
		return 0.;
	}
}



IRFFD_SphericalCoordinates::IRFFD_SphericalCoordinates(
)
{
}

IRFFD_SphericalCoordinates::~IRFFD_SphericalCoordinates(
)
{
}

REAL IRFFD_SphericalCoordinates::get_muMin(
		Data & gdata,
		const REAL pos[3]
)
{
	return -1.;
}

REAL IRFFD_SphericalCoordinates::get_muMax(
		Data & gdata,
		const REAL pos[3]
)
{
	return 1.;
}

REAL IRFFD_SphericalCoordinates::get_phiMin(
		Data & gdata,
		const REAL pos[3]
)
{
	return 0;
}

REAL IRFFD_SphericalCoordinates::get_phiMax(
		Data & gdata,
		const REAL pos[3]
)
{
	return 2.*pi;
}


RFFD_SphericalExtended::RFFD_SphericalExtended (
		const REAL radius,
		const vector3D& center
):
		SphericalExtended(radius, center),
		dVec(3),
		pos(3)
{
	int dims[2] = {3,3};
	rotatedBaseVectors.resize(dims);

	// Prevent matching in first comparison
	pos(0) = std::numeric_limits<double>::max();
}

RFFD_SphericalExtended::~RFFD_SphericalExtended () {}

REAL RFFD_SphericalExtended::get_muMin(
		Data & gdata,
		const REAL pos[3]
) {
	precompute(gdata, pos);
	assert(muMin <= 1);
	assert(muMin >= -1);
	return muMin;
}

void RFFD_SphericalExtended::get_dir(
		Data & gdata,
		const REAL pos[3],
		const REAL mu,
		const REAL phi,
		REAL dir[3]
) {
	REAL cosTheta = mu;
	REAL sinTheta = sqrt(1. - sqr(cosTheta));
	REAL cosPhi = cos(phi);
	REAL sinPhi = sin(phi);

	REAL vec[3];
	vec[0] = cosPhi * sinTheta;
	vec[1] = sinPhi * sinTheta;
	vec[2] = cosTheta;

	for (int iDir = 0; iDir < 3; ++iDir) {
		dir[iDir] = 0;
		for (int iBase = 0; iBase < 3; ++iBase) {
			dir[iDir] += rotatedBaseVectors(iDir, iBase) * vec[iBase];
		}
	}
}

REAL RFFD_SphericalExtended::fDirection(
		Data & gdata,
		const REAL pos[3],
		const REAL mu,
		const REAL phi
){
	precompute(gdata, pos);

	/*
		for a star we have a constant surface brightness: fDir = dn(theta, phi)/dOmega = C
	  normalization to 1 yields: 
			1 = int_0^thetaMax fDir(theta, phi) sin(theta) dTheta dPhi
			  = 2 * pi * C * int_muMin^1 dMu
			  = 2 * pi * C * (1 - muMin)
	*/

	if (mu >= muMin) {
		if(muSingular) {
			return 1. / (2. * pi);
		} else {
			return 1. / (2. * pi * (1. - muMin));
		}
	} else {
		return 0;
	}
}

REAL RFFD_SphericalExtended::fDirection(
		Data & gdata,
		const REAL pos[3],
		const REAL photonDirection[3]
) {
	precompute(gdata, pos);
	REAL mu = 0;
	for (int iDir = 0; iDir < 3; ++iDir) {
		mu += dVec(iDir) * photonDirection[iDir];
	}
	mu /= dNorm;
	return fDirection(gdata, pos, 0, mu);
}

bool RFFD_SphericalExtended::posChanged(
		const REAL pos[3]
) const
{
	static bool firstTime = true;
	if(firstTime) {
		firstTime = false;
		return true;
	}

	for (int iDir = 0; iDir < 3; ++iDir) {
		if (this->pos(iDir) != pos[iDir]) {
			return true;
		}
	}
	return false;
}

void RFFD_SphericalExtended::precompute(
		Data & gdata,
		const REAL pos[3]
)
{
	if (posChanged(pos)) {
		dNorm = 0;
		for (int iDir = 0; iDir < 3; ++iDir) {
			this->pos(iDir) = pos[iDir];
			dVec(iDir) = pos[iDir] - center[iDir];
			dNorm += sqr(dVec(iDir));
		}
		dNorm = sqrt(dNorm);

		if (dNorm < radius) {
			cout << dNorm << " " << radius << std::endl;
			throw UndefinedInsideSourceException();
		}

		muMin = sqrt(1. - sqr(radius/dNorm));

		if (1. - muMin < 1e-4) {
			muMin = 1.;
			muSingular = true;
		} else {
			muSingular = false;
		}

		// This is not the same mu and phi as for the other function
		// Here we compute the cartesion base vectors that the actual mu and phi will refer to
		REAL cosTheta = dVec(2)/dNorm;
		REAL sinTheta = sqrt(1. - sqr(cosTheta));
		REAL phi = atan2(dVec(1), dVec(0));
		REAL cosPhi = cos(phi);
		REAL sinPhi = sin(phi);

		// ex' = eTheta
		rotatedBaseVectors(0,0) = cosPhi * cosTheta;
		rotatedBaseVectors(1,0) = sinPhi * cosTheta;
		rotatedBaseVectors(2,0) = -sinTheta;

		// ey' = ePhi
		rotatedBaseVectors(0,1) = -sinPhi * sinTheta;
		rotatedBaseVectors(1,1) = cosPhi * sinTheta;
		rotatedBaseVectors(2,1) = 0;

		// ez' = er
		rotatedBaseVectors(0,2) = cosPhi * sinTheta;
		rotatedBaseVectors(1,2) = sinPhi * sinTheta;
		rotatedBaseVectors(2,2) = cosTheta;
	}

	assert(std::isfinite(dNorm));
	assert(muMin <= 1);
	assert(muMin >= -1);
}

double RFFD_SphericalExtended::getMuMin(Data & gdata, const double pos[3], const double direction[3]) {
	precompute(gdata, pos);
	double muEzDir = 0;
	for(int iDir = 0; iDir < 3; ++iDir) {
		muEzDir += rotatedBaseVectors(iDir, 2) * direction[iDir];
	}

	return cos(min(acos(muEzDir) + acos(muMin), pi)) ;
}


SphericalExtended::SphericalExtended(
		const REAL radius,
		const vector3D& center
):
	SphericalSymmetry(center)
{
	assert(radius >= 0);
	this->radius = radius;
}

REAL SphericalExtended::get_radius() const
{
	return radius;
}

FactorizedSphericalSource::FactorizedSphericalSource(const double radius, const vector3D& center):
	SphericalExtended(radius, center) {
	if (radius <= 0) {
		set_iRadiationFieldFactor_Direction(make_shared<RFFD_Radial>(center));
	} else {
		set_iRadiationFieldFactor_Direction(make_shared<RFFD_SphericalExtended>(radius, center));
	}
}


void FactorizedSphericalSource::set_center(const vector3D& center) {
	SphericalExtended::set_center(center);

	auto irffn = dynamic_pointer_cast<SphericalSymmetry>(get_iRadiationFieldFactor_NumberDensity());
	if (irffn) {
		irffn->set_center(center);
	}

	auto irffe = dynamic_pointer_cast<SphericalSymmetry>(get_iRadiationFieldFactor_Energy());
	if (irffe) {
		irffe->set_center(center);
	}

	auto irffd = dynamic_pointer_cast<SphericalSymmetry>(get_iRadiationFieldFactor_Direction());
	if (irffd) {
		irffd->set_center(center);
	}

}


///////////// COMPLETE FIELDS

ConstantBeamedMonochromatic::ConstantBeamedMonochromatic(
		normalisation * normHD,
		normalisation * normParticle,
		Quantity _photonSpectralNumberDensity,
		Quantity _photonEnergy,
		REAL _beamDirection[3]
)
{
	photonSpectralNumberDensity = _photonSpectralNumberDensity;
	photonSpectralNumberDensity_num = normParticle->get_num(normParticle->SPEC_NUM_DENS, photonSpectralNumberDensity);

	photonEnergy = _photonEnergy;
	photonEnergy_num = normParticle->get_num(normParticle->ENERGY, photonEnergy);

	for (int iDir = 0; iDir < 3; ++iDir) {
		beamDirection[iDir] = _beamDirection[iDir];
	}

	set_iRadiationFieldFactor_NumberDensity(make_shared<RFFN_Constant>(photonSpectralNumberDensity_num));
	set_iRadiationFieldFactor_Energy(make_shared<RFFE_Monochromatic>(photonEnergy_num));
	set_iRadiationFieldFactor_Direction(make_shared<RFFD_Constant>(beamDirection));
}

ConstantBeamedMonochromatic::~ConstantBeamedMonochromatic(
	)
{
}

NumArray<REAL> getNumArray(
		REAL a, REAL b, REAL c
)
{
	NumArray<REAL> arr(3);
	arr = {a, b, c};
	return arr;
}

MonochromaticPointSource::MonochromaticPointSource(
		normalisation * normHD,
		normalisation * normParticle,
		Quantity _photonEnergy,
		Quantity _luminosity,
		const vector3D& center_num
):
		FactorizedSphericalSource(0., center_num)
{
	photonEnergy = _photonEnergy;
	luminosity = _luminosity;

	photonEnergy_num = normParticle->get_num(normParticle->ENERGY, photonEnergy);
	Quantity radialNumberDensity = luminosity / (
				photonEnergy * 4. * pi * SpeedOfLight
				* sqr(normHD->normQuantity(normHD->LEN))
				* normParticle->normQuantity(normParticle->ENERGY)
			);
	radialNumberDensity_num = normParticle->get_num(normParticle->SPEC_NUM_DENS, radialNumberDensity);

	set_iRadiationFieldFactor_NumberDensity(make_shared<RFFN_Radial>(radialNumberDensity_num, center_num));
	set_iRadiationFieldFactor_Energy(make_shared<RFFE_Monochromatic>(photonEnergy_num));
	set_iRadiationFieldFactor_Direction(make_shared<RFFD_Radial>(center_num));
}

MonochromaticPointSource::~MonochromaticPointSource(
	)
{
}

MonochromaticExtendedSource::MonochromaticExtendedSource(
		normalisation * normHD,
		normalisation * normParticle,
		Quantity radius,
		Quantity photonEnergy,
		Quantity luminosity,
		const vector3D& center_num
):
		FactorizedSphericalSource(normHD->get_num(radius), center_num)
{
	this->photonEnergy = photonEnergy;
	this->luminosity = luminosity;

	photonEnergy_num = normParticle->get_num(normParticle->ENERGY, photonEnergy);

	Quantity radialNumberDensity = luminosity / (
			photonEnergy * 4. * pi * SpeedOfLight
			* sqr(normHD->normQuantity(normHD->LEN))
			* normParticle->normQuantity(normParticle->ENERGY)
	);
	radialNumberDensity_num = normParticle->get_num(normParticle->SPEC_NUM_DENS, radialNumberDensity);

	set_iRadiationFieldFactor_NumberDensity(make_shared<RFFN_Radial>(radialNumberDensity_num, center_num));
	set_iRadiationFieldFactor_Energy(make_shared<RFFE_Monochromatic>(photonEnergy_num));
}

MonochromaticExtendedSource::~MonochromaticExtendedSource(
	)
{
}

bool MonochromaticExtendedSource::isDefined(
		const Data & gdata,
		const REAL pos[3]
) const {
	if (distance(gdata, pos, center) < radius) {
		return false;
	} else {
		return true;
	}
}

BlackbodyPointSource::BlackbodyPointSource(
		normalisation * normHD,
		normalisation * normParticle,
		Quantity temperature,
		Quantity luminosity,
		const vector3D& center_num
):
		FactorizedSphericalSource(0., center_num)
{
	Quantity kbT = BoltzmannConstant * temperature;
	Quantity meanPhotonEnergy = kbT * sqr(sqr(pi)) / (30. * gsl_sf_zeta_int(3));

	Quantity radialNumberDensity = luminosity / (
			meanPhotonEnergy * 4. * pi * SpeedOfLight
			* sqr(normHD->normQuantity(normHD->LEN))
			* normParticle->normQuantity(normParticle->ENERGY)
	);
	REAL radialNumberDensity_num = normParticle->get_num(normParticle->SPEC_NUM_DENS, radialNumberDensity);
	REAL kbT_num = normParticle->get_num(kbT);

	set_iRadiationFieldFactor_NumberDensity(make_shared<RFFN_Radial>(radialNumberDensity_num, center_num));
	set_iRadiationFieldFactor_Energy(make_shared<RFFE_Blackbody>(kbT_num));
	set_iRadiationFieldFactor_Direction(make_shared<RFFD_Radial>(center_num));
}

BlackbodyPointSource::~BlackbodyPointSource(
	)
{
}

BlackbodyExtendedSource::BlackbodyExtendedSource(
		normalisation * normHD,
		normalisation * normParticle,
		Quantity radius,
		Quantity temperature,
		const vector3D& center_num
):
		FactorizedSphericalSource(normHD->get_num(radius), center_num)
{
	this->temperature = temperature;

	luminosity = StefanBoltzmannConstant * sqr(sqr(temperature)) * 4. * pi * sqr(radius);

	Quantity kbT = BoltzmannConstant * temperature;
	Quantity meanPhotonEnergy = kbT * sqr(sqr(pi)) / (30. * gsl_sf_zeta_int(3));

	Quantity radialNumberDensity = luminosity / (
			meanPhotonEnergy * 4. * pi * SpeedOfLight
			* sqr(normHD->normQuantity(normHD->LEN))
			* normParticle->normQuantity(normParticle->ENERGY)
	);

	REAL kbT_num = normParticle->get_num(kbT);
	REAL radialNumberDensity_num = normParticle->get_num(normParticle->SPEC_NUM_DENS, radialNumberDensity);

	set_iRadiationFieldFactor_NumberDensity(make_shared<RFFN_Radial>(radialNumberDensity_num, center_num));
	set_iRadiationFieldFactor_Energy(make_shared<RFFE_Blackbody>(kbT_num));
}

BlackbodyExtendedSource::~BlackbodyExtendedSource(
)
{
}

bool BlackbodyExtendedSource::isDefined(
		const Data & gdata,
		const REAL pos[3]
) const {
	if (distance(gdata, pos, center) < radius) {
		return false;
	} else {
		return true;
	}
}

#endif