#ifndef CRONOS_RADIATION_FIELD_SPECIAL_H
#define CRONOS_RADIATION_FIELD_SPECIAL_H

#include "RadiationField.H"
#include "MagneticField.H"
#include <gsl/gsl_sf.h>
#include <memory>
#include "marray.H"

using namespace CRONOS_CONSTANTS;


class UndefinedInsideSourceException: public runtime_error {
public:
	UndefinedInsideSourceException(): runtime_error("RadiationField undefined inside source") {}
};

class SphericalExtended: public SphericalSymmetry {
public:
	SphericalExtended(
			const REAL radius,
			const vector3D& center
	);

	REAL get_radius() const;

protected:
	REAL radius;
};


class RFFN_Constant: public IRadiationFieldFactor_NumberDensity {
public:
	RFFN_Constant (
			REAL numberDensity
	);

	REAL numberDensity (
			Data & gdata,
			const REAL pos[3]
	);

private:
	REAL theNumberDensity;
};


class RFFN_Radial: public IRadiationFieldFactor_NumberDensity, public SphericalSymmetry {
public:
	RFFN_Radial (
			REAL radialDensity,
			const vector3D& center
	);

	REAL numberDensity (
			Data & gdata,
			const REAL pos[3]
	);

private:
	REAL radialDensity;
};

class RFFE_Monochromatic: public IRadiationFieldFactor_Monochromatic {
public:
	RFFE_Monochromatic (
			REAL photonEnergy
	);

	REAL get_photonEnergy(
			Data & gdata,
			const REAL pos[3]
	);

private:
	REAL photonEnergy;
};


class RFFD_Radial: public IRadiationFieldFactor_Beamed, public SphericalSymmetry {
public:
	RFFD_Radial (
			const vector3D& center
	);

	void get_beamDirection (
			Data & gdata,
			const REAL pos[3],
			REAL beamDirection[3]
	);
};

class RFFD_Constant: public IRadiationFieldFactor_Beamed {
public:
	RFFD_Constant (
			const REAL beamDirection [3]
	);

	void get_beamDirection (
			Data & gdata,
			const REAL pos[3],
			REAL beamDirection[3]
	);

private:
	REAL beamDirection [3];
};



class RFFE_Blackbody: public IRadiationFieldFactor_Energy {
public:
	RFFE_Blackbody (
			REAL kbT
	);

	~RFFE_Blackbody ();

	REAL fEnergy(
			Data & gdata,
			const REAL pos[3],
			REAL energy
	);

	virtual double getEMax(
			Data & gdata,
			const REAL pos[3]
	);

	virtual double getEMin(
			Data & gdata,
			const REAL pos[3]
	);

private:
	REAL kbT, kbT_inv;
	REAL norm_inv;
};


class RFFN_SphericalExtended: public IRadiationFieldFactor_NumberDensity, public SphericalExtended {
public:
	RFFN_SphericalExtended (
			const REAL radialDensity,
			const vector3D& center,
			const REAL radius
	);

	~RFFN_SphericalExtended ();

	REAL numberDensity(
			Data & gdata,
			const REAL pos[3]
	);

private:
	REAL radialDensity;
	NumArray<REAL> pos, dVec;
	REAL dNorm, muMin;
};


class IRFF_Centered {

};

class IRFFD_SphericalCoordinates: public IRadiationFieldFactor_Direction {
public:
	IRFFD_SphericalCoordinates();

	virtual ~IRFFD_SphericalCoordinates();

	/** Limits for the field.
	 * 	mu = cos(theta)	... azimuth angle
	 * 		- default range [-1,1]
	 * 	phi 			... polar angle
	 * 		- default range [0, 2 pi]
	 *
	 * 	If the function is singular in a coordinate the limits will output the same value
	 * 	use this to check for that (if you want to integrate for example)
	 * 	In this case fDirection will output the integrated value.
	 */
	virtual REAL get_muMin(
			Data &gdata,
			const REAL pos[3]
	);

	virtual REAL get_muMax(
			Data &gdata,
			const REAL pos[3]
	);

	virtual REAL get_phiMin(
			Data &gdata,
			const REAL pos[3]
	);

	virtual REAL get_phiMax(
			Data &gdata,
			const REAL pos[3]
	);

	/** Calculates the cartesian direction
	 */
	virtual void get_dir(
			Data &gdata,
			const REAL pos[3],
			const REAL mu,
			const REAL phi,
			REAL dir[3]
	) = 0;

	virtual REAL fDirection(
			Data &gdata,
			const REAL pos[3],
			const REAL mu,
			const REAL phi
	) = 0;
};



class RFFD_SphericalExtended: public IRFFD_SphericalCoordinates, public SphericalExtended {
public:
	RFFD_SphericalExtended (
			const REAL radius,
			const vector3D& center
	);

	~RFFD_SphericalExtended ();

	REAL get_muMin(
			Data & gdata,
			const REAL pos[3]
	) final;

	void get_dir(
			Data & gdata,
			const REAL pos[3],
			const REAL mu,
			const REAL phi,
			REAL dir[3]
	) final;

	REAL fDirection(
			Data & gdata,
			const REAL pos[3],
			const REAL mu,
			const REAL phi
	) final;

	REAL fDirection(
			Data & gdata,
			const REAL pos[3],
			const REAL photonDirection[3]
	) final;

	double getMuMin(Data & gdata, const double pos[3], const double direction[3]);

private:
	bool posChanged(
			const REAL pos[3]
	) const;

	void precompute(
			Data & gdata,
			const REAL pos[3]
	);

	NumArray<REAL> pos, dVec;
	NumMatrix<REAL, 2> rotatedBaseVectors;
	REAL dNorm, muMin;
	bool muSingular;
};



///////////// COMPLETE FIELDS

class ConstantBeamedMonochromatic: public RadiationFieldFactorized  {
public:
	ConstantBeamedMonochromatic(
			normalisation * normHD,
			normalisation * normParticle,
			Quantity _photonNumberDensity,
			Quantity _photonEnergy,
			REAL _beamDirection[3]
		);

	~ConstantBeamedMonochromatic();

private:
	Quantity photonSpectralNumberDensity;
	REAL photonSpectralNumberDensity_num;

	Quantity photonEnergy;
	REAL photonEnergy_num;

	REAL beamDirection[3];
};

class FactorizedSphericalSource: public SphericalExtended, public RadiationFieldFactorized {
public:
	FactorizedSphericalSource(const double radius, const vector3D& center);
	void set_center(const vector3D& center) override;
};

using SphericalSource = SphericalExtended;


class MonochromaticPointSource: public FactorizedSphericalSource {
public:
	MonochromaticPointSource(
			normalisation * normHD,
			normalisation * normParticle,
			Quantity photonEnergy,
			Quantity luminosity,
			const vector3D& center_num
	);

	~MonochromaticPointSource();

private:
	Quantity photonEnergy;
	REAL photonEnergy_num;

	Quantity luminosity;
	REAL radialNumberDensity_num;
};


class MonochromaticExtendedSource: public FactorizedSphericalSource {
public:
	MonochromaticExtendedSource(
			normalisation * normHD,
			normalisation * normParticle,
			Quantity radius,
			Quantity photonEnergy,
			Quantity luminosity,
			const vector3D& center_num
	);

	~MonochromaticExtendedSource();

	bool isDefined(
			const Data & gdata,
			const REAL pos[3]
	) const final;

private:
	Quantity photonEnergy;
	REAL photonEnergy_num;

	Quantity luminosity;
	REAL radialNumberDensity_num;
};

class BlackbodyPointSource: public FactorizedSphericalSource {
public:
	BlackbodyPointSource(
			normalisation * normHD,
			normalisation * normParticle,
			Quantity temperature,
			Quantity luminosity,
			const vector3D& center_num
	);

	~BlackbodyPointSource();
};


class BlackbodyExtendedSource: public FactorizedSphericalSource {
public:
	BlackbodyExtendedSource(
		normalisation * normHD,
		normalisation * normParticle,
		Quantity radius,
		Quantity temperature,
		const vector3D& center_num
	);

	~BlackbodyExtendedSource();

	bool isDefined(
			const Data & gdata,
			const REAL pos[3]
	) const final;

private:
	Quantity temperature;

	Quantity luminosity;
	REAL radialNumberDensity_num;
};


#endif
