#include "SemilagrangeSolver.H"
#include <stdexcept>
#include <stdio.h>
#include <stdlib.h>
#include <cassert>

using namespace std;

/*------------------------------------------------------------
  Semilagrange Solver
  ------------------------------------------------------------*/

SemilagrangeSolver::SpaceSolver getSpaceSolver(int spatialOrder) {
	switch (spatialOrder) {
		case 1:
			return SemilagrangeSolver::SpaceSolver::Constant;
		case 2:
			return SemilagrangeSolver::SpaceSolver::Linear;
		default:
			throw std::invalid_argument("Only spatial order 1, 2 available");
	}
}

SemilagrangeSolver::TimeSolver getTimeSolver(int temporalOrder) {
	switch (temporalOrder) {
		case 1:
			return SemilagrangeSolver::TimeSolver::Euler;
		case 2:
			return SemilagrangeSolver::TimeSolver::Second;
		case 3:
			return SemilagrangeSolver::TimeSolver::Richardson;
		default:
			throw std::invalid_argument("Only temporal order 1, 2, 3 available");
	}
}

SemilagrangeSolver::SemilagrangeSolver(
		int spatialOrder,
		bool slopeLimited,
		int limiterType
): SemilagrangeSolver(getTimeSolver(1), getSpaceSolver(spatialOrder))
{
	if(slopeLimited)
		set_slopeLimiter(limiterType);
	showWarnings = true;
}

SemilagrangeSolver::SemilagrangeSolver(
		int temporalOrder,
		int spatialOrder
):
  SemilagrangeSolver(getTimeSolver(temporalOrder), getSpaceSolver(spatialOrder))
{
}

SemilagrangeSolver::SemilagrangeSolver(
	TimeSolver timeSolver,
	SpaceSolver spaceSolver
)
{
	this->spaceSolver = spaceSolver;
	this->timeSolver = timeSolver;

	lo = 0;
	hi = 0;
	mx = 0;
	loComp = 0;
	hiComp = 0;
	advectionFunction = 0;

	maxSubcycleDepth = 12;
	// Disable subcycling for now
	//maxSubcycleDepth = 1;
	reachedSubcycleDepth = 0;

	slopeLimiter = std::unique_ptr<limiter_func>(nullptr);
}

SemilagrangeSolver::~SemilagrangeSolver()
{
}

void SemilagrangeSolver::set_advectionFunction(
		Function<REAL> & advectionFunction
)
{
	Function<REAL> * advectionFunction_ptr = & advectionFunction;
	this->advectionFunction = advectionFunction_ptr;
	advectionFunctionStl = [=](REAL x) -> REAL {return advectionFunction_ptr->get_value(x);};
}


void SemilagrangeSolver::resize(const Grid1D & grid)
{
	lo = grid.getLo();
	hi = grid.getHi();
	mx = grid.get_mx();

	set_computationLimits(0, grid.get_mx());

	cellEdges_backpropagated.resize(grid.getLow(), grid.getHigh() + 1);
	accumulated_values.resize(grid.getLow(), grid.getHigh() + 1);
	parabolicCoeffs.resize(grid.getLow(), grid.getHigh());
}

void SemilagrangeSolver::set_computationLimits(int loComp, int hiComp)
{
	assert(loComp <= hiComp);
	assert(loComp >= 0);
	assert(hiComp <= mx);

	this->loComp = loComp;
	this->hiComp = hiComp;
}


void SemilagrangeSolver::singlestep(
		const Grid1D & grid,
		const BoundaryHandler1D & boundary,
		NumMatrix<REAL,1> & cellValues,
		const REAL dt
)
{
	assert(advectionFunction != 0);
	assert(cellValues.getLow(0) == lo);
	assert(cellValues.getHigh(0) == hi);

	cellEdges_backpropagated.clear();
	propagate_edges(grid, dt, timeSolver, cellEdges_backpropagated);

	boundary.do_BCsCellAve(cellValues, grid, advectionFunctionStl);

	unique_ptr<Interpolation1D> integralReconstruction;
	switch (spaceSolver) {
		case Constant:
			integralReconstruction = get_reconstruction_first(grid, boundary, cellEdges_backpropagated, cellValues);
			break;
		case Linear:
			integralReconstruction = get_reconstruction_second(grid, boundary, cellEdges_backpropagated, cellValues);
			break;
	}

	auto reconstruction = [&](REAL x) -> REAL {return (*integralReconstruction)(x);};
	compute_new_values(grid, reconstruction, cellValues);
}


void SemilagrangeSolver::compute_characteristics(
		const Grid1D & grid,
		NumMatrix<REAL,1> & positions,
		const REAL dt,
		TimeSolver timeSolver
) const
{
		for (int iPos = loComp; iPos <= hiComp + 1; ++iPos) {
			positions(iPos) = compute_characteristic(grid, positions(iPos), dt, timeSolver);
		}
}

REAL SemilagrangeSolver::compute_characteristic(
		const Grid1D & grid,
		const REAL position,
		const REAL dt,
		TimeSolver timeSolver
) const
{
	switch (timeSolver) {
		case Euler:
			return applyRungeKutta_first(position, dt);
		case Second:
			return applyRungeKutta_second(position, dt);
		case Richardson:
			return applyRichardson_second(grid, position, dt, 0, dt);
	}
}

REAL SemilagrangeSolver::applyRichardson_second(
		const Grid1D & grid,
		const REAL position,
		REAL dt,
		const REAL t,
		const REAL tend
) const
{
	// assume dt negative 
	if(t + 1e-3 * dt < tend) {
		return position;
	}
	if(t + dt < tend) {
		dt = tend - t;
	}

	static const double richardsonPosMax = grid.get_xL(grid.getHi() + 1);

	const double p1 = applyRungeKutta_first(position, dt);
	const double p2H = applyRungeKutta_first(position, 0.5*dt);
	const double p2 = applyRungeKutta_first(p2H, 0.5*dt);
	const double err = p2 - p1;
	const double p3 = p2 + err;

	// this assumes the advection velocity is strongly increasing with position
	if(p1 > richardsonPosMax) {
		return richardsonPosMax;
	}

	// assumes a logarthmical grid
	static const double richardsonTolRel = std::pow(grid.get_xL(1)/grid.get_xL(0), 1./10.) - 1;
	const double TOL = richardsonTolRel * p2;
	const double dtNew = 0.9 * dt * std::sqrt(0.5 * TOL/abs(err));
	
	if(abs(err) < TOL) {
			return applyRichardson_second(grid, p3, dtNew, t + dt, tend);
	} else {
		return applyRichardson_second(grid, position, dtNew, t, tend);
	}
}


REAL SemilagrangeSolver::applyRungeKutta_first(
		const REAL position,
		const REAL dt
) const
{
		// Explicit Euler step
		const REAL shift1 = dt * advectionFunction->get_value(position);
		return position + shift1;
}

REAL SemilagrangeSolver::applyRungeKutta_second(
		const REAL position,
		const REAL dt
) const
{
	constexpr REAL alpha = 2./3.; // Ralston
	// constexpr REAL alpha = 0.5; // Heun
	// constexpr REAL alpha = 1.; // Midpoint

	constexpr REAL butcher_a11 = alpha;
	constexpr REAL butcher_b2 = 0.5 / alpha;
	constexpr REAL butcher_b1 = 1. - butcher_b2;

	const REAL pos1 = position;
	const REAL shift1 = dt * advectionFunction->get_value(pos1);
	const REAL pos2 = pos1 + butcher_a11 * shift1;
	const REAL shift2 = dt * advectionFunction->get_value(pos2);
	return position + butcher_b1 * shift1 + butcher_b2 * shift2;
}



void SemilagrangeSolver::propagate_edges(
		const Grid1D & grid,
		const REAL dt,
		TimeSolver timeSolver,
		NumMatrix<REAL,1> & cellEdges_backpropagated
)
{
	long nSub = 1;
	int iSubLevel;

	for (iSubLevel = 0; iSubLevel < maxSubcycleDepth; ++iSubLevel)
	{
		for (int iEdge = loComp; iEdge <= hiComp + 1; iEdge++) {
			cellEdges_backpropagated(iEdge) = grid.get_xL(iEdge);
		}

		for (long iSub = 0; iSub < nSub; ++iSub) {
			compute_characteristics(grid, cellEdges_backpropagated, -dt / nSub, timeSolver);
		}

		if (check_crossing(grid, cellEdges_backpropagated)) {
			nSub *= 2;
		} else {
			break;
		}
	}
	reachedSubcycleDepth = std::max(reachedSubcycleDepth, iSubLevel);

	if (iSubLevel == maxSubcycleDepth) {
		cerr << iSubLevel << " " << maxSubcycleDepth << endl;
		throw std::runtime_error("Semilagrange Solver: Maximum amount of subcycles reached");
	}

	if(showWarnings) {
		if (cellEdges_backpropagated(loComp) < grid.get_xL(lo)) {
			cerr << "Warning: Characteristics leave the ghost domain on the left side "
				 << grid.get_xL(lo) << " " << cellEdges_backpropagated(loComp) << endl;
		}

		if (cellEdges_backpropagated(hiComp + 1) > grid.get_xL(hi+1)) {
			cerr << "Warning: Characteristics leave the ghost domain on the right side "
				 << grid.get_xL(hi+1) << " " << cellEdges_backpropagated(hiComp + 1) << endl;
		}
	}
}

bool SemilagrangeSolver::check_crossing(
		const Grid1D & grid,
		const NumMatrix<REAL,1> & cellEdges_backpropagated
) const
{
	for (int iEdge = loComp; iEdge <= hiComp; iEdge++) {
		if (cellEdges_backpropagated(iEdge + 1) < grid.get_xL(0)) continue;
		if (cellEdges_backpropagated(iEdge) > grid.get_xL(grid.get_mx()))	continue;

		if (cellEdges_backpropagated(iEdge) > cellEdges_backpropagated(iEdge + 1)) {
			return true;
		}
	}
	return false;
}


unique_ptr<Interpolation1D> SemilagrangeSolver::get_reconstruction_first(
		const Grid1D & grid,
		const BoundaryHandler1D & boundary,
		const NumMatrix<REAL,1> & cellEdges_backpropagated,
		NumMatrix<REAL,1> & values
)
{
	// Compute the integral at the cell edges
	// Go from high to low to avoid cancellation errors in the summation
	accumulated_values.clear();
	accumulated_values(hi+1) = 0;
	for (int i = hi; i >= lo; --i) {
		accumulated_values(i) = accumulated_values(i+1) + (values(i) * (grid.get_xL(i+1) - grid.get_xL(i)));
	}

	int extrapolationType;
	if (boundary.get_bcType(0) == 1) {
		extrapolationType = 2;
	} else {
		extrapolationType = 0;
	}

	return unique_ptr<Interpolation1D>(
			new PiecewiseLinear(grid.get_xL(), accumulated_values, extrapolationType)
	);
}

unique_ptr<Interpolation1D> SemilagrangeSolver::get_reconstruction_second(
		const Grid1D & grid,
		const BoundaryHandler1D & boundary,
		const NumMatrix<REAL,1> & cellEdges_backpropagated,
		NumMatrix<REAL,1> & values
)
{
	// Compute the integral at the cell edges
	// Go from high to low to avoid cancellation errors in the summation
	accumulated_values.clear();
	accumulated_values(hi+1) = 0;
	for (int i = hi; i >= lo; --i) {
		accumulated_values(i) = accumulated_values(i+1) + (values(i) * (grid.get_xL(i+1) - grid.get_xL(i)));
	}

	parabolicCoeffs.clear();
	REAL dp, dc, dm;
	for (int i = hi - 1; i > lo; --i) {
	
		const double distanceR = grid.get_xCen(i + 1) - grid.get_xCen(i);
		const double distanceL = grid.get_xCen(i) - grid.get_xCen(i - 1);
		dc = (
				  (values(i + 1) - values(i)) * sqr(distanceL) +
				  (values(i) - values(i - 1)) * sqr(distanceR)
				 ) / (
				  distanceL * distanceR * (distanceR + distanceL)
				 );
				 
		
		//dc = (values(i + 1) - values(i - 1)) / (grid.get_xCen(i + 1) - grid.get_xCen(i - 1));

		if (slopeLimiter) {
			dp = (values(i + 1) - values(i)) / (grid.get_xCen(i + 1) - grid.get_xCen(i));
			dm = (values(i) - values(i - 1)) / (grid.get_xCen(i) - grid.get_xCen(i - 1));
			// The minus sign is needed for the reversed direction of the integration
			parabolicCoeffs(i) = -0.5 * slopeLimiter->compute(dp, dc, dm);
		}
		else {
			parabolicCoeffs(i) = -0.5 * dc;
		}
	}

	// Use one-sided slopes for the border cells
	//parabolicCoeffs(lo) = -0.5 * (values(lo + 1) - values(lo)) / (grid.get_xCen(lo + 1) - grid.get_xCen(lo));
	//parabolicCoeffs(hi) = -0.5 * (values(hi) - values(hi - 1)) / (grid.get_xCen(hi) - grid.get_xCen(hi - 1));

	boundary.do_BCsSecond(parabolicCoeffs, grid, advectionFunctionStl);

	int extrapolationType;
	if (boundary.get_bcType(0) == 1) {
		extrapolationType = 2;
	} else {
		extrapolationType = 0;
	}

	// Interpolate the integral
	return unique_ptr<Interpolation1D>(
			new PiecewiseParabolic(grid.get_xL(), accumulated_values, parabolicCoeffs, extrapolationType)
	);
}

void SemilagrangeSolver::compute_new_values(
		const Grid1D & grid,
		const std::function<REAL (REAL)> & integral_reconstruction,
		NumMatrix<REAL,1> & values
) const
{
	REAL upper_integral = 0.;
	REAL lower_integral = integral_reconstruction(cellEdges_backpropagated(hiComp + 1));


	// the new cell value is given by the integral at t_n
	for (int iCell = hiComp; iCell >= loComp; --iCell)
	{
		upper_integral = lower_integral;
		lower_integral = integral_reconstruction(cellEdges_backpropagated(iCell));
		values(iCell) = (lower_integral - upper_integral)
						/ (grid.get_xL(iCell + 1) - grid.get_xL(iCell));
	}
}


void SemilagrangeSolver::reset_reachedSubcycleDepth (
)
{
	reachedSubcycleDepth = 0;
}

int SemilagrangeSolver::get_reachedSubcycleDepth (
)
{
	return reachedSubcycleDepth;
}

void SemilagrangeSolver::set_slopeLimiter(
		const string limiterType
)
{
	if (limiterType == "none" || limiterType == "") {
		slopeLimiter = unique_ptr<limiter_func>(nullptr);
	} else
	if (limiterType == "minmod") {
		slopeLimiter = unique_ptr<limiter_func>(new minmod);
	} else
	if (limiterType == "vanLeer") {
		slopeLimiter = unique_ptr<limiter_func>(new vanLeer);
	} else {
		throw std::invalid_argument("no such limiter");
	}
}

void SemilagrangeSolver::set_slopeLimiter(
		const int limiterType
)
{
	if (limiterType == 0) {
		set_slopeLimiter("none");
	} else
	if (limiterType == 1) {
		set_slopeLimiter("minmod");
	} else
	if (limiterType == 2) {
		set_slopeLimiter("vanLeer");
	} else {
		throw std::invalid_argument("no such limiter");
	}
}

void SemilagrangeSolver::disableWarnings()
{
	showWarnings = false;
}
