#ifndef SEMILAGRANGE_SOLVER_H
#define SEMILAGRANGE_SOLVER_H

#ifdef MODULE_TEST
#define REAL double
#else
#include "constants.H"
#endif

#include "matrix.H"
#include "CException.H"
#include "Functions.H"
#include "Interpolation1D.H"
#include "vanLeer.H"
#include "minmod.H"
#include "Grid1D.H"
#include "BoundaryHandler1D.H"
#include <memory>
#include <functional>


class SemilagrangeSolver {
public:

	enum TimeSolver {
		Euler, Second, Richardson
	};

	enum SpaceSolver {
		Constant, Linear
	};


	/** deprecated constructor */
	SemilagrangeSolver(
			int order,
			bool slopeLimited = false,
			int limiterType = 2
	);

  /** deprecated constructor */
	SemilagrangeSolver(
			int temporalOrder,
			int spatialOrder
	);

	SemilagrangeSolver(
			TimeSolver timeSolver,
			SpaceSolver spaceSolver
	);

	~SemilagrangeSolver();

	void set_advectionFunction(
			Function<REAL> & advectionFunction
	);

	void resize(
			const Grid1D & grid
	);

	void set_computationLimits(
			int loComp,
			int hiComp
	);

	void singlestep(
			const Grid1D & grid,
			const BoundaryHandler1D & boundary,
			NumMatrix<REAL,1> & cellValues,
			const REAL dt
	);

	// Subcycle statistics
	void reset_reachedSubcycleDepth (
	);

	int get_reachedSubcycleDepth(
	);


	/** limiterType:
	 *  0 <-> none
	 * 	1 <-> minmod
	 * 	2 <-> vanLeer
	 */
	void set_slopeLimiter(
			const string limiterType
	);

	void set_slopeLimiter(
			const int limiterType
	);

	void disableWarnings();

private:

	bool showWarnings;
	int lo, hi, mx;
	int loComp, hiComp;
	bool legacySubcycling;

	// Characteristics
	Function<REAL> * advectionFunction;
	function<REAL(REAL)> advectionFunctionStl;
	TimeSolver timeSolver;

	NumMatrix<REAL,1> cellEdges_backpropagated;
	int maxSubcycleDepth;
	int reachedSubcycleDepth;

	void compute_characteristics(
		const Grid1D & grid,
			NumMatrix<REAL,1> & positions,
			const REAL dt,
			TimeSolver timeSolver
	) const;

	void propagate_edges(
			const Grid1D & grid,
			const REAL dt,
			TimeSolver timeSolver,
			NumMatrix<REAL,1> & cellEdges_backpropagated
	);

	bool check_crossing(
			const Grid1D & grid,
			const NumMatrix<REAL,1> & cellEdges_backpropagated
	) const;

	REAL compute_characteristic(
		const Grid1D & grid,
		const REAL position,
		const REAL dt,
		TimeSolver timeSolver
	) const;

	REAL applyRungeKutta_first(
		const REAL position,
		const REAL dt
	) const;

	REAL applyRungeKutta_second(
		const REAL position,
		const REAL dt
	) const;

	REAL applyRichardson_second(
		const Grid1D & grid,
		const REAL position,
		REAL dt,
		const REAL t,
		const REAL tend
	) const;


	// Value Reconstruction

	SpaceSolver spaceSolver;
	NumMatrix<REAL,1> accumulated_values;

	std::unique_ptr<Interpolation1D> get_reconstruction_first(
			const Grid1D & grid,
			const BoundaryHandler1D & boundary,
			const NumMatrix<REAL,1> & cellEdges_backpropagated,
			NumMatrix<REAL,1> & cellValues
	);

	// Needed for second order
	std::unique_ptr<limiter_func> slopeLimiter;
	NumMatrix<REAL,1> parabolicCoeffs;

	std::unique_ptr<Interpolation1D> get_reconstruction_second(
			const Grid1D & grid,
			const BoundaryHandler1D & boundary,
			const NumMatrix<REAL,1> & cellEdges_backpropagated,
			NumMatrix<REAL,1> & cellValues
	);

	void compute_new_values(
			const Grid1D & grid,
			const std::function<REAL (REAL)> & reconstruction,
			NumMatrix<REAL,1> & values
	) const;
};


#endif