#include <iostream>
#include "Synchrotron.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)

using namespace std;
using namespace CRONOS_CONSTANTS;

SynchrotronProcess::SynchrotronProcess(
    normalisation* normHD, normalisation* normParticle,
    std::shared_ptr<MagneticField> magneticField)
    : RadiativeLossProcess(normHD, normParticle),
      synchrotronEmission(normHD, normParticle, magneticField)
{
  this->emissionField = &synchrotronEmission;

  Quantity synchrotronUnit =
      4. / 3. * ThomsonCrossSection * SpeedOfLight / (2. * VacuumPermeability);
  synchrotronUnit_num = synchrotronUnit.normalizeTo(
      1. / normParticle->normQuantity(normParticle->TIME) /
      (sqr(normHD->normQuantity(normHD->MAG_IND)) /*= magnetic field*/ /
       normParticle->normQuantity(
           normParticle->ENERGY) /*= particle reset energy*/));

  this->magneticField = magneticField;

  // Some remaining initialisations
  synchrotronFac_num = 0.;
}

SynchrotronProcess::~SynchrotronProcess() {}

std::string SynchrotronProcess::get_type() const { return "Synchrotron"; }

void SynchrotronProcess::set_particle(Particle& particle)
{
  LossProcess::set_particle(particle);
  synchrotronEmission.set_particle(particle);
}

void SynchrotronProcess::compute_constants(Data& gdata, int ix, int iy, int iz)
{
  LossProcess::compute_constants(gdata, ix, iy, iz);

  REAL magneticFieldStrength_COM =
      magneticField->get_magneticFieldStrength_COM(gdata, ix, iy, iz);

  synchrotronFac_num = sqr(magneticFieldStrength_COM) * synchrotronUnit_num /
                       particle->particleRestEnergy_num;

  // Debug output
  gdata.sync_fac(ix, iy, iz) = synchrotronFac_num;
}

void SynchrotronProcess::get_momentumLoss(REAL momentum, REAL& loss)
{
  loss -= synchrotronFac_num * momentum * sqrt(sqr(momentum) + 1.);
}

REAL SynchrotronProcess::get_momentumLoss_c2(REAL momentum)
{
  return synchrotronFac_num;
}

//
//	Synchrotron Emission
//

SynchrotronEmission::SynchrotronEmission(
    normalisation* normHD, normalisation* normParticle,
    std::shared_ptr<MagneticField> magneticField)
    : EmissionField_COM(normHD, normParticle)
{
  this->magneticField = magneticField;

  particle = NULL;

  Quantity energyCritFac = 3. / 2. * PlanckConstantReduced;
  energyCritFac_num = energyCritFac.normalizeTo(
      normParticle->normQuantity(normParticle->ENERGY) /
      (normParticle->normQuantity(normParticle->CHAR) *
       normHD->normQuantity(normHD->MAG_IND) /
       normParticle->normQuantity(normParticle->MASS)));

  Quantity singleEmissivityFac =
      sqrt(3.) /
      (8. * sqr(pi) * VacuumPermittivity * SpeedOfLight * PlanckConstantReduced);
  singleEmissivityFac /= (4. * pi);  // To account for isotrope particle distro
  singleEmissivityFac_num = singleEmissivityFac.normalizeTo(
      normParticle->normQuantity(normParticle->ENERGY) /
      (normParticle->normQuantity(normParticle->TIME) *
       normParticle->normQuantity(normParticle->ENERGY)) /
      (cube(normParticle->normQuantity(normParticle->CHAR)) *
       normHD->normQuantity(normHD->MAG_IND) /
       normParticle->normQuantity(normParticle->MASS)));

  const IntegrationType type =
      value_exists("SynchrotronIntegrationType")
          ? catParameterMap[svalue("SynchrotronIntegrationType")]
          : IntegrationType::BinnedCenter;
  setIntegrationType(type);
}

SynchrotronEmission::~SynchrotronEmission() {}

void SynchrotronEmission::set_particle(Particle& particle)
{
  this->particle = &particle;
}

void SynchrotronEmission::setIntegrationType(const IntegrationType type)
{
  switch (type) {
    case IntegrationType::BinnedCenter:
      emissivitySpectrumIntegrator =
          std::unique_ptr<MidpointIntegrator>(new MidpointIntegrator());
      ccout << "  - sync: using binned integration (midpoint rule)" << endl;
      break;
      ;
    case IntegrationType::GaussPowerlaw:
      emissivitySpectrumIntegrator = std::unique_ptr<GaussPowerlawIntegrator>(
          new GaussPowerlawIntegrator());
      ccout << "  - sync: using fixed integration (piecewise powerlaw reconst)"
            << endl;
      break;
      ;
  }
}

double SynchrotronEmission::compute_singleSpectrum(const double energy,
                                                 const double lorentzParticle,
                                                 const double magneticFieldStrength) const
{
  assert(particle != NULL);

  const double energyCrit_num = energyCritFac_num * particle->particleChargeAbs_num /
                        particle->particleRestEnergy_num *
                        magneticFieldStrength * sqr(lorentzParticle);

  const double x = energy / energyCrit_num;
  double emissivity = 0.;

  // Avoiding underflow
  if (x < 8e2) {
    emissivity = singleEmissivityFac_num *
                 cube(particle->particleChargeAbs_num) /
                 particle->particleRestEnergy_num * magneticFieldStrength *
                 gsl_sf_synchrotron_1(x);
  }

  return emissivity;
}


double SynchrotronEmission::compute_singleSpectrum(const double energy, const double lorentzParticle, const double magneticFieldStrength,
                                                 const double magneticFieldDir[3], const double photonDirection[3]) const
{
  assert(particle != NULL);

  // Fall back to the average single spectrum if no direction information is available
  bool hasMagneticDir = false;
  for (int iDir = 0; iDir < 3; iDir++) {
    if (abs(magneticFieldDir[iDir]) > 1e-10) {
      hasMagneticDir = true;
      break;
    }
  }

  double perpendicularMagneticFieldStrength = 0;

  if (hasMagneticDir) {
    double cosPitch = 0;
    for (int iDir = 0; iDir < 3; iDir++) {
      cosPitch += magneticFieldDir[iDir] * photonDirection[iDir];
    }
    const double sinPitch = sqrt(1. - sqr(cosPitch));
    perpendicularMagneticFieldStrength = sinPitch * magneticFieldStrength;
  } else  {
    constexpr double averagedFieldFactor = M_PI / 4.;
    perpendicularMagneticFieldStrength = averagedFieldFactor * magneticFieldStrength;
  }

  return compute_singleSpectrum(energy, lorentzParticle, perpendicularMagneticFieldStrength);
}

REAL SynchrotronEmission::get_photonEmissivity_COM(
    Data& gdata, int ix, int iy, int iz, const REAL photonEnergy_COM,
    const REAL photonDirection_COM[3])
{
  assert(particle != NULL);

  const double magneticFieldStrength_COM =
      magneticField->get_magneticFieldStrength_COM(gdata, ix, iy, iz);

  double magneticFieldDir_COM[3];
  magneticField->get_magneticFieldDir_COM(gdata, ix, iy, iz, magneticFieldDir_COM);

  particle->getSpectrum(ix, iy, iz, particleSpectrumBuffer);

  EmissivitySpectrumIntegrator::integrationLimits limits = {
      .gammaLoIdx = particle->getLo(),
      .gammaHiIdx = 0,
      .gammaLo = particle->get_momentumEdge(particle->getLo()),
      .gammaHi = 0};

  for (int ip = particle->getHi(); ip >= particle->getLo(); --ip) {
    limits.gammaHiIdx = ip + 1;
    if (particleSpectrumBuffer(ip) > 0) {
      break;
    }
  }
  limits.gammaHi = particle->get_momentumEdge(limits.gammaHiIdx);

  auto diffEmissivity = [&](double gammaParticle) {
    return compute_singleSpectrum(photonEnergy_COM, gammaParticle,
                                  magneticFieldStrength_COM, magneticFieldDir_COM, photonDirection_COM) /
           photonEnergy_COM;
  };

  return emissivitySpectrumIntegrator->integrate(
      particle->grid, particleSpectrumBuffer, limits, diffEmissivity);
}

#endif
