#include "StateInvertors.H"

#include <iostream>
using namespace std;


/* -------------------------------------------------------------------------------------------- */

primState Invertor1Dw::getPrim(const consState & cons, const ProblemType & Problem) {
  paramType params = computeParams(cons, Problem);
  set_functions(cons, params);
  return getPrim(cons, params); 
}

Invertor1Dw::paramType Invertor1Dw::computeParams(const consState & cons, const ProblemType & Problem) {  
  paramType params;
  params.momentum_sqr = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z); ;
  params.gamma_adiab = Problem.gamma;
  params.gamma_tilde = params.gamma_adiab/(params.gamma_adiab-1.);

  return params;
}

void Invertor1Dw::set_functions(const consState & cons, const paramType & params) {
  fun_f = [cons, params](double Wprime) -> double {
    const REAL u_sqr = params.momentum_sqr / (sqr(Wprime + cons.density) - params.momentum_sqr);
    const REAL lorentz_sqr = 1 + u_sqr;
    const REAL lorentz = sqrt(lorentz_sqr);
    const REAL chi = (Wprime - cons.density * u_sqr / (1. + lorentz)) / lorentz_sqr;

    return (Wprime - chi / params.gamma_tilde) - cons.tau;
  };

  fun_df = [cons, params](double Wprime) -> double {
    const REAL u_sqr = params.momentum_sqr / (sqr(Wprime + cons.density) - params.momentum_sqr);
    const REAL lorentz_sqr = 1 + u_sqr;
    const REAL lorentz = sqrt(lorentz_sqr);
    const REAL chi = (Wprime - cons.density * u_sqr / (1. + lorentz)) / lorentz_sqr;

    const REAL eos_dp_dchi = 1./params.gamma_tilde;
    const REAL eos_dp_drho = 0.;
    const REAL dchi_dL = 1./lorentz_sqr + lorentz * (cons.density + 2 * lorentz * chi) * params.momentum_sqr / cube(Wprime + cons.density);
    const REAL drho_dL = cons.density * lorentz * params.momentum_sqr / cube(Wprime + cons.density);

    return 1. - eos_dp_dchi * dchi_dL - eos_dp_drho * drho_dL;
  };

  fun_fdf = [cons, params](double Wprime, double & y, double & dy) -> void {
    const REAL u_sqr = params.momentum_sqr / (sqr(Wprime + cons.density) - params.momentum_sqr);
    const REAL lorentz_sqr = 1 + u_sqr;
    const REAL lorentz = sqrt(lorentz_sqr);
    const REAL chi = (Wprime - cons.density * u_sqr / (1. + lorentz)) / lorentz_sqr;

    const REAL eos_dp_dchi = 1./params.gamma_tilde;
    const REAL eos_dp_drho = 0.;
    const REAL dchi_dL = 1./lorentz_sqr + lorentz * (cons.density + 2 * lorentz * chi) * params.momentum_sqr / cube(Wprime + cons.density);
    const REAL drho_dL = cons.density * lorentz * params.momentum_sqr / cube(Wprime + cons.density);

    y = (Wprime - chi / params.gamma_tilde) - cons.tau;
    dy = 1. - eos_dp_dchi * dchi_dL - eos_dp_drho * drho_dL;
  };
}

primState Invertor1Dw::getPrim(const consState & cons, Invertor1Dw::paramType & params) {
	const REAL momentum_sqr = params.momentum_sqr;
  const REAL energy = cons.density + cons.tau;
  const REAL energy_sqr = sqr(energy);
  const REAL gamma_adiab = params.gamma_adiab;
  const REAL Press2Etherm = 1./(gamma_adiab - 1.);

  const REAL Wprime0 = (2*energy + sqrt(4*energy_sqr - 3*params.momentum_sqr))/3. - cons.density;
  REAL Wprime;
  try{
    Wprime = solve_fdf(Wprime0);
  } catch (GslRootSolverError & e) {
    throw InversionFailure(e.what());
  }

  const REAL pressure = Wprime - cons.tau;
	const REAL u_sqr =  momentum_sqr / (sqr(Wprime + cons.density) - momentum_sqr);

	const REAL lorentz = sqrt(1. + u_sqr);
	const REAL Dhgamma = Wprime + cons.density;

	// Overwrite the conservatives with the primitives
	const REAL Dh_inv = lorentz / Dhgamma;

	primState prim;
	prim.rho = cons.density / lorentz;
	prim.Etherm = Press2Etherm * pressure;
	prim.ux = cons.momentum_x * Dh_inv;
	prim.uy = cons.momentum_y * Dh_inv;
	prim.uz = cons.momentum_z * Dh_inv;

	#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	prim.specific_entropy = (params.gamma_adiab-1) * prim.Etherm * pow(prim.rho, -params.gamma_adiab);
	#endif

	return prim;
}



/* -------------------------------------------------------------------------------------------- */


/* -------------------------------------------------------------------------------------------- */

primState InvertorQuartic::getPrim(const consState & cons, const ProblemType & Problem) {
  paramType params = computeParams(cons, Problem);
  set_functions(cons, params);
  return getPrim(cons, params); 
}

InvertorQuartic::paramType InvertorQuartic::computeParams(const consState & cons, const ProblemType & Problem) {

	const REAL momentum_sqr = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);
  const REAL gamma_tilde = Problem.gamma/(Problem.gamma-1.);
	const REAL density_sqr = sqr(cons.density);
  const REAL energy = cons.density + cons.tau;
  const REAL energy_sqr = sqr(energy);

  paramType params;
  params.momentum = sqrt(momentum_sqr);
  params.gamma_adiab = Problem.gamma;
  params.gamma_tilde = gamma_tilde;

  // REAL coeff[5];
	// params.polynomialCoeff[4] = (momentum_sqr - energy_sqr) * sqr(gamma_tilde);
	// params.polynomialCoeff[3] = 2.*cons.density*energy*gamma_tilde;
	// params.polynomialCoeff[2] = gamma_tilde*(energy_sqr*gamma_tilde - 2.*momentum_sqr) - density_sqr;
	// params.polynomialCoeff[1] = -params.polynomialCoeff[3];
	// params.polynomialCoeff[0] = density_sqr + momentum_sqr;

  /* = (momentum_sqr - energy_sqr) written out to avoid truncation errors */
  const REAL temp1 = (((momentum_sqr - density_sqr) - 2. * cons.density * cons.tau) - sqr(cons.tau));
	params.polynomialCoeff[4] = temp1 * sqr(params.gamma_tilde);
	params.polynomialCoeff[3] = 2. * cons.density * energy * params.gamma_tilde;
  /* = (gamma_tilde * (energy_sqr*gamma_tilde - 2.*momentum_sqr) - density_sqr) written out to avoid truncation errors */
  const REAL temp2 = (gamma_tilde * (density_sqr*params.gamma_tilde - 2.*momentum_sqr) - density_sqr) + (2. * cons.density * cons.tau + sqr(cons.tau)) * sqr(params.gamma_tilde);
  params.polynomialCoeff[2] = temp2;
	params.polynomialCoeff[1] = -params.polynomialCoeff[3];
	params.polynomialCoeff[0] = density_sqr + momentum_sqr;

  return params;
}

void InvertorQuartic::set_functions(const consState & cons, const paramType & params) {
  fun_f = [cons, params](double lorentz) -> double {
    	int degree = params.polynomialCoeff.size() - 1;
      REAL val = 0;
      while(degree > 0){
        val += params.polynomialCoeff[degree];
        val *= lorentz;
        --degree;
      }
      val += params.polynomialCoeff[0];
      return val;
  };
}


primState InvertorQuartic::getPrim(const consState & cons, InvertorQuartic::paramType & params) {
	const REAL momentum_sqr = sqr(params.momentum);
  const REAL energy = cons.density + cons.tau;
  const REAL energy_sqr = sqr(energy);
  const REAL gamma_adiab = params.gamma_adiab;
  const REAL gamma_tilde = params.gamma_tilde;

  if ( energy_sqr - momentum_sqr < 1e-4 * energy_sqr ) {
		cerr << " Warning: The Numerical Quartic Inversion might yield wrong results. " << (energy_sqr - momentum_sqr) / energy_sqr << endl;
	}

	const REAL lorentzLower = max(sqrt(
			(-2.*momentum_sqr + energy_sqr*gamma_tilde + energy*sqrt(energy_sqr*sqr(gamma_tilde) - 4.*momentum_sqr*(gamma_tilde-1.))) /
			(2.*(energy_sqr-momentum_sqr)*gamma_tilde)
			),
      1.);

	const REAL lorentzUpper = 1./sqrt(1. - momentum_sqr/energy_sqr + 1e-15) + 1e-3;

  const REAL fLower = fun_f(lorentzLower);
  const REAL fUpper = fun_f(lorentzUpper);

  REAL lorentz;

  if (fLower * fUpper < 0) {
    try{
      lorentz = solve_f(lorentzLower, lorentzUpper);
    } catch (GslRootSolverError & e) {
      throw InversionFailure(e.what());
    }
  } else
  if (abs(fLower) < 1e-9)
  {
    lorentz = 1;
  } else {
    throw InversionFailure();
  }
  
	const REAL u = sqrt(sqr(lorentz)-1.);
	const REAL Dh_inv = u/(params.momentum + 1e-120);

	primState prim;
	prim.rho = cons.density / lorentz;
	prim.Etherm = energy - cons.density - params.momentum * u / lorentz;
	prim.ux = cons.momentum_x * Dh_inv;
	prim.uy = cons.momentum_y * Dh_inv;
	prim.uz = cons.momentum_z * Dh_inv;

	#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	prim.specific_entropy = (gamma_adiab-1) * prim.Etherm * pow(prim.rho, -gamma_adiab);
	#endif

	return prim;
}



/* -------------------------------------------------------------------------------------------- */

primState InvertorPolytropic::getPrim(const consState & cons, const ProblemType & Problem) {
  paramType params = computeParams(cons, Problem);
  set_functions(cons, params);
  return getPrim(cons, params); 
}

InvertorPolytropic::paramType InvertorPolytropic::computeParams(const consState & cons, const ProblemType & Problem) {
  paramType params;

  #if (EOS == POLYTROPIC)
  params.momentum_sqr = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);
  params.polytropic_K = Problem.polytropic_K;
  params.gamma_adiab = Problem.gamma;
  #endif

  return params;
}

void InvertorPolytropic::set_functions(const consState & cons, const paramType & params) {
  fun_f = [cons, params](double rho) -> double {
    const REAL PoverRho = params.polytropic_K * std::pow(rho, params.gamma_adiab - 1.);
    const REAL enthalpy = 1 + params.gamma_adiab / (params.gamma_adiab - 1.) * PoverRho;
    const REAL u_sqr = params.momentum_sqr / sqr(cons.density * enthalpy);
    const REAL lorentz_sqr = 1. + u_sqr;
    const REAL lorentz = sqrt(lorentz_sqr);

    return rho * lorentz - cons.density;
  };
}

primState InvertorPolytropic::getPrim(const consState & cons, paramType & params) {
  const REAL Press2Etherm = 1./(params.gamma_adiab - 1.);

  const REAL lorentzLower = 1;
  const REAL lorentzUpper = sqrt(1. + params.momentum_sqr/sqr(cons.density));

  const REAL rhoLower = cons.density / lorentzUpper;
  const REAL rhoUpper = cons.density / lorentzLower;

  const REAL fLower = fun_f(rhoLower);
  const REAL fUpper = fun_f(rhoUpper);

  REAL rho;

  if (fLower * fUpper < 0) {
    try{
      rho = solve_f(rhoLower, rhoUpper);
    } catch (GslRootSolverError & e) {
      throw InversionFailure(e.what());
    }
  } else
  if (abs(fLower) < 1e-12)
  {
    rho = rhoLower;
  } else 
  if (abs(fUpper) < 1e-12)
  {
    rho = rhoUpper;
  } else {
    throw InversionFailure();
  }

	const REAL pressure = params.polytropic_K * pow(rho, params.gamma_adiab);
	const REAL Dh_inv = 1./(cons.density * (1. + params.gamma_adiab/(params.gamma_adiab - 1.) * pressure/rho));

	primState prim;
	prim.rho = rho;
	prim.Etherm = Press2Etherm * pressure;
	prim.ux = cons.momentum_x * Dh_inv;
	prim.uy = cons.momentum_y * Dh_inv;
	prim.uz = cons.momentum_z * Dh_inv;

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
  prim.specific_entropy = params.polytropic_K;
#endif

	return prim;
}



#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
/* -------------------------------------------------------------------------------------------- */

InvertorEntropyPolytropic::paramType InvertorEntropyPolytropic::computeParams(const consState & cons, const ProblemType & Problem)  {
  paramType params;

  params.momentum_sqr = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);
  params.polytropic_K = cons.sigma/cons.density;
  params.gamma_adiab = Problem.gamma;

  return params;
}


/* -------------------------------------------------------------------------------------------- */

primState InvertorEntropy::getPrim(const consState & cons, const ProblemType & Problem) {
  paramType params = computeParams(cons, Problem);
  set_functions(cons, params);
  return getPrim(cons, params); 
}

InvertorEntropy::paramType InvertorEntropy::computeParams(const consState & cons, const ProblemType & Problem) {
  paramType params;

  params.momentum_sqr = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);
  params.sigma = cons.sigma;
  params.gamma_adiab = Problem.gamma;
  params.gamma_tilde = Problem.gamma/(Problem.gamma-1.);

  return params;
}

void InvertorEntropy::set_functions(const consState & cons, const paramType & params) {
  fun_f = [cons, params](double lorentz) -> double {
    const REAL Z = lorentz*(cons.density + params.gamma_tilde * params.sigma * std::pow(cons.density/lorentz, params.gamma_adiab - 1));
    const REAL Z_sqr = sqr(Z);

    return (Z_sqr- params.momentum_sqr)*sqr(lorentz) - Z_sqr;
  };
}

primState InvertorEntropy::getPrim(const consState & cons, InvertorEntropy::paramType & params) {
 	const REAL polytropic_K = cons.sigma/cons.density;
  const REAL energy = cons.density + cons.tau;
  const REAL energy_sqr = sqr(energy);
  const REAL momentum = sqrt(params.momentum_sqr);

	const REAL lorentzLower = max(sqrt(1. + params.momentum_sqr/(sqr(cons.density + polytropic_K * params.gamma_tilde * std::pow(cons.density, params.gamma_adiab))))
                              , 1.);
  const	REAL lorentzUpper = sqrt(1. + params.momentum_sqr/sqr(cons.density));

  const REAL fLower = fun_f(lorentzLower);
  const REAL fUpper = fun_f(lorentzUpper);

  REAL lorentz;

  if (fLower * fUpper < 0) {
    try {
      lorentz = solve_f(lorentzLower, lorentzUpper);
    } catch (GslRootSolverError & e) {
      throw InversionFailure(e.what());
    }
  } else
  if (abs(fLower) < 1e-9)
  {
    lorentz = 1;
  } else {
    // cout << lorentzLower << " " << lorentzUpper << endl;
    // cout << fLower << " " << fUpper << endl;
    throw InversionFailure();
  }

	const REAL u = sqrt(sqr(lorentz)-1.);
	const REAL Dh_inv = u/(momentum + 1e-120);

	primState prim;
	prim.specific_entropy = polytropic_K;
	prim.rho = cons.density / lorentz;

	prim.Etherm = 1./(params.gamma_adiab - 1.) * prim.specific_entropy * std::pow(prim.rho, params.gamma_adiab);
	prim.ux = cons.momentum_x * Dh_inv;
	prim.uy = cons.momentum_y * Dh_inv;
	prim.uz = cons.momentum_z * Dh_inv;

	return prim;
}


#endif

primState InvertorPluto::getPrim(const consState & cons, const ProblemType & Problem) {
  int iter;
  const int MAX_ITER = 20;
  double p, D_1, alpha, alpha2, lor2, lor, m, tol=1.e-14;
  double tau, theta, h, dh_dp, dh_dtau, gmmr;
  double yp, dyp, dp, scrh;
  double D, E, m2;

  D    = cons.density;
  E    = cons.tau + D;
  m2 = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);

  #if EOS == IDEAL
   gmmr = Problem.gamma/(Problem.gamma - 1.0);
  #endif
  m    = sqrt(m2);
  p    = m - E;
  p    = max(p, 1.e-18);
  D_1  = 1.0/D;
  for (iter = 0; iter < MAX_ITER; iter++) {

    alpha  = E + p;
    alpha2 = alpha*alpha;
    lor2   = 1.0 - m2/alpha2;

    lor2 = max(lor2,1.e-9);
    lor2 = 1.0/lor2;
    lor  = sqrt(lor2);

    tau   = lor*D_1;
    theta = p*tau;

    #if EOS == IDEAL
     h       = 1.0 + gmmr*theta;
     dh_dp   = gmmr*tau;
     dh_dtau = gmmr*p;
    #elif EOS == TAUB
     h       = 2.5*theta + sqrt(2.25*theta*theta + 1.0);
     scrh    = (5.0*h - 8.0*theta)/(2.0*h - 5.0*theta);
     dh_dp   = tau*scrh;
     dh_dtau = p*scrh;
    #endif

    yp  = D*h*lor - E - p;
    dyp = D*lor*dh_dp - m2*lor2*lor/(alpha2*alpha)*(lor*dh_dtau + D*h) - 1.0;
    dp  = yp/dyp;
    p  -= dp;
    if (fabs (dp) < tol*p) break;
  }

	primState prim;
	prim.rho = cons.density / lor;
	prim.Etherm = p/(Problem.gamma - 1.);
	prim.ux = cons.momentum_x / (D*h);
	prim.uy = cons.momentum_y / (D*h);
	prim.uz = cons.momentum_z / (D*h);

/* -- Recompute entropy consistently -- */

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
  #if EOS == IDEAL
  prim.specific_entropy = p/pow(prim.rho,Problem.gamma);
  #elif EOS == TAUB
  theta = p/prim.rho;  
  prim.specific_entropy = p/pow(prim.rho,5.0/3.0)*(1.5*theta + sqrt(2.25*theta*theta + 1.0);
  #endif
#endif

cout << prim << endl;
  return prim;
}


#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
primState InvertorEntropyPluto::getPrim(const consState & cons, const ProblemType & Problem) {

  int    k, done;
  const int MAX_ITER = 20;
  double v2, v2max, v2min, acc = 1.e-13;
  double h, W, W2, W3;
  double scrh, x, temp;
  double lor, lor2, rho, sigma, th;
  double dW_dlor;
  double fv2, dfv2, dv2, dv2old;
  double m2, D;

  D    = cons.density;
  m2 = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);

  sigma = cons.sigma/D;
  v2max = 1.0 - 1.e-8;
  v2min = 0.0;

  done = 0;

  v2     = 0.5*(v2max+v2min);
  dv2old = v2max-v2min;
  dv2    = dv2old;

  lor2 = 1.0/(1.0 - v2);
  lor  = sqrt(lor2);
  rho  = D/lor;

  #if EOS == IDEAL
   x  = pow(rho,Problem.gamma - 1.0);
   h  = 1.0 + Problem.gamma/(Problem.gamma - 1.0)*sigma*x;
   W  = D*h*lor;
   dW_dlor = (2 - Problem.gamma)*W/lor + (Problem.gamma - 1.0)*D;
  #elif EOS == TAUB
   x  = pow(rho,2.0/3.0);
   th = sigma*x/sqrt(1.0 + 3.0*sigma*x);
   h  = 2.5*th + sqrt(2.25*th*th + 1.0);
   W  = D*h*lor;

   scrh  = -2.0*x/(3.0*lor)*sigma*(2.0*sigma*x - 3.0*th*th);
   scrh /=  2.0*th*(1.0 + 3.0*sigma*x);
   dW_dlor = W/lor + D*lor*(5.0*h - 8.0*th)/(2.0*h - 5.0*th)*scrh;
  #endif

  W2    = W*W;
  W3    = W2*W;

  fv2  = v2 - m2/W2;
  dfv2 = 1.0 + m2/W3*dW_dlor*lor2*lor;

  for (k = 1; k < MAX_ITER; k++) {
    if ((((v2-v2max)*dfv2-fv2)*((v2-v2min)*dfv2-fv2) > 0.0)
        || (fabs(2.*fv2) > fabs(dv2old*dfv2))) {
       dv2old = dv2;
       dv2 = 0.5*(v2max-v2min);
       v2 = v2min+dv2;
       if (v2min == v2) done = 1;
    } else { 
       dv2old = dv2;
       dv2 = fv2/dfv2;
       temp = v2;
       v2 -= dv2;
       if (temp == v2) done = 1;
    }

    if (fabs(dv2) < acc*v2 || fabs(fv2) < acc) done = 1;
    
    lor2 = 1.0/(1.0 - v2);
    lor  = sqrt(lor2);

    rho = D/lor;

    #if EOS == IDEAL
     x  = pow(rho,Problem.gamma - 1.0);
     h  = 1.0 + Problem.gamma/(Problem.gamma - 1.0)*sigma*x;
     W  = D*h*lor;
     dW_dlor = (2 - Problem.gamma)*W/lor + (Problem.gamma - 1.0)*D;
    #elif EOS == TAUB
     x  = pow(rho,2.0/3.0);
     th = sigma*x/sqrt(1.0 + 3.0*sigma*x);
     h  = 2.5*th + sqrt(2.25*th*th + 1.0);
     W  = D*h*lor;

     scrh  = -2.0*x/(3.0*lor)*sigma*(2.0*sigma*x - 3.0*th*th);
     scrh /=  2.0*th*(1.0 + 3.0*sigma*x);
     dW_dlor = W/lor + D*lor*(5.0*h-8.0*th)/(2.0*h-5.0*th)*scrh;
    #endif

    W2    = W*W;

    if (done) break;

    W3    = W2*W;
    fv2 = v2 - m2/W2;

    if (fv2 < 0.0) v2min = v2;
    else           v2max = v2;

    dfv2 = 1.0 + m2/W3*dW_dlor*lor2*lor;
  }

  double prs;
  #if EOS == IDEAL
   prs = sigma*x*rho;
  #elif EOS == TAUB
   prs = th*rho;
  #endif


	primState prim;
	prim.rho = cons.density / lor;
	prim.Etherm = prs/(Problem.gamma - 1.);
	prim.ux = cons.momentum_x / (D*h);
	prim.uy = cons.momentum_y / (D*h);
	prim.uz = cons.momentum_z / (D*h);
  prim.specific_entropy = sigma;

  return prim;
}


#endif