#ifndef RHD_INVERTORS_H
#define RHD_INVERTORS_H

#include "constants.H"
#include "problem.H"
#include "state.H"
#include "GslWrapper.H"
#include <string>


class IStateInvertor {
public:
  virtual primState getPrim(const consState & cons, const ProblemType & Problem) = 0;
};


class InversionFailure: public std::runtime_error {
public:
  InversionFailure(): runtime_error("Inversion Failure") {};
  InversionFailure(std::string error): runtime_error("Inversion Failure: " + error) {};
};



/* STANDARD RHD INVERTORS --------------------------------------------------------------------- */

class Invertor1Dw: public IStateInvertor, public GslRootSolver {
public:
  using paramType = struct {REAL momentum_sqr; REAL gamma_adiab; REAL gamma_tilde; };
  virtual primState getPrim(const consState & cons, const ProblemType & Problem) override;

  virtual paramType computeParams(const consState & cons, const ProblemType & Problem);
  virtual primState getPrim(const consState & cons, paramType & params);

protected:
  virtual void set_functions(const consState & cons, const paramType & Problem);
};


class InvertorQuartic: public IStateInvertor, public GslRootSolver {
public:
  using paramType = struct {REAL momentum; REAL gamma_adiab; REAL gamma_tilde; std::array<REAL, 5> polynomialCoeff;};

  virtual primState getPrim(const consState & cons, const ProblemType & Problem) override;

  virtual paramType computeParams(const consState & cons, const ProblemType & Problem);
  virtual primState getPrim(const consState & cons, paramType & params);

protected:
  virtual void set_functions(const consState & cons, const paramType & Problem);
};


class InvertorPluto: 
    public IStateInvertor
{
public:
  primState getPrim(const consState & cons, const ProblemType & Problem) final;
};


/* POLYTROPIC INVERTOR -------------------------------------------------------------------------------- */

class InvertorPolytropic: public IStateInvertor, public GslRootSolver {
public:
  using paramType = struct {REAL momentum_sqr; REAL polytropic_K; REAL gamma_adiab;};

  virtual primState getPrim(const consState & cons, const ProblemType & Problem) override;

  virtual paramType computeParams(const consState & cons, const ProblemType & Problem);
  virtual primState getPrim(const consState & cons, paramType & params);

protected:
  virtual void set_functions(const consState & cons, const paramType & Problem);
};

/* ENTROPY INVERTORS ---------------------------------------------------------------------------- */

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)

class InvertorEntropyPolytropic: public InvertorPolytropic {
public:
  using paramType = InvertorPolytropic::paramType;
  virtual paramType computeParams(const consState & cons, const ProblemType & Problem) override;
};


class InvertorEntropy: public IStateInvertor, public GslRootSolver {
public:
  using paramType = struct {REAL momentum_sqr; REAL sigma; REAL gamma_adiab; REAL gamma_tilde;};

  virtual primState getPrim(const consState & cons, const ProblemType & Problem) override;

  virtual paramType computeParams(const consState & cons, const ProblemType & Problem);
  virtual primState getPrim(const consState & cons, paramType & params);

protected:
  virtual void set_functions(const consState & cons, const paramType & Problem);

};


class InvertorEntropyPluto: 
    public IStateInvertor
{
public:
  primState getPrim(const consState & cons, const ProblemType & Problem) final;
};

#endif



#endif