#include "RiemannSolverRHD.H"

#if (ENERGETICS == FULL)
HLLCSolver_RHD::HLLCSolver_RHD(const Data &gdata, int dir, int Fluid_Type) :
	RiemannSolverRHD(gdata, dir, Fluid_Type)
{
	veps = 1.e-120;

#if(FLUID_TYPE != CRONOS_MULTIFLUID)
	reset_Indices(gdata.fluid);
#else
	reset_Indices(gdata.fluids->fluids[0]);
#endif
	if(dir == 0) {
		this->qvPar = q_sx;
		this->qvP1  = q_sy;
		this->qvP2  = q_sz;
	} else if (dir == 1) {
		this->qvP2  = q_sx;
		this->qvPar = q_sy;
		this->qvP1  = q_sz;
	} else {
		this->qvP1  = q_sx;
		this->qvP2  = q_sy;
		this->qvPar = q_sz;
	}   
	gamma = value((char*)"Adiabatic_exponent");

}

/*-------------------------------------------------------------------
  HLLC Rieman solver according to paper

  Mignone, Bodo
  "An HLLC Riemann solver for relativistic flows"
  Mon. Not. R. Astron. Soc. 364, 126–136 (2005)

  doi:10.1111/j.1365-2966.2005.09546.x
  -----------------------------------------------------------------*/

void HLLCSolver_RHD::get_NumFlux(Data &gdata,
                             cronos::vector<double> &Pos,
                             phys_fields_1D &pfL,
                             phys_fields_1D &pfR,
                             fields_1D &fl, const int &dir, int iFluid) {

	// In case of multifluid run, need to re-evaluate indices:
#if(FLUID_TYPE == CRONOS_MULTIFLUID)
	if(dir == 0) {
		this->qvPar = q_sx;
		this->qvP1  = q_sy;
		this->qvP2  = q_sz;
	} else if (dir == 1) {
		this->qvP2  = q_sx;
		this->qvPar = q_sy;
		this->qvP1  = q_sz;
	} else {
		this->qvP1  = q_sx;
		this->qvP2  = q_sy;
		this->qvPar = q_sz;
	}
#endif

	int n_omInt(fl.get_num());


	REAL (Data::*getEdgL)(int) const;

	if (dir == 0) {
		getEdgL = &Data::getEdgL_x;
	} else if (dir == 1) {
		getEdgL = &Data::getEdgL_y;
	} else {
		getEdgL = &Data::getEdgL_z;
	}

	for (int i = -1; i <= gdata.mx[dir]+1; ++i){
		#ifdef USE_VALENCIA
			Pos.set(dir, (gdata.*getEdgL)(i));
			const REAL alpha = gdata.valencia->get_alpha(Pos.get(0), Pos.get(1), Pos.get(2));
			const REAL beta = gdata.valencia->get_betaCont(Pos.get(0), Pos.get(1), Pos.get(2), dir);
			const REAL v_interface = beta/alpha;
		#else
			const REAL v_interface = 0;
		#endif

		/*-------------------------------------------------------
		  Compute fluxes for regions outside the Riemann fan:
		  -----------------------------------------------------*/
		
		if(-fl.v_ch_m(i) >= v_interface) { // Riemann fan going right
			
			for(int q=0; q<n_omInt; ++q) {
				fl.flux[q](i) = pfR.flux[q](i-1) - v_interface * pfR.uCon[q](i-1);
			}
//#if(FLUID_TYPE==CRONOS_MULTIFLUID)
//			fl.ptotal[iFluid](i) = pfR.ptotal(i-1);
//#else
			fl.ptotal(i) = pfR.ptotal(i-1);
//#endif
			
		} else if (fl.v_ch_p(i) <= v_interface) { // Riemann fan going left
			
			for(int q=0; q<n_omInt; ++q) {
				fl.flux[q](i) = pfL.flux[q](i) - v_interface * pfL.uCon[q](i);
			}
//#if(FLUID_TYPE==CRONOS_MULTIFLUID)
//			fl.ptotal[iFluid](i) = pfL.ptotal(i);
//#else
			fl.ptotal(i) = pfL.ptotal(i);
//#endif
			
		} else {
		
			/*-------------------------------------------------------
			  The rest is taking place inside the Riemann fan
			  -------------------------------------------------------*/

			// If carbuncle test is used -> check whether cells i and i-1 are
			// flagged as lying within the vicinity of a strong shock
			if(gdata.use_carbuncleFlag) {
				// if both cells are flagged -> use hll for all variables instead
				if(fl.carbuncle_flag(i)*fl.carbuncle_flag(i-1) > 0) {
					for(int q=0; q<fl.get_num(); ++q) {
						REAL fac = 1./(fl.v_ch_p(i)+fl.v_ch_m(i)+veps);

						fl.flux[q](i) = (fl.v_ch_m(i)*pfL.flux[q](i) +
										 fl.v_ch_p(i)*pfR.flux[q](i-1) -
										 fl.v_ch_m(i)*fl.v_ch_p(i)*(pfL.uCon[q](i  )-
												 	 	 	 	 	pfR.uCon[q](i-1)))*fac;
					}
					continue;
				}
			}



			REAL uConL[n_omInt], uConR[n_omInt];
			REAL uPriL[n_omInt], uPriR[n_omInt];
//			REAL uConSL[n_omInt], uPriSL[n_omInt];
			REAL uConSL[n_omInt];
//			REAL uConSR[n_omInt], uPriSR[n_omInt];
			REAL uConSR[n_omInt];

			// Saving array values:
				
			for(int q=0; q<n_omInt; ++q) {
				/* 
				   Be aware of the two differnt meanings of "L" For
				   hllc "L" means "on the Left hand side of the cell
				   interface" For the reconstruction it means on the
				   left side of the cell
				*/
				uConR[q] = pfL.uCon[q](i  );
				uConL[q] = pfR.uCon[q](i-1);
				uPriR[q] = pfL.uPri[q](i  );
				uPriL[q] = pfR.uPri[q](i-1);
			}


			REAL densityL = uConL[q_rho];
			REAL densityR = uConR[q_rho];

			REAL pThermL(pfR.ptherm(i-1));
			REAL pThermR(pfL.ptherm(i  ));

			REAL WL = sqrt(1 + sqr(uPriL[q_sx]) + sqr(uPriL[q_sy]) + sqr(uPriL[q_sz]));
			REAL WR = sqrt(1 + sqr(uPriR[q_sx]) + sqr(uPriR[q_sy]) + sqr(uPriR[q_sz]));

			REAL vParL = uPriL[qvPar] / WL;
			REAL vParR = uPriR[qvPar] / WR;

			REAL mParL = uConL[qvPar];
			REAL mParR = uConR[qvPar];
			
			REAL EL = uConL[q_Eges];
			REAL ER = uConR[q_Eges];

			// Convert Tau -> Energy
			EL += densityL;
			ER += densityR;

			// Characteristic velocities
			REAL vCharL = -fl.v_ch_m(i);
			REAL vCharR =  fl.v_ch_p(i);

			// Just the signal velocities relative to background flow
			REAL vSigL = vCharL - vParL;
			REAL vSigR = vCharR - vParR;





			// Velocity of contact discontinuity S*
			// Eq. (18)
			REAL vCharS = 0;

			REAL AL = vCharL * EL - mParL;
			REAL AR = vCharR * ER - mParR;
			REAL BL = mParL * vSigL - pThermL;
			REAL BR = mParR * vSigR - pThermR;

			REAL c_2 = vCharL * AR - vCharR * AL;
			REAL c_1 = (AR - AL + vCharL * BR - vCharR * BL);
			REAL c_0 = BR - BL;

			if (c_2 == 0) {
				vCharS = c_0 / c_1;
			} else {
				vCharS = (c_1 - sqrt(sqr(c_1) - 4. * c_2 * c_0)) / (2. * c_2);
			}


			// Pressure of contact discontinuity pS
			// Eq. (17)
			REAL pSL =  (vCharS * AL - BL) / (1 - vCharL * vCharS + veps);
			REAL pSR =  (vCharS * AR - BR) / (1 - vCharR * vCharS + veps);
			REAL pS = 0.5*(pSL + pSR);


			// Starred quantities according to Eq. (16)
			REAL facL = 1 / (vCharL - vCharS + veps);
			REAL facR = 1 / (vCharR - vCharS + veps);

			// Density in middle region
			uConSL[q_rho] = uConL[q_rho]*vSigL * facL;
			uConSR[q_rho] = uConR[q_rho]*vSigR * facR;

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
			// Entropy in middle region (analog to the density)
			uConSL[q_Eadd] = uConL[q_Eadd] * vSigL * facL;
			uConSR[q_Eadd] = uConR[q_Eadd] * vSigR * facR;
#endif

			// Momenta in middle region
			uConSL[qvPar] = (mParL * vSigL + pS - pThermL) * facL;
			uConSL[qvP1 ] = uConL[qvP1] * vSigL * facL;
			uConSL[qvP2 ] = uConL[qvP2] * vSigL * facL;

			uConSR[qvPar] = (mParR * vSigR + pS - pThermR) * facR;
			uConSR[qvP1 ] = uConR[qvP1] * vSigR * facR;
			uConSR[qvP2 ] = uConR[qvP2] * vSigR * facR;


			// Overall energy in middle region
			uConSL[q_Eges] = (EL*vSigL + pS*vCharS - pThermL*vParL) * facL;
			uConSR[q_Eges] = (ER*vSigR + pS*vCharS - pThermR*vParR) * facR;

			// Convert Energy -> Tau
			uConSL[q_Eges] -= uConSL[q_rho];
			uConSR[q_Eges] -= uConSR[q_rho];


			// Fluxes from Eq (13)
			if(vCharS >= v_interface) {

				for(int q=0; q<n_omInt; ++q) {
					fl.flux[q](i) = pfR.flux[q](i-1) + vCharL*(uConSL[q] -
					                                           uConL[q])
													 - v_interface * uConSL[q];
					fl.ptotal(i) = pfR.ptherm(i-1);
				}
				
			} else {
				
				for(int q=0; q<n_omInt; ++q) {
					fl.flux[q](i) = pfL.flux[q](i  ) + vCharR*(uConSR[q] -
					                                           uConR[q])
													 - v_interface * uConSR[q];
					fl.ptotal(i) = pfL.ptherm(i  );
				}
				
			}
		}

	}
}


void HLLCSolver_RHD::get_NumFlux(Data &gdata, phys_fields_0D &pfM,
		phys_fields_0D &pfP, num_fields_0D &f_num, int dir, int iFluid) {
	//! Compute numerical flux from phys fluxes using hllc solver
	/*
	 * @PARAM pfM variables and fluxes on left-hand side of interface
	 * @PARAM pfP variables and fluxes on right-hand side of interface
	 */

	throw CException(" 0D Methods are not available for RHD yet ");
}

#endif
