#include "integration.H"

template <typename T>
int AdaptiveIntegrator<T>::get_iterations(
	) const
{
	return iterations;
}

template <typename T>
T Integrator<T>::integrate(
		Function<T> * integrand,
		T lowerLimit,
		T upperLimit
)
{

	return integrate(*integrand, lowerLimit, upperLimit);
}

template <typename T>
T ClosedIntegrator<T>::integrate(
		Function<T> * integrand,
		T lowerLimit,
		T upperLimit,
		T lowerVal,
		T upperVal
)
{
	return integrate(*integrand, lowerLimit, upperLimit, lowerVal, upperVal);
}

template <typename T>
T TrapezoidRule<T>::integrate(
		Function<T> & integrand,
		T lowerLimit,
		T upperLimit
		)
{
	T lowerVal = integrand(lowerLimit);
	T upperVal = integrand(upperLimit);

	return integrate(
			integrand,
			lowerLimit,
			upperLimit,
			lowerVal,
			upperVal
		);
}

template <typename T>
T TrapezoidRule<T>::integrate(
				Function<T> & integrand,
				T lowerLimit,
				T upperLimit,
				T lowerVal,
				T upperVal
			)
{
	return 0.5 * (upperVal + lowerVal) * (upperLimit - lowerLimit);
}



template <typename T>
T AdaptiveIntegrator<T>::integrate(
		Function<T> * integrand,
		T lowerLimit,
		T upperLimit,
		T tolerance
)
{
	return integrate(*integrand, lowerLimit, upperLimit, tolerance);
}

template <typename T>
T AdaptiveIntegrator<T>::integrate(
		Function<T> * integrand,
		T lowerLimit,
		T upperLimit,
		T lowerVal,
		T upperVal,
		T tolerance
)
{
	return integrate(*integrand, lowerLimit, upperLimit, lowerVal, upperVal,tolerance);
}

template <typename T>
AdaptiveSubdivisionIntegration<T>::AdaptiveSubdivisionIntegration(
		ClosedIntegrator<T> * integrator
	) :	AdaptiveIntegrator<T>(integrator->order + 1)
{
	this->integrator = integrator;
	T p2 = std::pow(2, integrator->order);
	extrapolationFac = 1. / (p2 - 1.);
	errorFac = p2 / (p2 - 1.);
}


template <typename T>
T AdaptiveSubdivisionIntegration<T>::integrate(
		Function<T> & integrand,
		T lowerLimit,
		T upperLimit,
		T tolerance
	)
{
	this->iterations = 0;
	T singleIntegralValue = integrator->integrate(integrand, lowerLimit, upperLimit);
	return subdiv_integrate( integrand, lowerLimit, upperLimit, singleIntegralValue, tolerance);
}

#include <iostream>

template <typename T>
T AdaptiveSubdivisionIntegration<T>::subdiv_integrate(
		Function<T> & integrand,
		T lowerLimit,
		T upperLimit,
		T singleIntegralValue,
		T tolerance
	)
{
	T midLimit = 0.5 * (lowerLimit + upperLimit);
	T midVal = integrand(midLimit);

	T subIntegral1 = integrator->integrate( integrand, lowerLimit, midLimit );
	T subIntegral2 = integrator->integrate( integrand, midLimit, upperLimit );
	T subIntegralValue = subIntegral1 + subIntegral2;
	T integralDiff = subIntegralValue - singleIntegralValue;

	if ( std::abs(integralDiff) * errorFac > tolerance )
	{
		this->iterations++;
		subIntegral1 = subdiv_integrate( integrand, lowerLimit, midLimit, subIntegral1, tolerance);
		subIntegral2 = subdiv_integrate( integrand, midLimit, upperLimit, subIntegral2, tolerance);
		subIntegralValue = subIntegral1 + subIntegral2;
		return subIntegralValue;
	}
	else
	{
		// Richardson Extrapolation
		return subIntegralValue + integralDiff * extrapolationFac;
	}
}

template <typename T>
T AdaptiveSubdivisionIntegration<T>::integrate(
		Function<T> & integrand,
		T lowerLimit,
		T upperLimit,
		T lowerVal,
		T upperVal,
		T tolerance
	)
{
	this->iterations = 0;
	T singleIntegralValue = integrator->integrate( integrand, lowerLimit, upperLimit, lowerVal, upperVal );
	return subdiv_integrate( integrand, lowerLimit, upperLimit, lowerVal, upperVal, singleIntegralValue, tolerance);
}

template <typename T>
T AdaptiveSubdivisionIntegration<T>::subdiv_integrate(
		Function<T> & integrand,
		T lowerLimit,
		T upperLimit,
		T lowerVal,
		T upperVal,
		T singleIntegralValue,
		T tolerance
	)
{
	T midLimit = 0.5 * (lowerLimit + upperLimit);
	T midVal = integrand(midLimit);

	T subIntegral1 = integrator->integrate( integrand, lowerLimit, midLimit, lowerVal, midVal );
	T subIntegral2 = integrator->integrate( integrand, midLimit, upperLimit, midVal, upperVal );
	T subIntegralValue = subIntegral1 + subIntegral2;
	T integralDiff = subIntegralValue - singleIntegralValue;

	if ( std::abs(integralDiff) * errorFac > tolerance )
	{
		this->iterations++;
		subIntegral1 = subdiv_integrate( integrand, lowerLimit, midLimit, lowerVal, midVal, subIntegral1, tolerance);
		subIntegral2 = subdiv_integrate( integrand, midLimit, upperLimit, midVal, upperVal, subIntegral2, tolerance);
		subIntegralValue = subIntegral1 + subIntegral2;
		return subIntegralValue;
	}
	else
	{
		// Richardson Extrapolation
		return subIntegralValue + integralDiff * extrapolationFac;
	}
}



template <typename T>
AdaptiveTrapezSubdivisionIntegration<T>::AdaptiveTrapezSubdivisionIntegration(
		) :
		AdaptiveSubdivisionIntegration<T>(new TrapezoidRule<T>)
{
}


template <typename T>
AdaptiveTrapezSubdivisionIntegration<T>::~AdaptiveTrapezSubdivisionIntegration(
		)
{
	delete AdaptiveSubdivisionIntegration<T>::integrator;
}

template <typename T, int n>
Quadrature<T,n>::Quadrature(int p)
		: Integrator<T>(p)
{
	set_weights();
	set_abscissae();
}

template <typename T, int n>
T Quadrature<T, n>::integrate(
		Function<T> & integrand,
		T lowerLimit,
		T upperLimit
)
{
	T dM = 0.5 * (upperLimit - lowerLimit);
	T dP = 0.5 * (upperLimit + lowerLimit);
	T result = 0;
	for (int iStep = 0; iStep < n; ++iStep) {
		T x = dM * abscissae[iStep] + dP;
		result += weights[iStep] * integrand(x);
	}
	return dM * result;
}


template <typename T>
void GaussianQuadrature<T, 2>::set_weights() {
	T * weights = this->weights;
	weights[0] = weights[1] = 1.0000000000000000000000000;
}

template <typename T>
void GaussianQuadrature<T, 2>::set_abscissae() {
	T * abscissae = this->abscissae;
	abscissae[0] = 0.5773502691896257645091488;
	abscissae[1] = -abscissae[0];
}

template <typename T>
void GaussianQuadrature<T, 4>::set_weights() {
	T * weights = this->weights;
	weights[0] = weights[2] = 0.6521451548625461426269361;
	weights[1] = weights[3] = 0.3478548451374538573730639;
}

template <typename T>
void GaussianQuadrature<T, 4>::set_abscissae() {
	T * abscissae = this->abscissae;
	abscissae[0] = 0.339981043584856264802665;
	abscissae[1] = 0.8611363115940525752239465;
	abscissae[2] = -abscissae[0];
	abscissae[3] = -abscissae[1];
}

template <typename T>
void GaussianQuadrature<T, 6>::set_weights() {
	T * weights = this->weights;
	weights[0] = weights[3] = 0.4679139345726910473898703;
	weights[1] = weights[4] = 0.3607615730481386075698335;
	weights[2] = weights[5] = 0.1713244923791703450402961;
}

template <typename T>
void GaussianQuadrature<T, 6>::set_abscissae() {
	T * abscissae = this->abscissae;
	abscissae[0] = 0.2386191860831969086305017;
	abscissae[1] = 0.6612093864662645136613996;
	abscissae[2] = 0.9324695142031520278123016;
	abscissae[3] = -abscissae[0];
	abscissae[4] = -abscissae[1];
	abscissae[5] = -abscissae[2];
}



// ---------------------------------------------
// Precompilations

template class Integrator<float>;
template class AdaptiveIntegrator<float>;
template class TrapezoidRule<float>;
template class AdaptiveSubdivisionIntegration<float>;
template class AdaptiveTrapezSubdivisionIntegration<float>;

template class Integrator<double>;
template class AdaptiveIntegrator<double>;
template class TrapezoidRule<double>;
template class AdaptiveSubdivisionIntegration<double>;
template class AdaptiveTrapezSubdivisionIntegration<double>;


template class GaussianQuadrature<float, 2>;
template class GaussianQuadrature<float, 4>;
template class GaussianQuadrature<float, 6>;

template class GaussianQuadrature<double, 2>;
template class GaussianQuadrature<double, 4>;
template class GaussianQuadrature<double, 6>;





