#ifndef CRONOS_INTEGRATION_H
#define CRONOS_INTEGRATION_H

#include <cmath>

#include "Functions.H"


template <typename T>
class Integrator {
public:
	Integrator(int p): order(p) {};
	virtual ~Integrator() {};

	T integrate(
			Function<T> * integrand,
			T lowerLimit,
			T upperLimit
	);

	virtual T integrate(
			Function<T> & integrand,
			T lowerLimit,
			T upperLimit
	) = 0;

	const int order;
};


template <typename T>
class ClosedIntegrator: public Integrator<T> {
public:
	ClosedIntegrator(int p): Integrator<T>(p) {};
	virtual ~ClosedIntegrator() {};

	T integrate(
			Function<T> * integrand,
			T lowerLimit,
			T upperLimit,
			T lowerVal,
			T upperVal
	);

	virtual T integrate(
			Function<T> & integrand,
			T lowerLimit,
			T upperLimit,
			T lowerVal,
			T upperVal
	) = 0;

	virtual T integrate(
			Function<T> & integrand,
			T lowerLimit,
			T upperLimit
	) = 0;
};


template <typename T>
class AdaptiveIntegrator {
public:
	virtual ~AdaptiveIntegrator() {};

	T integrate(
			Function<T> * integrand,
			T lowerLimit,
			T upperLimit,
			T tolerance
	);

	T integrate(
			Function<T> * integrand,
			T lowerLimit,
			T upperLimit,
			T lowerVal,
			T upperVal,
			T tolerance
	);

	virtual T integrate(
				Function<T> & integrand,
				T lowerLimit,
				T upperLimit,
				T tolerance
			) = 0;

	virtual T integrate(
				Function<T> & integrand,
				T lowerLimit,
				T upperLimit,
				T lowerVal,
				T upperVal,
				T tolerance
			) = 0;

	int get_iterations(
		) const;

	const int order;

protected:
	int iterations;
	AdaptiveIntegrator(int p): order(p) {};
};



template <typename T>
class TrapezoidRule: public ClosedIntegrator<T> {
public:
	TrapezoidRule() : ClosedIntegrator<T>(2) {};

	T integrate(
				Function<T> & integrand,
				T lowerLimit,
				T upperLimit
			);

	T integrate(
				Function<T> & integrand,
				T lowerLimit,
				T upperLimit,
				T lowerVal,
				T upperVal
			);

};


template <typename T, int n>
class Quadrature: public Integrator<T> {
public:
	Quadrature(int p);

	virtual T integrate(
			Function<T> & integrand,
			T lowerLimit,
			T upperLimit
	);

protected:
	virtual void set_weights() = 0;
	virtual void set_abscissae() = 0;

	T weights[n];
	T abscissae[n];
};

template <typename T, int n>
class GaussianQuadrature_Base: public Quadrature<T, n> {
public:
	GaussianQuadrature_Base(): Quadrature<T,n>(2 * n) {};
};

template <typename T, int n>
class GaussianQuadrature: public GaussianQuadrature_Base<T, n> {};

template <typename T>
class GaussianQuadrature<T, 2>: public GaussianQuadrature_Base<T, 2>{
public:
	void set_weights() final;
	void set_abscissae() final;
};

template <typename T>
class GaussianQuadrature<T, 4>: public GaussianQuadrature_Base<T, 4>{
public:
	void set_weights() final;
	void set_abscissae() final;
};

template <typename T>
class GaussianQuadrature<T, 6>: public GaussianQuadrature_Base<T, 6>{
public:
	void set_weights() final;
	void set_abscissae() final;
};


template <typename T>
class AdaptiveSubdivisionIntegration: public AdaptiveIntegrator<T> {
public:
	AdaptiveSubdivisionIntegration(
			ClosedIntegrator<T> * integrator
		);

	T integrate(
			Function<T> & integrand,
			T lowerLimit,
			T upperLimit,
			T tolerance
		);

	T integrate(
			Function<T> & integrand,
			T lowerLimit,
			T upperLimit,
			T lowerVal,
			T upperVal,
			T tolerance
		);

protected:
	ClosedIntegrator<T> * integrator;

private:
	T errorFac;
	T extrapolationFac;

	T subdiv_integrate(
			Function<T> & integrand,
			T lowerLimit,
			T upperLimit,
			T singleIntegralValue,
			T tolerance
		);

	T subdiv_integrate(
			Function<T> & integrand,
			T lowerLimit,
			T upperLimit,
			T lowerVal,
			T upperVal,
			T singleIntegralValue,
			T tolerance
		);


};


template <typename T>
class AdaptiveTrapezSubdivisionIntegration: public AdaptiveSubdivisionIntegration<T> {
public:
	AdaptiveTrapezSubdivisionIntegration();
	~AdaptiveTrapezSubdivisionIntegration();
};


#endif
