#include "rhd_implicit_trafo.H"
#include <iostream>
#include <cassert>


void RHD_TransCons2P_implicit::set_x(REAL pressure_) {
	pressure = pressure_;
	energy_tilde = energy + pressure;
	gamma_rel_sqr = 1. / (1. - momentum_sqr/(energy_tilde*energy_tilde));
	gamma_rel = sqrt(gamma_rel_sqr);
}

REAL RHD_TransCons2P_implicit::get_value() {
	return energy_tilde - gamma_rel*(density + pressure*gamma_tilde*gamma_rel);
}

REAL RHD_TransCons2P_implicit::get_slope() {
	return 1. + gamma_rel_sqr * (gamma_rel * momentum_sqr/cube(energy_tilde) * (density+ 2.*pressure*gamma_tilde*gamma_rel) - gamma_tilde);
}

void RHD_TransCons2P_implicit::set_parameters(REAL density_, REAL momentum_sqr_, REAL energy_, REAL gamma_adiab_) {
	energy = energy_;
	density = density_;
	momentum_sqr = momentum_sqr_;
	gamma_tilde = gamma_adiab_/(gamma_adiab_-1.);
}

// ---------------------------------------------------------------------------------------------------------

void RHD_TransCons2Wprime_implicit::set_x(REAL Wprime_) {
	Wprime = Wprime_;
	u_sqr = momentum_sqr / (sqr(Wprime + density) - momentum_sqr);
	gamma_rel_sqr = 1 + u_sqr;
	gamma_rel = sqrt(gamma_rel_sqr);
	chi = (Wprime - density * u_sqr / (1. + gamma_rel)) / gamma_rel_sqr;
}

REAL RHD_TransCons2Wprime_implicit::get_value() {
	return Wprime - tau - get_pressure();
}


REAL RHD_TransCons2Wprime_implicit::get_slope() {
	eos_dp_dchi = 1./gamma_tilde;
	eos_dp_drho = 0.;
	dchi_dL = 1./gamma_rel_sqr + gamma_rel * (density + 2 * gamma_rel * chi) * momentum_sqr / cube(Wprime + density);
	drho_dL = density * gamma_rel * momentum_sqr / cube(Wprime + density);

	return 1. - eos_dp_dchi * dchi_dL - eos_dp_drho * drho_dL;
}

REAL RHD_TransCons2Wprime_implicit::get_pressure() {
	return chi / gamma_tilde;
}

void RHD_TransCons2Wprime_implicit::set_parameters(REAL density_, REAL momentum_sqr_, REAL tau_, REAL gamma_adiab_) {
	tau = tau_;
	density = density_;
	momentum_sqr = momentum_sqr_;
	gamma_tilde = gamma_adiab_/(gamma_adiab_-1.);
}


// ---------------------------------------------------------------------------------------------------------

void RHD_TransConsP2E_implicit::set_x(REAL energy_) {
	energy = energy_;
	energy_tilde = energy + pressure;
	gamma_rel = 1. / sqrt(1. - momentum_sqr/sqr(energy_tilde));
}

REAL RHD_TransConsP2E_implicit::get_value() {
	return energy_tilde - gamma_rel*(density + pressure*gamma_tilde*gamma_rel);
}

REAL RHD_TransConsP2E_implicit::get_slope() {
	return 1. + cube(gamma_rel/energy_tilde)*momentum_sqr * (density+ 2.*pressure*gamma_tilde*gamma_rel);
}

void RHD_TransConsP2E_implicit::set_parameters(REAL density_, REAL momentum_sqr_, REAL pressure_, REAL gamma_adiab_) {
	pressure = pressure_;
	density = density_;
	momentum_sqr = momentum_sqr_;
	gamma_tilde = gamma_adiab_/(gamma_adiab_-1.);
}


// ---------------------------------------------------------------------------------------------------------

void RHD_TransEntropy2Lorentz_implicit::set_x(REAL lorentz_) {
	lorentz = lorentz_;
	Z = lorentz*(density + gamma_tilde * sigma * std::pow(density/lorentz, gamma_adiab_red));

	lorentz_sqr = sqr(lorentz);
	Z_sqr = sqr(Z);
}

REAL RHD_TransEntropy2Lorentz_implicit::get_value() {
	return (Z_sqr-momentum_sqr)*sqr(lorentz) - Z_sqr;
}

REAL RHD_TransEntropy2Lorentz_implicit::get_slope() {
	REAL ZoverW = Z/lorentz;
	REAL dZdW = ZoverW - gamma_adiab_red*(ZoverW - density);
	return 2.*(lorentz*(Z_sqr - momentum_sqr) + (lorentz_sqr - 1.)*Z*dZdW);
}

void RHD_TransEntropy2Lorentz_implicit::set_parameters(REAL density_, REAL momentum_sqr_, REAL sigma_, REAL gamma_adiab_) {
	density = density_;
	momentum_sqr = momentum_sqr_;
	sigma = sigma_;
	gamma_tilde = gamma_adiab_/(gamma_adiab_-1.);
	gamma_adiab_red = gamma_adiab_-1.;
}


// ---------------------------------------------------------------------------------------------------------
void RHD_TransCons2Prim_polytropic_implicit::set_x(REAL rho_) {
	// Mind that rho can become negative during the iteration and the slope is not defined in that case
	rho = rho_ > 0. ? rho_ : 0.;
	PoverRho = polytropic_K * std::pow(rho, gamma_adiab - 1.);
	enthalpy = 1 + gamma_tilde * PoverRho;
	u_sqr = momentum_sqr / sqr(density * enthalpy);
	lorentz_sqr = 1 + u_sqr;
	lorentz = sqrt(lorentz_sqr);
}

REAL RHD_TransCons2Prim_polytropic_implicit::get_value() {
	return rho * lorentz - density;
}

REAL RHD_TransCons2Prim_polytropic_implicit::get_slope() {
	return lorentz - u_sqr/(enthalpy*lorentz) * gamma_adiab * PoverRho;
}

void RHD_TransCons2Prim_polytropic_implicit::set_parameters(REAL density_, REAL momentum_sqr_, REAL gamma_adiab_, REAL polytropic_K_) {
	density = density_;
	momentum_sqr = momentum_sqr_;
	polytropic_K = polytropic_K_;
	gamma_tilde = gamma_adiab_/(gamma_adiab_-1.);
	gamma_adiab = gamma_adiab_;
}

