#include "state.H"

bool primState::isNan() const {
  return ( 
    cronos_isnan(rho)
    || cronos_isnan(ux)
    || cronos_isnan(uy)
    || cronos_isnan(uz)
    || cronos_isnan(Etherm)
    #if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
      || cronos_isnan(specific_entropy)
    #endif
  );
}

bool primState::isOk() const {
  if (rho < 0) {
    return false;
  }
  else if (Etherm < 0) {
    return false;
  }
  #if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
  else if (specific_entropy < 0) {
    return false;
  }
  #endif
  else {
    return true;
  }
}

REAL primState::computeLorentz() const {
  return sqrt(1. + sqr(ux) + sqr(uy) + sqr(uz));
}

bool consState::isNan() const {
  return ( 
    cronos_isnan(density)
    || cronos_isnan(momentum_x)
    || cronos_isnan(momentum_y)
    || cronos_isnan(momentum_z)
    || cronos_isnan(tau)
    #if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
      || cronos_isnan(sigma)
    #endif
  );
}

bool consState::isOk() const {
  if (density < 0) {
    return false;
  }

  #if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
  else if (sigma < 0 && tau < 0) {
    return false;
  }
  #else
  else if (tau < 0) {
    return false;
  }
  #endif

  else {
    return true;
  }
}

#if(FLUID_TYPE == CRONOS_RHD)

consState Transformations::readCons(const Data & gdata, int ix, int iy, int iz) {
  consState cons;
  cons.density = gdata.om[q_rho](ix, iy, iz);
  cons.momentum_x = gdata.om[q_sx](ix, iy, iz);
  cons.momentum_y = gdata.om[q_sy](ix, iy, iz);
  cons.momentum_z = gdata.om[q_sz](ix, iy, iz);
  cons.tau = gdata.om[q_Eges](ix, iy, iz);
  #if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
  cons.sigma = gdata.om[q_Eadd](ix, iy, iz);
  #endif

  return cons;
}

void Transformations::writePrim(const Data & gdata, int ix, int iy, int iz, const primState & prim) {
  gdata.om[q_rho](ix, iy, iz) = prim.rho;
  gdata.om[q_sx](ix, iy, iz) = prim.ux;
  gdata.om[q_sy](ix, iy, iz) = prim.uy;
  gdata.om[q_sz](ix, iy, iz) = prim.uz;
  gdata.om[q_Eges](ix, iy, iz) = prim.Etherm;
  #if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
  gdata.om[q_Eadd](ix, iy, iz) = prim.specific_entropy;
  #endif
}

void Transformations::writeCons(const Data & gdata, int ix, int iy, int iz, const consState & cons) {
  gdata.om[q_rho](ix, iy, iz) = cons.density;
  gdata.om[q_sx](ix, iy, iz) = cons.momentum_x;
  gdata.om[q_sy](ix, iy, iz) = cons.momentum_y;
  gdata.om[q_sz](ix, iy, iz) = cons.momentum_z;
  gdata.om[q_Eges](ix, iy, iz) = cons.tau;
  #if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
  gdata.om[q_Eadd](ix, iy, iz) = cons.sigma;
  #endif
}

#endif

std::ostream & operator<<(std::ostream & os, const primState& prim)
{
  os
    << "rho = " << prim.rho
    << ", ux = " << prim.ux 
    << ", uy = " << prim.uy
    << ", uz = " << prim.uz
    << ", Eth = " << prim.Etherm
    #if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
      << ", s = " << prim.specific_entropy
    #endif
    ;
  return os;
}

std::ostream & operator<<(std::ostream & os, const consState& cons)
{
  os
    << "D = " << cons.density
    << ", mx = " << cons.momentum_x 
    << ", my = " << cons.momentum_y 
    << ", mz = " << cons.momentum_z 
    << ", tau = " << cons.tau
    #if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
      << ", sigma = " << cons.sigma 
    #endif
    ;
  return os;
}

consState computeCons(const primState & prim, const ProblemType & Problem) {
  consState cons;
  const REAL u_sqr = sqr(prim.ux) + sqr(prim.uy) + sqr(prim.uz);
  const REAL lorentz = sqrt(1. + u_sqr);
  const REAL pressure = (Problem.gamma - 1.) * prim.Etherm;
  const REAL enthalpy = 1. + Problem.gamma * prim.Etherm / prim.rho;
  const REAL D = prim.rho * lorentz;
  const REAL Dh = D * enthalpy;

  cons.density = D;
  cons.momentum_x = Dh * prim.ux;
  cons.momentum_y = Dh * prim.uy;
  cons.momentum_z = Dh * prim.uz;
  cons.tau = D * (enthalpy * lorentz - 1.) - pressure;

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
  // Computation of entropy for dual energy corrections
  cons.sigma = D * prim.specific_entropy;
#endif

  return cons;
}