#ifndef REAL
#define REAL double
#endif

#include "root_finder.H"
#include "state.H"

primState pressureFlush(const consState & cons, const ProblemType & Problem)
{
	static const REAL pressure = value_exists("smallPress") ? value("smallPress") : 1e-8;
	const REAL momentum_sqr = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);
	
	const REAL energy = cons.tau + cons.density;
	const REAL Dhgamma = energy + pressure;

	const REAL u_sqr =  momentum_sqr / (sqr(Dhgamma) - momentum_sqr);
	const REAL lorentz = sqrt(1. + u_sqr);

	// Overwrite the conservatives with the primitives
	REAL Dh_inv = lorentz / Dhgamma;

	primState prim;
	prim.rho = cons.density / lorentz;
	prim.Etherm = pressure / (Problem.gamma - 1);
	prim.ux = cons.momentum_x * Dh_inv;
	prim.uy = cons.momentum_y * Dh_inv;
	prim.uz = cons.momentum_z * Dh_inv;

	#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	prim.specific_entropy = (Problem.gamma-1) * prim.Etherm * pow(prim.rho, -Problem.gamma);
	#endif

	return prim;
}


consState normalStateCorrection_0D(consState cons, const ProblemType & Problem)
{
	const REAL momentum_sqr = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);
	const REAL momentum = sqrt(momentum_sqr);
	// Overwrite the conservatives with the primitives
	REAL tau_min = sqrt(sqr(cons.density) + momentum_sqr) - cons.density;
	tau_min *= 1. + 1e-3;
	cons.tau = tau_min;
	return cons;
}

consState normalStateCorrection_1D(consState cons, const ProblemType & Problem)
{
	const REAL momentum_sqr = sqr(cons.momentum_x) + sqr(cons.momentum_y) + sqr(cons.momentum_z);
	const REAL momentum = sqrt(momentum_sqr);

	const REAL density_tilde = cons.density / cons.tau;
	const REAL momentum_tilde = momentum /cons. tau;

	if (momentum_tilde < 1.) {
		exit(1.);
	}

	REAL coeff[4];
	coeff[3] = 1.;
	coeff[2] = 0.;
	coeff[1] = 1. - 2.*density_tilde;
	coeff[0] = -2.*momentum_tilde;

	PolynomialProblem<4> poly;
	poly.set_coefficients(coeff);

	NewtonMethod newtonSolver;
	newtonSolver.set_problem(poly);
	REAL momentum_tilde_star = newtonSolver.solve(momentum_tilde, 1e-8);
	REAL density_tilde_star = 0.5*(sqr(momentum_tilde_star) - 1.);

	const REAL eps = 1e-3;
	momentum_tilde_star *= 1. - eps;
	density_tilde_star += eps;

	cons.tau = cbrt(cons.tau * momentum * cons.density / (momentum_tilde_star * density_tilde_star));
	cons.density = cons.tau * density_tilde_star;

	const REAL momentum_fac = cons.tau * momentum_tilde_star / momentum;
	cons.momentum_x *= momentum_fac;
	cons.momentum_y *= momentum_fac;
	cons.momentum_z *= momentum_fac;

	return cons;
}

#if(FLUID_TYPE == CRONOS_RHD)

consState Transformations::CorrectUnphysicalState_RHD(const consState & cons, const ProblemType & Problem)
{
#if(RHD_STATE_CORRECTION == NORMAL_CORRECTION_1D)
	return normalStateCorrection_1D(cons, Problem);
#elif(RHD_STATE_CORRECTION == NORMAL_CORRECTION_0D)
	return normalStateCorrection_0D(cons, Problem);
#elif(RHD_STATE_CORRECTION == FLUSH)
	primState prim = pressureFlush(cons, Problem);
	return computeCons(prim, Problem);
#endif
}

#endif