#include "SemilagrangeSolver.H"
#include "Functions.H"
#include <iostream>
#include <cmath>

using namespace std;

#ifndef REAL
#define REAL double
#endif


template <typename T>
class LinearFunction: public Function<T> {
public:
	LinearFunction(T a): a(a) {};

protected:
	T a;
	T _get_value() {return a * this->x;};
};

template <typename T>
class ConstantFunction: public Function<T> {
public:
	ConstantFunction(T c): c(c) {};

protected:
	T c;
	T _get_value() {return c;};
};



int PowerlawAdvectionTest() {
	cerr << " -------------------------------------- " << endl;
	cerr << "   Powerlaw Advection Test " << endl;
	LinearFunction<REAL> fun(-1);
	SemilagrangeSolver solver;
	solver.set_advectionFunction(&fun);


	int nBins = 1000;
	REAL endFac = 1024;
	float t = 1.;
	int nSteps = 100;


	REAL logSpace = std::pow(endFac, 1./nBins);

	int lo = 0;
	int hi = nBins - 1;
	int delta = nBins;

	int loCen[1] = {lo - delta};
	int hiCen[1] = {hi + delta};

	NumMatrix<REAL,1> centers(loCen, hiCen);
	NumMatrix<REAL,1> values(loCen, hiCen);

	int loEdge[1] = {lo - delta};
	int hiEdge[1] = {hi + 1 + delta};
	NumMatrix<REAL,1> edges(loEdge, hiEdge);

	solver.resize(loCen[0], hiCen[0]);

	// define gridding
	edges(0) = 10;
	for (int i = lo; i <= hi; ++i) {
		edges(i+1) = edges(i) * logSpace;
		centers(i) = 0.5 * (edges(i) + edges(i+1));
	}

	// Boundary cells
	for (int i = lo - 1; i >= loCen[0] ; --i) {
		edges(i) = edges(i+1) / logSpace;
		centers(i) = 0.5 * (edges(i) + edges(i+1));
	}

	for (int i = hi + 1; i <= hiCen[0] ; ++i) {
		edges(i+1) = edges(i) * logSpace;
		centers(i) = 0.5 * (edges(i) + edges(i+1));
	}

	// set some values
	for (int i = lo; i <= hi; ++i) {
		values(i) = pow(centers(i), -2.);
	}

	for (int i = lo; i <= hi; ++i) {
		cout << centers(i) << " ";
	}
	cout << endl;

	for (int i = lo; i <= hi; ++i) {
		cout << values(i) << " ";
	}
	cout << endl;


	REAL integral = 0;
	for (int i = loCen[0]; i <= hiCen[0]; ++i) {
		integral += values(i) * (edges(i+1) - edges(i));
	}
	cerr << " integral: " << integral << endl;

	cerr << " ------------------- " << endl;


	float dt = t/nSteps;
	for (int i = 0; i < nSteps; ++i) {
		solver.singlestep(centers, edges, dt, values);
	}

	for (int i = lo; i <= hi; ++i) {
		cout << values(i) << " ";
	}
	cout << endl;

	integral = 0;
	for (int i = loCen[0]; i <= hiCen[0]; ++i) {
		integral += values(i) * (edges(i+1) - edges(i));
	}
	cerr << " integral: " << integral << endl;
}

int LinearAdvectionBlockTest() {
	cerr << " -------------------------------------- " << endl;
	cerr << "   Linear Advection Block Test " << endl;
	ConstantFunction<REAL> fun(1.);
	SemilagrangeSolver solver;
	solver.set_advectionFunction(&fun);

	int nBins = 1000;
	REAL size = 10;
	REAL blockSize = 5;

	REAL linSpace = size/nBins;

	int lo = 0;
	int hi = nBins - 1;
	int delta = 5;

	int loCen[1] = {lo - delta};
	int hiCen[1] = {hi + delta};

	NumMatrix<REAL,1> centers(loCen, hiCen);
	NumMatrix<REAL,1> values(loCen, hiCen);

	int loEdge[1] = {lo - delta};
	int hiEdge[1] = {hi + 1 + delta};
	NumMatrix<REAL,1> edges(loEdge, hiEdge);

	solver.resize(loCen[0], hiCen[0]);

	// define gridding
	edges(0) = 0.5;
	for (int i = lo; i <= hi; ++i) {
		edges(i+1) = edges(i) + linSpace;
		centers(i) = 0.5 * (edges(i) + edges(i+1));
	}

	// Boundary cells
	for (int i = lo - 1; i >= loCen[0] ; --i) {
		edges(i) = edges(i+1) - linSpace;
		centers(i) = 0.5 * (edges(i) + edges(i+1));
	}

	for (int i = hi + 1; i <= hiCen[0] ; ++i) {
		edges(i+1) = edges(i) + linSpace;
		centers(i) = 0.5 * (edges(i) + edges(i+1));
	}


	// set some values
	values.clear();
	for (int i = lo; i <= hi; ++i) {
		if (centers(i) > 0.5 * (size - blockSize) && centers(i) < 0.5 * (size + blockSize)) {
			values(i) = 1.;
		}
	}

	for (int i = lo; i <= hi; ++i) {
		cout << centers(i) << " ";
	}
	cout << endl;
	for (int i = lo; i <= hi; ++i) {
		cout << values(i) << " ";
	}
	cout << endl;


	REAL integral = 0;
	for (int i = loCen[0]; i <= hiCen[0]; ++i) {
		integral += values(i) * (edges(i+1) - edges(i));
	}
	cerr << " integral: " << integral << endl;

	cerr << " ------------------- " << endl;

	float t = 1.;
	int nSteps = 1000;
	float dt = t/nSteps;

	for (int i = 0; i < nSteps; ++i) {
		solver.singlestep(centers, edges, dt, values);
	}

	for (int i = lo; i <= hi; ++i) {
		cout << values(i) << " ";
	}
	cout << endl;

	integral = 0;
	for (int i = loCen[0]; i <= hiCen[0]; ++i) {
		integral += values(i) * (edges(i+1) - edges(i));
	}
	cerr << " integral: " << integral << endl;
}


int LinearAdvectionTest() {
	cerr << " -------------------------------------- " << endl;
	cerr << "   Linear Advection Test " << endl;
	ConstantFunction<REAL> fun(-1.);
	SemilagrangeSolver solver;
	solver.set_advectionFunction(&fun);

	int nBins = 1000;
	REAL size = 10;

	REAL linSpace = size/nBins;

	int lo = 0;
	int hi = nBins - 1;
	int delta = 5;

	int loCen[1] = {lo - delta};
	int hiCen[1] = {hi + delta};

	NumMatrix<REAL,1> centers(loCen, hiCen);
	NumMatrix<REAL,1> values(loCen, hiCen);

	int loEdge[1] = {lo - delta};
	int hiEdge[1] = {hi + 1 + delta};
	NumMatrix<REAL,1> edges(loEdge, hiEdge);

	solver.resize(loCen[0], hiCen[0]);

	// define gridding
	edges(0) = 0.5;
	for (int i = lo; i <= hi; ++i) {
		edges(i+1) = edges(i) + linSpace;
		centers(i) = 0.5 * (edges(i) + edges(i+1));
	}

	// Boundary cells
	for (int i = lo - 1; i >= loCen[0] ; --i) {
		edges(i) = edges(i+1) - linSpace;
		centers(i) = 0.5 * (edges(i) + edges(i+1));
	}

	for (int i = hi + 1; i <= hiCen[0] ; ++i) {
		edges(i+1) = edges(i) + linSpace;
		centers(i) = 0.5 * (edges(i) + edges(i+1));
	}


	// set some values
	values.clear();
	for (int i = lo; i <= hi; ++i) {
			values(i) = centers(i);
	}

	for (int i = lo; i <= hi; ++i) {
		cout << centers(i) << " ";
	}
	cout << endl;
	for (int i = lo; i <= hi; ++i) {
		cout << values(i) << " ";
	}
	cout << endl;


	REAL integral = 0;
	for (int i = loCen[0]; i <= hiCen[0]; ++i) {
		integral += values(i) * (edges(i+1) - edges(i));
	}
	cerr << " integral: " << integral << endl;

	cerr << " ------------------- " << endl;

	float t = 1.;
	int nSteps = 1000;
	float dt = t/nSteps;

	for (int i = 0; i < nSteps; ++i) {
		solver.singlestep(centers, edges, dt, values);
	}

	for (int i = lo; i <= hi; ++i) {
		cout << values(i) << " ";
	}
	cout << endl;

	integral = 0;
	for (int i = loCen[0]; i <= hiCen[0]; ++i) {
		integral += values(i) * (edges(i+1) - edges(i));
	}
	cerr << " integral: " << integral << endl;
}

int main() {

//	PowerlawAdvectionTest();
//	LinearAdvectionBlockTest();
	LinearAdvectionTest();

}
