
#ifndef CRONOS_RECONST_H
#define CRONOS_RECONST_H = 1

#include "matrix.H"
#include "data.H"
#include "limiter.H"
#include "fields_local.H"
#include <vector>

class SingleReconstruction{
public:
	SingleReconstruction(){}
	SingleReconstruction(const Data &, int dir, int substep);
	SingleReconstruction(const Data &, const CronosFluid &fluid, int dir, int qReconst, int substep);
	virtual ~SingleReconstruction();

	/**
	 * Compute reconstruction of cell-centered variable
	 * */
	virtual void computeNormal(const Data &, NumMatrix<REAL,1> &, NumMatrix<REAL,1> &,
			NumMatrix<REAL,1> &, NumMatrix<REAL,1> &) = 0;

	/**
	 * Compute reconstruction of face-centered variable in direction of respective cell face
	 * */
	virtual void computePar(const Data &, NumMatrix<REAL,1> &, NumMatrix<REAL,1> &,
	                NumMatrix<REAL,1> &) = 0;

	/**
	 * Compute reconstruction of face-centered variable in perpendicular direction
	 * */
	virtual void computePerp(const Data &, NumMatrix<REAL,1> &, NumMatrix<REAL,1> &,
			NumMatrix<REAL,1> &, NumMatrix<REAL,1> &, NumMatrix<REAL,1> &) = 0;

	virtual void perpareDerivs(const Data& gdata, int ix, int iy, int iz) {};

	virtual void get_Vals_EW(const Data &, phys_fields_0D &xFieldsW,
			phys_fields_0D &xFieldsE, int, int, int) = 0;
	virtual void get_Vals_SN(const Data &, phys_fields_0D &xFieldsS,
			phys_fields_0D &xFieldsN, int, int, int) = 0;
	virtual void get_Vals_BT(const Data &, phys_fields_0D &xFieldsB,
			phys_fields_0D &xFieldsT, int, int, int) = 0;
protected:
	/**
	 * Compute left-handed, centred, and right-handed derivatives in given direction
	 * */
	void getDeriv(const Data &, const NumMatrix<REAL,1> &, int);

	/**
	 * Compute left-handed, centred, and right-handed derivatives in all spatial directions
	 * --> block-structured version
	 * */
	void getDerivs(const Data &, int ix, int iy, int iz);
	int dir;
	int qReconst;
	int substep;

	NumMatrix<REAL, 1> derivPerp, derivM;
	REAL deriv_x, deriv_y, deriv_z;
	REAL dudxp_q, dudx0_q, dudxm_q;
	REAL dudyp_q, dudy0_q, dudym_q;
	REAL dudzp_q, dudz0_q, dudzm_q;

	REAL dudxp, dudx0, dudxm;
};



class Reconstruction{
public:
	Reconstruction(){}
	Reconstruction(const Data &, int dir, const CronosFluid &fluid, int substep=-1);
	Reconstruction(const Data &, int dir, int num, int substep=-1);
	virtual ~Reconstruction();

	/**
	 * Compute reconstruction in case of 1D arrays
	 * */
	void compute(const Data &, fields_1D &, phys_fields_1D &, phys_fields_1D &);

	/**
	 * Compute reconstruction for block-structured code
	 * */
	void compute(const Data &, std::vector<phys_fields_0D*> &allFields, int, int, int);

	/**
	 * Compute reconstruction in case of 1D arrays
	 * alternative function parameters
	 * */
	void compute(const Data &, NumMatrix<REAL,1> [], NumMatrix<REAL,1> [],
	             NumMatrix<REAL,1> [], NumMatrix<REAL,1> []);

protected:
	void set_singleReconstructions(const Data &);
	void set_singleReconstructions(const Data &, const CronosFluid &fluid);

	int dir;
	int substep;
	std::vector<int>::iterator iter;
	std::vector<int> ListNormal;
	std::vector<int> ListParallel;
	std::vector<int> ListPerp;

	std::vector<SingleReconstruction*> ListReconstructionNormal;
	std::vector<SingleReconstruction*> ListReconstructionPar;
	std::vector<SingleReconstruction*> ListReconstructionPerp;

};




class Reconstruction2D{
public:
	Reconstruction2D(const Data &, const int & dir);
	void compute(const Data &, fields_2D &,
	             phys_fields_2D &, phys_fields_2D &,
	             phys_fields_2D &, phys_fields_2D &);
	virtual void computeNormal(const Data &,
	                   NumMatrix<REAL,2> &, NumMatrix<REAL,2> &,
	                   NumMatrix<REAL,2> &, NumMatrix<REAL,2> &,
	                   NumMatrix<REAL,2> &, NumMatrix<REAL,2> &,
	                   NumMatrix<REAL,2> &) = 0;
	virtual void computePerp(const Data &, NumMatrix<REAL,2> &, NumMatrix<REAL,2> &,
	                 const int &,
	                 NumMatrix<REAL,2> &, NumMatrix<REAL,2> &,
	                 NumMatrix<REAL,2> &, NumMatrix<REAL,2> &) = 0;
protected:
	void getDeriv(const Data &, const NumMatrix<REAL,2> &,
	              int dir, int i, int j);
	REAL dudxp, dudx0, dudxm;
	REAL dudyp, dudy0, dudym;
	int dir0, dir1;
};
#endif
