#include "gridgen.H"
#include <iomanip>
#include <cmath>
#include <vector>
#include "DissipationMHD.H"

using namespace std;

REAL HyperbolicSolver::singlestep(Data &gdata, gridFunc &gfunc,
                                  ProblemType &Problem, int n)
{
  //! Block structured version of singlestep
  /*! 
   * This version of singlestep solves all directions simultaneously
   * with future applications like AMR in mind
   */

  if(eos == (EquationOfState*)0) {
	eos = new EquationOfState(Problem);
  }
  if(sources == NULL) {
	sources = new SourceTerms(gdata, Problem, IntegrateA, thermal);
  }

// ----------------------------------------------------------------
// Checking for negative Temperatures
// ----------------------------------------------------------------

	if(gdata.rank == 0) {
		if(ENERGETICS == FULL && TempNeg !=0) {
			cout << " Negative Temperatures: " << TempNeg << endl;
			TempNeg = 0;
		}
	}

// ----------------------------------------------------------------
// Start the clock:
// ----------------------------------------------------------------

	gettimeofday(&tick, 0);
	gettimeofday(&tin1, 0);
	cstart = clock();  

// ----------------------------------------------------------------
// Relevant declarations
// ----------------------------------------------------------------

//	nom = new NumMatrix<REAL,3> [n_omInt];
//	for (int q = 0; q < n_omInt; ++q) {
//#if (FLUID_TYPE == CRONOS_MHD)
//		if((gdata.om[q].getName() == "B_x" ||
//		    gdata.om[q].getName() == "B_y" ||
//		    gdata.om[q].getName() == "B_z")) {
//			nom[q].resize(Index::set(0,0,0),
//			              Index::set(0,0,0));
//		} else {
//#endif
//			nom[q].resize(Index::set(0,0,0),
//			              Index::set(gdata.mx[0],gdata.mx[1],gdata.mx[2]));
//#if (FLUID_TYPE == CRONOS_MHD)
//		}
//}
//#endif
	for (int q = 0; q < n_omInt; ++q) {
		gdata.nom[q].clear();
	}

	// Prepare user fields if necessary
	#if (OMS_USER == TRUE)

//	nom_user = new NumMatrix<REAL,3> [n_omIntUser];
//	for (int q = 0; q < n_omIntUser; ++q) {
//		nom_user[q].resize(Index::set(0,0,0),
//		                   Index::set(gdata.mx[0],gdata.mx[1],gdata.mx[2]));
//		nom_user[q].clear();
//	}
	for (int q = 0; q < n_omIntUser; ++q) {
		gdata.nom_user[q].clear();
	}

#endif

// ---------------------------------------------------------------	      
//      Trafo of variables to conservative form
//----------------------------------------------------------------

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	if(n_omInt > 7) {
		Trafo->computeEntropyFromE(gdata, gfunc, Problem);
	}
#endif

#if (USE_COROTATION == CRONOS_ON)
	// Transform to inertial frame velocity
	Trafo->TransCorotToInert(gdata, gfunc, Problem);
#endif

	Trafo->TransPrim2Cons(gdata, gfunc, Problem);

	// User defined transform
	Problem.TransPrim2Cons(gdata);

// ---------------------------------------------------------------	      
//      Saving old variables for Runge-Kutta (conservative form)
//----------------------------------------------------------------

	if (n == 0) {
		for (int q = 0; q < n_omInt; ++q){
#if(FLUID_TYPE == CRONOS_HYDRO || FLUID_TYPE == CRONOS_RHD)
			gdata.om[n_Omega+q] = gdata.om[q];         // save om_n
#elif (FLUID_TYPE == CRONOS_MHD)
			if((q < q_Bx || q > q_Bz) || !IntegrateA) {
				gdata.om[n_Omega+q] = gdata.om[q];         // save om_n
			} else {
				gdata.om[n_Omega+q] = gdata.om[n_omInt+N_ADD+q-q_Bx];
			}
#endif
		}

#if (OMS_USER == TRUE)
		// Saving user-variables (if necessary)
		for (int q = 0; q < n_omIntUser; ++q) {
			gdata.om_user[n_OmegaUser+q] = gdata.om_user[q];
		}
#endif

	}

// ---------------------------------------------------------------	      
//      Trafo primitive variables for reconstruction
//----------------------------------------------------------------

	Trafo->TransCons2Prim(gdata, gfunc, Problem);

	// User defined transform
	Problem.TransCons2Prim(gdata);

	// Check for nans in primitive variables
	gdata.CheckNan(1);

	// Compute the carbuncle-flag if necessary
	if(gdata.use_carbuncleFlag) {
		RiemannX->compute_carbuncleFlag(gdata);
	}

	REAL cfl_lin(0.), cfl_eta(0.);

	// range of ghost cells
	int n_ghost[3];
	for(int dir=0; dir<3; ++dir) {
		if(gdata.mx[dir]<2) {
			n_ghost[dir] = 0;
		} else {
			n_ghost[dir] = 3;
		}
	}

//	phys_fields_0D physVals[9];


	std::vector<phys_fields_0D*> physVals;
	for(int inum=0; inum<9; ++inum) {
		physVals.push_back(new phys_fields_0D(gdata, inum, gdata.fluid));
	}

	std::vector<num_fields_0D*> numVals;
	for(int idir=0; idir<3; ++idir) {
		numVals.push_back(new num_fields_0D(gdata, gdata.fluid));
	}

	phys_fields_0D physVals_oldE(gdata, 1, gdata.fluid);
	phys_fields_0D physVals_oldN(gdata, 3, gdata.fluid);
	phys_fields_0D physVals_oldT(gdata, 5, gdata.fluid);
	num_fields_0D numVals_oldE(gdata, gdata.fluid), numVals_oldN(gdata, gdata.fluid), numVals_oldT(gdata, gdata.fluid);

	Reconstruction reconst(gdata, gdata.fluid);

	cronos::vector<REAL> pos(0,0,0);
	cronos::vector<int> ipos(0,0,0);

	int n_omInt = gdata.fluid.get_N_OMINT();
	NumArray<REAL> uPriOld_E(n_omInt);
	NumArray<REAL> flux_numOld_E(n_omInt);
	double ptotalOld_E;

	NumMatrix<REAL,2> uPriOld_N(Index::set(0,-n_ghost[0]),
			Index::set(n_omInt,gdata.mx[0]+n_ghost[0]));
	NumMatrix<REAL,2> uPriCurr_N(Index::set(0,-n_ghost[0]),
				Index::set(n_omInt,gdata.mx[0]+n_ghost[0]));
	NumMatrix<REAL,2> flux_numOld_N(Index::set(0,-n_ghost[0]),
			Index::set(n_omInt,gdata.mx[0]+n_ghost[0]));
	NumMatrix<REAL,2> flux_numCurr_N(Index::set(0,-n_ghost[0]),
				Index::set(n_omInt,gdata.mx[0]+n_ghost[0]));
	NumMatrix<REAL,1>  ptotalOld_N(Index::set(-n_ghost[0]),
			Index::set(gdata.mx[0]+n_ghost[0]));
	NumMatrix<REAL,1>  ptotalCurr_N(Index::set(-n_ghost[0]),
			Index::set(gdata.mx[0]+n_ghost[0]));

	NumMatrix<REAL,3> uPriOld_T(Index::set(0,-n_ghost[0],-n_ghost[1]),
			Index::set(n_omInt,gdata.mx[0]+n_ghost[0],gdata.mx[1]+n_ghost[1]));
	NumMatrix<REAL,3> uPriCurr_T(Index::set(0,-n_ghost[0],-n_ghost[1]),
			Index::set(n_omInt,gdata.mx[0]+n_ghost[0],gdata.mx[1]+n_ghost[1]));
	NumMatrix<REAL,3> flux_numOld_T(Index::set(0,-n_ghost[0],-n_ghost[1]),
			Index::set(n_omInt,gdata.mx[0]+n_ghost[0],gdata.mx[1]+n_ghost[1]));
	NumMatrix<REAL,3> flux_numCurr_T(Index::set(0,-n_ghost[0],-n_ghost[1]),
			Index::set(n_omInt,gdata.mx[0]+n_ghost[0],gdata.mx[1]+n_ghost[1]));
	NumMatrix<REAL,2> ptotalOld_T(Index::set(-n_ghost[0],-n_ghost[1]),
			Index::set(gdata.mx[0]+n_ghost[0],gdata.mx[1]+n_ghost[1]));
	NumMatrix<REAL,2> ptotalCurr_T(Index::set(-n_ghost[0],-n_ghost[1]),
			Index::set(gdata.mx[0]+n_ghost[0],gdata.mx[1]+n_ghost[1]));

	// DEBUG: Storage-Kram auslagern

	// Loop over grid-block
	for (int iz = -n_ghost[2]+1; iz <= gdata.mx[2]+n_ghost[2]-1; ++iz){
		pos.set(2, gdata.getCen_z(iz));
		ipos.set(2, iz);
		// Store 'previous current' data in old array
		uPriOld_T = uPriCurr_T;
		flux_numOld_T = flux_numCurr_T;
		ptotalOld_T = ptotalCurr_T;

//		if(iz==62 || iz==61) {
//			cout << " Restored stuff " << flux_numCurr_T(4,62,0) << endl;
//			cout << " " << flux_numCurr_T(4,62,1) << endl;
//		}


		for (int iy = -n_ghost[1]+1; iy <= gdata.mx[1]+n_ghost[1]-1; ++iy){
			pos.set(1, gdata.getCen_z(iy));
			ipos.set(1, iy);
			// Store 'previous current' data in old array
			uPriOld_N = uPriCurr_N;
			flux_numOld_N = flux_numCurr_N;
			ptotalOld_N = ptotalCurr_N;
			for (int ix = -n_ghost[0]+1; ix <= gdata.mx[0]+n_ghost[0]-1; ++ix){
				pos.set(0, gdata.getCen_z(ix));
				ipos.set(0, ix);

				// Reconstruction at given position
				reconst.compute(gdata, physVals, ix, iy, iz);

//				if(ix==62 && iy==0 && (iz==62 || iz==61 || iz==63 || iz==64 || iz==102)) {
//					cout << " Reconst: ";
//					cout << physVals[4]->uPri(4) << " ";
//					cout << physVals[5]->uPri(4) << " ";
//					cout << gdata.om[4](ix,iy,iz) << " ";
//					cout << endl;
//				}

				// Load old data
				physVals_oldE.uPri = uPriOld_E;
				for(int q_ind=0; q_ind<n_omInt; ++q_ind) {
					physVals_oldN.uPri(q_ind) = uPriOld_N(q_ind, ix);
					physVals_oldT.uPri(q_ind) = uPriOld_T(q_ind, ix, iy);
				}

				// Load old flux values:
				numVals_oldE.flux_num = flux_numOld_E;
				numVals_oldE.ptotal_num = ptotalOld_E;
				for(int q_ind=0; q_ind<n_omInt; ++q_ind) {
					numVals_oldN.flux_num(q_ind) = flux_numOld_N(q_ind,ix);
					numVals_oldT.flux_num(q_ind) = flux_numOld_T(q_ind,ix,iy);
				}
				numVals_oldN.ptotal_num = ptotalOld_N(ix);
				numVals_oldT.ptotal_num = ptotalOld_T(ix,iy);



				// Compute conservative variables and physical fluxes
				// x-faces
				// i,j,k - West
				Trafo->get_Cons(gdata, Problem, *eos, *physVals[0], ix, iy, iz, 0);
				PhysFlux->get_PhysFlux(gdata, Problem, *physVals[0], ix, iy, iz, 0);
				// i-1,j,k - East
				Trafo->get_Cons(gdata, Problem, *eos, physVals_oldE, ix-1, iy, iz, 1);
				PhysFlux->get_PhysFlux(gdata, Problem, physVals_oldE, ix-1, iy, iz, 1);

				// y-faces
				// i,j,k - South
				Trafo->get_Cons(gdata, Problem, *eos, *physVals[2], ix, iy, iz, 2);
				PhysFlux->get_PhysFlux(gdata, Problem, *physVals[2], ix, iy, iz, 2);
				// i,j-1,k - North
				Trafo->get_Cons(gdata, Problem, *eos, physVals_oldN, ix, iy-1, iz, 3);
				PhysFlux->get_PhysFlux(gdata, Problem, physVals_oldN, ix, iy-1, iz, 3);

				// z-faces
				// i,j,k - bottom
				Trafo->get_Cons(gdata, Problem, *eos, *physVals[4], ix, iy, iz, 4);
				PhysFlux->get_PhysFlux(gdata, Problem, *physVals[4], ix, iy, iz, 4);
				// i,j,k-1 - top
				Trafo->get_Cons(gdata, Problem, *eos, physVals_oldT, ix, iy, iz-1, 5);
				PhysFlux->get_PhysFlux(gdata, Problem, physVals_oldT, ix, iy, iz-1, 5);


				// Possible flux modification by user
				// --> later maybe

				if(ix>=-1 && iy>=-1 && iz>=-1){
				// Compute characteristic velocities and numerical fluxes
				// at i-1/2,j,k
				RiemannX->get_vChar(gdata, Problem, *physVals[0], physVals_oldE,
						*numVals[0], ix, iy, iz, 0, cfl_lin);
				RiemannX->get_NumFlux(gdata, physVals_oldE, *physVals[0], *numVals[0], 0);

				// Char at i,j-1/2,k
				RiemannY->get_vChar(gdata, Problem, *physVals[2], physVals_oldN,
						*numVals[1], ix, iy, iz, 1, cfl_lin);
				RiemannY->get_NumFlux(gdata, physVals_oldN, *physVals[2], *numVals[1], 1);

				// vChar at i,j,k-1/2
				RiemannZ->get_vChar(gdata, Problem, *physVals[4], physVals_oldT,
						*numVals[2], ix, iy, iz, 2, cfl_lin);
				RiemannZ->get_NumFlux(gdata, physVals_oldT, *physVals[4], *numVals[2], 2);


				if(ix==31 && iy==31 && iz==29) {
					cout << " flux " <<  numVals[2]->flux_num(4) << " ";
					cout << physVals[4]->flux_phys(4) << " ";
					cout << physVals_oldT.flux_phys(4) << " ";
					cout << physVals[4]->uCon(4) << " ";
					cout << physVals_oldT.uCon(4) << " ";
					cout << endl;
				}

//				if(ix==62 && iy==0 && (iz==62 || iz==61 || iz==63 || iz==64 || iz==102)) {
//					cout << " num flux " << numVals[2]->flux_num(4) << " ";
//					cout << numVals[2]->v_ch_m << " ";
//					cout << numVals[2]->v_ch_p << " ";
//					cout << physVals[4]->flux_phys(4) << " ";
//					cout << physVals_oldT.flux_phys(4) << " ";
//					cout << physVals[4]->uCon(4) << " ";
//					cout << physVals_oldT.uCon(4) << " ";
//					cout << endl;
//				}


				// Save fluxes


				// Apply changes of normal variables
				// DEBUG: need to reload the data
				// x-direction - applied at ix-1,iy,iz
				if(ix>=1 && ix<=gdata.mx[0]+1 && iy>=0 && iy<=gdata.mx[1] && iz>=0 && iz<=gdata.mx[2]) {
//					if(ix==61 && iy==0 && iz==61) {
//						cout << " Bef x " << gdata.nom[4](62,0,61) << endl;
//					}
					get_Changes(gdata, numVals_oldE, *numVals[0], gdata.nom,
							ix-1, iy, iz, 0, RiemannX->get_Fluid_Type());
//					if(ix==61 && iy==0 && iz==61) {
//						cout << " Aft x " << gdata.nom[4](62,0,61) << endl;
//					}
				}

				if(ix>=0 && ix<=gdata.mx[0] && iy>=1 && iy<=gdata.mx[1]+1 && iz>=0 && iz<=gdata.mx[2]) {
//					if(ix==62 && iy==1 && iz==61) {
//						cout << " Bef y " << gdata.nom[4](62,0,61) << endl;
//					}
					get_Changes(gdata, numVals_oldN, *numVals[1], gdata.nom,
							ix, iy-1, iz, 1, RiemannX->get_Fluid_Type());
//					if(ix==62 && iy==1 && iz==61) {
//						cout << " Aft y " << gdata.nom[4](62,0,61) << endl;
//					}
				}
				if(ix>=0 && ix<=gdata.mx[0] && iy>=0 && iy<=gdata.mx[1] && iz>=1 && iz<=gdata.mx[2]+1) {
					get_Changes(gdata, numVals_oldT, *numVals[2], gdata.nom,
							ix, iy, iz-1, 2, RiemannX->get_Fluid_Type());
					if(ix==31 && iy==31 && iz==29) {
						cout << " z-change " <<  numVals[2]->flux_num(4) << " " << numVals_oldT.flux_num(4)<< endl;
					}
				}
//				get_Changes(gdata, numVals[0].flux_num, numVals[1].flux_num,
//						numVals[2].flux_num, numVals[3].flux_num,
//						numVals[4].flux_num, numVals[5].flux_num, gdata.nom, numVals[0].ptotal_num,
//						ix, iy, iz, RiemannX->get_Fluid_Type());

				}

				// Apply changes for magnetic field
				// DEBUG: to be done (much?) later

				// Save old data in x-direction
				// Reconstruction data
				uPriOld_E = physVals[1]->uPri;

				// Save old data in y-and z-direction
				// Reconstruction data:
				for(int q_ind=0; q_ind<n_omInt; ++q_ind) {
					uPriCurr_N(q_ind, ix) = physVals[3]->uPri(q_ind);
					uPriCurr_T(q_ind, ix, iy) = physVals[5]->uPri(q_ind);
				}


				// Save old fluxes and ptotal
				flux_numOld_E = numVals[0]->flux_num;
				ptotalOld_E = numVals[0]->ptotal_num;
//				flux_numOld_E = numVals_oldE.flux_num;
//				ptotalOld_E = numVals_oldE.ptotal_num;

				for(int q_ind=0; q_ind<n_omInt; ++q_ind) {
					flux_numCurr_N(q_ind,ix) = numVals[1]->flux_num(q_ind);
					ptotalCurr_N(ix) = numVals[1]->ptotal_num;
					flux_numCurr_T(q_ind,ix,iy) = numVals[2]->flux_num(q_ind);
					ptotalCurr_T(ix,iy) = numVals[2]->ptotal_num;
				}
//
//				if(ix==62 && iy==0 && (iz==62 || iz==61)) {
//					cout << " Stored stuff " << flux_numCurr_T(4,ix,iy) << endl;
//				}

//				if(gdata.nom[4](62,0,61) > 1.) {
//					cout << " error at " << ix << " " << iy << " " << iz << endl;
//					cout << gdata.nom[4](62,0,61) << endl;
//					exit(3);
//				}

			}
		}
	}

//	cout << " gdata.nom(0) ";
//	for(int ix=0; ix<128; ++ix) {
//		cout << ix << " " << gdata.nom[0](ix,64,64) << endl;
//	}
//
//	cout << " err " << gdata.nom[4](63,63,60) << endl;

	gettimeofday(&tin5, 0);

// ----------------------------------------------------------------
//   Check for errors:
// ----------------------------------------------------------------

	for(int q = 0; q<n_omInt; ++q) {
		CheckNan(gdata.nom[q],q, 0, 1,"nom");
	}

// ----------------------------------------------------------------
//   Compute Courant number
// ----------------------------------------------------------------

	REAL cfl = compute_cfl(gdata, Problem, cfl_eta, cfl_lin, n);
	
// ----------------------------------------------------------------
//   Geometrical source terms:
// ----------------------------------------------------------------

#if (GEOM != CARTESIAN)
	sources->src_Geom(gdata, Problem, gdata.nom);
#endif

// ----------------------------------------------------------------
//   Geometrical source terms:
// ----------------------------------------------------------------

#if (USE_COROTATION == CRONOS_ON)
	Trafo->src_Corotating(gdata, Problem, gdata.nom);
#endif

// ----------------------------------------------------------------
//   User defined source terms:
// ----------------------------------------------------------------

#if (USE_COROTATION == CRONOS_ON)
	// Transform to co-rotating frame velocity for user src
	Trafo->TransInertToCorot(gdata, gfunc, Problem);
#endif

	Problem.src_User(gdata, gdata.nom, gdata.nom_user);

#if (USE_COROTATION == CRONOS_ON)
	// Transform back to inertial frame
	Trafo->TransCorotToInert(gdata, gfunc, Problem);
#endif

// ----------------------------------------------------------------
//   Check for errors again before applying the changes:
// ----------------------------------------------------------------

	for(int q = 0; q<n_omInt; ++q) {
		CheckNan(gdata.nom[q],q, 0, 2,"nom");
	}

// ----------------------------------------------------------------
//   Transform to conservative variables:
// ----------------------------------------------------------------

	Trafo->TransPrim2Cons(gdata, gfunc, Problem);
	
	Problem.TransPrim2Cons(gdata);

// ----------------------------------------------------------------
//   Determine domain to be integrated and apply changes:
// ----------------------------------------------------------------

	for (int q = 0; q < n_omInt; ++q) {
		
		TimeIntegratorGeneric[q]->Substep(gdata, Problem, gdata.nom[q], gdata.om, n);

	}

#if (OMS_USER == TRUE)

	for (int q=0; q<n_omIntUser; ++q) {

		TimeIntegratorUser[q]->Substep(gdata, Problem, gdata.nom_user[q],
		                              gdata.om_user, n);
		
	}

#endif

//	delete [] nom;
//#if (OMS_USER == TRUE)
//	delete [] nom_user;
//#endif

	gettimeofday(&tock, 0);
	cstep = clock();

// ----------------------------------------------------------------
//   Determine time at intermediate steps:
// ----------------------------------------------------------------

#if (RK_STEPS == 3) 
	// Time must come before boundary for time dependent boundaries
	if (n == 0) gdata.time += gdata.dt;      // First trial step with t* = t+dt
	if (n == 1) gdata.time -= 0.5*gdata.dt;  // Second trial step with t* = t+0.5dt
	if (n == 2) gdata.time += 0.5*gdata.dt;      // Last Runge-Kutta step - next time step
#else
	if (n == 0) gdata.time += gdata.dt;
#endif

// ----------------------------------------------------------------
//   Transformation to primitive variables and apply bcs
// ----------------------------------------------------------------

	gettimeofday(&tock2, 0);
#if (USE_ANGULAR_MOMENTUM == TRUE)
	Trafo->TransAngMom2Vel(gdata, gfunc, Problem);
#else
	Trafo->TransMomen2Vel(gdata, gfunc, Problem);
#endif

#if (FLUID_TYPE == CRONOS_MHD)
	if(IntegrateA) {
		compute_B(gdata, gfunc, Problem);
	}
#endif

#if (USE_COROTATION == CRONOS_ON)
	// Transform to co-rotating frame velocity for BCs
	Trafo->TransInertToCorot(gdata, gfunc, Problem);
#endif


#if(ENERGETICS == FULL)
	int q_max = q_Eges;
#else
	int q_max = n_omInt;
#endif

	// Boundary conditions
	for(int q=0; q<q_max; ++q) {
		gfunc.boundary(gdata, Problem, gdata.om[q],B,q);
	}

	if(ENERGETICS == FULL) {

#if (USE_COROTATION == CRONOS_ON)
		// Transform back for energy trafo
		Trafo->TransCorotToInert(gdata, gfunc, Problem);
#endif

		if(thermal) {
#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
			Trafo->TransE2Eth(gdata, gfunc, Problem, n, true);
#else
			Trafo->TransE2Eth(gdata, gfunc, Problem);
#endif
		} else {
			Trafo->TransE2T(gdata, gfunc, Problem);
		}

		for(int q=q_Eges; q<n_omInt; ++q) {
			gfunc.boundary(gdata, Problem, gdata.om[q],B,q);
		}

#if (USE_COROTATION == CRONOS_ON)
		// Transform to co-rotating frame velocity for BCs
		Trafo->TransInertToCorot(gdata, gfunc, Problem);
#endif

	}

	Problem.TransCons2Prim(gdata);

#if (OMS_USER == TRUE)

	for(int q=0; q<n_omIntUser; ++q) {
		gfunc.boundary(gdata, Problem, gdata.om_user[q],B,q);
	}
#endif

// ----------------------------------------------------------------
//   Runtime estimates
// ----------------------------------------------------------------

	gettimeofday(&tock2, 0);
	cend = clock();

	timeval tstepOld = tstep;
	if(n == RK_STEPS-1) {
		gettimeofday(&tstep, 0);
	}

	REAL delt = ((tock.tv_sec + tock.tv_usec/1.e6) - 
	             (tick.tv_sec + tick.tv_usec/1.e6));

	REAL delt2 = ((tock2.tv_sec + tock2.tv_usec/1.e6) - 
	              (tock.tv_sec + tock.tv_usec/1.e6));

	REAL dtStep = ((tstep.tv_sec + tstep.tv_usec/1.e6) -
	               (tstepOld.tv_sec + tstepOld.tv_usec/1.e6));

	if(gdata.rank == 0) {
		if(n == RK_STEPS-1 && Problem.get_Info() && Problem.checkout(5)) {
			cout << "------------------------------------------------------" << endl;
			cout << " Time needed for substeps:  " << delt << " " << delt2 << endl;
			cout << "             for full step: " << dtStep << endl;
//       cout << " CPU Cycle times: " << gdata.rank << " ";
//       cout << (1.*(cstep-cstart))/CLOCKS_PER_SEC << " ";
//       cout << (1.*(cend-cstep))/CLOCKS_PER_SEC << endl;
		}

	}


// ----------------------------------------------------------------
//   Test physical state
// ----------------------------------------------------------------

#if (USE_COROTATION == CRONOS_ON)
	// Transform to inertial frame velocity for phystest
	Trafo->TransCorotToInert(gdata, gfunc, Problem);
#endif

	phystest(gdata, gfunc, Problem, n);

#if (USE_COROTATION == CRONOS_ON)
	// Transform back to co-rotating frame
	Trafo->TransInertToCorot(gdata, gfunc, Problem);
#endif

	// Computing div B only at end of timestep
	if(Problem.mag && n == RK_STEPS-1){
		compute_divB(gdata, gfunc, Problem);
	}

//	cout << " my cfl: " << cfl << endl;

	return cfl;

}
