#ifndef CRONOS_ADDITIONAL_H
#define CRONOS_ADDITIONAL_H

#include "RadiationField.H"
#include "MagneticField.H"

using namespace CRONOS_CONSTANTS;


REAL distance(
		const Data & gdata,
		int ix,int iy, int iz,
		REAL xCen_num, REAL yCen_num, REAL zCen_num,
		int dir = -1
		)
{
	switch(dir) {
#if (GEOM == CYLINDRICAL)
		case 0 :	return gdata.getCen_x(ix);
		case 1 : 	return 0;
		case 2 : 	return gdata.getCen_z(iz) - zCen_num;
#else
		case 0 :	return gdata.getCen_x(ix) - xCen_num;
	    case 1 : 	return gdata.getCen_y(iy) - yCen_num;
	    case 2 : 	return gdata.getCen_z(iz) - zCen_num;
#endif
	    default: 	return sqrt(sqr(distance(gdata, ix, iy, iz, xCen_num, yCen_num, zCen_num, 0)) +
								sqr(distance(gdata, ix, iy, iz, xCen_num, yCen_num, zCen_num, 1)) +
								sqr(distance(gdata, ix, iy, iz, xCen_num, yCen_num, zCen_num, 2)) );
	}
}


class ToroidalMagneticField: public MagneticField_idealMHD_LAB {
public:
	ToroidalMagneticField(
			normalisation * normHD,
			Quantity magneticFieldStrength,
			REAL torusAxis[3],
			REAL xCen_num, REAL yCen_num, REAL zCen_num
		):
		MagneticField_idealMHD_LAB(normHD)
	{
		this->magneticFieldStrength = magneticFieldStrength;
		this->magneticFieldStrength_num = normHD->get_num(magneticFieldStrength);

		for (int iDir = 0; iDir < 3; ++iDir) {
			this->torusAxis[iDir] = torusAxis[iDir];
		}

		this->xCen_num = xCen_num;
		this->yCen_num = yCen_num;
		this->zCen_num = zCen_num;
	}

	virtual	~ToroidalMagneticField(
		)
	{
	}

	// All arguments in COM Frame
	virtual REAL get_magneticFieldStrength_LAB(
			Data & gdata,
			int ix, int iy, int iz
		)
	{
		REAL distance_inv = 1. / distance(gdata, ix, iy, iz, xCen_num, yCen_num, zCen_num);
		REAL magneticFieldStrength_LAB = magneticFieldStrength_num * distance_inv;
		return magneticFieldStrength_LAB;
	}

	virtual void get_magneticFieldDir_LAB(
			Data & gdata,
			int ix, int iy, int iz,
			REAL magneticFieldDir_LAB[3]
		)
	{

		REAL distance_inv = 1. / distance(gdata, ix, iy, iz, xCen_num, yCen_num, zCen_num);

		REAL toroid[3];
		REAL toroid_norm=0;

		REAL r_rel[3];
		for (int iDir = 0; iDir < 3; ++iDir) {
			r_rel[iDir] = distance(gdata, ix, iy, iz, xCen_num, yCen_num, zCen_num, iDir) * distance_inv;
		}

		for (int iDir = 0; iDir < 3; ++iDir) {
			toroid[iDir] = torusAxis[(iDir + 1)%3] * r_rel[(iDir + 2)%3] - r_rel[(iDir + 1)%3] * torusAxis[(iDir + 2)%3];
			toroid_norm += sqr(toroid[iDir]);
		}

		toroid_norm = 1. / sqrt(toroid_norm);

		for (int iDir = 0; iDir < 3; ++iDir) {
			magneticFieldDir_LAB[iDir] = toroid[iDir] * toroid_norm;
		}
	}

protected:
	Quantity magneticFieldStrength;
	REAL magneticFieldStrength_num;
	REAL torusAxis[3];
	REAL xCen_num, yCen_num, zCen_num;

};



class UndirectedThemalPartitionMagneticField: public MagneticField_idealMHD_COM {
public:
	UndirectedThemalPartitionMagneticField(
			normalisation * normHD,
			REAL thermalEnergyFraction
	):
			MagneticField_idealMHD_COM(normHD)
	{
		this->thermalEnergyFraction = thermalEnergyFraction;
		magneticFieldFactor = normHD->get_num(normHD->MAG_IND,
				sqrt(thermalEnergyFraction * 2. * VacuumPermeability * normHD->normQuantity(normHD->E_DENS))
		);
		ccout << "  Using mag field fac " << magneticFieldFactor << endl;

	}

	virtual	~UndirectedThemalPartitionMagneticField(
	)
	{
	}

	// All arguments in COM Frame
	virtual REAL get_magneticFieldStrength_COM(
			Data & gdata,
			int ix, int iy, int iz
	)
	{
		assert(gdata.om[gdata.fluid.get_q_Eges()].getName() == "Etherm");
		REAL magneticFieldStrength_LAB = magneticFieldFactor * sqrt(gdata.om[gdata.fluid.get_q_Eges()](ix, iy, iz));
		return magneticFieldStrength_LAB;
	}

	virtual void get_magneticFieldDir_LAB(
			Data & gdata,
			int ix, int iy, int iz,
			REAL magneticFieldDir_LAB[3]
	)
	{
		// Directionless
		for (int iDir = 0; iDir < 3; ++iDir) {
			magneticFieldDir_LAB[iDir] = 0;
		}
	}

	virtual void get_magneticFieldDir_COM(
			Data & gdata,
			int ix, int iy, int iz,
			REAL magneticFieldDir_COM[3]
	)
	{
		// Directionless
		for (int iDir = 0; iDir < 3; ++iDir) {
			magneticFieldDir_COM[iDir] = 0;
		}
	}

protected:
	REAL thermalEnergyFraction;
	REAL magneticFieldFactor;
};



#endif
