#ifndef CRONOS_GENERICBINARY_H
#define CRONOS_GENERICBINARY_H

#if (FLUID_TYPE != CRONOS_RHD)
#error "check constants.H settings"
#endif

#include "problem.H"
#include "randgen.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)
#include <gsl/gsl_sf.h>
#include "Adiabatic.H"
#include "InverseCompton.H"
#include "Synchrotron.H"
#include "additional.H"
#include "Absorption.H"
#endif



 using namespace CRONOS_CONSTANTS;

/*
  Alle Benutzerklassen erben einfach von der Oberklasse Problemtype -
  die kann ich noch in specific.C unterbringen.

  Hier können beliebige benötigte Subroutinen eingebunden werden -
  damit stören sie nicht im Hauptcode
 */

class RHD_NORM : public normalisation {
public:
	virtual void setup();
};

void RHD_NORM::setup() {

	set_definition(LEN, AstronomicalUnit);
	set_definition(MASS, ProtonMass);
	set_definition(VEL, SpeedOfLight);
	set_definition(TEMP);
	set_definition(MAG_IND, Tesla);
	set_definition(NUM_DENS);

	finish_setup();
}


class GenericBinary: public ProblemType {
public:
	GenericBinary (Data &);
	virtual void init_fields(Data &, int [], int []);
	//virtual void src_User(Data &, NumMatrix<REAL,3> [], NumMatrix<REAL,3> []);
	virtual void reset_User(Data &);
	virtual void get_PhysFluxUser(Data &, cronos::vector<REAL> &, phys_fields_1D &, phys_fields_1D &, int, REAL);

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	virtual void init_particleTransport(Data &);
#endif


private:

	void init_analyticCD(Data &, int, int, int);
	void init_nearest(Data &, int, int, int);
	void init_vacuum(Data &, int, int, int);

	REAL distanceStar(Data &, int, int, int);
	REAL distanceStar(Data &, int, int, int, int);
	REAL distancePulsar(Data &, int, int, int);
	REAL distancePulsar(Data &, int, int, int, int);

	void set_star(Data &, int, int, int, REAL);
	void set_pulsar(Data &, int, int, int, REAL);
	void set_noFlux(Data &, NumMatrix<REAL,3> [], NumMatrix<REAL,3> [], int, int, int);

	REAL get_cdCondition(REAL r, REAL z);


	Quantity xStar, yStar, zStar, rStar, rInjStar, mlrStar, vStarWind, uStarWind, luminosityStar;
	Quantity xPulsar, yPulsar, zPulsar, rInjPulsar, mlrPulsar, vPulsarWind, uPulsarWind;
	Quantity rhoStar, rhoPulsar;

	Quantity temperatureStar;

	REAL xStar_num, yStar_num, zStar_num, rStar_num, rInjStar_num;
	REAL xPulsar_num, yPulsar_num, zPulsar_num, rInjPulsar_num;
	REAL rhoStar_num, rhoPulsar_num;
	REAL uStarWind_num, uPulsarWind_num;

	REAL pulsarAxis[3];

	REAL WStarWind, WPulsarWind;

	REAL fac_pStar, fac_pPulsar;

	REAL eta;
	Quantity distanceStarPulsar;
	REAL distanceStarPulsar_num;
	REAL eps;

	bool hasStar, hasPulsar;

	string initialisation;


#if (EOS != POLYTROPIC)
	REAL polytropic_K;
#endif

	REAL polytropic_K_star, polytropic_K_pulsar;

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	Quantity magneticField;
	REAL magneticField_num;

	Quantity radiationEDens;
	REAL radiationEDens_num;

	Quantity photonEnergy;
	REAL photonEnergy_num;
#endif

};

GenericBinary::GenericBinary(Data &gdata)  : ProblemType(gdata)
{

	TrafoNorm = new RHD_NORM();
	TrafoNorm->setup();

	eps = 1e-250;

	//
	//	Star Parameters
	//


	hasStar = !value_exists("noStar");

	if (hasStar) {

	xStar 			= value("xStar") * AstronomicalUnit;
	yStar 			= value("yStar") * AstronomicalUnit;
	zStar 			= value("zStar") * AstronomicalUnit;
	rStar	 		= value("rStar") * SolarRadius;
	rInjStar 		= value("rInjStar") * AstronomicalUnit;

	vStarWind 		= value("vStarWind") * 1e3 * Meter / Second;

	mlrStar 		= value("mlrStar") * SolarMass / Year;

	luminosityStar	= value("luminosityStar") * SolarLuminosity;
	temperatureStar = value("temperatureStar") * Kelvin;
	fac_pStar 		= value("fac_pStar");

	WStarWind 		= 1./sqrt(1. - sqr(vStarWind.normalizeTo(SpeedOfLight)));
	uStarWind 		= vStarWind * WStarWind;
	rhoStar 		= mlrStar / (uStarWind * 4 * pi * sqr(rInjStar));


	}




	//
	//	Pulsar Parameters
	//
	

	hasPulsar = !value_exists("noPulsar");

	if (hasPulsar) {

	xPulsar 		= value("xPulsar") * AstronomicalUnit;
	yPulsar 		= value("yPulsar") * AstronomicalUnit;
	zPulsar 		= value("zPulsar") * AstronomicalUnit;
	rInjPulsar 		= value("rInjPulsar") * AstronomicalUnit;

	vPulsarWind 	= value("vPulsarWind") * SpeedOfLight;

	WPulsarWind 	= 1./sqrt(1. - sqr(vPulsarWind.normalizeTo(SpeedOfLight)));
	uPulsarWind 	= vPulsarWind * WPulsarWind;

	pulsarAxis[0] = 0.; pulsarAxis[1] = 0.; pulsarAxis[2] = 1.;


	if (value_exists("mlrPulsar")) {
		eta 		= 0;
		mlrPulsar 	= value("mlrPulsar") * SolarMass / Year;
	} else {
		eta = value("eta");
		mlrPulsar	 = eta * mlrStar * uStarWind / uPulsarWind;
	}

	Quantity Lsd = eta * mlrStar * vStarWind * SpeedOfLight;

	if (gdata.rank == 0) {
		std::cout << " ---- Pulsar Parameters ---- " << std::endl;
		std::cout << "MLR	\t" << mlrPulsar << std::endl;
		std::cout << "Lsd	\t" << Lsd.normalizeTo(Erg / Second) << " ergs / s" << std::endl;
	}

	rhoPulsar 	= mlrPulsar / (uPulsarWind * 4. * pi * sqr(rInjPulsar));
	fac_pPulsar = value("fac_pPulsar");

	}


	//
	//	General Parameters
	//

	if (value_exists("initialisation")) {
		initialisation = svalue("initialisation");
	} else {
		initialisation = "analyticCD";
	}


	distanceStarPulsar = sqrt(sqr(xStar - xPulsar) + sqr(yStar - yPulsar) + sqr(zStar - zPulsar));

#if (EOS != POLYTROPIC)
	polytropic_K = value("polytropic_K");
#endif

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	radiationEDens = luminosityStar / (4. * pi * SpeedOfLight * sqr(rInjStar));
	photonEnergy = BoltzmannConstant * temperatureStar * sqr(sqr(pi)) / (30. * gsl_sf_zeta_int(3));

	magneticField = value("magneticField") * Gauss;
#endif



	// ----------------------------------------------------------------------------------------------


	rhoStar_num 	= TrafoNorm->get_num(TrafoNorm->DENS, rhoStar);
	rhoPulsar_num 	= TrafoNorm->get_num(TrafoNorm->DENS, rhoPulsar);

	rInjStar_num 		= TrafoNorm->get_num(rInjStar);
	rInjPulsar_num 	= TrafoNorm->get_num(rInjPulsar);

	uStarWind_num	= TrafoNorm->get_num(uStarWind);
	uPulsarWind_num	= TrafoNorm->get_num(uPulsarWind);

	xStar_num 		= TrafoNorm->get_num(xStar);
	yStar_num 		= TrafoNorm->get_num(yStar);
	zStar_num 		= TrafoNorm->get_num(zStar);
	xPulsar_num		= TrafoNorm->get_num(xPulsar);
	yPulsar_num		= TrafoNorm->get_num(yPulsar);
	zPulsar_num		= TrafoNorm->get_num(zPulsar);

	distanceStarPulsar_num = TrafoNorm->get_num(distanceStarPulsar);

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	magneticField_num = TrafoNorm->get_num(magneticField);
	photonEnergy_num = TrafoNorm->get_num(photonEnergy);
	radiationEDens_num	= TrafoNorm->get_num(TrafoNorm->E_DENS, radiationEDens);
#endif

	// assume star surface temp to be 1000 K
	// use resulting polytropic k for the global initialisation
	// could be different for pulsar ? effect ?
//	polytropic_K_star = pow(rhoStar_num, 1. - gamma) * TrafoNorm->get_num(BoltzmannConstant / ProtonMass * temperatureStar);
//	polytropic_K_pulsar = pow(rhoPulsar_num, 1. - gamma) * TrafoNorm->get_num(BoltzmannConstant / ElectronMass * 100. * Kelvin);


	polytropic_K_star = std::pow(rhoStar_num, 1. - gamma) * fac_pStar;
	polytropic_K_pulsar = std::pow(rhoPulsar_num, 1. - gamma) * fac_pPulsar;

//	polytropic_K_star = polytropic_K;
//	polytropic_K_pulsar = polytropic_K;
	if (gdata.rank == 0) {
		std::cout << polytropic_K_star << " " << rhoStar_num << " " << gamma << std::endl;
	}

}

REAL GenericBinary::get_cdCondition(REAL r, REAL z0) {
	REAL eta_sqrt = sqrt(eta);
	REAL cdSlope = sin(pi * eta_sqrt) / (1. + 2. * eta_sqrt - cos(pi * eta_sqrt));
	REAL rmax = distanceStarPulsar_num * tan(0.5 * pi * eta_sqrt);

	if (z0 < 0.) {
		return 1.;
	} else
	if (z0 < distanceStarPulsar_num) {
		return atan(r/z0) - eta_sqrt * atan(r / (distanceStarPulsar_num - z0));
	} else {
		return (distanceStarPulsar_num - z0) + cdSlope * (r - rmax);
	}
}


void GenericBinary::init_analyticCD(Data &gdata, int ix, int iy, int iz) {
	REAL dStar = distanceStar(gdata, ix, iy, iz);
	REAL dPulsar = distancePulsar(gdata, ix, iy, iz);

	REAL z0 = 0.5 * ((sqr(dStar) - sqr(dPulsar)) / distanceStarPulsar_num + distanceStarPulsar_num);
	REAL r = sqrt(sqr(dStar) - sqr(z0));

	if (get_cdCondition(r, z0) > 0) {
		set_star(gdata, ix, iy, iz, dStar);
	}
	else {
		set_pulsar(gdata, ix, iy, iz, dPulsar);
	}
}


void GenericBinary::init_nearest(Data &gdata, int ix, int iy, int iz) {
	REAL dStar = distanceStar(gdata, ix, iy, iz);
	REAL dPulsar = distancePulsar(gdata, ix, iy, iz);

	if ( hasStar && hasPulsar ) {
		if (dStar < dPulsar ){
			set_star(gdata, ix, iy, iz, dStar);
		}
		else {
			set_pulsar(gdata, ix, iy, iz, dPulsar);
		}
	} else
	if ( !hasStar && hasPulsar ) {
		set_pulsar(gdata, ix, iy, iz, dPulsar);
	} else
	if ( hasStar && !hasPulsar ) {
		set_star(gdata, ix, iy, iz, dStar);
	}
}


void GenericBinary::init_vacuum(Data &gdata, int ix, int iy, int iz) {
	bool isOutsideCell = true;

	if(hasStar) {
		REAL dStar = distanceStar(gdata, ix, iy, iz);
		if (dStar < rInjStar_num){
			set_star(gdata, ix, iy, iz, dStar);
			isOutsideCell = false;
		}
	}

	if(hasPulsar) {
		REAL dPulsar = distancePulsar(gdata, ix, iy, iz);
		if (dPulsar < rInjPulsar_num){
			set_pulsar(gdata, ix, iy, iz, dPulsar);
			isOutsideCell = false;
		}
	}

	if(isOutsideCell) {
		gdata.om[q_Eges](ix, iy, iz) = 1e-10;
		gdata.om[q_rho](ix, iy, iz) = 1e-8;
		gdata.om[q_sx](ix, iy, iz) = 0.;
		gdata.om[q_sy](ix, iy, iz) = 0.;
		gdata.om[q_sz](ix, iy, iz) = 0.;
	}
}


void GenericBinary::init_fields(Data &gdata, int ibeg[3], int iend[3])
{	

	void (GenericBinary::*init_func)(Data &, int, int, int);

	if (initialisation == "analyticCD") {
		if (hasStar && hasPulsar) {
			init_func = &GenericBinary::init_analyticCD;
		} else {
			throw CException(" analytic CD only available if star & pulsar are available");
		}
	} else
	if (initialisation == "nearest") {
		init_func = &GenericBinary::init_nearest;
	} else {
		init_func = &GenericBinary::init_vacuum;
	}



	for(int ix=ibeg[0]; ix<=iend[0]; ++ix) {
		for(int iy=ibeg[1]; iy<=iend[1]; ++iy) {
			for(int iz=ibeg[2]; iz<=iend[2]; ++iz) {

				(this->*init_func)(gdata, ix, iy, iz);

#if (PARTICLE_TRANSPORT == CRONOS_ON)
				for (int ip = omb_particle; ip <= ome_particle; ip++) {
					gdata.om_user[ip](ix, iy, iz) = 0;
				}
#endif

			}
		}
	}
}




//void GenericBinary::src_User(Data &gdata, NumMatrix<REAL,3> * nom,
//                         NumMatrix<REAL,3> * nom_user)

void GenericBinary::reset_User(Data &gdata)
{
	for(int iz=0; iz<=gdata.mx[2]; ++iz) {
		for(int iy=0; iy<=gdata.mx[1]; ++iy) {
			for(int ix=0; ix<=gdata.mx[0]; ++ix) {
				REAL dStar = distanceStar(gdata, ix, iy, iz);
				REAL dPulsar = distancePulsar(gdata, ix, iy, iz);
				REAL dPulsar_inv = 1./(dPulsar + 1e-20);

				if (dStar < rInjStar_num && hasStar){
					//std::cout << ix << "\t" << iy << "\t " << iz << "\t" << "inside Star" << std::endl;
					set_star(gdata, ix, iy, iz, dStar);
//					set_noFlux(gdata, nom, nom_user, ix, iy, iz);
				}

				else

				if (dPulsar < rInjPulsar_num && hasPulsar) {
					//std::cout << ix << "\t" << iy << "\t " << iz << "\t" << "inside pulsar" << std::endl;
					set_pulsar(gdata, ix, iy, iz, dPulsar);
//					set_noFlux(gdata, nom, nom_user, ix, iy, iz);
				}
//				else {
//					set_gravityFlux(gdata, nom, ix, iy, iz, dStar, dPulsar);
//				}

			}
		}
	}
}

void GenericBinary::set_star(Data &gdata, int ix, int iy, int iz, REAL dStar) {
	REAL dStar_inv = 1./(dStar + eps);

	gdata.om[q_rho](ix, iy, iz) = rhoStar_num * sqr(rInjStar_num * dStar_inv);
	for (int dir = 0; dir < 3; ++dir) {
		gdata.om[q_sx + dir](ix, iy, iz) = distanceStar(gdata, ix, iy, iz, dir) * dStar_inv * uStarWind_num;
	}
	gdata.om[q_Eges](ix, iy, iz) = std::pow(gdata.om[q_rho](ix, iy, iz), gamma) * polytropic_K_star / (gamma - 1.);

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	gdata.om[q_Eadd](ix, iy, iz) = polytropic_K_star;
#endif

	gdata.om_user[0](ix, iy, iz) = gdata.om[q_rho](ix, iy, iz) *
			sqrt(1 + sqr(gdata.om[q_sx](ix, iy, iz)) + sqr(gdata.om[q_sy](ix, iy, iz)) + sqr(gdata.om[q_sz](ix, iy, iz)))
			* 1;
	gdata.om_user[1](ix, iy, iz) = 0;
}

void GenericBinary::set_pulsar(Data &gdata, int ix, int iy, int iz, REAL dPulsar) {
	REAL dPulsar_inv = 1./(dPulsar + eps);

	gdata.om[q_rho](ix, iy, iz) = rhoPulsar_num * sqr(rInjPulsar_num * dPulsar_inv);
	for (int dir = 0; dir < 3; ++dir) {
		gdata.om[q_sx + dir](ix, iy, iz) = distancePulsar(gdata, ix, iy, iz, dir) * dPulsar_inv * uPulsarWind_num;
	}
	gdata.om[q_Eges](ix, iy, iz) = std::pow(gdata.om[q_rho](ix, iy, iz), gamma) * polytropic_K_pulsar / (gamma - 1.);

#if(CRSWITCH_DUAL_ENERGY == CRONOS_ON)
	gdata.om[q_Eadd](ix, iy, iz) = polytropic_K_pulsar;
#endif

	gdata.om_user[0](ix,iy,iz) = 0;
	gdata.om_user[1](ix, iy, iz) =  gdata.om[q_rho](ix, iy, iz) *
			sqrt(1 + sqr(gdata.om[q_sx](ix, iy, iz)) + sqr(gdata.om[q_sy](ix, iy, iz)) + sqr(gdata.om[q_sz](ix, iy, iz)))
			* 1;

}


inline REAL GenericBinary::distanceStar(Data &gdata, int ix, int iy, int iz)
{
	return sqrt(
				sqr(distanceStar(gdata, ix, iy, iz, 0)) +
				sqr(distanceStar(gdata, ix, iy, iz, 1)) +
				sqr(distanceStar(gdata, ix, iy, iz, 2))
			);
}

inline REAL GenericBinary::distanceStar(Data &gdata, int ix, int iy, int iz, int dir)
{

	switch(dir) {
#if (GEOM == CYLINDRICAL)
		case 0 :	return gdata.getCen_x(ix);
		case 1 : 	return 0;
		case 2 : 	return gdata.getCen_z(iz) - zStar_num;
#else
		case 0 :	return gdata.getCen_x(ix) - xStar_num;
	    case 1 : 	return gdata.getCen_y(iy) - yStar_num;
	    case 2 : 	return gdata.getCen_z(iz) - zStar_num;
#endif
	    default: 	return distanceStar(gdata, ix, iy, iz);
	}
}

inline REAL GenericBinary::distancePulsar(Data &gdata, int ix, int iy, int iz)
{
	return sqrt(
				sqr(distancePulsar(gdata, ix, iy, iz, 0)) +
				sqr(distancePulsar(gdata, ix, iy, iz, 1)) +
				sqr(distancePulsar(gdata, ix, iy, iz, 2))
			);
}

inline REAL GenericBinary::distancePulsar(Data &gdata, int ix, int iy, int iz, int dir)
{

	switch(dir) {
#if (GEOM == CYLINDRICAL)
		case 0 :	return gdata.getCen_x(ix);
		case 1 : 	return 0;
		case 2 : 	return gdata.getCen_z(iz) - zPulsar_num;
#else
		case 0 :	return gdata.getCen_x(ix) - xPulsar_num;
	    case 1 : 	return gdata.getCen_y(iy) - yPulsar_num;
	    case 2 : 	return gdata.getCen_z(iz) - zPulsar_num;
#endif
	    default: 	return distancePulsar(gdata, ix, iy, iz);
	}
}


void GenericBinary::set_noFlux(Data &gdata, NumMatrix<REAL,3> * nom, NumMatrix<REAL,3> * nom_user,
		int ix, int iy, int iz)
{
	nom[q_rho](ix,iy,iz) = 0.;
	nom[q_sx](ix,iy,iz) = 0.;
	nom[q_sy](ix,iy,iz) = 0.;
	nom[q_sz](ix,iy,iz) = 0.;
	nom[q_Eges](ix,iy,iz) = 0.;
	nom_user[0](ix,iy,iz) = 0.;
	nom_user[1](ix,iy,iz) = 0.;
}


void GenericBinary::get_PhysFluxUser(Data &gdata, cronos::vector<REAL> &iPos,
		phys_fields_1D &pf, phys_fields_1D &pfUser, int dir, REAL shift)
{
	REAL v, u2, W;

	for (int i = -1; i <= gdata.mx[dir]+1; ++i)
	{
		u2 = sqr(pf.uPri[q_sx](i)) + sqr(pf.uPri[q_sy](i)) + sqr(pf.uPri[q_sz](i));
		W = sqrt(1. + u2);
		v = pf.uPri[q_sx + dir](i)/W;


		pfUser.flux[0](i) = pfUser.uPri[0](i) * v;
		pfUser.flux[1](i) = pfUser.uPri[1](i) * v;
	}

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	particleTransport.get_PhysFlux(gdata, iPos, pf, pfUser, dir, shift);
#endif

}

#if (PARTICLE_TRANSPORT == CRONOS_ON)

void GenericBinary::init_particleTransport(Data & gdata)
{
	// Defining normalisations
	particleTransport.set_normHD(TrafoNorm);
	normalisation * normHD = TrafoNorm;
	normalisation * normParticle = & particleTransport.get_normParticle();

	// Adding Particles
	particleTransport.add_particle(new Particle(gdata, normParticle, 0, 2));


	// Adding Loss Processes
	particleTransport.add_lossProcess(new Adiabatic(normHD, normParticle));


	if (!value_exists("magneticFieldType")) {
		throw CException("No magnetic field type specified");
	}

	MagneticField * pulsarMagneticField;

	if (svalue("magneticFieldType") == "toroidal") {
		ccout << " Using MagneticField: ToroidalMagneticField " << endl;
		pulsarMagneticField = new ToroidalMagneticField(normHD, magneticField, pulsarAxis, xPulsar_num, yPulsar_num,
														zPulsar_num);
	} else
	if (svalue("magneticFieldType") == "thermalPartition") {
		ccout << " Using MagneticField: UndirectedThemalPartitionMagneticField " << endl;
		pulsarMagneticField = new UndirectedThemalPartitionMagneticField(normHD, value("magneticEnergyPartitionFraction"));
	} else {
		throw CException("Wrong magnetic field type");
	}

	particleTransport.add_lossProcess(new SynchrotronProcess(normHD, normParticle, pulsarMagneticField));
	particleTransport.add_magneticField(pulsarMagneticField);


	IRadiationField * iStellarPhotonField = new MonochromaticPointSource(normHD, normParticle,
																photonEnergy,
																luminosityStar,
																xStar, yStar, zStar
																);

	RadiationField * stellarPhotonField = new RadiationField();
	stellarPhotonField->set_iRadiationField(iStellarPhotonField);

	particleTransport.add_lossProcess(new InverseComptonProcess(normHD, normParticle, stellarPhotonField));


	// Postprocesses

	if (value_exists("absorptionType")) {
		IRadiationField * iStellarPhotonField_absorption;

		if (svalue("absorptionType") == "MonochromaticPointSource") {
			ccout << " Using MonochromaticPointSource absorption " << endl;
			Quantity meanPhotonEnergy = BoltzmannConstant * temperatureStar * sqr(sqr(pi)) / (30. * gsl_sf_zeta_int(3));
			iStellarPhotonField_absorption = new MonochromaticPointSource(
						normHD, normParticle, meanPhotonEnergy, luminosityStar,
						xStar, yStar, zStar
			);

		} else
		if (svalue("absorptionType") == "BlackbodyPointSource") {
			ccout << " Using BlackbodyPointSource absorption " << endl;
			iStellarPhotonField_absorption = new BlackbodyPointSource(
					normHD, normParticle, temperatureStar, luminosityStar,
					xStar, yStar, zStar
			);
		} else {
			throw CException(" Choose a suitable absorption type ");
		}

		RadiationField * stellarPhotonField_absorption = new RadiationField();
		stellarPhotonField_absorption->set_iRadiationField(iStellarPhotonField_absorption);

		EmissionPostProcess * absorption = new Absorption(stellarPhotonField_absorption);
		particleTransport.add_emissionPostProcess(absorption);
	} else {
		ccout << " Using NO absorption " << endl;
	}
}

#endif

#endif
