import numpy as np
import matplotlib.pyplot as plt
import scipy.optimize as opt

from CronosPyLib.ProjHandler import ProjHandler
from matplotlib_utils import multipage, gridShow

fieldFigs = {}

def doWork(proj):
    assert(proj.cronosData[0].infos['TestCase'] == b'Linear wave decay')
    assert(proj.cronosData[0].infos['fluid_type'] == b'rhd')

    ns = proj.cronosData[0].infos['Nx']
    init = proj.cronosData[0].fluids[0]
    final = proj.cronosData[-1].fluids[0]

    ts = proj.cronosData[-1]['Data/WavePars/']['times']
    ts = np.array(ts)
    rms = lambda key: proj.cronosData[-1]['Data/WavePars/rms'][key]/proj.cronosData[-1]['Data/WavePars/rms'][key][0]

    def model(time, gamma, offset):
        return np.exp(-gamma * time) * offset

    popts = []

    if proj.rootPath.find('first') > 0:
        selection = np.array(ts) < 35
    else:
        selection = np.array(ts) > 100

    fields = ['rho', 'v_x', 'v_y', 'v_z', 'Etherm']

    for field in fields:
        try:
            fig = fieldFigs[proj.rootPath][field]
            plt.figure(fig.number)
        except:
            fig = plt.figure()
            try:
                fieldFigs[proj.rootPath][field] = fig
            except:
                fieldFigs[proj.rootPath] = {}
                fieldFigs[proj.rootPath][field] = fig

        plt.title(f'{proj.rootPath} {field}', fontsize=7)
        plt.semilogy(ts, rms(field), label=f'{proj.name} {field}')
        popt, pcov = opt.curve_fit(model, ts[selection], rms(field)[selection], p0=[1e-2, 1])
        print(f'{field}:\tgamma = {popt[0]} \toffset = {popt[1]}')
        popts.append(popt)
        plt.semilogy(ts, model(ts, *popt), 'k--')
        plt.legend()

    popt = np.mean(popts, axis=0)
    print(f'Mean\tgamma = {popt[0]} \toffset = {popt[1]}')
    print(f'-------------------------------------------------------------------------------')

    return ns[0], popt[0]


def run_once(f):
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return f(*args, **kwargs)
    wrapper.has_run = False
    return wrapper

if __name__ == '__main__':
    filesSmall = [
        './LinearWaves/small/DecayingRHDWave_64',
        './LinearWaves/small/DecayingRHDWave_128',
        './LinearWaves/small/DecayingRHDWave_256',
        './LinearWaves/small/DecayingRHDWave_512'
    ]

    filesMedium = [
        './LinearWaves/medium/DecayingRHDWave_32',
        './LinearWaves/medium/DecayingRHDWave_64',
        './LinearWaves/medium/DecayingRHDWave_128',
        './LinearWaves/medium/DecayingRHDWave_256'
    ]

    filesMediumFirst = [
        './LinearWaves/medium_first/DecayingRHDWave_32',
        './LinearWaves/medium_first/DecayingRHDWave_64',
        './LinearWaves/medium_first/DecayingRHDWave_128',
        './LinearWaves/medium_first/DecayingRHDWave_256'
    ]

    filesBig = [
        './LinearWaves/big/DecayingRHDWave_64',
        './LinearWaves/big/DecayingRHDWave_128',
        './LinearWaves/big/DecayingRHDWave_256',
        './LinearWaves/big/DecayingRHDWave_512'
    ]

    #filesList = [filesSmall, filesMedium, filesBig]

    filesList = [filesMediumFirst, filesMedium]

    orderFig = plt.figure()

    for files in filesList:

        nxs = []
        gammas = []

        for iFile, file in enumerate(files):
            proj = ProjHandler(file, skip=0)
            nx_loc, gammas_loc = doWork(proj)

            nxs.append(nx_loc)
            gammas.append(gammas_loc)

        nxs = np.array(nxs)
        gammas = np.array(gammas)

        plt.figure(orderFig.number)
        plt.loglog(nxs, gammas, label=proj.rootPath)

    plt.loglog(nxs, 0.02 * np.power(nxs/64, -2), 'k:')
    plt.loglog(nxs, 0.1 * np.power(nxs/64, -1), 'k:')
    plt.grid(True)
    plt.legend(fontsize=7)

    multipage("order.pdf")
    gridShow(2)
    plt.show()