#ifndef CRONOS_RHD_ADIABATIC_TEST_H
#define CRONOS_RHD_ADIABATIC_TEST_H

#include "problem.H"
#include "randgen.H"
#if (PARTICLE_TRANSPORT == CRONOS_ON)
#include "ParticleTransport.H"
#include "Adiabatic.H"
#endif

// using namespace std;

class RHD_AdiabaticTest: public ProblemType {
public:
	RHD_AdiabaticTest (Data &);
	virtual void init_fields(Data &, int [], int []);
	REAL getUx_temporal(REAL, REAL);
	REAL getRho_temporal(REAL, REAL);

	REAL getUx_spatial(REAL, REAL);
	REAL getRho_spatial(REAL, REAL);

	void src_User(Data &gdata, NumMatrix<REAL,3> nom[],
	                         NumMatrix<REAL,3> nom_user[]);

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	void init_particleTransport(Data &);
#endif

private:
	REAL rho0, p, W0, c;
	REAL adiabatic_exponent;

	REAL (RHD_AdiabaticTest::*getUx)(REAL, REAL);
	REAL (RHD_AdiabaticTest::*getRho)(REAL, REAL);
};

RHD_AdiabaticTest::RHD_AdiabaticTest(Data &gdata)  : ProblemType(gdata)
{
	rho0 = value("rho0");
	p = value("p");
	
	W0 = value("W0");
	c = value("c");


	adiabatic_exponent = value("Adiabatic_exponent");

	if (svalue("test_type") == "temporal") {
		getUx  = &RHD_AdiabaticTest::getUx_temporal;
		getRho = &RHD_AdiabaticTest::getRho_temporal;
	} else
	if (svalue("test_type") == "spatial") {
		getUx  = &RHD_AdiabaticTest::getUx_spatial;
		getRho = &RHD_AdiabaticTest::getRho_spatial;
	} else
	{
		throw CException(" No such test Type ");
	}

}

REAL RHD_AdiabaticTest::getUx_temporal(REAL time, REAL x)
{

	return sqrt(sqr(exp(c * time)*W0) - 1.);
}

REAL RHD_AdiabaticTest::getRho_temporal(REAL time, REAL x)
{

	return exp(-c * time)*rho0;
}


REAL RHD_AdiabaticTest::getUx_spatial(REAL time, REAL x)
{

	return sqrt(sqr(
				exp(c * x)*W0
			) - 1.);
}

REAL RHD_AdiabaticTest::getRho_spatial(REAL time, REAL x)
{

	return exp(-c * x)*rho0;
}

void RHD_AdiabaticTest::init_fields(Data &gdata, int ibeg[3], int iend[3])
{	

	for(int ix=ibeg[0]; ix<=iend[0]; ++ix) {
		REAL rho =  (this->*getRho)(gdata.time, gdata.getCen_x(ix));
		REAL ux =  (this->*getUx)(gdata.time, gdata.getCen_x(ix));

		for(int iy=ibeg[1]; iy<=iend[1]; ++iy) {
			for(int iz=ibeg[2]; iz<=iend[2]; ++iz) {

				gdata.om[q_rho](ix, iy, iz) = rho;
				gdata.om[q_sx](ix, iy, iz) = ux;
				gdata.om[q_sy](ix, iy, iz) = 0.;
				gdata.om[q_sz](ix, iy, iz) = 0.;
				gdata.om[q_Eges](ix, iy, iz) = p/(adiabatic_exponent - 1);
			}
		}
	}

}

void RHD_AdiabaticTest::src_User(Data &gdata, NumMatrix<REAL,3> nom[],
                         NumMatrix<REAL,3> nom_user[])
{
	for (int qq=q_rho; qq <= q_Eges; ++qq) {
		nom[qq].clear();
	}

	for(int ix=-1; ix<=gdata.mx[0]+1; ++ix) {

		REAL rho = (this->*getRho)(gdata.time + gdata.dt, gdata.getCen_x(ix));
		REAL ux = (this->*getUx)(gdata.time + gdata.dt, gdata.getCen_x(ix));

		for(int iy=-1; iy<=gdata.mx[1]+1; ++iy) {
			for(int iz=-1; iz<=gdata.mx[2]+1; ++iz) {
				gdata.om[q_rho](ix, iy, iz) = rho;
				gdata.om[q_sx](ix, iy, iz) = ux;
				gdata.om[q_sy](ix, iy, iz) = 0.;
				gdata.om[q_sz](ix, iy, iz) = 0.;
				gdata.om[q_Eges](ix, iy, iz) = p/(adiabatic_exponent - 1);
			}
		}
	}

}


#if (PARTICLE_TRANSPORT == CRONOS_ON)
using namespace CRONOS_CONSTANTS;
class RHD_NORM : public normalisation {
public:
	virtual void setup();
};

void RHD_NORM::setup() {

	set_definition(LEN);
	set_definition(MASS);
	set_definition(VEL, SpeedOfLight);
	set_definition(TEMP);
	set_definition(MAG_IND);
	set_definition(NUM_DENS);

	finish_setup();
}



void RHD_AdiabaticTest::init_particleTransport(Data & gdata)
{
	// Defining normalisations
	normalisation * normHD =  new RHD_NORM();
	normHD->setup();

	particleTransport.set_normHD(normHD);
	normalisation * normParticle = & particleTransport.get_normParticle();


	// Adding Particles
	Particle * particle = new Particle(gdata, normParticle, 0, 0);
	particleTransport.add_particle(particle);

	for(int iz=-3; iz<=gdata.mx[2]+3; ++iz) {
		for(int iy=-3; iy<=gdata.mx[1]+3; ++iy) {
			for(int ix=-3; ix<=gdata.mx[0]+3; ++ix) {
				if (gdata.getCen_x(ix) > 0.45 && gdata.getCen_x(ix) < 0.55)
					particle->set_powerlaw(1, 2, ix, iy, iz);
			}
		}
	}


//	if (gdata.rank == 0)
//		particleTransport.get_particle(0)->set_powerlaw(1, spectralIndex, 50,0,0);


	// Adding Loss Processes
	particleTransport.add_lossProcess(new Adiabatic(normHD, normParticle));
}
#endif


#endif
