#ifndef CRONOS_ABSORPTION_TEST_H
#define CRONOS_ABSORPTION_TEST_H

#if (PARTICLE_TRANSPORT != CRONOS_ON || FLUID_TYPE != CRONOS_RHD)
#error "check constants.H settings"
#endif

#include "problem.H"
#include "randgen.H"

#include "ParticleTransport.H"
#include "Adiabatic.H"
#include "InverseCompton.H"
#include "Synchrotron.H"
#include "RadiationField_special.H"
#include "MagneticField_special.H"
#include "DummyEmissionProcess.H"
#include "Absorption.H"

using namespace CRONOS_CONSTANTS;


using namespace CRONOS_CONSTANTS;
class AbsorptionTest_NORM : public normalisation {
public:
	virtual void setup();
};

void AbsorptionTest_NORM::setup() {

	set_definition(LEN, SolarRadius);
	set_definition(MASS);
	set_definition(VEL, SpeedOfLight);
	set_definition(TEMP);
	set_definition(MAG_IND);
	set_definition(NUM_DENS);

	finish_setup();
}

class AbsorptionTest: public ProblemType {
public:
	AbsorptionTest (Data &);
	virtual void init_fields(Data &, int [], int []);
	virtual void src_User(Data &, NumMatrix<REAL,3> [], NumMatrix<REAL,3> []);

	virtual void init_particleTransport(Data &);


private:
	REAL rho, ux, p;

	Quantity particlePowerlawConstant;
	REAL particleSpectralIndex;

	Quantity sourceTemperature, sourceRadius, sourceDistance;
	REAL sourceRadius_num, sourceDistance_num;
	string photonFieldType;
	};

AbsorptionTest::AbsorptionTest(Data &gdata)  : ProblemType(gdata)
{
	// Defining normalisations
	TrafoNorm = new AbsorptionTest_NORM();
	TrafoNorm->setup();

	rho = value("rho");
	ux = value("ux");
	p = value("p");

	particlePowerlawConstant = value("particlePowerlawConstant") / cube(Meter);
	particleSpectralIndex = value("particleSpectralIndex");

	sourceTemperature = value("sourceTemperature") * Kelvin;
	sourceRadius = value("sourceRadius") * SolarRadius;
	sourceDistance = value("sourceDistance") * SolarRadius;

	sourceRadius_num = TrafoNorm->get_num(sourceRadius);
	sourceDistance_num = TrafoNorm->get_num(sourceDistance);

	photonFieldType = svalue("photonFieldType");
}


void AbsorptionTest::init_fields(Data &gdata, int ibeg[3], int iend[3])
{

	for(int ix=ibeg[0]; ix<=iend[0]; ++ix) {
		for(int iy=ibeg[1]; iy<=iend[1]; ++iy) {
			for(int iz=ibeg[2]; iz<=iend[2]; ++iz) {
				gdata.om[q_rho](ix, iy, iz) = rho;
				gdata.om[q_sx](ix, iy, iz) = ux;
				gdata.om[q_sy](ix, iy, iz) = 0.;
				gdata.om[q_sz](ix, iy, iz) = 0.;
				gdata.om[q_Eges](ix, iy, iz) = p/(gamma - 1);
			}
		}
	}
}

void AbsorptionTest::src_User(Data &gdata, NumMatrix<REAL,3> nom[],
								 NumMatrix<REAL,3> nom_user[])
{
	// Completely disable the hydro step
	for (int qq=q_rho; qq <= q_Eges; ++qq) {
		nom[qq].clear();
	}

	for (int qq=0; qq < N_OMINT_USER; ++qq) {
		nom_user[qq].clear();
	}

	for(int ix=-1; ix<=gdata.mx[0]+1; ++ix) {
		for(int iy=-1; iy<=gdata.mx[1]+1; ++iy) {
			for(int iz=-1; iz<=gdata.mx[2]+1; ++iz) {
				gdata.om[q_rho](ix, iy, iz) = rho;
				gdata.om[q_sx](ix, iy, iz) = ux;
				gdata.om[q_sy](ix, iy, iz) = 0.;
				gdata.om[q_sz](ix, iy, iz) = 0.;
				gdata.om[q_Eges](ix, iy, iz) = p/(gamma - 1);
			}
		}
	}

}



void AbsorptionTest::init_particleTransport(Data & gdata)
{

	normalisation * normHD = this->TrafoNorm;

	particleTransport.set_normHD(normHD);
	normalisation * normParticle = & particleTransport.get_normParticle();


	// Adding Particles
	Particle * particle = new Particle(gdata, normParticle, 0, 0);
	particleTransport.add_particle(particle);

	REAL particlePowerlawConstant_num = normParticle->get_num(normParticle->NUM_DENS, particlePowerlawConstant);

	for(int iz=-3; iz<=gdata.mx[2]+3; ++iz) {
		for(int iy=-3; iy<=gdata.mx[1]+3; ++iy) {
			for(int ix=-3; ix<=gdata.mx[0]+3; ++ix) {
				particle->set_powerlaw(particlePowerlawConstant_num, particleSpectralIndex, ix, iy, iz);
			}
		}
	}


	// Adding Loss Processes
	Quantity diffEmissivity = 1.e-40 / (cube(Meter) * Second * ElectronVolt * 4. * pi);
	particleTransport.add_lossProcess(new DummyEmissionProcess(normHD, normParticle, diffEmissivity));

	std::shared_ptr<RadiationField> photonField;
	Quantity luminosity = StefanBoltzmannConstant * sqr(sqr(sourceTemperature)) * 4. * pi * sqr(sourceRadius);
	Quantity meanPhotonEnergy = BoltzmannConstant * sourceTemperature * sqr(sqr(pi)) / (30. * gsl_sf_zeta_int(3));

	vector3D center{0, 0, -sourceDistance_num};

	if (photonFieldType == "MonochromaticPointSource") {
		photonField = make_shared<MonochromaticPointSource>(normHD, normParticle,
													meanPhotonEnergy, luminosity,
													center
		);

	} else
	if (photonFieldType == "MonochromaticExtended") {
		photonField = make_shared<MonochromaticExtendedSource>(normHD, normParticle,
												sourceRadius,
												meanPhotonEnergy, luminosity,
												center
		);

	} else
	if (photonFieldType == "BlackbodyPointSource") {
		photonField = make_shared<BlackbodyPointSource>(normHD, normParticle,
										sourceTemperature, luminosity,
										center
		);

	} else
	if (photonFieldType == "BlackbodyExtended") {
		photonField = make_shared<BlackbodyExtendedSource>(normHD, normParticle,
										sourceRadius, sourceTemperature,
										 center
		);
	} else
//	if (photonFieldType == "ConstantBeamedMonochromatic") {
//		iPhotonField = new ConstantBeamedMonochromatic(normHD, normParticle,
//													   sourceTemperature, luminosity,
//													   0, 0, -sourceDistance_num
//		);
//
//	} else
	{
		throw CException("Invalid radiation field type");
	}

	EmissionPostProcess * absorption = new Absorption(photonField);
	particleTransport.add_emissionPostProcess(absorption);
}


#endif
