#ifndef CRONOS_RHD_EMISSION_TEST_H
#define CRONOS_RHD_EMISSION_TEST_H

#if (PARTICLE_TRANSPORT != CRONOS_ON || FLUID_TYPE != CRONOS_RHD)
#error "check constants.H settings"
#endif

#include "problem.H"
#include "randgen.H"


#include "ParticleTransport.H"
#include "Adiabatic.H"
#include "InverseCompton.H"
#include "Synchrotron.H"
#include "RadiationField_special.H"
#include "MagneticField_special.H"

// using namespace std;

class RHD_Emission_Test: public ProblemType {
public:
	RHD_Emission_Test (Data &);
	virtual void init_fields(Data &, int [], int []);

	void src_User(Data &gdata, NumMatrix<REAL,3> nom[],
	                         NumMatrix<REAL,3> nom_user[]);

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	void init_particleTransport(Data &);
#endif

private:
	REAL rho, ux, p;

	Quantity particlePowerlawConstant;
	REAL particleSpectralIndex;

	Quantity magneticFieldEnergyDensity, magneticFieldStrength;
	REAL phiMag, thetaMag;
	bool noSynchrotron;

	Quantity photonEnergy, photonEnergyDensity;
	REAL phiRad, thetaRad;
	 bool noInverseCompton;
};

RHD_Emission_Test::RHD_Emission_Test(Data &gdata)
	: ProblemType(gdata)
{
	rho = value("rho");
	ux = value("ux");
	p = value("p");

	particlePowerlawConstant = value("particlePowerlawConstant") / cube(Meter);
	particleSpectralIndex = value("particleSpectralIndex");

	noSynchrotron = value_exists("noSynchrotron");
	noInverseCompton = value_exists("noInverseCompton");

	if(!noSynchrotron) {
		phiMag = value("phiMag") * pi / 180.;
		thetaMag = value("thetaMag") * pi / 180.;
		magneticFieldEnergyDensity = value("magneticFieldEnergyDensity") * ElectronVolt / cube(Meter);
		magneticFieldStrength = sqrt(magneticFieldEnergyDensity * 2. * VacuumPermeability);
		ccout << magneticFieldStrength.normalizeTo(Gauss) << " " << Gauss << endl;
	}

	if(!noInverseCompton) {
		phiRad = value("phiRad") * pi / 180.;
		thetaRad = value("thetaRad") * pi / 180.;
		photonEnergy = value("photonEnergy") * ElectronVolt;
		photonEnergyDensity = value("photonEnergyDensity") * ElectronVolt / cube(Meter);
	}
}


void RHD_Emission_Test::init_fields(Data &gdata, int ibeg[3], int iend[3])
{	

	for(int ix=ibeg[0]; ix<=iend[0]; ++ix) {
		for(int iy=ibeg[1]; iy<=iend[1]; ++iy) {
			for(int iz=ibeg[2]; iz<=iend[2]; ++iz) {
				gdata.om[q_rho](ix, iy, iz) = rho;
				gdata.om[q_sx](ix, iy, iz) = ux;
				gdata.om[q_sy](ix, iy, iz) = 0.;
				gdata.om[q_sz](ix, iy, iz) = 0.;
				gdata.om[q_Eges](ix, iy, iz) = p/(gamma - 1);
			}
		}
	}

	for (int qq=0; qq < N_OMINT_USER; ++qq) {
		gdata.om_user[qq].clear();
	}

}

void RHD_Emission_Test::src_User(Data &gdata, NumMatrix<REAL,3> nom[],
                         NumMatrix<REAL,3> nom_user[])
{
	// Completely disable the hydro step
	for (int qq=q_rho; qq <= q_Eges; ++qq) {
		nom[qq].clear();
	}

	for (int qq=0; qq < N_OMINT_USER; ++qq) {
		nom_user[qq].clear();
	}

	for(int ix=-1; ix<=gdata.mx[0]+1; ++ix) {
		for(int iy=-1; iy<=gdata.mx[1]+1; ++iy) {
			for(int iz=-1; iz<=gdata.mx[2]+1; ++iz) {
				gdata.om[q_rho](ix, iy, iz) = rho;
				gdata.om[q_sx](ix, iy, iz) = ux;
				gdata.om[q_sy](ix, iy, iz) = 0.;
				gdata.om[q_sz](ix, iy, iz) = 0.;
				gdata.om[q_Eges](ix, iy, iz) = p/(gamma - 1);
			}
		}
	}

}


using namespace CRONOS_CONSTANTS;
class RHD_NORM : public normalisation {
public:
	virtual void setup();
};

void RHD_NORM::setup() {

	set_definition(LEN, value("NormLen") * Meter);
	set_definition(MASS);
	set_definition(VEL, SpeedOfLight);
	set_definition(TEMP);
	set_definition(MAG_IND);
	set_definition(NUM_DENS);

	finish_setup();
}



void RHD_Emission_Test::init_particleTransport(Data & gdata)
{

	// Defining normalisations
	normalisation * normHD =  new RHD_NORM();
	normHD->setup();

	particleTransport.set_normHD(normHD);
	normalisation * normParticle = & particleTransport.get_normParticle();


	// Adding Particles
	Particle * particle = new Particle(gdata, normParticle, 0, 0);
	particleTransport.add_particle(particle);

	REAL particlePowerlawConstant_num = normParticle->get_num(normParticle->NUM_DENS, particlePowerlawConstant);

	for(int iz=-3; iz<=gdata.mx[2]+3; ++iz) {
		for(int iy=-3; iy<=gdata.mx[1]+3; ++iy) {
			for(int ix=-3; ix<=gdata.mx[0]+3; ++ix) {
					particle->set_powerlaw(particlePowerlawConstant_num, particleSpectralIndex, ix, iy, iz);
			}
		}
	}


	// Synchrotron
	if (!noSynchrotron) {
		REAL magneticFieldAxis[3] = {cos(phiMag) * sin(thetaMag), sin(phiMag) * sin(thetaMag), cos(thetaMag)};


		auto magneticField = make_shared<ConstantMagneticField>(normHD, magneticFieldStrength, magneticFieldAxis);

		particleTransport.add_lossProcess(new SynchrotronProcess(normHD, normParticle, magneticField));
	}


	// Inverse Compton
	if (!noInverseCompton) {
		Quantity photonNumberDensity = photonEnergyDensity / photonEnergy;
		Quantity photonSpectralNumberDensity = photonNumberDensity / (
				normParticle->normQuantity(normParticle->ENERGY));

		REAL radiationDir[3] = {cos(phiRad) * sin(thetaRad), sin(phiRad) * sin(thetaRad), cos(thetaRad)};

		auto photonField = make_shared<ConstantBeamedMonochromatic>(normHD, normParticle,
																		photonSpectralNumberDensity, photonEnergy,
																		radiationDir);

		particleTransport.add_lossProcess(new InverseComptonProcess(normHD, normParticle, photonField));
	}
}


#endif
