#!/usr/bin/env python3

import sys

from CronosPyLib.ProjHandler import ProjHandler
from CronosPyLib.EmissionSky import EmissionSky
import os
import numpy as np

from astropy.constants import c
import astropy as ap
import astropy.units as u
import naima
from naima.models import Synchrotron, PowerLaw, InverseCompton

import matplotlib.pyplot as plt

from cxlib.matplotlib_utils import savefigMultipage



if __name__ == '__main__':
	rootDir = os.getcwd()
	proj = ProjHandler(os.path.join(rootDir, 'RHD_Emission_Test_static'))

	def work_old():
		electron_dist = PowerLaw(amplitude=1e0 / (511 * u.keV), e_0=511 * u.keV, alpha=1.5)

		sky = proj.emissionData[0].skys[0]
		nEnergies = len(sky.energies)
		nChannels = len(sky.channels)

		Eemin = proj.cat['momentumb'] * u.eV
		Eemax = proj.cat['momentume'] * u.eV


		fig, ax = plt.subplots()
		# SYNCROTRON stuff
		cronosSED = sky.getFlux("Synchrotron")[:, 0, 0] * sky.energies * sky.energies
		ax.loglog(sky.energies, cronosSED, label='cronos - SYN')

		magneticFieldEnergyDensity = proj.cat['magneticFieldEnergyDensity'] * u.eV / u.m ** 3
		mag = np.sqrt(magneticFieldEnergyDensity * 2 * ap.constants.mu0)
		# SYN = Synchrotron(electron_dist, B=6.34564e-9 * u.G, Eemax=100 * u.TeV, Eemin=100 * u.MeV)
		SYN = Synchrotron(electron_dist, B=mag, Eemax=Eemax, Eemin=Eemin)

		naimaSED = SYN.flux(sky.energies * u.eV, distance=1 * u.meter) * sky.energies * sky.energies
		naimaSED *= 1e4  # cm^-2 -> m^-2
		ax.loglog(sky.energies, naimaSED, 'k-.', label="naima - SYN")

		# IC stuff

		# At this point naima has only isotropic IC for monochromatic fields implemented
		# Let therefore compare the solid angle averaged result of cronos

		deltaPhi = sky.phis[1] - sky.phis[0]
		deltaTheta = sky.thetas[1] - sky.thetas[0]

		deltaPhis = np.full_like(sky.phis, deltaPhi)
		deltaThetas = np.full_like(sky.thetas, deltaTheta)

		deltaPhiss, deltaThetass = np.meshgrid(deltaPhis, deltaThetas)
		_, thetass = np.meshgrid(sky.phis, sky.thetas)

		dA = deltaPhiss * deltaThetass * np.sin(thetass)

		cronosSED = np.transpose(sky.getFlux("InverseCompton"))
		cronosSED = np.sum(cronosSED * dA[:, :, np.newaxis], axis=(0, 1)) / (4. * np.pi)
		cronosSED = cronosSED * sky.energies * sky.energies

		ax.loglog(sky.energies, cronosSED, label='cronos - IC (avg)')

		photonEnergyDensity = proj.cat['photonEnergyDensity'] * u.eV / u.m ** 3
		photonEnergy = proj.cat['photonEnergy'] * u.eV
		IC = InverseCompton(electron_dist, seed_photon_fields=[['MONO', photonEnergy, photonEnergyDensity, 0 * u.deg],
																													 ['THERM', 11604. * u.K, photonEnergyDensity]],
												Eemax=Eemax, Eemin=Eemin)

		print(IC.seed_photon_fields)

		naimaSED = IC.flux(sky.energies * u.eV, distance=1 * u.meter, seed='MONO') * sky.energies * sky.energies
		naimaSED *= 1e4  # cm^-2 -> m^-2
		ax.loglog(sky.energies, naimaSED, 'k-.', label="naima - IC (iso)")

		naimaSED = IC.flux(sky.energies * u.eV, distance=1 * u.meter, seed='THERM') * sky.energies * sky.energies
		naimaSED *= 1e4  # cm^-2 -> m^-2
		ax.loglog(sky.energies, naimaSED, 'g-.', label="naima - IC (iso) (therm)")

		# cosmetics
		ax.grid(True)
		ax.set_xscale('log')
		ax.set_yscale('log')
		ax.set_xlabel("Energy / eV")
		ax.set_ylabel("SED / eV m^{-2} s^{-1}")
		ax.set_ylim((1e-35, 1e-5))
		ax.legend(loc='lower left')

	def work():
		electron_dist = PowerLaw(amplitude=1e0 / (511 * u.keV), e_0=511 * u.keV, alpha=1.5)

		Eemin = proj.cat['momentumb'] * u.eV
		Eemax = proj.cat['momentume'] * u.eV

		emission = proj.emissionData[-1].emissions[0]
		energies = emission.energies
		directions = emission.directions
		spectrum = emission.reductions[0]

		fig, ax = plt.subplots()
		# SYNCROTRON stuff

		cronosSED = spectrum["Synchrotron"][0] * energies * energies
		ax.loglog(energies, cronosSED, label='cronos - SYN')

		magneticFieldEnergyDensity = proj.cat['magneticFieldEnergyDensity'] * u.eV / u.m ** 3
		mag = np.sqrt(magneticFieldEnergyDensity * 2 * ap.constants.mu0)
		# SYN = Synchrotron(electron_dist, B=6.34564e-9 * u.G, Eemax=100 * u.TeV, Eemin=100 * u.MeV)
		SYN = Synchrotron(electron_dist, B=mag, Eemax=Eemax, Eemin=Eemin)

		naimaSED = SYN.flux(energies * u.eV, distance=1 * u.meter) * energies * energies
		naimaSED *= 1e4  # cm^-2 -> m^-2
		ax.loglog(energies, naimaSED, 'k-.', label="naima - SYN")

		# IC stuff

		# At this point naima has only isotropic IC for monochromatic fields implemented
		# Let therefore compare the solid angle averaged result of cronos
		deltaPhi = directions.phi[1] - directions.phi[0]
		deltaTheta = directions.theta[1] - directions.theta[0]
		dA = deltaPhi * deltaTheta * np.sin(directions.dir2thetaphi[:,0])

		cronosSED = spectrum["InverseCompton"] * dA[:, np.newaxis]
		cronosSED = np.sum(cronosSED, axis=(0)) / (4. * np.pi)
		cronosSED = cronosSED * energies * energies

		ax.loglog(energies, cronosSED, label='cronos - IC (avg)')

		photonEnergyDensity = proj.cat['photonEnergyDensity'] * u.eV / u.m ** 3
		photonEnergy = proj.cat['photonEnergy'] * u.eV
		IC = InverseCompton(electron_dist, seed_photon_fields=[['MONO', photonEnergy, photonEnergyDensity, 0 * u.deg],
																													 ['THERM', 11604. * u.K, photonEnergyDensity]],
												Eemax=Eemax, Eemin=Eemin)

		print(IC.seed_photon_fields)

		naimaSED = IC.flux(energies * u.eV, distance=1 * u.meter, seed='MONO') * energies * energies
		naimaSED *= 1e4  # cm^-2 -> m^-2
		ax.loglog(energies, naimaSED, 'k-.', label="naima - IC (iso)")

		naimaSED = IC.flux(energies * u.eV, distance=1 * u.meter, seed='THERM') * energies * energies
		naimaSED *= 1e4  # cm^-2 -> m^-2
		ax.loglog(energies, naimaSED, 'g-.', label="naima - IC (iso) (therm)")

		# cosmetics
		ax.grid(True)
		ax.set_xscale('log')
		ax.set_yscale('log')
		ax.set_xlabel("Energy / eV")
		ax.set_ylabel("SED / eV m^{-2} s^{-1}")
		ax.set_ylim((1e-35, 1e-5))
		ax.legend(loc='lower left')

	try:
		work()
	except:
		work_old()

	savefigMultipage("naima_comparison.pdf")
	plt.show()





