#!/usr/bin/env python3

import sys

from CronosPyLib.ProjHandler import ProjHandler
import os
import numpy as np

import astropy.constants as c
import astropy.units as u

import matplotlib.pyplot as plt
from cxlib.matplotlib_utils import savefigMultipage

from matplotlib import rc
rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
## for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
rc('text', usetex=True)

def latex_float(f):
    float_str = "{0:.3g}".format(f)
    if "e" in float_str:
        base, exponent = float_str.split("e")
        return r"{0} \times 10^{{{1}}}".format(base, int(exponent))
    else:
        return float_str

def getAnalyticalSolution(lossCoefficient, lossIndex, f0):
    lossIndexRed = lossIndex - 1
    def solution(x,t):
        xPrime = np.power(np.power(x, -lossIndexRed) - lossCoefficient * t * lossIndexRed, -1/lossIndexRed)
        return f0(xPrime) * np.power(xPrime / x, lossIndex)

    return solution


if __name__ == '__main__':
    rootDir = os.getcwd()
    proj = ProjHandler(os.path.join(rootDir, 'SynchrotronTest'))

    particleSpectralIndex = proj.cat['particleSpectralIndex']
    particlePowerlawConstant = proj.cat['particlePowerlawConstant']

    magneticFieldStrength = proj.cat['magneticFieldStrength'] * u.T
    magneticFieldEnergyDensity = magneticFieldStrength**2 / (2 * c.mu0)

    simulationTimeUnit = u.m / c.c

    # \dot{gamma} = coeff * gamma^2
    # coeff in simulation units 1/[TimeUnit]
    lossCoefficient = (4. / 3. * c.sigma_T / (c.m_e * c.c) * magneticFieldEnergyDensity).to(1/simulationTimeUnit)

    maximumGamma = proj.cat['momentume'] * (u.eV / (c.m_e * c.c**2)).to(1)
    def initialValue(gamma):
        if gamma < maximumGamma:
            return particlePowerlawConstant * np.power(gamma, -particleSpectralIndex)
        else:
            return 0
    initialValue = np.vectorize(initialValue)
    analyticalSolution = getAnalyticalSolution(lossCoefficient.value, 2, initialValue)

    for data in proj.cronosData:
        time = data.infos["time"]

        particle = data.particles[0]
        gammas = particle.momenta

        # edgesMod = np.power(particle.momentumEdges, 1 - particleSpectralIndex)
        # edgesIntegrals = (edgesMod[1:] - edgesMod[:-1]) / (1 - particleSpectralIndex)
        # gammas = np.power( edgesIntegrals /
        #     particle.momentumWidths, -1/particleSpectralIndex)

        gammas = gammas[3:-3]

        spectrum = particle.get_spectrum(0,0,0)

        anaSol = analyticalSolution(gammas, time)
        err = spectrum - anaSol
        relErr = np.abs(err) / anaSol

        plt.figure(1)
        # plt.loglog(gammas, spectrum, label=f't={(time * simulationTimeUnit).to(u.s):.1E}')

        timeLabel = r'${0}\, \mathrm{{s}}$'.format(latex_float((time * simulationTimeUnit).to(u.s).value))
        plt.loglog(gammas, spectrum, label=timeLabel)
        plt.loglog(gammas, anaSol, 'k:')

        plt.figure(2)
        plt.loglog(gammas, relErr, label=timeLabel)

    plt.figure(1)
    plt.grid(True)
    plt.legend(loc='best')
    plt.xlabel(r'$\gamma$')
    plt.ylabel(r'$\mathcal{N}$')

    plt.figure(2)
    plt.grid(True)
    plt.legend(loc='best')
    plt.xlabel(r'$\gamma$')
    plt.ylabel(r'$\Delta \mathcal{N}_\mathrm{rel}$')


    savefigMultipage("SynchrotronTest.pdf")

    for n in plt.get_fignums():
        plt.figure(n)
        plt.savefig(f'{proj.name}.{n}.png')

    # plt.show()
