#ifndef CRONOS_RHD_RIEMANN_PROBLEM_H
#define CRONOS_RHD_RIEMANN_PROBLEM_H

#include "problem.H"
#include "randgen.H"

// using namespace std;

/*
  Alle Benutzerklassen erben einfach von der Oberklasse Problemtype -
  die kann ich noch in specific.C unterbringen.

  Hier können beliebige benötigte Subroutinen eingebunden werden -
  damit stören sie nicht im Hauptcode
 */

class RHD_RiemannProblem: public ProblemType {
public:
	RHD_RiemannProblem (Data &);
	virtual void init_fields(Data &, int [], int []);
	void WriteToH5(Hdf5Stream &);

private:
	REAL rhoL, rhoR;
	REAL vxL, vxR;
	REAL vtL, vtR;
	REAL uxL, uxR;
	REAL utL, utR;
	REAL pL, pR;
	REAL adiabatic_exponent;
};

RHD_RiemannProblem::RHD_RiemannProblem(Data &gdata)  : ProblemType(gdata)
{
	rhoL = value((char*)"rhoL");
	rhoR = value((char*)"rhoR");

	vxL = value((char*)"uxL");
	vtL = value((char*)"utL");
	vxR = value((char*)"uxR");
	vtR = value((char*)"utR");
	
	REAL WL = 1./sqrt(1. - sqr(vxL) - sqr(vtL));
	REAL WR = 1./sqrt(1. - sqr(vxR) - sqr(vtR));

	uxL = vxL * WL;
	utL = vtL * WL;
	uxR = vxR * WR;
	utR = vtR * WR;

	pL = value((char*)"pL");
	pR = value((char*)"pR");
	
	adiabatic_exponent = value((char*)"Adiabatic_exponent");
}




void RHD_RiemannProblem::init_fields(Data &gdata, int ibeg[3], int iend[3])
{	


	for(int ix=ibeg[0]; ix<=iend[0]; ++ix) {
		REAL rho(0.);
		REAL ux(0.);
		REAL ut(0.);
		REAL p(0.);

		if (gdata.getCen_x(ix) > 0.5) {
			rho = rhoR;
			ux = uxR;
			ut = utR;
			p = pR;
		} else {
			rho = rhoL;
			ux = uxL;
			ut = utL;
			p = pL;
		}

		for(int iy=ibeg[1]; iy<=iend[1]; ++iy) {
			for(int iz=ibeg[2]; iz<=iend[2]; ++iz) {

				gdata.om[q_rho](ix, iy, iz) = rho;
				gdata.om[q_sx](ix, iy, iz) = ux;
				gdata.om[q_sy](ix, iy, iz) = ut;
				gdata.om[q_sz](ix, iy, iz) = 0.;
				gdata.om[q_Eges](ix, iy, iz) = p/(adiabatic_exponent - 1);
			}
		}
	}

}

void RHD_RiemannProblem::WriteToH5(Hdf5Stream &h5out) {
	// Indicate that we are using a test case
	h5out.AddGlobalAttr("TestCase", "rhd 1d riemann problem");

	// supply normalised values necessary for the analytical solution
	// Use own group for that purpose
	hid_t group = h5out.AddGroup("Data/RHDPars");

	h5out.AddGlobalAttr("rhoL", rhoL, group);
	h5out.AddGlobalAttr("uxL", vxL, group);
	h5out.AddGlobalAttr("utL", vtL, group);
	h5out.AddGlobalAttr("pL", pL, group);
	h5out.AddGlobalAttr("rhoR", rhoR, group);
	h5out.AddGlobalAttr("uxR", vxR, group);
	h5out.AddGlobalAttr("utR", vtR, group);
	h5out.AddGlobalAttr("pR", pR, group);
	h5out.AddGlobalAttr("gamma", adiabatic_exponent, group);

	// Now close the new group
	h5out.CloseGroup(group);
}


#endif
