#ifndef CRONOS_RHD_RIEMANN_PROBLEM_H
#define CRONOS_RHD_RIEMANN_PROBLEM_H

#include "problem.H"
#include "randgen.H"

// using namespace std;

/*
  Alle Benutzerklassen erben einfach von der Oberklasse Problemtype -
  die kann ich noch in specific.C unterbringen.

  Hier können beliebige benötigte Subroutinen eingebunden werden -
  damit stören sie nicht im Hauptcode
 */

class RHD_RiemannProblem_DSA: public ProblemType {
public:
	RHD_RiemannProblem_DSA (Data &);
	virtual void init_fields(Data &, int [], int []);
	virtual void get_PhysFluxUser(Data &, cronos::vector<REAL> &, phys_fields_1D &, phys_fields_1D &, int, REAL);
	virtual void get_particleAcceleration(Data & gdata,
			cronos::vector<int> & ipos,
			NumMatrix<REAL,1> &cellCenters,
			NumMatrix<REAL,1> &cellEdges,
			NumMatrix<REAL,1> &values,
			NumMatrix<REAL,1> &velocities
			);

	void WriteToH5(Hdf5Stream &);

private:
	REAL rhoL, rhoR;
	REAL vxL, vxR;
	REAL vtL, vtR;
	REAL uxL, uxR;
	REAL utL, utR;
	REAL pL, pR;
	REAL adiabatic_exponent;
	REAL kapZZ, kapPZ;
};

RHD_RiemannProblem_DSA::RHD_RiemannProblem_DSA(Data &gdata)  : ProblemType(gdata)
{
	rhoL = value((char*)"rhoL");
	rhoR = value((char*)"rhoR");

	vxL = value((char*)"uxL");
	vtL = value((char*)"utL");
	vxR = value((char*)"uxR");
	vtR = value((char*)"utR");
	
	REAL WL = 1./sqrt(1. - sqr(vxL) - sqr(vtL));
	REAL WR = 1./sqrt(1. - sqr(vxR) - sqr(vtR));

	uxL = vxL * WL;
	utL = vtL * WL;
	uxR = vxR * WR;
	utR = vtR * WR;

	pL = value((char*)"pL");
	pR = value((char*)"pR");
	
	adiabatic_exponent = value((char*)"Adiabatic_exponent");

	kapZZ = 1.;
	kapPZ = 0.;
}




void RHD_RiemannProblem_DSA::init_fields(Data &gdata, int ibeg[3], int iend[3])
{	


	for(int ix=ibeg[0]; ix<=iend[0]; ++ix) {
		REAL rho(0.);
		REAL ux(0.);
		REAL ut(0.);
		REAL p(0.);

		if (gdata.getCen_x(ix) > 0.5) {
			rho = rhoR;
			ux = uxR;
			ut = utR;
			p = pR;
		} else {
			rho = rhoL;
			ux = uxL;
			ut = utL;
			p = pL;
		}

		for(int iy=ibeg[1]; iy<=iend[1]; ++iy) {
			for(int iz=ibeg[2]; iz<=iend[2]; ++iz) {

				gdata.om[q_rho](ix, iy, iz) = rho;
				gdata.om[q_sx](ix, iy, iz) = ux;
				gdata.om[q_sy](ix, iy, iz) = ut;
				gdata.om[q_sz](ix, iy, iz) = 0.;
				gdata.om[q_Eges](ix, iy, iz) = p/(adiabatic_exponent - 1);

				for (int species=0; species < N_PARTICLE_SPECIES; species++) {
					int omb = omb_particle + species * N_MOMENTUM_BINS;
					int ome = omb + N_MOMENTUM_BINS - 1;

					for (int ip = omb; ip <= ome; ++ip) {
						if (ip == omb) {
							gdata.om_user[ip](ix, iy, iz) = 0;
						} else {
							gdata.om_user[ip](ix, iy, iz) = 0;
						}
					}
				}
			}
		}
	}

}


void RHD_RiemannProblem_DSA::get_PhysFluxUser(Data &gdata, cronos::vector<REAL> &iPos,
		phys_fields_1D &pf, phys_fields_1D &pfUser, int dir, REAL shift)
{
	REAL v, u2, W;
	int omb, ome;

	for (int i = -1; i <= gdata.mx[dir]+1; ++i)
	{
		u2 = sqr(pf.uPri[q_sx](i)) + sqr(pf.uPri[q_sy](i)) + sqr(pf.uPri[q_sz](i));
		W = sqrt(1. + u2);
		v = pf.uPri[q_sx + dir](i)/W;

		for (int ip = omb_particle; ip <= ome_particle; ip++) {
			pfUser.flux[ip](i) = pfUser.uPri[ip](i) * v;
		}
	}

}

void RHD_RiemannProblem_DSA::get_particleAcceleration(Data & gdata,
			cronos::vector<int> & ipos,
			NumMatrix<REAL,1> &cellCenters,
			NumMatrix<REAL,1> &cellEdges,
			NumMatrix<REAL,1> &values,
			NumMatrix<REAL,1> &velocities
			)
{
}

void RHD_RiemannProblem_DSA::WriteToH5(Hdf5Stream &h5out) {
	// Indicate that we are using a test case
	h5out.AddGlobalAttr("TestCase", "rhd 1d riemann problem");

	// supply normalised values necessary for the analytical solution
	// Use own group for that purpose
	hid_t group = h5out.AddGroup("Data/RHDPars");

	h5out.AddGlobalAttr("rhoL", rhoL, group);
	h5out.AddGlobalAttr("uxL", vxL, group);
	h5out.AddGlobalAttr("utL", vtL, group);
	h5out.AddGlobalAttr("pL", pL, group);
	h5out.AddGlobalAttr("rhoR", rhoR, group);
	h5out.AddGlobalAttr("uxR", vxR, group);
	h5out.AddGlobalAttr("utR", vtR, group);
	h5out.AddGlobalAttr("pR", pR, group);
	h5out.AddGlobalAttr("gamma", adiabatic_exponent, group);

	// Now close the new group
	h5out.CloseGroup(group);
}

#endif
