#ifndef CRONOS_RHD_RIEMANN_PROBLEM_H
#define CRONOS_RHD_RIEMANN_PROBLEM_H

#include "problem.H"
#include "randgen.H"

#if (PARTICLE_TRANSPORT == CRONOS_ON)
#include "ParticleTransport.H"
#include "Adiabatic.H"
#endif

#include <gsl/gsl_multiroots.h>


// using namespace std;

/*
  Alle Benutzerklassen erben einfach von der Oberklasse Problemtype -
  die kann ich noch in specific.C unterbringen.

  Hier können beliebige benötigte Subroutinen eingebunden werden -
  damit stören sie nicht im Hauptcode
 */

class RHD_SingleShock: public ProblemType {
public:
	RHD_SingleShock (Data &);
	virtual void init_fields(Data &, int [], int []);

	void WriteToH5(Hdf5Stream &);

#if (PARTICLE_TRANSPORT == CRONOS_ON)
	void init_particleTransport(Data &);
#endif


private:
	REAL rhoA, rhoB;
	REAL uxA, vxA;
	REAL utA, vtA;
	REAL uxB;
	REAL utB;
	REAL pA, pB;
	REAL vShock;
};


REAL lorentz(double ux, double ut)
{
	return sqrt(1. + sqr(ux) + sqr(ut));
}

double gamma_adiab;
REAL h(double theta)
{
	return 1. + gamma_adiab / (gamma_adiab - 1.) * theta;
}
REAL dh(double theta)
{
	return gamma_adiab / (gamma_adiab - 1.);
}

struct rankine_hugoniot_params { double rhoA; double uxA; double utA; double pA; double vShock; };

int rankine_hugoniot (const gsl_vector * x, void * p, gsl_vector * f) {
	struct rankine_hugoniot_params * params = (struct rankine_hugoniot_params *)p;
	const double rhoA 		= (params->rhoA);
	const double uxA 		= (params->uxA);
	const double utA 		= (params->utA);
	const double pA 		= (params->pA);
	const double vShock 	= (params->vShock);

	const double hA 		= h(pA/rhoA);
	const double WA 		= lorentz(uxA, utA);

	const double rhoB 		= gsl_vector_get(x,0);
	const double uxB 		= gsl_vector_get(x,1);
	const double utB 		= gsl_vector_get(x,2);
	const double pB 		= gsl_vector_get(x,3);

	const double hB 		= h(pB/rhoB);
	const double WB 		= lorentz(uxB, utB);

	gsl_vector_set(f, 0, rhoA*uxA - rhoA*vShock*WA + rhoB*(-uxB + vShock*WB));
	gsl_vector_set(f, 1, pA - pB + hA*rhoA*(uxA*uxA) - hB*rhoB*(uxB*uxB) - hA*rhoA*uxA*vShock*WA + hB*rhoB*uxB*vShock*WB);
	gsl_vector_set(f, 2, hA*rhoA*utA*(uxA - vShock*WA) + hB*rhoB*utB*(-uxB + vShock*WB));
	gsl_vector_set(f, 3, rhoA*uxA*(-1 + hA*WA) - rhoB*uxB*(-1 + hB*WB) - vShock*(-pA + pB + rhoA*WA*(-1 + hA*WA) - rhoB*WB*(-1 + hB*WB)));

//	cout  << "x val:\t\t"
//		   << gsl_vector_get (x, 0)
//	<< " " << gsl_vector_get (x, 1)
//	<< " " << gsl_vector_get (x, 2)
//	<< " " << gsl_vector_get (x, 3) << endl;

	//	cout  << "func val:\t"
	//		   << gsl_vector_get (f, 0)
	//	<< " " << gsl_vector_get (f, 1)
	//	<< " " << gsl_vector_get (f, 2)
	//	<< " " << gsl_vector_get (f, 3) << endl;

	return GSL_SUCCESS;
}


int rankine_hugoniot_df (const gsl_vector * x, void *p, gsl_matrix * J)
{
	struct rankine_hugoniot_params * params = (struct rankine_hugoniot_params *)p;
	const double rhoA 		= (params->rhoA);
	const double uxA 		= (params->uxA);
	const double utA 		= (params->utA);
	const double pA 		= (params->pA);
	const double vShock 	= (params->vShock);

	const double hA 		= h(pA/rhoA);
	const double dhA 		= dh(pA/rhoA);
	const double WA 		= lorentz(uxA, utA);

	const double rhoB 		= gsl_vector_get(x,0);
	const double uxB 		= gsl_vector_get(x,1);
	const double utB 		= gsl_vector_get(x,2);
	const double pB 		= gsl_vector_get(x,3);

	const double hB 		= h(pB/rhoB);
	const double dhB 		= dh(pB/rhoB);
	const double WB 		= lorentz(uxB, utB);

	gsl_matrix_set (J, 0, 0, -uxB + vShock*WB);
	gsl_matrix_set (J, 0, 1, rhoB*(-1 + (uxB*vShock)/WB));
	gsl_matrix_set (J, 0, 2, (rhoB*utB*vShock)/WB);
	gsl_matrix_set (J, 0, 3, 0);
	gsl_matrix_set (J, 1, 0, -(((-(dhB*pB) + hB*rhoB)*uxB*(uxB - vShock*WB))/rhoB));
	gsl_matrix_set (J, 1, 1, (hB*rhoB*(uxB*uxB*vShock - 2*uxB*WB + vShock*(WB*WB)))/WB);
	gsl_matrix_set (J, 1, 2, (hB*rhoB*utB*uxB*vShock)/WB);
	gsl_matrix_set (J, 1, 3, -1 - dhB*(uxB*uxB) + dhB*uxB*vShock*WB);
	gsl_matrix_set (J, 2, 0, -(((-(dhB*pB) + hB*rhoB)*utB*(uxB - vShock*WB))/rhoB));
	gsl_matrix_set (J, 2, 1, (hB*rhoB*utB*(uxB*vShock - WB))/WB);
	gsl_matrix_set (J, 2, 2, (hB*rhoB*(utB*utB*vShock + WB*(-uxB + vShock*WB)))/WB);
	gsl_matrix_set (J, 2, 3, dhB*utB*(-uxB + vShock*WB));
	gsl_matrix_set (J, 3, 0, -(((uxB - vShock*WB)*(-(dhB*pB*WB) + rhoB*(-1 + hB*WB)))/rhoB));
	gsl_matrix_set (J, 3, 1, (rhoB*(-(hB*(uxB*uxB)) + WB - hB*(WB*WB) + uxB*vShock*(-1 + 2*hB*WB)))/WB);
	gsl_matrix_set (J, 3, 2, (rhoB*utB*(-vShock - hB*(uxB - 2*vShock*WB)))/WB);
	gsl_matrix_set (J, 3, 3, -(dhB*uxB*WB) + vShock*(-1 + dhB*(WB*WB)));

	return GSL_SUCCESS;
}

int
rankine_hugoniot_fdf (const gsl_vector * x, void *params,
                gsl_vector * f, gsl_matrix * J)
{
	rankine_hugoniot(x, params, f);
	rankine_hugoniot_df(x, params, J);

	return GSL_SUCCESS;
}

RHD_SingleShock::RHD_SingleShock(Data &gdata)  : ProblemType(gdata)
{
	rhoA = value("rhoA");
	uxA = value("uxA");
	utA = value("utA");
	pA = value("pA");
	vShock = value("vShock");
	
	gamma_adiab = gamma;

	REAL WA = lorentz(uxA, utA);
	vxA = uxA / WA;
	vtA = utA / WA;

	REAL vRelA = (vShock - vxA) / (1. - vShock * vxA);
	REAL WRelA = 1. / sqrt(1. - sqr(vRelA));

	REAL WxA = lorentz(uxA, 0);
	REAL WShock = 1. / sqrt(1. - sqr(vShock));

	int status;
	size_t i, iter = 0;

	const size_t n = 4;
	struct rankine_hugoniot_params p = {rhoA, uxA, utA, pA, vShock};
	gsl_multiroot_function_fdf func = {&rankine_hugoniot,
								  &rankine_hugoniot_df,
								  &rankine_hugoniot_fdf,
								  n, &p};

	// Start value: from empirical fiddling
	double x_init[4] = {
			rhoA * (3. * WRelA + sqrt(WxA/WShock)),
			1.15 * WShock - 0.85 / cube(WShock),
			utA / WRelA,
			rhoA * 0.7 * sqr(WRelA)
	};

	gsl_vector *x = gsl_vector_alloc(n);

	for(int i=0; i < n; ++i) {
		gsl_vector_set (x, i, x_init[i]);
	}

	const gsl_multiroot_fdfsolver_type *solverType;
	gsl_multiroot_fdfsolver *solver;
	solverType = gsl_multiroot_fdfsolver_newton;
	solver = gsl_multiroot_fdfsolver_alloc (solverType, n);
	gsl_multiroot_fdfsolver_set (solver, &func, x);

	do
	{
		iter++;
		status = gsl_multiroot_fdfsolver_iterate(solver);

		cout << iter << ":\t"
				<< " " << gsl_vector_get (solver->x, 0)
				<< " " << gsl_vector_get (solver->x, 1)
				<< " " << gsl_vector_get (solver->x, 2)
				<< " " << gsl_vector_get (solver->x, 3)


				<< " | " << gsl_vector_get (solver->dx, 0)
				<< " " << gsl_vector_get (solver->dx, 1)
				<< " " << gsl_vector_get (solver->dx, 2)
				<< " " << gsl_vector_get (solver->dx, 3)


				<< " | " << gsl_vector_get (solver->f, 0)
				<< " " << gsl_vector_get (solver->f, 1)
				<< " " << gsl_vector_get (solver->f, 2)
				<< " " << gsl_vector_get (solver->f, 3)
		<< endl;

		if (status)   /* check if solver is stuck */
			break;

		status = gsl_multiroot_test_residual(solver->f, 1e-6);
	}
	while (status == GSL_CONTINUE && iter < 1000);

	if (status != GSL_SUCCESS) {
		cout << gsl_strerror(status) << endl;
		exit(1.);
	}

	rhoB	= gsl_vector_get (solver->x, 0);
	uxB		= gsl_vector_get (solver->x, 1);
	utB		= gsl_vector_get (solver->x, 2);
	pB		= gsl_vector_get (solver->x, 3);

	if (abs(rhoA - rhoB)/rhoA < 1e-2) {
		cout << " iteration converged to the upstream state:"
				<< " " << rhoB
				<< " " << uxB
				<< " " << utB
				<< " " << pB
				<< endl;
		exit(1.);
	}

	gsl_multiroot_fdfsolver_free(solver);
	gsl_vector_free(x);

//	exit(1.);

}




void RHD_SingleShock::init_fields(Data &gdata, int ibeg[3], int iend[3])
{	


	for(int ix=ibeg[0]; ix<=iend[0]; ++ix) {
		REAL rho(0.);
		REAL ux(0.);
		REAL ut(0.);
		REAL p(0.);

		if (gdata.getCen_x(ix) < 0) {
			rho = rhoA;
			ux = uxA;
			ut = utA;
			p = pA;
		} else {
			rho = rhoB;
			ux = uxB;
			ut = utB;
			p = pB;
		}

		for(int iy=ibeg[1]; iy<=iend[1]; ++iy) {
			for(int iz=ibeg[2]; iz<=iend[2]; ++iz) {

				gdata.om[q_rho](ix, iy, iz) = rho;
				gdata.om[q_sx](ix, iy, iz) = ux;
				gdata.om[q_sy](ix, iy, iz) = ut;
				gdata.om[q_sz](ix, iy, iz) = 0.;
				gdata.om[q_Eges](ix, iy, iz) = p/(gamma - 1.);

				for (int species=0; species < N_PARTICLE_SPECIES; species++) {
					int omb = omb_particle + species * N_MOMENTUM_BINS;
					int ome = omb + N_MOMENTUM_BINS - 1;

					for (int ip = omb; ip <= ome; ++ip) {
						if (ip == omb) {
							gdata.om_user[ip](ix, iy, iz) = 0;
						} else {
							gdata.om_user[ip](ix, iy, iz) = 0;
						}
					}
				}
			}
		}
	}

}


void RHD_SingleShock::WriteToH5(Hdf5Stream &h5out) {
	// Indicate that we are using a test case
	// h5out.AddGlobalAttr("TestCase", "rhd 1d single shock");

	// supply normalised values necessary for the analytical solution
	// Use own group for that purpose
//	hid_t group = h5out.AddGroup("Data/RHDPars");
//
//	h5out.AddGlobalAttr("rhoL", rhoL, group);
//	h5out.AddGlobalAttr("uxL", vxL, group);
//	h5out.AddGlobalAttr("utL", vtL, group);
//	h5out.AddGlobalAttr("pL", pL, group);
//	h5out.AddGlobalAttr("rhoR", rhoR, group);
//	h5out.AddGlobalAttr("uxR", vxR, group);
//	h5out.AddGlobalAttr("utR", vtR, group);
//	h5out.AddGlobalAttr("pR", pR, group);
//	h5out.AddGlobalAttr("gamma", adiabatic_exponent, group);
//
//	// Now close the new group
//	h5out.CloseGroup(group);
}

#if (PARTICLE_TRANSPORT == CRONOS_ON)
using namespace CRONOS_CONSTANTS;
class RHD_NORM : public normalisation {
public:
	virtual void setup();
};

void RHD_NORM::setup() {

	set_definition(LEN);
	set_definition(MASS);
	set_definition(VEL, SpeedOfLight);
	set_definition(TEMP);
	set_definition(MAG_IND);
	set_definition(NUM_DENS);

	finish_setup();
}



void RHD_SingleShock::init_particleTransport(Data & gdata)
{
	// Defining normalisations
	normalisation * normHD =  new RHD_NORM();
	normHD->setup();

	particleTransport.set_normHD(normHD);
	normalisation * normParticle = & particleTransport.get_normParticle();


	// Adding Particles
	Particle * particle = new Particle(gdata, normParticle, 0, 0);
	particleTransport.add_particle(particle);

	// Adding Loss Processes
	particleTransport.add_lossProcess(new Adiabatic(normHD, normParticle));
}
#endif


#endif
