#include "GslWrapper.H"


GslRootSolver::GslRootSolver() {
  solver_f = gsl_root_fsolver_alloc(gsl_root_fsolver_brent);
  solver_fdf = gsl_root_fdfsolver_alloc(gsl_root_fdfsolver_newton);

  gsl_params = {.f = &fun_f, .df = &fun_df, .fdf = &fun_fdf};
  function_f = {.function = &gsl_adapter_f, .params = &gsl_params};
  function_fdf = {.f = &gsl_adapter_f, .df = &gsl_adapter_df, .fdf = &gsl_adapter_fdf, .params = &gsl_params};
}

GslRootSolver::~GslRootSolver() {
  gsl_root_fsolver_free(solver_f);
  gsl_root_fdfsolver_free(solver_fdf);
}

void GslRootSolver::check_gslStatus(const int status) const {
  if (status > 0) {
    gsl_set_error_handler(oldHandler);
    std::string error = gsl_strerror(status);
    throw GslRootSolverError(error);
  }
}

double GslRootSolver::solve_f (double xLower, double xUpper) { 
  oldHandler = gsl_set_error_handler_off();
  int status = gsl_root_fsolver_set(solver_f, &function_f, xLower, xUpper);
  check_gslStatus(status);

  int iter = 0;
  double x;
  do {
    iter++;
    status = gsl_root_fsolver_iterate (solver_f);
    x = gsl_root_fsolver_root(solver_f);
    xLower = gsl_root_fsolver_x_lower(solver_f);
    xUpper = gsl_root_fsolver_x_upper(solver_f);
    status = gsl_root_test_interval(xLower, xUpper, 0., tolRel);
  }
  while (status == GSL_CONTINUE && iter < max_iter);

  check_gslStatus(status);
  gsl_set_error_handler(oldHandler);
  return x;
}


double GslRootSolver::solve_fdf (double x0) {
  oldHandler = gsl_set_error_handler_off();
  int status = gsl_root_fdfsolver_set(solver_fdf, &function_fdf, x0);
  check_gslStatus(status);

  int iter = 0;
  double x = x0;
  do {
    iter++;
    status = gsl_root_fdfsolver_iterate (solver_fdf);
    x0 = x;
    x = gsl_root_fdfsolver_root (solver_fdf);
    status = gsl_root_test_delta (x, x0, 0, tolRel);
  }
  while (status == GSL_CONTINUE && iter < max_iter);

  check_gslStatus(status);
  gsl_set_error_handler(oldHandler);
  return x;
}

double GslRootSolver::gsl_adapter_f(double x, void * params) {
  gsl_params_T * funs = static_cast<gsl_params_T *>(params);
  return funs->f->operator()(x);
}

double GslRootSolver::gsl_adapter_df(double x, void * params) {
  gsl_params_T * funs = static_cast<gsl_params_T *>(params);
  return funs->df->operator()(x);
}

void GslRootSolver::gsl_adapter_fdf(double x, void * params, double * y, double * dy) {
  gsl_params_T * funs = static_cast<gsl_params_T *>(params);
  return funs->fdf->operator()(x, *y, *dy);
}