#include "HaloBuffer.H"
// Necessary for < C++ 11
#include <cmath>


using namespace std;

Buffer::Buffer(
	)
{
	maxSize = 0;
	currentSize = 0;
	buffer = 0;
	fields.clear();
}

Buffer::~Buffer(
	)
{
	delete [] buffer;
}

void Buffer::resize(
	int maxSize
	)
{
	clear();
	delete [] buffer;

	this->maxSize = maxSize;
	buffer = new REAL[maxSize];

	// For debug purposes
	for (int i=0; i<maxSize; ++i) {
		buffer[i] = -i - 1;
	}
}

void Buffer::add_field(
		int lo[3],
		int hi[3]
	)
{
	DataLink<REAL, 3> field(&buffer[currentSize], lo, hi);

	currentSize += field.getSize();
	assert(currentSize <= maxSize);
	fields.push_back(field);
}

void Buffer::clear(
	)
{
	fields.clear();
	currentSize = 0;
}

int  Buffer::get_currentSize(
	) const
{
	return currentSize;
}
int  Buffer::get_maxSize(
	) const
{
	return maxSize;
}

int Buffer::get_nFields(
	) const
{
	return fields.size();
}

DataLink<REAL, 3> & Buffer::operator()(
		int iField
	)
{
	return fields[iField];
}

Buffer::operator REAL*()
{
	return buffer;
}


Buffer::operator const REAL*() const
{
	return buffer;
}




HaloBuffer::HaloBuffer(
		int mxCell[3],
		int rim,
		int numMax
	)
{
	for(int dir=0; dir<3; ++dir) {
		this->mxCell[dir] = mxCell[dir];
	}
	this->numMax = numMax;
	this->rim = rim;
	this->nFields = 0;
}

HaloBuffer::~HaloBuffer(
	)
{
}

Buffer & HaloBuffer::operator()(
		int shiftx,
		int shifty,
		int shiftz
	)
{
	return buffer[get_index(shiftx, shifty, shiftz)];
}

int HaloBuffer::get_index(
		int shiftx,
		int shifty,
		int shiftz
	) const
{
	assert(	shiftx >= -1 && shiftx <= 1 &&
			shifty >= -1 && shifty <= 1 &&
			shiftz >= -1 && shiftz <= 1	);

	return 9 * shiftz + 3 * shifty + shiftx + 13;
}

void HaloBuffer::init_buffers(
	)
{
	// The 0,0,0 buffer needs special treatment
	int iBuffCen = get_index(0, 0, 0);

	// Do the resizing
	for(int shiftz=-1; shiftz <= 1; ++shiftz) {
		for(int shifty=-1; shifty <= 1; ++shifty) {
			for(int shiftx=-1; shiftx <= 1; ++shiftx) {
				int iBuff = get_index(shiftx, shifty, shiftz);

				// Dont resize the central buffer
				if (iBuff == iBuffCen) {
					continue;
				}

				// Get the extents
				int maxSize = numMax * compute_maxSingleSize(rim, shiftx, shifty, shiftz);

				// resize the respective buffer
				buffer[iBuff].resize(maxSize);
			}
		}
	}
}

void HaloBuffer::add_field(
	)
{
	int loDelta[3] = { 0,0,0 };
	int hiDelta[3] = { 0,0,0 };
	add_field(loDelta, hiDelta);
}

void HaloBuffer::add_field(
		int loDelta[3],
		int hiDelta[3]
	)
{
	assert(nFields < numMax);
	++nFields;

	for (int dir = 0; dir < 3; ++dir) {
		assert(abs(loDelta[dir]) <= rim);
		assert(abs(hiDelta[dir]) <= rim);
	}

	// The 0,0,0 buffer needs special treatment
	int iBuffCen = get_index(0, 0, 0);

	int loBuff[3];
	int hiBuff[3];

	// add the field to every buffer
	for (int shiftz = -1; shiftz <= 1; ++shiftz) {
		for (int shifty = -1; shifty <= 1; ++shifty) {
			for (int shiftx = -1; shiftx <= 1; ++shiftx) {
				int iBuff = get_index(shiftx, shifty, shiftz);

				// skip the central buffer 
				if (iBuff == iBuffCen) {
					continue;
				}

				// Get the buffer extents
				get_extent(loDelta, hiDelta, shiftx, shifty, shiftz, loBuff, hiBuff);
				
				// add the field
				buffer[iBuff].add_field(loBuff, hiBuff);
			}
		}
	}
}

void HaloBuffer::clear(
	)
{
	for (int shiftz = -1; shiftz <= 1; ++shiftz) {
		for (int shifty = -1; shifty <= 1; ++shifty) {
			for (int shiftx = -1; shiftx <= 1; ++shiftx) {
				int iBuff = get_index(shiftx, shifty, shiftz);
				buffer[iBuff].clear();
			}
		}
	}
	nFields = 0;
}

int HaloBuffer::get_size(
		int shiftx,	int shifty,	int shiftz
	) const
{
	int iBuff = get_index(shiftx, shifty, shiftz);
	return buffer[iBuff].get_currentSize();
}

int HaloBuffer::get_maxSize(
		int shiftx, int shifty, int shiftz
	) const
{
	int iBuff = get_index(shiftx, shifty, shiftz);
	return buffer[iBuff].get_maxSize();
}

bool HaloBuffer::is_face(
		int shiftx,	int shifty,	int shiftz
	) const
{
	return (abs(shiftx) + abs(shifty) + abs(shiftz) == 1);
}

bool HaloBuffer::is_edge(
		int shiftx,	int shifty,	int shiftz
	) const
{
	return (abs(shiftx) + abs(shifty) + abs(shiftz) == 2);
}

bool HaloBuffer::is_corner(
		int shiftx,	int shifty,	int shiftz
	) const
{
	return (abs(shiftx) + abs(shifty) + abs(shiftz) == 3);
}




//
//	Derived Buffer Types
//

SndBuffer::SndBuffer(
		int mxCell[3],
		int rim,
		int numMax
	):
		HaloBuffer(mxCell, rim, numMax)
{
	init_buffers();
}

void SndBuffer::get_extent(
		int loDelta[3], int hiDelta[3],
		int shiftx, int shifty, int shiftz,
		int loBuff[3], int hiBuff[3]
	) const
{
	int shift[3] = { shiftx, shifty, shiftz };

	// Compute the buffer limits
	for(int dir=0; dir < 3; ++dir) {

		if(shift[dir] ==  -1) {
			loBuff[dir] = hiDelta[dir];
			hiBuff[dir] = rim - 1;
		} else
		if(shift[dir] ==  0) {
			loBuff[dir] = loDelta[dir];
			hiBuff[dir] = mxCell[dir] + hiDelta[dir];
		} else
		if(shift[dir] ==  1) {
			loBuff[dir] = mxCell[dir] - rim + 1;
			hiBuff[dir] = mxCell[dir] + loDelta[dir];
		}
	}
}

int SndBuffer::compute_maxSingleSize(
	int rim,
	int shiftx, int shifty, int shiftz
) const
{
	int shift[3] = { shiftx, shifty, shiftz };
	int maxSize = 1;

	// Compute the buffer limits
	for(int dir=0; dir < 3; ++dir) {

		if(shift[dir] ==  -1) {
			maxSize *= 2 * rim;
		} else
		if(shift[dir] ==  0) {
			maxSize *= mxCell[dir] + 2 * rim;
		} else
		if(shift[dir] ==  1) {
			maxSize *= 2 * rim;
		}
	}
	return maxSize;
}


RcvBuffer::RcvBuffer(
		int mxCell[3],
		int rim,
		int numMax
	):
		HaloBuffer(mxCell, rim, numMax)
{
	init_buffers();
}

void RcvBuffer::get_extent(
		int loDelta[3], int hiDelta[3],
		int shiftx, int shifty, int shiftz,
		int loBuff[3], int hiBuff[3]
	) const
{
	int shift[3] = { shiftx, shifty, shiftz };

	// Compute the buffer limits

	for(int dir=0; dir < 3; ++dir) {

		if(shift[dir] ==  -1) {
			loBuff[dir] = -rim;
			hiBuff[dir] = loDelta[dir] - 1;
		} else
		if(shift[dir] ==  0) {
			loBuff[dir] = loDelta[dir];
			hiBuff[dir] = mxCell[dir] + hiDelta[dir];
		} else
		if(shift[dir] ==  1) {
			loBuff[dir] = mxCell[dir] + hiDelta[dir] + 1;
			hiBuff[dir] = mxCell[dir] + rim;
		}
	}

}

int RcvBuffer::compute_maxSingleSize(
	int rim,
	int shiftx, int shifty, int shiftz
) const
{
	int shift[3] = { shiftx, shifty, shiftz };
	int maxSize = 1;

	// Compute the buffer limits
	for(int dir=0; dir < 3; ++dir) {

		if(shift[dir] ==  -1) {
			maxSize *= 2 * rim;
		} else
		if(shift[dir] ==  0) {
			maxSize *= mxCell[dir] + 2 * rim;
		} else
		if(shift[dir] ==  1) {
			maxSize *= 2 * rim;
		}
	}
	return maxSize;
}
