#ifndef CRONOS_BOUNDARY_H
#define CRONOS_BOUNDARY_H

#include "data.H"
#include "problem.H"
#include "HaloBuffer.H"
#include "DataLink.H"

#include <stdlib.h>


class Boundary {
public:
	/**
	 * Constructor for boundary class
	 * @param num number of fields to be treated simultaneously
	 * */
	Boundary(Data &, int num);
//	void boundary(Data &, ProblemType &);
//	void boundary(Data &, ProblemType &, NumMatrix<double,3> &omb, int);
//	void boundary_old(Data &, ProblemType &, NumMatrix<double,3> &omb, int, int, int iFluid=0);
//
//	/*!
//	 * Optimised boundary condition routine for MPI parallel case - trying to avoid stalls by
//	 * MPI communication
//	 * @param gdata Grid and Array data class
//	 * @param problem Class containing the user setup
//	 * @param omb NumMatrix array holding the fields, for which bcs are to be computed
//	 * @param rim number of ghost cells to be taken into account
//	 * @param q index of omb field in global data array
//	 * @param iFluid index of fluid considered (only applicable in multifluid simulations)
//	 * */
//	void boundary(Data &gdata, ProblemType &problem, NumMatrix<double,3> &omb,
//			int rim, int q, int iFluid=0);


	/*!
	 * Boundary conditions with efficient handling of MPI communication
	 *
	 * */
	void boundary(Data &gdata, ProblemType &problem, NumMatrix<double,3> omb[], int num, int iFluid=0, bool non_generic=false);


	/*! internal MPI-boundary conditions only
	 * BCs will NOT be applied to the real boundaries of the domain
	 * routine needed, e.g., for the magnetic vector potential
	 * @param gdata Grid and Array data class
	 * @param problem Class containing the user setup
	 * @param omb NumMatrix array holding the fields, for which bcs are to be computed
	 * */
	void prep_boundaries(Data &, ProblemType &);
	int get_bc_Type(int);
private:
//	void bc_Periodic(Data &gdata, ProblemType &, NumMatrix<REAL,3> &omb,
//	                 int dir, int q, int rim, bool internal_only=false);


	/*!
	 * Determine size of send buffer in send direction
	 * @param gdata Grid and Array data class
	 * @param range_min begin of buffer
	 * @param end of buffer
	 * @param dir Cartesian direction for boundary communication
	 * @param leftToRight data transfer direction (1 for left to right, 0 else)
	 * @param q index of omb field in global data array
	 * @param rim number of ghost cells to be taken into account
	 * */
	void get_bcBuffRange(Data &gdata, int &range_min, int & range_max,
			int dir, int leftToRight, int q, int rim );


	/*!
	 * Periodic boundary conditions for non-mpi version of the code, only
	 * */
	void bc_Periodic_serial(Data &gdata, ProblemType &, NumMatrix<REAL,3> &omb,
			int dir, int q, int rim);


	void bc_AxisCyl(Data &gdata, ProblemType &Problem,
            NumMatrix<REAL,3> &omb, int q, int rim);
	void bc_AxisSph(Data &gdata, ProblemType &Problem,
			NumMatrix<REAL,3> &omb, int top, int q, int rim);
	void bc_Axis(Data &gdata, ProblemType &Problem,
	             NumMatrix<REAL,3> &omb, int dir, int top, int q, int rim);
	void get_AxisPartners(Data &gdata, int top=0);
	void do_AxisValCorrectionCyl(Data &gdata, ProblemType &Problem);
	void do_AxisValCorrectionSph(Data &gdata, ProblemType &Problem, bool top);


	void apply_bc_outer(Data &, ProblemType &, NumMatrix<double,3> &,
            int, int, int, int);

	void bc_Extrapolate(Data &, ProblemType &, NumMatrix<double,3> &,
	                    int, int, int, int);
	void bc_Outflow(Data &, ProblemType &, NumMatrix<double,3> &,
	                int, int, int, int);
	void bc_Reflecting(Data &, ProblemType &, NumMatrix<double,3> &,
	                   int, int, int, int);
	void bc_Fixed(Data &, ProblemType &, NumMatrix<double,3> &,
	              int, int, int, int);
	void bc_Fixed_general(NumMatrix<REAL,3> &, NumMatrix<REAL,3> &,
	                      int imin[3], int imax[3]);
	void ExchangePositions(Data &, int top=0);
#ifdef parallel
	void bc_MPI_Init(Data &gdata, NumMatrix<double,3> omb[], int num);

	/**
	* Initialize Buffer Fields (-> extents)
	* */
	void bc_MPI_InitBufferFields(Data &gdata, NumMatrix<double, 3> omb[], int num);

	/**
	 * Start all receive operations
	 * */
	void bc_MPI_RecvAll(Data &gdata, int num);

	/**
	 * Send single MPI buffer
	 * */
	void bc_MPI_sendBuff(Data &gdata, Buffer &sBuff, int num, int shift0, int shift1, int shift2);

	/**
	 * Check status of send buffer and wait
	 * */
	void bc_MPI_checkSendBuff(Buffer &sBuff, int shift0, int shift1, int shift2);

	/**
	 * Copy data from omb fields into send buffer
	 * */
	void fill_MPI_SendBuff(NumMatrix<double,3> omb[], Buffer & sBuff, int num, int shift0, int shift1, int shift2);
	/**
	 * Fill send buffers and send them via MPI
	 * */
	void process_MPI_Send(Data &gdata, NumMatrix<double,3> omb[], int num, int shift0, int shift1, int shift2);

	/**
	 * Finalize MPI communication -> check and write receive buffers
	 * */
	void bc_MPI_Finalize(NumMatrix<double,3> omb[], int num);

	/**
	 * Check status of receive buffer and wait
	 * */
	void bc_MPI_checkRecvBuff(Buffer &rBuff, int shift0, int shift1, int shift2);

	/**
	 * Copy data from receive buffer to omb field
	 * */
	void write_MPI_RecvBuff(NumMatrix<double,3> omb[], Buffer & rBuff, int num, int shift0, int shift1, int shift2);

	/**
	 * Check status of receive buffer and write to omb fields
	 * */
	void process_MPI_Recv(NumMatrix<double,3> omb[], int num, int shift0, int shift1, int shift2);

	void get_AxisData(Data &gdata, 
			NumMatrix<REAL,3> &myBound,
			NumMatrix<REAL,3> &global_bcData, int top=0);
	void store_AxisData(Data &gdata, int lbound[], int ubound[],
	                    int rank_local,
	                    NumMatrix<REAL,3> &local_bcData,
	                    NumMatrix<REAL,3> &global_bcData);
	void ExchangeData(Data &);
	MPI_Request requests_send[27], requests_recv[27];

	std::unique_ptr<SndBuffer> SendBuff;
	std::unique_ptr<RcvBuffer> RecvBuff;

#endif
	int bc_Type[6];
	NumMatrix<int,1> AxisPartnersPhi[2];  // array of mapping partners
	NumMatrix<REAL,1> PhiPos[2];
	NumMatrix<REAL,1> cosPos[3], sinPos[3];
	// NumMatrix<REAL,3> BoundVals[2*DIM];
	// NumMatrix<REAL,3> BoundVals_User[2*DIM];
	NumMatrix<REAL,3> bcVals_xb[N_OMINT], bcVals_xe[N_OMINT];
	NumMatrix<REAL,3> bcVals_yb[N_OMINT], bcVals_ye[N_OMINT];
	NumMatrix<REAL,3> bcVals_zb[N_OMINT], bcVals_ze[N_OMINT];

//	NumMatrix<double,3> Recv_x, Recv_y, Recv_z, Recv_empty;
//	NumMatrix<double,3> Send_x, Send_y, Send_z, Send_empty;
	/**
	 * Identifier of boundary type in Cartesian directions
	 * (-1 -> left; +1 -> right)
	 * values: 0 -> out boundary; 1 -> MPI boundary
	 * */
	NumMatrix<int,3> is_halo;

//	NumArray<REAL> RecvArr_x, RecvArr_y, RecvArr_z, RecvArr_empty;
//	NumArray<REAL> SendArr_x, SendArr_y, SendArr_z, SendArr_empty;
//	NumArray<REAL> SendArr_buff, RecvArr_buff;
//	NumArray<REAL> RecvArrSmall_x, RecvArrSmall_y, RecvArrSmall_z;
//	NumArray<REAL> SendArrSmall_x, SendArrSmall_y, SendArrSmall_z;
//	NumArray<REAL> RecvArr_xLR, RecvArr_yLR, RecvArr_zLR;
//	NumArray<REAL> RecvArr_xRL, RecvArr_yRL, RecvArr_zRL;
//	NumArray<REAL> SendArr_xLR, SendArr_yLR, SendArr_zLR;
//	NumArray<REAL> SendArr_xRL, SendArr_yRL, SendArr_zRL;
//
//	NumArray<REAL> SendArrRL, SendArrLR, RecvArrRL, RecvArrLR;

	NumArray<int> i_min, i_max;

//#ifdef parallel
//	MPI_Request requests_xRL[2], requests_xLR[2];
//	MPI_Request requests_yRL[2], requests_yLR[2];
//	MPI_Request requests_zRL[2], requests_zLR[2];
//	MPI_Request requests_RL[2], requests_LR[2];
//#endif

#if (OMS_USER == TRUE)
	NumMatrix<REAL,3> bcVals_User_xb[N_OMINT], bcVals_User_xe[N_OMINT];
	NumMatrix<REAL,3> bcVals_User_yb[N_OMINT], bcVals_User_ye[N_OMINT];
	NumMatrix<REAL,3> bcVals_User_zb[N_OMINT], bcVals_User_ze[N_OMINT];
#endif
	int n_om, n_omInt, n_omIntAll;
	int n_omUser, n_omIntUser;
	int q_Bx, q_By, q_Bz;
	int rimDefault;
	int numMax;
};

#endif
